/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *% Copyright (C) 1993 by WATCOM International Corp. All rights      %
 *% reserved. No part of this software may be reproduced in any form %
 *% or by any means - graphic, electronic or mechanical, including   %
 *% photocopying, recording, taping or information storage and	     %
 *% retrieval systems - except with the written permission of WATCOM %
 *% International Corp. 					     %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *
 * Modified:	By:		Reason:
 * ---------	---		-------
 * 93-03-08	LifeBoat Inc.	386 support
--------------------------------------------------------Version-11.0---------
 * 95-05-22	Greg Bentz	variety.h and dll linkage
 * 95-10-06	M. Hildebrand	Changed _bios prefix to __nec98_bios.
 *
 */

#include "variety.h"
#include <dos.h>
#include <bios98.h>
#include "nonibm.h"

/*
 * FUNCTION: __nec98_bios_keybrd
 * DESCRIPTION: uses int 18h to access bios keyboard functions
 *	_KEYBRD_READ - read a byte from the keyboard buffer.
 *		-if no keystroke is available wait until one is
 *	_KEYBRD_READY - check if a keystroke is available
 *		-if one is return it without removing it from the buffer
 *	_KEYBRD_SHIFTSTATUS - get the status of the shift keys
 *	_KEYBRD_INITIALIZE - initialize the keyboad buffer
 *	_KEYBRD_INPUTSTATUS - get status of the 16 key groups
 *
 *	The parameter __buf is ignored except for _KEYBRD_INPUTSTATUS
 *
 * RETURNS:
 * _KEYBRD_READ - ASCII value in low byte
 *		- Scan Code in high byte
 * _KEYBRD_READY - 0 if no character in buffer
 *		 - same as _KEYBRD_READ otherwise
 * _KEYBRD_SHIFTSTATUS - low order 5 bits are significant
 *			- bit 0 	shift key pressed
 *			- bit 1 	caps key pressed
 *			- bit 2 	kana key pressed
 *			- bit 3 	graph key pressed
 *			- bit 4 	ctrl key pressed
 * _KEYBRD_INITIALIZE - no meaningful return value
 * _KEYBRD_INPUTSTATUS - status of the 16 key groups placed in __buf
 */

_WCRTLINK unsigned short __nec98_bios_keybrd( unsigned __cmd, unsigned char * __buf )
{
    union REGS r;
    int i;

    if( !__NonIBM )  return( 0 );	// fail if not a NEC 98 machine

#if defined(__386__)
    switch( r.h.ah = __cmd ) {
      case _KEYBRD_READ:
	int386( 0x18, &r, &r );
	break;
      case _KEYBRD_READY:
	int386( 0x18, &r, &r );
	if( !r.h.bh ) r.x.eax = 0;
	break ;
      case _KEYBRD_SHIFTSTATUS:
	int386( 0x18, &r, &r );
	r.h.ah = 0;
	break;
      case _KEYBRD_INITIALIZE:
	int386( 0x18, &r, &r );
	break;
      default:
	for( i = 0; i < 16; i++ ){
	    r.h.al = i;
	    int386( 0x18, &r, &r );
	    __buf[i] = r.h.ah;
	}
    }
    return r.w.ax;
#else
    switch( r.h.ah = __cmd ) {
      case _KEYBRD_READ:
	int86( 0x18, &r, &r );
	break;
      case _KEYBRD_READY:
	int86( 0x18, &r, &r );
	if( !r.h.bh ) r.w.ax = 0;
	break ;
      case _KEYBRD_SHIFTSTATUS:
	int86( 0x18, &r, &r );
	r.h.ah = 0;
	break;
      case _KEYBRD_INITIALIZE:
	int86( 0x18, &r, &r );
	break;
      default:
	for( i = 0; i < 16; i++ ){
	    r.h.al = i;
	    int86( 0x18, &r, &r );
	    __buf[i] = r.h.ah;
	}
    }
    return r.w.ax;
#endif
}
