;=========================================================================
;==	Name:		FDFS						==
;==	Operation:	Float double to float single conversion		==
;==	Inputs:		AX;BX;CX;DX	double precision float		==
;==	Outputs:	DX;AX		single precision float		==
;==	Volatile:	CX, DX destroyed				==
;==	Date:		By:		Reason:				==
;==	85-10-02	David Yach	initial implementation		==
;==	86-09-24	F.W.Crigger	chg result to DX:AX from AX:DX	==
;==	88-11-29	F.W.Crigger	check exponent for 0, allow -0	==
;==	92-11-05	F.W.Crigger	set overflow to true IEEE inf	==
;==----------------------------------------------------Version 11.0-----==
;==	25-nov-95	Greg Bentz	wasm export support             ==
;=========================================================================
include mdef.inc
include struct.inc

	modstart	fdfs

	xdefp	__FDFS

	defpe	__FDFS
	test	ax,07ff0h	; check exponent
	je	retzero		; if exponent = 0 then just return 0
	_shl	cx,1		; shift number over
	_rcl	bx,1		; ...
	_rcl	ax,1		; shift out sign bit
	mov	dh,0		; put 0 in dh
	rcr	dh,1		; save sign bit in dh
	add	ch,20h		; round floating point number
	adc	bx,0		; ...
	adc	ax,0		; increment exponent if need be
	je	oflow		; overflow if exponent went to 0
	cmp	ax,2*16*(03ffh+80h) ; check for maximum exponent
	jae	oflow		; overflow if above or equal
	cmp	ax,2*16*(03ffh-7eh) ; check for minimum exponent
	jb	uflow		; underflow if below
	sub	ax,2*16*(03ffh-7fh) ; correct bias
	_shl	cx,1		; do rest of shift
	_rcl	bx,1		; ...
	_rcl	ax,1		; ...
	_shl	cx,1		; ...
	_rcl	bx,1		; ...
	_rcl	ax,1		; ...
	or	ah,dh		; put in sign bit
	mov	dx,bx		; put low part of mantissa in correct reg
	xchg	ax,dx		; flip results	FWC 86-09-24
	ret			; return

oflow:	mov	ax,7f80h	; return maximum possible number
	or	ah,dh		; put in sign bit
	sub	dx,dx		; ...
	xchg	ax,dx		; flip results	FWC 86-09-24
	ret			; and return

uflow:
retzero:sub	ax,ax
	mov	dx,ax		; return ax;dx = 0
	ret
	endproc __FDFS

	endmod
	end
