;=========================================================================
;==	Name:		FDFS						==
;==	Operation:	Float double to float single conversion		==
;==	Inputs:		EDX:EAX		double precision float		==
;==	Outputs:	EAX		single precision float		==
;==	Volatile:	EDX destroyed					==
;==	Date:		By:		Reason:				==
;==	85-10-02	David Yach	initial implementation		==
;==	86-09-24	F.W.Crigger	chg result to DX:AX from AX:DX	==
;==	88-11-29	F.W.Crigger	check exponent for 0, allow -0	==
;==	89-03-10	F.W.Crigger	Adapted for 386			==
;==	92-11-05	F.W.Crigger	set overflow to true IEEE inf	==
;==----------------------------------------------------Version 11.0-----==
;==	25-nov-95	Greg Bentz	wasm export support             ==
;=========================================================================
include mdef.inc
include struct.inc

	modstart	fdfs

	xdefp	__FDFS

	defpe	__FDFS
	push	EBX		; save EBX
	test	EDX,07ff00000h	; check exponent
	je	short retzero	; if exponent = 0 then just return 0
	sub	EBX,EBX		; set to 0
	_shl	EAX,1		; shift number over
	_rcl	EDX,1		; ...
	rcr	EBX,1		; save sign
	add	EAX,20000000h	; round floating point number
	adc	EDX,0		; ...
	je	oflow		; overflow if exponent went to 0
	cmp	EDX,(03ffh+80h) shl 21 ; check for maximum exponent
	jae	oflow		; overflow if above or equal
	cmp	EDX,(03ffh-7eh) shl 21 ; check for minimum exponent
	jb	uflow		; underflow if below
	sub	EDX,(03ffh-7fh) shl 21 ; correct bias
	_shl	EAX,1		; do rest of shift
	_rcl	EDX,1		; ...
	_shl	EAX,1		; ...
	_rcl	EDX,1		; ...
	or	EDX,EBX		; put in sign bit
	mov	EAX,EDX		; get result into EAX
	pop	EBX		; restore EBX
	ret			; return

oflow:	mov	EAX,7F800000h	; return maximum possible number
	or	EAX,EBX		; put in sign bit
	pop	EBX		; restore EBX
	ret			; and return

uflow:
retzero:sub	EAX,EAX		; set result to 0
	pop	EBX		; restore EBX
	ret
	endproc __FDFS

	endmod
	end
