/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  17-mar-87	F.W.Crigger	Initial implementation
  19-apr-87	F.W.Crigger	Added error checking
  29-jan-88	F.W.Crigger	strtol - set *endptr to nptr if conversion
				fails.
------------------------------------------------------------Version 6.5-----
  27-oct-88	F.W.Crigger	strtol and strtoul merged into _stol so that
				specs of ANSI standard can be achieved.
------------------------------------------------------------Version 8.0-----
  27-nov-90	A.F.Scian	made _stol static
  10-Jun-92	D. Wellstood	Unicode support
  08-sep-94	Greg Bentz	add check for overflow on multiply
--------------------------------------------------------Version-11.0---------
  23-feb-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  22-may-95	Greg Bentz	variety.h and dll linkage
  21-sep-95	M. Hildebrand	Added _u version, which calls wide version.
  03-jan-96	Greg Bentz	remove stubs that are now aliases
*/
#include "variety.h"
#include "widechar.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include "seterrno.h"

/*
 * this table is the largest value that can safely be multiplied
 * by the associated base without overflowing
 */
static unsigned long nearly_overflowing[] = {
    ULONG_MAX / 2,  ULONG_MAX / 3,  ULONG_MAX / 4,  ULONG_MAX / 5,
    ULONG_MAX / 6,  ULONG_MAX / 7,  ULONG_MAX / 8,  ULONG_MAX / 9,
    ULONG_MAX / 10, ULONG_MAX / 11, ULONG_MAX / 12, ULONG_MAX / 13,
    ULONG_MAX / 14, ULONG_MAX / 15, ULONG_MAX / 16, ULONG_MAX / 17,
    ULONG_MAX / 18, ULONG_MAX / 19, ULONG_MAX / 20, ULONG_MAX / 21,
    ULONG_MAX / 22, ULONG_MAX / 23, ULONG_MAX / 24, ULONG_MAX / 25,
    ULONG_MAX / 26, ULONG_MAX / 27, ULONG_MAX / 28, ULONG_MAX / 29,
    ULONG_MAX / 30, ULONG_MAX / 31, ULONG_MAX / 32, ULONG_MAX / 33,
    ULONG_MAX / 34, ULONG_MAX / 35, ULONG_MAX / 36
};

#define hexstr(p) (p[0] == '0' && (p[1] == 'x' || p[1] == 'X'))

static unsigned long int _stol( const CHAR_TYPE *nptr,CHAR_TYPE **endptr,int base,int who)
    {
	const CHAR_TYPE *p;
	const CHAR_TYPE *startp;
	int digit;
	unsigned long int value;
	unsigned long int prev_value;
	CHAR_TYPE sign;
	char overflow;		/*overflow is used as a flag so it does not
				 *need to be of type CHAR_TYPE */

	if( endptr != NULL ) *endptr = (CHAR_TYPE *)nptr;
	p = nptr;
	while( __F_NAME(isspace,iswspace)(*p) ) ++p;
	sign = *p;
	if( sign == '+'  ||  sign == '-' ) ++p;
	if( base == 0 ) {
	    if( hexstr(p) )	 base = 16;
	    else if( *p == '0' ) base = 8;
	    else		 base = 10;
	}
	if( base < 2  ||  base > 36 ) {
	    __set_errno( EDOM );
	    return( 0 );
	}
	if( base == 16 ) {
	    if( hexstr(p) )  p += 2;	/* skip over '0x' */
	}
	startp = p;
	overflow = 0;
	value = 0;
	for(;;) {
	    digit = radix_value( *p );
	    if( digit >= base ) break;
	    if( value > nearly_overflowing[base-2] ) overflow = 1;
	    prev_value = value;
	    value = value * base + digit;
	    if( value < prev_value ) overflow = 1;
	    ++p;
	}
	if( p == startp )  p = nptr;
	if( endptr != NULL ) *endptr = (CHAR_TYPE *)p;
	if( who == 1 ) {
	    if( value >= 0x80000000 ) {
		if( value == 0x80000000  &&  sign == '-' ) {
		    ;  /* OK */
		} else {
		    overflow = 1;
		}
	    }
	}
	if( overflow ) {
	    __set_errno( ERANGE );
	    if( who == 0 )     return( ULONG_MAX );
	    if( sign == '-' )  return( LONG_MIN );
	    return( LONG_MAX );
	}
	if( sign == '-' )  value = - value;
	return( value );
    }


_WCRTLINK unsigned long int __F_NAME(strtoul,wcstoul)( const CHAR_TYPE *nptr,
		CHAR_TYPE **endptr, int base )
    {
	return( _stol( nptr, endptr, base, 0 ) );
    }


_WCRTLINK long int __F_NAME(strtol,wcstol)( const CHAR_TYPE *nptr, CHAR_TYPE **endptr, int base )
    {
	return( _stol( nptr, endptr, base, 1 ) );
    }

static int radix_value( CHAR_TYPE c )
    {
	if( c >= '0'  &&  c <= '9' ) return( c - '0' );
	c = __F_NAME(tolower,towlower)(c);
	if( c >= 'a'  &&  c <= 'i' ) return( c - 'a' + 10 );
	if( c >= 'j'  &&  c <= 'r' ) return( c - 'j' + 19 );
	if( c >= 's'  &&  c <= 'z' ) return( c - 's' + 28 );
	return( 37 );
    }
