/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %       Copyright (C) 1991, by WATCOM Systems Inc. All rights     %
   %       reserved. No part of this software may be reproduced      %
   %       in any form or by any means - graphic, electronic or      %
   %       mechanical, including photocopying, recording, taping     %
   %       or information storage and retrieval systems - except     %
   %       with the written permission of WATCOM Systems Inc.        %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
   Date		By		Reason
   ----		--		------
   91/10/08	G. Coschi 	initial implementation
   91/12/15	Craig Eisler	multiple windowing support
   27-mar-92	Craig Eisler	different call to NewWindow
   21-oct-92	Craig Eisler	ANSI function names
--------------------------------------------------------Version-11.0---------
   20-feb-95	M.P.Hildebrand	Added multibyte character support.
   10-apr-95	M. Hildebrand	If __IsDBCS is true, then try to find a DBCS
				font; otherwise, find a fixed-size font.
   21-apr-95	M. Hildebrand	Now all fonts are queried correctly when
				choosing a font to use.
*/

#include "variety.h"
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#define	INCL_WIN
#define	INCL_GPI
#include <wos2.h>
#include "win.h"
#include "pmmenu.h"

static	char		*WatcomClass = { "WATCOM" };
static	HWND		ClientWindow;
static	HMQ		hMessageQueue = { NULL };
static	FATTRS		FontAttrs;

#define DISPLAY(x)	WinMessageBox( HWND_DESKTOP, NULL, x, "Error", 0, MB_APPLMODAL | MB_NOICON | MB_OK | MB_MOVEABLE );

extern void _SetWinMenuHandle( HWND hmenu );
extern HWND _GetWinMenuHandle( void );

_WCRTLINK int	__InitDefaultWin() {
//==================================

    ULONG	style;
    RECTL	rcl;
    HWND	menhdl;
    HWND	submenu;
    MENUITEM	*menudesc;
    MENUITEM	menus;
    SWP		swp;

    _ClassName = WatcomClass;

    _AnchorBlock = WinInitialize( 0 );
    if( _AnchorBlock == 0 ) return( 0 );
    hMessageQueue = WinCreateMsgQueue( _AnchorBlock, 0 );
    if( hMessageQueue == 0 ) return( 0 );
    if( !WinRegisterClass( _AnchorBlock, _ClassName, (PFNWP)_MainDriver,
                           CS_SIZEREDRAW, 0 ) ) {
	return( 0 );
    }

    _InitMainWindowData( 0 );

    style = FCF_TITLEBAR | FCF_SYSMENU | FCF_SIZEBORDER | FCF_MINMAX |
	    FCF_SHELLPOSITION | FCF_TASKLIST;
    _MainFrameWindow = WinCreateStdWindow( HWND_DESKTOP,
    			WS_VISIBLE | WS_CLIPCHILDREN,
    			&style, _ClassName, "", 0, NULL, 0, &ClientWindow );

    if( _MainFrameWindow == 0 ) return( 0 );
    WinSendMsg( _MainFrameWindow, WM_SETICON,
    	MPFROMLONG( WinQuerySysPointer( HWND_DESKTOP, SPTR_APPICON, TRUE ) ), 0 );
    WinQueryWindowRect( _MainWindow, &rcl );
    WinSetWindowPos( _MainFrameWindow, HWND_TOP,
    		rcl.xLeft, 
		rcl.yBottom, 
    		rcl.xLeft - rcl.xRight, 
		rcl.yTop - rcl.yBottom,
    		SWP_SHOW | SWP_SIZE | SWP_MOVE | SWP_MAXIMIZE );
    
    _InitFunctionPointers();
    _MainWindow = ClientWindow;
    WinUpdateWindow( _MainWindow );
    
    menudesc = &menus;
    menhdl = WinCreateMenu( _MainFrameWindow, NULL );
    WinSetWindowBits( menhdl, QWL_STYLE, MS_ACTIONBAR, MS_ACTIONBAR );
    submenu = WinCreateMenu( menhdl, NULL );

    menudesc->afStyle = MIS_TEXT;
    menudesc->afAttribute = 0;
    menudesc->hwndSubMenu = NULL;
    menudesc->hItem = 0;

    menudesc->iPosition = 0;
    menudesc->id = DID_FILE_SAVE;
    if ( MIT_ERROR == (BOOL)WinSendMsg( submenu, ( ULONG )MM_INSERTITEM, MPFROMP( menudesc ), MPFROMP( "Save As..." ) ) ) abort();
    menudesc->iPosition = 1;
    menudesc->id = DID_FILE_CLEAR;
    if ( MIT_ERROR == (BOOL)WinSendMsg( submenu, ( ULONG )MM_INSERTITEM, MPFROMP( menudesc ), MPFROMP( "Set Lines Between Clears..." ) ) ) abort();
    menudesc->afStyle = MIS_SEPARATOR;
    menudesc->iPosition = 2;
    menudesc->id = 0;
    if ( MIT_ERROR == (BOOL)WinSendMsg( submenu, ( ULONG )MM_INSERTITEM, MPFROMP( menudesc ), MPFROMP( "Exit" ) ) ) abort();
    menudesc->afStyle = MIS_TEXT;
    menudesc->iPosition = 3;
    menudesc->id = DID_FILE_EXIT;
    if ( MIT_ERROR == (BOOL)WinSendMsg( submenu, ( ULONG )MM_INSERTITEM, MPFROMP( menudesc ), MPFROMP( "Exit" ) ) ) abort();
    menudesc->iPosition = 0;
    menudesc->id = DID_MAIN_FILE;
    menudesc->hwndSubMenu = submenu;
    menudesc->afStyle = MIS_TEXT | MIS_SUBMENU;
    if ( MIT_ERROR == (BOOL)WinSendMsg( menhdl, ( ULONG )MM_INSERTITEM, MPFROMP( menudesc ), MPFROMP( "~File" ) ) ) abort();
    
    submenu = WinCreateMenu( menhdl, NULL );
    menudesc->afStyle = MIS_TEXT;
    menudesc->iPosition = 0;
    menudesc->id = DID_EDIT_CLEAR;
    menudesc->hwndSubMenu = NULL;
    if ( MIT_ERROR == (BOOL)WinSendMsg( submenu, ( ULONG )MM_INSERTITEM, MPFROMP( menudesc ), MPFROMP( "Clear" ) ) ) abort();
    menudesc->iPosition = 1;
    menudesc->id = DID_EDIT_COPY;
    if ( MIT_ERROR == (BOOL)WinSendMsg( submenu, ( ULONG )MM_INSERTITEM, MPFROMP( menudesc ), MPFROMP( "Copy" ) ) ) abort();
    menudesc->iPosition = 1;
    menudesc->id = DID_MAIN_EDIT;
    menudesc->hwndSubMenu = submenu;
    menudesc->afStyle = MIS_SUBMENU;
    if ( MIT_ERROR == (BOOL)WinSendMsg( menhdl, ( ULONG )MM_INSERTITEM, MPFROMP( menudesc ), MPFROMP( "~Edit" ) ) ) abort();

    submenu = WinCreateMenu( menhdl, NULL );
    _SetWinMenuHandle( submenu );
    menudesc->iPosition = 2;
    menudesc->afAttribute = 0;
    menudesc->id = DID_MAIN_WIND;
    menudesc->hwndSubMenu = submenu;
    menudesc->afStyle = MIS_SUBMENU;
    if ( MIT_ERROR == (BOOL)WinSendMsg( menhdl, ( ULONG )MM_INSERTITEM, MPFROMP( menudesc ), MPFROMP( "~Windows" ) ) ) abort();

    submenu = WinCreateMenu( menhdl, NULL );
    menudesc->afStyle = MIS_TEXT;
    menudesc->iPosition = 0;
    menudesc->id = DID_HELP_ABOUT;
    menudesc->hwndSubMenu = NULL;
    if ( MIT_ERROR == (BOOL)WinSendMsg( submenu, ( ULONG )MM_INSERTITEM, MPFROMP( menudesc ), MPFROMP( "About" ) ) ) abort();
    menudesc->iPosition = 3;
    menudesc->id = DID_MAIN_HELP;
    menudesc->hwndSubMenu = submenu;
    menudesc->afStyle = MIS_SUBMENU;
    if ( MIT_ERROR == (BOOL)WinSendMsg( menhdl, ( ULONG )MM_INSERTITEM, MPFROMP( menudesc ), MPFROMP( "~Help" ) ) ) abort();
    
    WinSendMsg( _MainFrameWindow, ( ULONG )WM_UPDATEFRAME, 0, 0 );
    _NewWindow( "Standard IO", 0,1,2,-1 );
    WinUpdateWindow( _MainFrameWindow );
    WinQueryWindowPos( _MainWindow, &swp );
    swp.fl |= SWP_MAXIMIZE;
    WinSetWindowPos( _MainWindow, HWND_TOP, swp.x, swp.y, swp.cx, swp.cy, SWP_MAXIMIZE | SWP_ACTIVATE | SWP_MOVE | SWP_SHOW | SWP_SIZE );
    
    return( 1 );
}

_WCRTLINK void	__FiniDefaultWin() {
//==================================

    if( _MainWindow != 0 ) {
        WinDestroyWindow( _MainWindow );
    }
    if( hMessageQueue != 0 ) {
        WinDestroyMsgQueue( hMessageQueue );
    }
    WinTerminate( _AnchorBlock );
}

void	_CreateFont( LPWDATA w ) {
//===============================

    LONG		cFonts = { 0 };
    LONG		num_fonts;
    FONTMETRICS	*	fonts;
    int			i;
    unsigned		font_size;
    unsigned		font_selected;
    HPS			ps;
#ifdef _MBCS
    int			fontChosenFlag = 0;
#endif

    /*** Prepare to search for a suitable console font ***/
    ps = WinGetPS( w->hwnd );
    num_fonts = GpiQueryFonts( ps, QF_PUBLIC, NULL, &cFonts,
                               sizeof( FONTMETRICS ), NULL );
    fonts = _MemAlloc( num_fonts * sizeof( FONTMETRICS ) );
    GpiQueryFonts( ps, QF_PUBLIC, NULL, &num_fonts, sizeof( FONTMETRICS ),
		   fonts );
    font_size = UINT_MAX;
    font_selected = 0;

    #ifdef _MBCS
	if( __IsDBCS ) {
	    /*** Try to find a DBCS font ***/
	    for( i = 0; i < num_fonts; ++i ) {
		if( ! (fonts[i].fsType & (FM_TYPE_DBCS|FM_TYPE_FIXED)) )  continue;
		if( fonts[i].lEmHeight > 10 )  continue;
		font_size = fonts[i].lEmHeight;
		font_selected = i;
		fontChosenFlag = 1;
		break;
	    }
	    /*** Try to find a regular font if can't find DBCS ***/
	    if( fontChosenFlag == 0 ) {
		for( i = 0; i < num_fonts; ++i ) {
		    if( ! (fonts[i].fsType & FM_TYPE_DBCS) )  continue;
		    if( fonts[i].lEmHeight > 10 )  continue;
		    font_size = fonts[i].lEmHeight;
		    font_selected = i;
		    fontChosenFlag = 1;
		    break;
		}
	    }
	}

	/*** If !__IsDBCS or if DBCS font not found, find a fixed font ***/
	if( fontChosenFlag == 0 ) {
	    for( i = 0; i < num_fonts; ++i ) {
		if( ! (fonts[i].fsType & FM_TYPE_FIXED) )  continue;
		if( fonts[i].lEmHeight > 10 )  continue;
		font_size = fonts[i].lEmHeight;
		font_selected = i;
		break;
	    }
	}
    #else
	/*** Try to find a suitable font ***/
	for( i = 0; i < num_fonts; ++i ) {
	    if( ! (fonts[i].fsType & FM_TYPE_FIXED) )  continue;
	    if( fonts[i].lEmHeight > 10 )  continue;
	    font_size = fonts[i].lEmHeight;
	    font_selected = i;
	    break;
        }
    #endif

    /*** Set up the chosen font ***/
    FontAttrs.usRecordLength = sizeof( FATTRS );
    FontAttrs.fsSelection = 0;
    FontAttrs.lMatch = fonts[font_selected].lMatch;
    strcpy( FontAttrs.szFacename, fonts[font_selected].szFacename );
    FontAttrs.idRegistry = fonts[font_selected].idRegistry;
    FontAttrs.usCodePage = 0;
    FontAttrs.lMaxBaselineExt = 0;
    FontAttrs.lAveCharWidth = 0;
    FontAttrs.fsType = 0;
    FontAttrs.fsFontUse = 0;
    w->xchar = fonts[font_selected].lAveCharWidth;
    w->ychar = fonts[font_selected].lMaxBaselineExt + SPACE_BETWEEN_LINES;
    w->base_offset = fonts[font_selected].lMaxDescender;
    _MemFree( fonts );
    WinReleasePS( ps );
}


void	_SelectFont( HPS ps ) {
//============================
    
    GpiCreateLogFont( ps, NULL, FIXED_FONT, &FontAttrs );
    GpiSetCharSet( ps, FIXED_FONT );
}
