/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %       Copyright (C) 1991, by WATCOM Systems Inc. All rights     %
   %       reserved. No part of this software may be reproduced      %
   %       in any form or by any means - graphic, electronic or      %
   %       mechanical, including photocopying, recording, taping     %
   %       or information storage and retrieval systems - except     %
   %       with the written permission of WATCOM Systems Inc.        %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
   Date		By		Reason
   ----		--		------
   01-aug-91	Craig Eisler	initial implementation
---------------------------------------------------------Version 9.0--------
   15-dec-91	Craig Eisler	multiple window support
---------------------------------------------------------Version 9.x--------
   27-mar-92	Craig Eisler	NT version
   21-oct-92	Craig Eisler	ANSI function names
   03-dec-93	Greg Bentz	call flushall() when execution complete
   12-may-94	Alex Brodksy	Added more functinaly
   25-may-94	Greg Bentz	change Yield() to Sleep(0) for Win32
--------------------------------------------------------Version-11.0---------
   20-feb-95	M.P.Hildebrand	Added multibyte character support.
   20-feb-95	Alex Brodsky /	Fixed bug which caused the default windowing
   		M.P.Hildebrand	system to eat too much CPU time.
   10-apr-95	M. Hildebrand	Fixed assorted bugs.
   12-apr-95	M. Hildebrand	Use FAR_mb* rather than redefining _fmb*.
   18-apr-95	M. Hildebrand	Fixed a _dwShutDown bug.
   17-may-96	M. Hildebrand	_CloseWindow removes window data.
*/

#include "variety.h"
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "win.h"
/*
 * _SetMyDC - set display context for windows
 */
HFONT _SetMyDC( HDC dc, DWORD bkclr, DWORD txtclr )
{
    SetBkColor( dc, bkclr );
    SetTextColor( dc, txtclr );
    return( SelectObject( dc, _FixedFont ) );

} /* _SetMyDC */

/*
 * _MessageLoop
*/
int _MessageLoop( BOOL doexit )
{
    MSG		msg;
    WORD	rc=1;

    while( PeekMessage( &msg, NULL, NULL, NULL, PM_NOYIELD | PM_NOREMOVE ) ) {
	rc = GetMessage( &msg, NULL, NULL, NULL );
	if( !rc ) {
	    if( doexit ) {
		_WindowExitRtn = NULL;
		exit( msg.wParam );
	    }
	    break;
	}
	TranslateMessage( &msg );
	DispatchMessage( &msg );
    }
    #if defined(__NT__)
	// Yield() is obsolete under Win32
	Sleep(0);
    #else
	Yield();
    #endif
    return( rc );

} /* _MessageLoop */

/*
 * _BlockingMessageLoop
*/
int _BlockingMessageLoop( BOOL doexit )
{
    MSG		msg;
    WORD	rc=1;

    rc = GetMessage( &msg, NULL, NULL, NULL );
    if( !rc ) {
	if( doexit ) {
	    _WindowExitRtn = NULL;
	    exit( msg.wParam );
	}
    } else {
	TranslateMessage( &msg );
	DispatchMessage( &msg );
    }
    return( _MessageLoop( doexit ) );

} /* _BlockingMessageLoop */

/*
 * _OutOfMemory - display out of memory message
 */
void _OutOfMemory( void )
{
    MessageBox( NULL,"Out Of Memory!","SYSTEM ERROR", MB_OK );

} /* _OutOfMemory */

/*
 * _ExecutionComplete - display execution complete message
 */
void _ExecutionComplete( void )
{
    flushall();
//    MessageBox( NULL, "Execution Complete", "", MB_OK | MB_ICONEXCLAMATION );

} /* _ExecutionComplete */

int 	_SetConTitle( LPWDATA w, char *title ) {
//==============================================

    SetWindowText( w->hwnd, title );
    return( 1 );
}
    
int 	_SetAppTitle( char *title ) {
//===================================

    SetWindowText( _MainWindow, title );
    return( 1 );
}

int	_ShutDown( void ) {
//=========================

    flushall();
    DestroyWindow( _MainWindow );
    _MainWindowDestroyed = 1;
    return( 0 );
}


int	_CloseWindow( LPWDATA w ) {
//=================================

    if( w->destroy ) {
	_DestroyAWindow( w );
    	DestroyWindow( w->hwnd );
    }
    return( 0 );
}


/*
 * _NewCursor - change the cursor type
 */
void _NewCursor( LPWDATA w, cursors type )
{
    if( w->hascursor ) {
	DestroyCaret();
	w->hascursor = FALSE;
    }
    if( type == KILL_CURSOR ) return;
    w->CaretType = type;
    switch( type ) {
    case SMALL_CURSOR:
	CreateCaret( w->hwnd, NULL, 0, w->ychar );
	w->hascursor = TRUE;
	break;
    case FAT_CURSOR:
	CreateCaret( w->hwnd, NULL, 4, w->ychar );
	w->hascursor = TRUE;
	break;
    case ORIGINAL_CURSOR:
	break;
    }

} /* _NewCursor */

/*
 * _DisplayCursor - show the current cursor position
 */
void _DisplayCursor( LPWDATA w )
{
    HDC			dc;
    SIZE		size;

    dc = GetDC( w->hwnd );
    SelectObject( dc, _FixedFont );
    #ifdef _MBCS
	#ifdef __NT__
	    GetTextExtentPoint32( dc, w->tmpbuff->data,
				  FAR_mbsnbcnt(w->tmpbuff->data,w->buffoff+w->curr_pos),
				  &size );
	#else
	    GetTextExtentPoint( dc, w->tmpbuff->data,
				FAR_mbsnbcnt(w->tmpbuff->data,w->buffoff+w->curr_pos),
				&size );
	#endif
    #else
	#ifdef __NT__
	    GetTextExtentPoint32( dc, w->tmpbuff->data, w->buffoff+w->curr_pos,
	    			  &size );
	#else
	    GetTextExtentPoint( dc, w->tmpbuff->data, w->buffoff+w->curr_pos,
	    			&size );
	#endif
    #endif    
    SetCaretPos( size.cx+1, (w->LastLineNumber-w->TopLineNumber)*w->ychar );
    ReleaseDC( w->hwnd, dc );
    ShowCaret( w->hwnd );
    
} /* _DisplayCursor */

/*
 * _SetInputMode - set whether or not we are in input mode
 */
void _SetInputMode( LPWDATA w, BOOL val  )
{
    WORD cmd;

    w->InputMode = val;
    if( w->InputMode ) cmd = MF_GRAYED;
    else cmd = MF_ENABLED;
    EnableMenuItem( _SubMenuEdit, MSG_FLUSH, cmd );

} /* _SetInputMode */

/*
 * _ShowWindowActive - do windows-specific stuff to make a window active
 */
void _ShowWindowActive( LPWDATA w, LPWDATA last )
{
    if( last != NULL ) {
	CheckMenuItem( _SubMenuWindows, MSG_WINDOWS+last->handles[0],
		 MF_UNCHECKED | MF_BYCOMMAND );
	SendMessage( last->hwnd, WM_NCACTIVATE, FALSE, 0L );
	if( last->CaretType != ORIGINAL_CURSOR ) {
	    _NewCursor( last, KILL_CURSOR );
	}
    }
    if( w != NULL ) {
	CheckMenuItem( _SubMenuWindows, MSG_WINDOWS+w->handles[0],
		 MF_CHECKED | MF_BYCOMMAND );
	ShowWindow( w->hwnd, SW_SHOW );
	SendMessage( w->hwnd, WM_NCACTIVATE, TRUE, 0L );
	BringWindowToTop( w->hwnd );
	if( w->CaretType != ORIGINAL_CURSOR ) {
	    _NewCursor( w, w->CaretType );
	    _DisplayCursor( w );
	}
    }

} /* _ShowWindowActive */
