/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %       Copyright (C) 1991, by WATCOM Systems Inc. All rights     %
   %       reserved. No part of this software may be reproduced      %
   %       in any form or by any means - graphic, electronic or      %
   %       mechanical, including photocopying, recording, taping     %
   %       or information storage and retrieval systems - except     %
   %       with the written permission of WATCOM Systems Inc.        %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
   Date		By		Reason
   ----		--		------
   01-aug-91	Craig Eisler	initial implementation
---------------------------------------------------------Version 9.0--------
   15-oct-91	G. Coschi	modifications for OS/2 PM
   15-dec-91	Craig Eisler	multiple window support
---------------------------------------------------------Version 9.x--------
   27-mar-92	Craig Eisler	NT version
   21-oct-92	Craig Eisler	ANSI function names
--------------------------------------------------------Version-11.0---------
*/

#include "variety.h"
#if defined( __OS2__ )
  #define INCL_WIN
  #include <wos2.h>
#endif
#include "win.h"

/*
 * _MovePageUp - move up a page
 */
void _MovePageUp( LPWDATA w )
{

    if( w->TopLineNumber == 1 )  return;
    w->CurrentLineNumber -= (w->height-1);
    w->TopLineNumber -= (w->height-1);
    if( (signed long) w->CurrentLineNumber < 1 ) {
	w->CurrentLineNumber = 1;
    }
    if( (signed long) w->TopLineNumber < 1 ) {
	w->TopLineNumber = 1;
    }
    _PositionScrollThumb( w );
    _DisplayAllLines( w, TRUE );

} /* _MovePageUp */

/*
 * _MovePageDown - move down a page
 */
void _MovePageDown( LPWDATA w )
{
    DWORD	ll;

    ll = _GetLastLineNumber( w );
    w->CurrentLineNumber += (w->height-1);
    w->TopLineNumber += (w->height-1);
    if( w->CurrentLineNumber > ll ) {
	w->CurrentLineNumber = ll;
	w->TopLineNumber = w->CurrentLineNumber - (w->height/2);
	if( (signed long) w->TopLineNumber < 1 ) w->TopLineNumber = 1;
    }
    _PositionScrollThumb( w );
    _DisplayAllLines( w, TRUE );

} /* _MovePageDown */

/*
 * _MoveLineUp - move up one line
 */
void _MoveLineUp( LPWDATA w )
{
    LPLDATA	lp;

    if( w->TopLineNumber != 1 ) {
	w->CurrentLineNumber--;
	w->TopLineNumber--;
	_ShiftWindow( w, -1 );
	lp = _GetLineDataPointer( w, w->TopLineNumber );
	_DisplayLineInWindow( w, 1, lp->data );
	_PositionScrollThumb( w );
    }

} /* _MoveLineUp */

/*
 * _MoveLineDown - move down one line
 */
void _MoveLineDown( LPWDATA w )
{
    LPLDATA	lp;
    DWORD	bot;
    DWORD	ll;

    ll = _GetLastLineNumber( w );
    if( w->CurrentLineNumber != ll ) {
	bot = w->TopLineNumber + w->height-1;
	if( bot < ll ) {
	    _ShiftWindow( w, 1 );
	    lp = _GetLineDataPointer( w, bot );
	    if( lp != NULL ) {
		_DisplayLineInWindow( w, w->height-1, lp->data );
	    }
	    w->CurrentLineNumber++;
	    w->TopLineNumber++;
	    _PositionScrollThumb( w );
	}
    }

} /* _MoveLineDown */

/*
 * _MoveToLine - move to a line, given a thumb position
 */
void _MoveToLine( LPWDATA w, DWORD line, BOOL center )
{
    int		height;
    DWORD	bottom;
    DWORD	ll;

    height = w->height-2;
    bottom = w->TopLineNumber + height;
    ll = _GetLastLineNumber( w );
    if( (line >= w->TopLineNumber) && (line <= bottom) && !center ) return;
    if( center || bottom < ll ) {
	w->TopLineNumber = line - height/2;
    } else {
	if( bottom > ll ) bottom = ll;
	if( bottom > line ) {
	    w->TopLineNumber = line - height;
	}
    }
    if( (signed long) w->TopLineNumber < 1L ) w->TopLineNumber = 1;
    w->CurrentLineNumber = line;
    _PositionScrollThumb( w );
    _DisplayAllLines( w, TRUE );

} /* _MoveToLine */
