/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  10-aug-89	John Dahms	defined for OS/2
  15-feb-90	F.W.Crigger	Use date/time of LastWrite vs Creation
  06-apr-90	F.W.Crigger	Added _opendir with second parm for attribute
  12-apr-90	D.J.Gaudet	Get new handle from DosFindFirst, and use
				DosFindClose to allow multiple opendir()s
				Call _set_errno when error occurs
  16-apr-90	D.J.Gaudet	When in real mode use interrupts because the
				DosFind functions don't work the same
  05-apr-91	F.W.Crigger	_opendir - POSIX conformance, i.e if name is
				just a directory name, then append "\*.*"
				Allow * and ? in name for past compatibility
--------------------------------------------------Version 8.5----------
  09-jul-91	G. Coschi	support OS/2 2.0
  12-dec-91	A.F.Scian	fixed POSIX second call to __opendir to use
				pathname instead of dirname
--------------------------------------------------Version 9.0----------
  03-jul-92	D.J.Gaudet	HPFS filename support
  19-oct-94	Greg Bentz	must call DosFindClose for every call to
				DosFindFirst
  01-feb-95	M.J.Carmody	trying to fix foo\ and .. kind of problems
--------------------------------------------------------Version-11.0---------
  27-feb-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  22-may-95	Greg Bentz	variety.h and dll linkage
  20-oct-95	M. Hildebrand	Added readdir and _wreaddir.
  18-mar-96	Greg Bentz	handle "*."
*/

#include "variety.h"
#include "widechar.h"
#include <errno.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <sys\types.h>
#include <direct.h>
#include <dos.h>
#define INCL_ERRORS
#include <wos2.h>
#include "tinyio.h"
#include "rtdata.h"
#include "strdup.h"
#include "seterrno.h"
#ifdef __WIDECHAR__
    #include <mbstring.h>
#endif

#if defined(__WARP__)
  #define FF_LEVEL	1
  #define FF_BUFFER	FILEFINDBUF3
#else
  #define FF_LEVEL	0
  #define FF_BUFFER	FILEFINDBUF
#endif

#define _DIR_ISFIRST		0
#define _DIR_NOTFIRST		1
#define _DIR_MAX_FOR_CLOSE_OK	2	/* dummy value used by closedir */
#define _DIR_CLOSED		3

#define SEEK_ATTRIB (_A_SUBDIR | _A_HIDDEN | _A_SYSTEM | _A_RDONLY | _A_ARCH)

/* Since dptr->d_dta is not used under OS/2, we will store the handle for
 * this directory scan in it.  This macro simply does the appropriate casting
 * for us. DJG
 */
#define HANDLE_OF(dirp) 	( *( HDIR * )( &( dirp )->d_dta[ 0 ] ) )

/* we'll use this to copy from a FILEFINDBUF to a DIR in copydir() */
struct name {
    char buf[ NAME_MAX + 1 ];
};

static	void		copydir( DIR_TYPE *dirp, FF_BUFFER *dir_buff ) {
/*****************************************************************/

#ifdef __WIDECHAR__
    wchar_t		wcs[_MAX_PATH];
#endif

    /*** Copy the structure data ***/
    dirp->d_attr = dir_buff->attrFile;
    dirp->d_time = *(unsigned short *)&dir_buff->ftimeLastWrite;
    dirp->d_date = *(unsigned short *)&dir_buff->fdateLastWrite;
    dirp->d_size = dir_buff->cbFile;
    *(struct name *)dirp->d_name = *(struct name *)dir_buff->achName;

#ifdef __WIDECHAR__
    mbstowcs( wcs, (char*)dirp->d_name, _MAX_PATH );	/* convert string */
    wcscpy( dirp->d_name, wcs );			/* copy string */
#endif
}


static	int	is_directory( const CHAR_TYPE *name ) {
/*****************************************************/
    if( name[0] == '\0' )return( 0 );
    while( name[1] != '\0' ){
	if( name[0] == '*' || name[0] == '?' ) {
	    return( 0 );
	}
	++name;
    }
    if( name[0] == '\\' || name[0] == '/' || name[0] == '.' ){
	return( 1 );
    }
    return( 0 );
}


_WCRTLINK DIR_TYPE	*__F_NAME(__opendir,_w__opendir)( const CHAR_TYPE *dirname,
		int attr, DIR_TYPE *dirp ) {
/*************************************************************************/

    USHORT	    rc;
    FF_BUFFER	    dir_buff;
    HFILE	    handle;
    OS_UINT	    searchcount;
    
    /*** Convert a wide char string to a multibyte string ***/
    #ifdef __WIDECHAR__    
	char	    mbcsName[MB_CUR_MAX*_MAX_PATH];

	if( wcstombs( mbcsName, dirname, MB_CUR_MAX*(wcslen(dirname)+1) ) == (size_t)-1 )
	    return( NULL );
    #endif    

#if defined(__OS2_286__)
    if( _RWD_osmode == OS2_MODE ) {
#endif
	handle = ~0;		/* we want our own handle */
	searchcount = 1;	/* only one at a time */
	#ifndef __WIDECHAR__
	    rc = DosFindFirst( (PSZ)dirname, &handle, attr, (PVOID)&dir_buff,
			sizeof( dir_buff ), &searchcount, FF_LEVEL );
	#else
	    rc = DosFindFirst( (PSZ)mbcsName, &handle, attr, (PVOID)&dir_buff,
			sizeof( dir_buff ), &searchcount, FF_LEVEL );
	#endif

	if( rc != 0 ) {
	    __set_errno_dos( rc );
	    return( NULL );
	}
	if( searchcount != 1 ) {
	    DosFindClose( handle );
	    __set_errno_dos( ERROR_NO_MORE_FILES );
	    return( NULL );
	}
	HANDLE_OF( dirp ) = handle;	/* store our handle	*/
	copydir( dirp, &dir_buff );	/* copy in other fields */

#if defined(__OS2_286__)
    } else {		/* real mode */
	DIR_TYPE	    buf;

	TinySetDTA( buf.d_dta );    /* set our DTA */
	#ifndef __WIDECHAR__
	    rc = TinyFindFirst( dirname, attr );
	#else
	    rc = TinyFindFirst( mbcsName, attr );
	#endif
	if( rc > 0 ) {
	    __set_errno_dos( rc );
	    return( NULL );
	}
	*dirp = buf;			/* copy to new memory */
    }
#endif

    dirp->d_first = _DIR_ISFIRST;	/* indicate we have 1st name */
    return( dirp );
}


_WCRTLINK DIR_TYPE	*__F_NAME(_opendir,_w_opendir)( const CHAR_TYPE *dirname, int attr ) {
/*************************************************************************************/

    DIR_TYPE 	    *dirp;
    int 	    i;
    auto CHAR_TYPE  pathname[_MAX_PATH+6];
    const CHAR_TYPE *    dirnameStart = dirname;

    dirp = malloc( sizeof( DIR_TYPE ) );
    HANDLE_OF( dirp ) = 0;		    /* initialize handle    */
    if( dirp == NULL ) {
	__set_errno_dos( ERROR_NOT_ENOUGH_MEMORY );
	return( NULL );
    }
    if( ! is_directory( dirname ) ) {
	if( __F_NAME(__opendir,_w__opendir)( dirname, attr, dirp ) == NULL ) {
	    free( dirp );
	    return( NULL );
	}
    }else{
	dirp->d_attr = _A_SUBDIR;
    }
    if( dirp->d_attr & _A_SUBDIR ) {				/* 05-apr-91 */
	for( i = 0; i < _MAX_PATH; i++ ) {
	    pathname[i] = *dirname;
	    if( *dirname == '\0' ) {
		if( i != 0  &&	 pathname[i-1] != '\\' && pathname[i-1] != '/' ){
		    pathname[i++] = '\\';
		}
		#ifndef __WIDECHAR__
		    strcpy( &pathname[i], "*.*" );
		#else
		    wcscpy( &pathname[i], L"*.*" );
		#endif
		if( HANDLE_OF( dirp ) != 0 ) {
		    DosFindClose( HANDLE_OF(dirp) );
		}
		if( __F_NAME(__opendir,_w__opendir)( pathname, attr, dirp ) == NULL ) {
		    free( dirp );
		    return( NULL );
		}
		break;
	    }
	    if( *dirname == '*' ) break;
	    if( *dirname == '?' ) break;
	    ++dirname;
	}
    }
    if( dirp != NULL )  dirp->d_openpath = __F_NAME(__clib_strdup,__clib_wcsdup)( dirnameStart );
    return( dirp );
}


_WCRTLINK DIR_TYPE	*__F_NAME(opendir,_wopendir)( const CHAR_TYPE *dirname ) {
/*************************************************************************/

    return( __F_NAME(_opendir,_w_opendir)( dirname, SEEK_ATTRIB ) );
}


_WCRTLINK DIR_TYPE *__F_NAME(readdir,_wreaddir)( DIR_TYPE *dirp ) {
/******************************************************/

    USHORT  rc;

    if( dirp == NULL ) {
	return( NULL );
    }
    if( dirp->d_first == _DIR_CLOSED )  return( NULL );
    if( dirp->d_first == _DIR_ISFIRST ) {	/* if we already have one */
	dirp->d_first = _DIR_NOTFIRST;
	return( dirp );
    }

#if defined(__OS2_286__)
    if( _RWD_osmode == OS2_MODE )	   /* protected mode */
#endif
	{

	FF_BUFFER	dir_buff;
	OS_UINT 	searchcount = 1;

	rc = DosFindNext( HANDLE_OF( dirp ), (PVOID)&dir_buff,
	    sizeof( dir_buff ), &searchcount );
	if( rc == ERROR_NO_MORE_FILES ) {
	    return( NULL );
	}
	if( rc != 0 ) {
	    __set_errno_dos( rc );
	    return( NULL );
	}

	if( searchcount != 1 ) {
	    __set_errno_dos( ERROR_NO_MORE_FILES );
	    return( NULL );
	}

	copydir( dirp, &dir_buff );

	}
#if defined(__OS2_286__)
    else {		/* real mode */
	TinySetDTA( dirp->d_dta );

	rc = TinyFindNext();
	if( TINY_ERROR( rc ) == 18 ) { // E_nomore files
	    return( NULL );
	} else if( TINY_ERROR( rc ) ) {
	    __set_errno_dos( TINY_INFO( rc ) );
	    return( NULL );
	}
    }
#endif
    return( dirp );
}


_WCRTLINK int	__F_NAME(closedir,_wclosedir)( DIR_TYPE *dirp ) {
/**********************************************************/

    if( dirp == NULL || dirp->d_first > _DIR_MAX_FOR_CLOSE_OK  ) {/* error */
	__set_errno_dos( ERROR_INVALID_HANDLE );
	return( -1 );
    }

#if defined(__OS2_286__)
    if( _RWD_osmode == OS2_MODE )
#endif
	DosFindClose( HANDLE_OF( dirp ) );

    dirp->d_first = _DIR_CLOSED;
    if( dirp->d_openpath != NULL )  free( dirp->d_openpath );
    free( dirp );

    return( 0 );			/* ok */
}


_WCRTLINK void __F_NAME(rewinddir,_wrewinddir)( DIR_TYPE *dirp )
/*********************************************************/
{
    CHAR_TYPE *		openpath;
    DIR_TYPE *		newDirp;

    /*** Get the name of the directory before closing it ***/
    if( dirp->d_openpath == NULL )  return;	/* can't continue if NULL */
    openpath = __F_NAME(__clib_strdup,__clib_wcsdup)( dirp->d_openpath ); /* store path */
    if( openpath == NULL ) {
	dirp->d_first = _DIR_CLOSED;	/* so reads won't work any more */
	return;
    }

    /*** Reopen the directory ***/
    __F_NAME(closedir,_wclosedir)( dirp );		/* close directory */
    newDirp = __F_NAME(opendir,_wopendir)( openpath );  /* open it again */
    if( newDirp == NULL ) {
	dirp->d_first = _DIR_CLOSED;	/* so reads won't work any more */
	return;
    }

    /*** Clean up and go home ***/
    free( openpath );			    /* don't need this any more */
    memcpy( dirp, newDirp, sizeof(DIR_TYPE) );   /* copy into user buffer */
}
