/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  21-jul-87	H.W.Hillis	Initial implementation
  25-nov-87	M.Glaum 	Added big data code
  03-dec-87	F.W.Crigger	Do extra checking in closedir.
  16-jan-89	F.W.Crigger	Look for SYSTEM and HIDDEN files as well
------------------------------------------------------------Version 7.0------
  15-mar-89	F.W.Crigger	Adapted for 386
  11-oct-89	F.W.Crigger	set sregs.es for running in protect mode
  06-apr-90	F.W.Crigger	Added _opendir with second parm for attribute
------------------------------------------------------------Version 8.0------
  02-jan-91	A.F.Scian	removed use of intdosx() functions
				changed to use TinyIO
  05-apr-91	F.W.Crigger	_opendir - POSIX conformance, i.e if name is
				just a directory name, then append "\*.*"
				Allow * and ? in name for past compatibility
------------------------------------------------------------Version 9.0------
  03-jul-92	D.J.Gaudet	handle dirname[0] == 0 in root_directory()
  18-jun-93	John Dahms	all library allocations use lib_xxx routines
  19-feb-94	F.W.Crigger	#define __OS2__ so we get 256 byte name in
				directory structure
  01-feb-95	M.J.Carmody	trying to fix foo\ and .. kind of problems
--------------------------------------------------------Version-11.0---------
  22-may-95	Greg Bentz	variety.h and dll linkage
  18-mar-96	Greg Bentz	handle "*."
*/
#include "variety.h"
#include "widechar.h"
#define __OS2__
#include <stdlib.h>
#include <string.h>
#include <direct.h>
#include "liballoc.h"
#include "tinyio.h"
#include "seterrno.h"

#define SEEK_ATTRIB (TIO_HIDDEN | TIO_SYSTEM | TIO_SUBDIRECTORY)

static	int	is_directory( const char *name ) {
/**************************************************/
    if( name[0] == '\0' )return( 0 );
    while( name[1] != '\0' ){
	if( name[0] == '*' || name[0] == '?' ) {
	    return( 0 );
	}
	++name;
    }
    if( name[0] == '\\' || name[0] == '/' || name[0] == '.' ){
	return( 1 );
    }
    return( 0 );
}

_WCRTLINK DIR_TYPE *_opendir( const char *name, char attr )
{
    DIR_TYPE 	*parent;
    int 	i;
    tiny_ret_t	rc;
    auto char	pathname[_MAX_PATH+6];

    parent = lib_malloc( sizeof( *parent ) );
    if( parent == NULL ) {
	return( parent );
    }
    if( ! is_directory( name ) ) {
	rc = TinyFindFirstDTA( name, attr, &parent->d_dta );
	if( TINY_ERROR( rc ) ) {
	    __set_errno_dos( TINY_INFO( rc ) );
	    lib_free( parent );
	    return( NULL );
	}
    } else {
	parent->d_attr = _A_SUBDIR;
    }
    if( parent->d_attr & _A_SUBDIR ) {			/* 05-apr-91 */
	for( i = 0; i < _MAX_PATH; i++ ) {
	    pathname[i] = *name;
	    if( *name == '\0' ) {
		if( i != 0  &&	pathname[i-1] != '\\' && pathname[i-1] != '/' ){
		    pathname[i++] = '\\';
		}
		strcpy( &pathname[i], "*.*" );
		rc = TinyFindFirstDTA( pathname, attr, &parent->d_dta );
		if( TINY_ERROR( rc ) ) {
		    __set_errno_dos( TINY_INFO( rc ) );
		    lib_free( parent );
		    return( NULL );
		}
		break;
	    }
	    if( *name == '*' ) break;
	    if( *name == '?' ) break;
	    ++name;
	}
    }
    parent->d_first = 1;
    return( parent );
}


_WCRTLINK DIR_TYPE *opendir( const char *name )
{
    return( _opendir( name, SEEK_ATTRIB ) );
}


_WCRTLINK DIR_TYPE *readdir( DIR_TYPE *parent )
{
    tiny_ret_t rc;

    if( parent == NULL ) {
	return( NULL );
    }
    if( parent->d_first ) {
	parent->d_first = 0;
	return( parent );
    }
    rc = TinyFindNextDTA( &parent->d_dta );
    if( TINY_ERROR( rc ) == 18 ) { // E_nomore files
	return( NULL );
    } else if( TINY_ERROR( rc ) ) {
	__set_errno_dos( TINY_INFO( rc ) );
	return( NULL );
    }
    return( parent );
}


_WCRTLINK int closedir( DIR_TYPE *dirp )
{
    if( dirp == NULL ) {
	return( 1 );
    }
    if( dirp->d_first > 1 ) {
	return( 1 );
    }
    TinyFindCloseDTA( &dirp->d_dta );
    dirp->d_first = 2;
    lib_free( dirp );
    return( 0 );
}
