/*
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %       Copyright (C) 1992, by WATCOM Systems Inc. All rights     %
  %       reserved. No part of this software may be reproduced      %
  %       in any form or by any means - graphic, electronic or      %
  %       mechanical, including photocopying, recording, taping     %
  %       or information storage and retrieval systems - except     %
  %       with the written permission of WATCOM Systems Inc.        %
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  06-mar-92     Craig Eisler	defined for NT
  28-jan-94	Greg Bentz	added getcmd2
  25-may-94	Greg Bentz	change getcmd2 to _bgetcmd
--------------------------------------------------------Version-11.0---------
  22-may-95	Greg Bentz	variety.h and dll linkage
  19-sep-95	T. Schiller	merged NT, 16/32-bit OS/2, Netware, OSI and
				Win386 versions into one file.
  19-oct-95	Greg Bentz	allow for embedded whitespace in WIN32 .exe
  				file name
  13-dec-95	M. Hildebrand	Made OS_GET_CMD_LINE static.
*/
#include "variety.h"
#include <limits.h>
#include <ctype.h>
#include <string.h>
#if defined(__NT__)
    #include <windows.h>
#elif defined(__OS2__)
    #define INCL_DOSMISC
    #include <wos2.h>
#endif

#ifdef __NT__

    static char *OS_GET_CMD_LINE(void)
    {
	char *cmd = GetCommandLine();
	if( *cmd == '"' ) {
	    cmd++;
	    while (*cmd && *cmd != '"') {
		cmd++;
	    }
	    if( *cmd ) cmd++;
	} else {
	    while (*cmd && !isspace(*cmd)) {
		++cmd;
	    }
	}
	return cmd;
    }

#elif defined(__NETWARE__) || defined(__OSI__)

    extern char *_LpCmdLine;
    #define OS_GET_CMD_LINE()    _LpCmdLine

#elif defined(__WIN386__)

    extern char _WCFAR *_wincmdptr;
    #define OS_GET_CMD_LINE()    _wincmdptr

#elif defined(__OS2_286__)

    #include <dos.h>
    static char _WCFAR *OS_GET_CMD_LINE(void)
    {
	unsigned short  envseg, cmdoff;
	char _WCFAR      *cmd;

	if (DosGetEnv( &envseg, &cmdoff))
	    return NULL;

	cmd = MK_FP(envseg, 0);
	if (!*cmd) ++cmd;	/* adjust if null environment space */
	while (*cmd) { 	/* skip over environment space */
	    while(*cmd) ++cmd;
	    ++cmd;
	}
	++cmd;
	while(*cmd) ++cmd;/* skip over first copy of program name */
	++cmd;

	while( *cmd ) ++cmd;	/* skip over second copy of program name */
	++cmd;

	return cmd;
    }

#elif defined(__WARP__)

    static char *OS_GET_CMD_LINE(void)
    {
	PTIB  tib;
	PPIB  pib;
	char *cmd;

	if (DosGetInfoBlocks(&tib, &pib))
	    return NULL;

	cmd = pib->pib_pchcmd;
	while (*cmd) ++cmd;	/* skip over second copy of program name */
	++cmd;
	return cmd;
    }

#else

    #error *** This Module Does Not Support That Platform ***

#endif

_WCRTLINK int _bgetcmd( char *buffer, int len )
{
    int  cmdlen;
#if defined(__WIN386__) || defined(__OS2_286__)
    char _WCFAR *cmd;
    char _WCFAR *tmp;
#else
    char *cmd;
    char *tmp;
#endif

    if (buffer && (len > 0))
	*buffer = 0x00;

    cmd = OS_GET_CMD_LINE();
    if (!cmd)
	return 0;

    while (isspace(*cmd))
	++cmd;

    for (cmdlen = 0, tmp = cmd; *tmp; ++tmp, ++cmdlen)
	;

    if (!buffer || (len <= 0))
	return cmdlen;

    len--;
    len = (len < cmdlen) ? len : cmdlen;

    while (len)
    {
	*buffer++ = *cmd++;
	--len;
    }
    //strncpy(buffer, cmd, len);
    buffer[len] = 0x00;

    return cmdlen;
} /* _bgetcmd() */


_WCRTLINK char *getcmd( char *buffer )
{
    _bgetcmd (buffer, INT_MAX);
    return (buffer);
} /* getcmd() */

