/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %     Copyright (C) 1995, by Watcom International Inc.  All rights    %
   %     reserved.  No part of this software may be reproduced or        %
   %     used in any form or by any means - graphic, electronic or       %
   %     mechanical, including photocopying, recording, taping or        %
   %     information storage and retrieval systems - except with the     %
   %     written permission of Watcom International Inc.                 %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   Date		By		Reason
   ----		--		------
--------------------------------------------------------Version-11.0---------
   14-dec-95	M. Hildebrand	Initial implementation.
   20-dec-95	M. Hildebrand	Mapping like baseXXXXXXXX --> baseXXa12345.
   10-jun-96	M. Hildebrand	Fixed problem with large PIDs.
*/

#include "variety.h"
#include "widechar.h"
#include <io.h>
#include <malloc.h>
#include <mbstring.h>
#include <process.h>
#include <stdio.h>
#include <string.h>
#ifdef __NT__
    #include <windows.h>
#endif


_WCRTLINK CHAR_TYPE *__F_NAME(_mktemp,_wmktemp)( CHAR_TYPE *template )
{
    CHAR_TYPE		letter;
    unsigned		pid;
    CHAR_TYPE *		xs;

    /*** Ensure the template is valid ***/
    if( !is_valid_template( template, &xs ) ) {
	return( NULL );
    }

    /*** Get the process/thread ID ***/
    #ifdef __NT__
	#ifdef __SW_BM
	    pid = GetCurrentThreadId();	    /* thread ID for multi-thread */
	#else
	    pid = GetCurrentProcessId();    /* process ID for single-thread */
	#endif
    #else
	pid = getpid();
    #endif
    pid %= 100000;	/* first few digits tend to be repeated under 95 */

    /*** Try to build a unique filename ***/
    for( letter='a'; letter<='z'; letter++ ) {
	#ifdef __WIDECHAR__
	    _bwprintf( xs, wcslen(xs) + 1, L"%c%05u", letter, pid );
	#else
	    _bprintf( xs, strlen(xs) + 1, "%c%05u", letter, pid );
	#endif
	if( __F_NAME(access,_waccess)( template, F_OK )  !=  0 ) {
	    return( template );
	}
    }
    return( NULL );
}


static int is_valid_template( CHAR_TYPE *template, CHAR_TYPE **xs )
{
    int			len;
    CHAR_TYPE *		p;

    /*** Ensure the last 6 characters form the string "XXXXXX" ***/
    len = __F_NAME(_mbslen,wcslen)( template );
    if( len < 6 )  return( 0 );		/* need room for "XXXXXX" */
    #ifdef __WIDECHAR__
	p = template + len - 6;
	if( wcscmp( p, L"XXXXXX" ) )  return( 0 );
    #else
	p = _mbsninc( template, len-6 );
	if( _mbscmp( p, "XXXXXX" ) )  return( 0 );
    #endif
    *xs = p;

    return( 1 );
}
