/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  20-mar-87	F.W.Crigger	Initial implementation
---------------------------------------------------------Version 6.5------
  08-jul-88	F.W.Crigger	Make sure file does not exist
  16-oct-90	J.B.Schueler	Preserve errno value across this function
  19-apr-91	B.J. Stecher	stuck in pid (and nid on QNX) for uniqueness
  12-jul-91	J.B.Schueler	getpid defined in unistd.h for PenPoint
  16-Jun-92	D. Wellstood	Unicode support
  26-oct-92	J.B.Schueler	added "const" keyword to tmp_ext
  16-dec-92	B.J. Stecher	use __tmpdir function for QNX
  11-feb-93	A. Gupta	defined getpid() as GetThreadID() for
  				Netware since GetThreadID() returns a pointer
				to a thread struct (and will be unique)
  18-jun-93	John Dahms  	all library allocations use lib_xxx routines
  11-nov-93	Greg Bentz	rationalize exit_with_msg
  28-jan-94	Greg Bentz	change exit_with_msg to __fatal_runtime_error
  27-feb-95	Greg Bentz	modify to suport big PIDs and merge with QNX
--------------------------------------------------------Version-11.0---------
  02-mar-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  22-may-95	Greg Bentz	variety.h and dll linkage
  21-sep-95	M. Hildebrand	Added _u version, which calls wide version.
  03-jan-96	Greg Bentz	remove stubs that are now aliases
*/

#include "variety.h"
#include "widechar.h"
#ifndef __QNX__
    #include <io.h>
#endif
#include <errno.h>
#include <process.h>
#include <stdio.h>
#include <string.h>
#ifdef __QNX__
    #include <unistd.h>
#endif
#include "rtdata.h"
#include "exitwmsg.h"
#include "liballoc.h"
#include "seterrno.h"

/*
    U's     are unique filename letters for the process
    P's     are the pid
    N's     are the nid
    .tmp    is a literal string

    On QNX, the name is of the form (only use the 16 low order bits of the nid):

	_PPPPPPP.NNNN.UUU

    On other systems, the name looks like:

	_PPPPPPP.UUU
*/

#if defined(__PENPOINT__)
  void __alloctmpnambuf( void );
#elif !defined(__NETWARE__)
  static CHAR_TYPE _tmpname[L_tmpnam];
#endif

#if defined(__NETWARE__)
 #include "thread.h"
 #define getpid()	GetThreadID()
 extern int		GetThreadID( void );
#elif defined(__QNX__)
 extern char *__tmpdir( char * );
#endif

#if defined(__PENPOINT__)
static void __alloctmpnambuf()
{
    if( _RWD_tmpnambuf != NULL ) return;

    _RWD_tmpnambuf = lib_malloc( L_tmpnam * sizeof( wchar_t ) );
    if( _RWD_tmpnambuf == NULL ) {
	__fatal_runtime_error(
		"Not enough memory to allocate tmpnam buffer\r\n", 1 );
    }
    _RWD_tmpnambuf[0] = NULLCHAR;
}
#endif

static CHAR_TYPE *__F_NAME(putbits,_wputbits)( CHAR_TYPE *p, unsigned val )
{
    if( val > 0x1f ) {
	p = __F_NAME(putbits,_wputbits)( p, val >> 5 );
	val &= 0x1f;
    }
    if( val >= 10 ) {
	*p = val - 10 + 'A';
    } else {
	*p = val + '0';
    }
    return( p + 1 );
}

static int init_name()
{
    CHAR_TYPE	*p;

    p = (CHAR_TYPE *)_RWD_tmpnambuf;
    #if defined(__QNX__)
	p = __tmpdir( p );
    #endif
    *p++ = '_';
    p = __F_NAME(putbits,_wputbits)( p, getpid() );
    #if defined(__QNX__)
	*p++ = '.';
	p = __F_NAME(putbits,_wputbits)( p, (unsigned short)getnid() );
    #endif
    *p++ = '.';
    *p++ = 'A';
    *p++ = 'A';
    *p   = 'A' - 1;
    return( p - (CHAR_TYPE *)_RWD_tmpnambuf );
}


_WCRTLINK CHAR_TYPE *__F_NAME(tmpnam,_wtmpnam)( CHAR_TYPE *buf )
{
    int i, err;

    err = _RWD_errno;
    #if defined(__PENPOINT__)
	__alloctmpnambuf();
    #endif
    if( _RWD_tmpnambuf[0] == NULLCHAR ) {
	i = init_name();
    } else {
	i = __F_NAME(strlen,wcslen)( _RWD_tmpnambuf ) - 1;
    }
    for(;;) {
	while( _RWD_tmpnambuf[i] == 'Z' ) {
	    _RWD_tmpnambuf[i] = 'A';
	    --i;
	    if( _RWD_tmpnambuf[i] == '.' ) {
		i = init_name();
	    }
	}
	_RWD_tmpnambuf[i]++;
	if( __F_NAME(access,_waccess)( (CHAR_TYPE *)_RWD_tmpnambuf, 0 ) != 0 ) break;
    }
    if( buf != NULL ) {
	__F_NAME(strcpy,wcscpy)( buf, (CHAR_TYPE *)_RWD_tmpnambuf );
    } else {
	buf = (CHAR_TYPE *)_RWD_tmpnambuf;
    }
    __set_errno( err );
    return( buf );
}
