/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1995, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  02-oct-95	T. Schiller	Initial implementation
*/
#include "variety.h"
#include <math.h>
#include <float.h>
#include "rtdata.h"

#if defined(__AXP__)

/*
 * FPCR Trap Disable Flags
 */
#define FPCR_INEXACT	    0x40000000
#define FPCR_UNDERFLOW	    0x20000000
#define FPCR_OVERFLOW	    0x00080000
#define FPCR_ZERODIVIDE	    0x00040000
#define FPCR_INVALID	    0x00020000

extern unsigned long _GetFPCR(void);
extern void          _SetFPCR(unsigned long);

static unsigned int MapToCW(unsigned long fpcr)
{
    unsigned int cw;

    /*
     * The rounding bits are identical but in the highword of the fpcr.
     */
    cw = (fpcr >> 16) & ~(_MCW_RC);

    if (fpcr & FPCR_INEXACT)
	cw &= ~_EM_INEXACT;

    if (fpcr & FPCR_ZERODIVIDE)
	cw &= ~_EM_ZERODIVIDE;

    if (fpcr & FPCR_OVERFLOW)
	cw &= ~_EM_OVERFLOW;

    if (fpcr & FPCR_UNDERFLOW)
	cw &= ~_EM_UNDERFLOW;

    if (fpcr & FPCR_INVALID)
	cw &= ~_EM_INVALID;

    return cw;
} /* MapToCW() */


static unsigned long MapFromCW(unsigned int cw)
{
    unsigned long fpcr = 0L;

    /*
     * The rounding bits are identical but in the highword of the fpcr.
     */
    fpcr = (cw & ~_MCW_RC) << 16;

    if (!(cw & _EM_INEXACT))
	fpcr |= FPCR_INEXACT;

    if (!(cw & _EM_INVALID))
	fpcr |= FPCR_INVALID;

    if (!(cw & _EM_ZERODIVIDE))
	fpcr |= FPCR_ZERODIVIDE;

    if (!(cw & _EM_OVERFLOW))
	fpcr |= FPCR_OVERFLOW;

    if (!(cw & _EM_UNDERFLOW))
	fpcr |= FPCR_UNDERFLOW;

    return fpcr;
} /* MapFromCW() */
#endif


_WCRTLINK unsigned _controlfp(unsigned new, unsigned mask)
{
#if defined(_M_IX86)
    return _controlfp(new, mask);
#elif defined(__AXP__)
    unsigned int  cw;

    cw = MapToCW(_GetFPCR());

    if (mask)
    {
	cw = (cw & ~mask) | (new & mask);
	_SetFPCR(MapFromCW(cw));
    }

    return cw;
#elif defined(__PPC__)
    // No idea yet
#endif
} /* _controlfp() */
