/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1989, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  09-aug-89     Johh Dahms      defined for OS/2
--------------------------------------------------------Version-11.0---------
  06-jan-95	M.P.Hildebrand	Increasing file size now pads new space with
  				zero bytes.  Also, added logic for setting
				errno if necessary.
  28-may-95	Greg Bentz	variety.h and dll linkage
  26-jul-96	M. Hildebrand	Use rtcheck.h support.
*/
#include "variety.h"
#include <errno.h>
#include <io.h>
#include <stdio.h>
#include <string.h>
#include <wos2.h>
#include "fileacc.h"
#include "rtcheck.h"
#include "seterrno.h"


_WCRTLINK int chsize( int handle, long size )
{
    long	curOffset, endOffset;
    char	buf[512];
    long	diff;
    unsigned	amount;
    int		retCode;

    __handle_check( handle, -1 );

    /*** Prepare to change the size ***/
    _AccessFileH( handle );
    curOffset = lseek( handle, 0L, SEEK_CUR );	/* get current offset */
    if( curOffset == -1 ) {
	_ReleaseFileH( handle );
    	return( -1 );
    }
    endOffset = lseek( handle, 0L, SEEK_END );	/* get file size */
    if( endOffset == -1 ) {
	_ReleaseFileH( handle );
    	return( -1 );
    }

    if( size > endOffset ) {
	/*** Increase file size ***/
    	diff = size - endOffset;		/* number of bytes to pad */
    	memset( buf, 0, 512 );			/* zero buffer */
    	do {
	    if( diff >= 512 ) {
	    	amount = 512;
	    } else {
	    	amount = (unsigned)diff;
	    }
	    retCode = write( handle, buf, amount );
	    if( retCode != amount ) {
	        __set_errno( ENOSPC );
	        retCode = -1;
		break;
	    }
	    diff -= amount;
	} while( diff != 0 );
	if( retCode != -1 ) retCode = 0;
    } else {
	/*** Shrink file ***/
	retCode = DosNewSize( handle, size );
	if( retCode != 0 ) {
	    __set_errno_dos( retCode );
	    retCode = -1;
	}
	if( curOffset > size ) curOffset = size;
    }

    /*** Clean up and go home ***/
    curOffset = lseek( handle, curOffset, SEEK_SET );
    if( curOffset == -1 ) retCode = -1;
    _ReleaseFileH( handle );
    return( retCode );
}
