/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1987 by WATCOM Systems Inc. All rights      %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  01-nov-87     F.W.Crigger     Initial implementation, adapted from stat.c
  14-dec-90     John Dahms      adapted for the new tinyio.h
  24-may-91	A.F.Scian	added code to check for R/W handle attributes
  				using the _iomode array (MSC does this)
  22-nov-92	F.W.Crigger	change reference to _iomode array
  08-mar-93	A. Gupta	changed our stat struct (in stat.h)
  03-mar-94	Greg Bentz	add _fstat()
  31-aug-94	Alex Brodsky	Fill in additional information in stat struct
--------------------------------------------------------Version-11.0---------
  28-may-95	Greg Bentz	variety.h and dll linkage
  24-nov-95	M. Hildebrand	Added _wfstat.
  03-jan-96	Greg Bentz	remove stubs that are now aliases
  26-jul-96	M. Hildebrand	Use rtcheck.h support.
*/
#include "variety.h"
#include <stdio.h>
#include <stddef.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <errno.h>
#include "tinyio.h"
#include "iomode.h"
#include "rtcheck.h"
#include "seterrno.h"
#include "widechar.h"


extern time_t _d2ttime();


#ifdef __WIDECHAR__
_WCRTLINK int _wfstat( int handle, struct _wstat *buf )
#else
_WCRTLINK int fstat( int handle, struct stat *buf )
#endif
    {
	unsigned	iomode_flags;
        union {
            tiny_ret_t rc;
            tiny_file_stamp_t stamp;
        } u;

	__handle_check( handle, -1 );

        u.rc = TinyGetDeviceInfo( handle );
        if( TINY_ERROR(u.rc) ) {
            return( __set_errno_dos( TINY_INFO(u.rc) ) );
        }
        /* isolate drive number */
        buf->st_dev = TINY_INFO(u.rc) & TIO_CTL_DISK_DRIVE_MASK;
        buf->st_rdev = buf->st_dev;

        buf->st_nlink = 1;
        buf->st_uid = buf->st_gid = 0;
        buf->st_ino = handle;
	buf->st_mode = 0;
	iomode_flags = __IOMode( handle );
        if( iomode_flags & _READ ) {
	    buf->st_mode |= S_IRUSR | S_IRGRP | S_IROTH;
	}
        if( iomode_flags & _WRITE ) {
            buf->st_mode |= S_IWUSR | S_IWGRP | S_IWOTH;
        }
        if( TINY_INFO(u.rc) & TIO_CTL_DEVICE ) {
            buf->st_size = 0;
            buf->st_atime = buf->st_ctime = buf->st_mtime = 0;
            buf->st_mode |= S_IFCHR;
        } else {                /* file */
            u.rc = TinyGetFileStamp( handle );
            if( TINY_ERROR(u.rc) ) {
                return( __set_errno_dos( TINY_INFO(u.rc) ) );
            }
            buf->st_mtime = _d2ttime( u.stamp.date, u.stamp.time );
            buf->st_atime = buf->st_ctime = buf->st_mtime;
            buf->st_size = filelength( handle );
            buf->st_mode |= S_IFREG;
        }
	buf->st_btime = buf->st_mtime;
    	buf->st_attr = 0;
    	buf->st_archivedID = 0;
    	buf->st_updatedID = 0;
    	buf->st_inheritedRightsMask = 0;
    	buf->st_originatingNameSpace = 0;
    	buf->st_name[0] = '\0';
        return( 0 );
    }

