/*
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %       Copyright (C) 1992, by WATCOM Systems Inc. All rights     %
  %       reserved. No part of this software may be reproduced      %
  %       in any form or by any means - graphic, electronic or      %
  %       mechanical, including photocopying, recording, taping     %
  %       or information storage and retrieval systems - except     %
  %       with the written permission of WATCOM Systems Inc.        %
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  05-mar-92     Craig Eisler	defined for NT
  12-sep-92     G. Coschi	check for character mode files (i.e CON)
  15-oct-92	Craig Eisler	use iomode array indices as handles
  22-nov-92	F.W.Crigger	change reference to _iomode array
  02-dec-92     G. Coschi	check for pipes
  08-mar-93	A. Gupta	changed our stat struct (in stat.h)
  30-sep-93     G. Coschi	don't call GetFileSize() for unknown file
  				types
  08-oct-93	A.F.Scian	fixed to call __MakeDOSDT so that times
  				in struct are local rather than GMT
  03-mar-94	Greg Bentz	add _fstat()
  12-apr-94	F.W.Crigger	buf->st_mode wasn't initialized !!!
  31-aug-94	Alex Brodsky	Fill in additional information in stat struct
  08-dec-94	A.F.Scian	optimized setting of st_atime and st_ctime
  				fields for FAT file systems (all times are
				the same)
--------------------------------------------------------Version-11.0---------
  28-may-95	Greg Bentz	variety.h and dll linkage
  24-nov-95	M. Hildebrand	Added _wfstat.
  03-jan-96	Greg Bentz	remove stubs that are now aliases
  04-jul-96	M. Hildebrand	Added __int64 support.
  23-jul-96	M. Hildebrand	Fixed handling of st_attr and st_mode.
  25-jul-96	M. Hildebrand	Fixed other st_mode problems.
  26-jul-96	M. Hildebrand	Use rtcheck.h support.
*/
#include "variety.h"
#include "int64.h"
#include "widechar.h"
#include <windows.h>
/* gross hack for building 11.0 libraries with 10.6 compiler */
#ifndef __WATCOM_INT64__
    #include <limits.h>		/* a gross hack to make a gross hack work */
    #define __WATCOM_INT64__
    #define __int64		double
#endif
/* most includes should go after this line */
#include <stdio.h>
#include <stddef.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <errno.h>
#include <direct.h>
#include "defwin.h"
#include "ntex.h"
#include "fileacc.h"
#include "iomode.h"
#include "rtcheck.h"
#include "seterrno.h"

extern time_t _d2ttime();


#ifdef __WIDECHAR__
 #ifdef __INT64__
  _WCRTLINK int _wfstati64( int hid, struct _wstati64 *buf )
 #else
  _WCRTLINK int _wfstat( int hid, struct _wstat *buf )
 #endif
#else
 #ifdef __INT64__
  _WCRTLINK int _fstati64( int hid, struct _stati64 *buf )
 #else
  _WCRTLINK int fstat( int hid, struct stat *buf )
 #endif
#endif
{
    DWORD			size;
#ifdef __INT64__
    DWORD			highorder;
    INT_TYPE			tmp;
#endif
    WORD			d;
    WORD			t;
    WORD			md;
    WORD			mt;
    DWORD			ftype;
    FILETIME			ctime, atime, mtime;
    HANDLE			h;
    unsigned			iomode_flags;
    BOOL			osrc;
    BY_HANDLE_FILE_INFORMATION	fileinfo;
    
    __handle_check( hid, -1 );
    h = __FileHandleIDs[ hid ];
    _AccessFileH( hid );

    buf->st_mode = 0;				/* 12-apr-94 */
    #ifdef DEFAULT_WINDOWING
	if( _WindowsIsWindowedHandle != NULL ) {
	    if( _WindowsIsWindowedHandle( hid ) != NULL ) {
		buf->st_mode |= S_IFCHR;	/* it's a console */
	    }
	}
    #endif
    iomode_flags = __IOMode( hid );
    if( iomode_flags & _READ ) {
	buf->st_mode |= S_IRUSR | S_IRGRP | S_IROTH;
    }
    if( iomode_flags & _WRITE ) {
	buf->st_mode |= S_IWUSR | S_IWGRP | S_IWOTH;
    }
    ftype = GetFileType( h );
    if( (iomode_flags & _ISTTY) || /* for default windows */
        (ftype == FILE_TYPE_CHAR) ||
	(ftype == FILE_TYPE_PIPE) ||
	 /*
	    FILE_TYPE_UNKNOWN is returned when standard output handle
	    is specified and it's not a console application - also we
	    don't want to call GetFileSize()
	 */
	(ftype == FILE_TYPE_UNKNOWN) ) {
	#ifdef __INT64__
	    _clib_U32ToU64( 0L, tmp );
	    buf->st_size = GET_REALINT64(tmp);
	#else
	    buf->st_size = 0;
	#endif
	buf->st_atime = buf->st_ctime = buf->st_mtime = 0;
	buf->st_attr = 0;
	buf->st_mode |= S_IRUSR | S_IRGRP | S_IROTH;
	if( ftype == FILE_TYPE_PIPE ) {
	    buf->st_mode |= S_IFFIFO;
	} else if( ftype == FILE_TYPE_CHAR ) {
	    buf->st_mode |= S_IFCHR;
	}
	buf->st_dev = buf->st_rdev = 1;
    } else {
	/*** Try to get attributes (can reasonably fail; see Win32 docs) ***/
	osrc = GetFileInformationByHandle( h, &fileinfo );
	if( osrc == TRUE ) {
	    buf->st_attr = fileinfo.dwFileAttributes;
	    if( fileinfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
		buf->st_mode |= S_IFDIR;
	    } else {
		buf->st_mode |= S_IFREG;
	    }
	} else {
	    buf->st_attr = 0;		/* cannot be determined */
	}

	/*** Get the file size ***/
	if( buf->st_mode & S_IFDIR ) {
	    #ifdef __INT64__
		_clib_U32ToU64( 0L, tmp );
		buf->st_size = GET_REALINT64(tmp);
	    #else
		buf->st_size = 0;
	    #endif
	} else {
	    size = GetFileSize( h, __I64NAME(NULL,&highorder) );
	    if( size == -1 ) {
		_ReleaseFileH( hid );
		return( __set_errno_nt() );
	    }
	    #ifdef __INT64__
		MAKE_INT64(tmp,highorder,size);
		buf->st_size = GET_REALINT64(tmp);
	    #else
		buf->st_size = size;
	    #endif
	}

	/*** Get the file time ***/
	if( !GetFileTime( h, &ctime, &atime, &mtime ) ) {
	    _ReleaseFileH( hid );
	    return( __set_errno_nt() );
	}
	__MakeDOSDT( &mtime, &md, &mt );
	buf->st_mtime = _d2ttime( md, mt );
	if(( ctime.dwLowDateTime == mtime.dwLowDateTime ) &&
	   ( ctime.dwHighDateTime == mtime.dwHighDateTime )) {
	    buf->st_ctime = buf->st_mtime;
	} else {
	    __MakeDOSDT( &ctime, &d, &t );
	    if( d == md && t == mt ) {
		buf->st_ctime = buf->st_mtime;
	    } else {
		buf->st_ctime = _d2ttime( d, t );
	    }
	}
	if(( atime.dwLowDateTime == mtime.dwLowDateTime ) &&
	   ( atime.dwHighDateTime == mtime.dwHighDateTime )) {
	    buf->st_atime = buf->st_mtime;
	} else {
	    __MakeDOSDT( &atime, &d, &t );
	    if( d == md && t == mt ) {
		buf->st_atime = buf->st_mtime;
	    } else {
		buf->st_atime = _d2ttime( d, t );
	    }
	}

	buf->st_dev = buf->st_rdev = 0;
    }
    _ReleaseFileH( hid );

    buf->st_nlink = 1;
    buf->st_ino = hid;
    buf->st_uid = buf->st_gid = 0;

    buf->st_btime = buf->st_mtime;
    buf->st_archivedID = 0;
    buf->st_updatedID = 0;
    buf->st_inheritedRightsMask = 0;
    buf->st_originatingNameSpace = 0;
    buf->st_name[0] = '\0';
    return( 0 );
}
