/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %     Copyright (C) 1994, by WATCOM International Inc.  All rights    %
   %     reserved.  No part of this software may be reproduced or        %
   %     used in any form or by any means - graphic, electronic or       %
   %     mechanical, including photocopying, recording, taping or        %
   %     information storage and retrieval systems - except with the     %
   %     written permission of WATCOM International Inc.                 %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
   Date		By		Reason
   ----		--		------
   26-jan-94	Greg Bentz	defined
   06-apr-95	Greg Bentz	run286 doesn't support OS/2 API call
--------------------------------------------------------Version-11.0---------
  28-may-95	Greg Bentz	variety.h and dll linkage
  25-oct-95	M. Hildebrand	Use POSIX-level flag bits, not streamio bits.
  26-oct-95	M. Hildebrand	Added _open_osfhandle for MS compatability.
  26-jul-96	M. Hildebrand	Use rtcheck.h support.
*/
#include "variety.h"
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include "openmode.h"
#include "iomode.h"
#include "rtcheck.h"
#include "rtdata.h"
#include "seterrno.h"
#include "fileacc.h"
#if defined(__OS2__)
 #define INCL_DOS
 #include <wos2.h>
#elif defined(__NT__)
 #include <windows.h>
 #include "ntex.h"
#endif
 

static int check_mode( int handle, int mode ) {
/*********************************************/

    int flags;
    int __errno;

    handle=handle;mode=mode;flags=flags;
    __errno = 0;
    #if defined(__QNX__)
	if( (flags = fcntl( handle, F_GETFL )) == -1 ) {
	    return( -1 );
	}
    
	__errno = EOK;
	if( (flags & O_APPEND) && !(mode & _APPEND) ) {
	    __errno = EACCES;
	}
	if( (flags & O_ACCMODE) == O_RDONLY ) {
	    if( ( (mode&O_RDWR) && !(mode&O_RDONLY) ) || (mode&O_WRONLY) ) {
		__errno = EACCES;
	    }
	} else if( (flags & O_ACCMODE) == O_WRONLY ) {
	    if( ( (mode&O_RDWR) && !(mode&O_WRONLY) ) || (mode&O_RDONLY) ) {
		__errno = EACCES;
	    }
	}
    #elif defined(__WARP__)
	{
	    ULONG   state;
	    int	    rc;
     
	    rc = DosQueryFHState((HFILE)handle, &state);
    
	    if( rc == 0 ) {
		if( (state&OPENMODE_ACCESS_MASK) == OPENMODE_ACCESS_RDONLY ) {
		    if( ( (mode&O_RDWR) && !(mode&O_RDONLY) ) || (mode&O_WRONLY) ) {
			__errno = EACCES;
		    }
		}
		if( (state&OPENMODE_ACCESS_MASK) == OPENMODE_ACCESS_WRONLY ) {
		    if( ( (mode&O_RDWR) && !(mode&O_WRONLY) ) || (mode&O_RDONLY) ) {
			__errno = EACCES;
		    }
		}
	    } else {
		__errno = EACCES;
	    }
	}
    #elif defined(__OS2_286__)
	// there is support under OS/2 1.x, but Pharlap's run286 doesn't
	// support the DosQFHandState() function
    #elif defined(__NT__)
	// there doesn't appear to be any support under NT for interrogating
	// how a file handle was opened
    #else
	// other systems are MS-DOS and NETWARE...
	// these systems have no support for interrogating how a file handle
	// was opened.
    #endif
    if( __errno == EACCES ) {
	__set_errno( __errno );
	return( -1 );
    }
    return(0);
}

_WCRTLINK int _hdopen( int handle, int mode )
{
#if !defined(__QNX__)
    int		rwmode;
    unsigned	io_mode;
#endif

    if( check_mode( handle, mode ) ) {
	return( -1 );
    }
    #if defined(__NETWARE__)
    {
	extern void __NewFileHandle( int *handle );
	int new_handle;

	__NewFileHandle( &new_handle );
	if( new_handle == -1 ) {
	    return( -1 );
	}
	__FileHandles[ new_handle ] = handle;
	handle = new_handle;
    }
    #elif defined(__NT__)
	handle = _InitAccessH( handle );
    #elif !defined(__QNX__)
	_InitAccessH( handle );
    #endif
    #ifndef __QNX__
	rwmode = mode & ( O_RDONLY | O_WRONLY | O_RDWR );
	io_mode = __IOMode( handle );
	io_mode &= ~(_READ|_WRITE|_APPEND|_BINARY);
	if( rwmode == O_RDWR )	io_mode |= _READ | _WRITE;
	if( rwmode == O_RDONLY) io_mode |= _READ;
	if( rwmode == O_WRONLY) io_mode |= _WRITE;
	if( mode & O_APPEND )	io_mode |= _APPEND;
	if( mode & O_BINARY ) {
	   io_mode |= _BINARY;
	}
	__SetIOMode( handle, io_mode );
    #endif

    return( handle );
}
