/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1988, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  10-may-88	F.W.Crigger	Initial implementation
  22-aug-90	J.B.Schueler	Changed non-block return of _set_errno(rc)
  ---		---		Version 8.0
  21-nov-90	A.F.Scian	changed busy-wait loop to a call to sleep()
				(OS can execute other tasks while waiting!)
   2-mar-93	B.J. Stecher	made QNX specific version
  03-mar-94	Greg Bentz	add _locking
--------------------------------------------------------Version-11.0---------
  28-may-95	Greg Bentz	variety.h and dll linkage
  03-jan-96	Greg Bentz	remove stubs that are now aliases
  16-jul-96	B.J. Stecher	made QNX version agree with doc.
  26-jul-96	M. Hildebrand	Use rtcheck.h support.
*/
#include "variety.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef __QNX__
#include <fcntl.h>
#else
#include <dos.h>
#endif
#include <time.h>
#include <errno.h>
#include <sys/locking.h>
#include "rtcheck.h"
#include "rtdata.h"
#include "seterrno.h"

#ifdef __QNX__
_WCRTLINK int (locking)( int handle, int mode, long nbytes )
{
    int 		cmd;
    struct flock 	flock_buff;
    int			ret;
    int			tries;

    __handle_check( handle, -1 );

    flock_buff.l_whence = SEEK_SET;
    flock_buff.l_start = lseek( handle, 0L, SEEK_CUR );
    flock_buff.l_len = nbytes;

    switch( mode ) {
    case LK_UNLCK:	/* unlock */
	cmd = F_SETLK;
	flock_buff.l_type = F_UNLCK;
	break;
    case LK_LOCK:	/* blocking read/write lock */
	cmd = F_SETLK;
	flock_buff.l_type = F_WRLCK;
	break;
    case LK_NBLCK:	/* non-blocking read/write lock */
	cmd = F_SETLK;
	flock_buff.l_type = F_WRLCK;
	break;
    case LK_RLCK:	/* blocking read lock */
        mode = LK_LOCK; /* so we only have to check for 1 thing in loop */
	cmd = F_SETLK;
	flock_buff.l_type = F_RDLCK;
	break;
    case LK_NBRLCK:	/* non-blocking read lock */
	cmd = F_SETLK;
	flock_buff.l_type = F_RDLCK;
	break;
    default:
        errno = ENOTSUP;
	return( -1 );
    }
    tries = 10;
    for( ;; ) {
        ret = fcntl( handle, cmd, &flock_buff );
	if( ret != -1 ) break;
	if( errno != EAGAIN ) break;
	if( mode != LK_LOCK ) break;
	if( --tries == 0 ) break;
	sleep( 1 );
    }
    if( errno == EAGAIN ) errno = EDEADLK;
    return( ret );
}

#else

_WCRTLINK int (locking)( int handle, int mode, long nbytes )
{
    unsigned long offset;
    int retry_count;
    int rc;

    __handle_check( handle, -1 );

    offset = lseek( handle, 0L, SEEK_CUR );
    if( mode == LK_UNLCK ) return( unlock( handle, offset, nbytes ) );
    for( retry_count = 0; retry_count < 10; ++retry_count ) {
	rc = lock( handle, offset, nbytes );
	if( rc == 0 )  return( 0 );
	if( mode == LK_NBLCK  ||  mode == LK_NBRLCK ) {
	    return( rc );				/* JBS 22-aug-90 */
	}
	sleep( 1 ); 				/* wait 1 second */
    }
    __set_doserrno( rc );
    __set_errno( EDEADLOCK );
    return( -1 );
}
#endif
