/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  18-aug-87     F.W.Crigger     Initial implementation
--------------------------------------------------Version 6.0----------
  11-apr-88     F.W.Crigger     use tinyio
--------------------------------------------------Version 8.0----------
  02-jan-90     A.F.Scian       updated to use POSIX-compliant TinyIO
  22-sep-92	F.W.Crigger	return rc vs TINY_INFO_SEEK(rc)
  15-jul-93	Greg Bentz	made use of TinyLSeek()
  28-jul-93	Greg Bentz	make 32 bit Windows version use intdos()
  03-mar-94	Greg Bentz	add _lseek
--------------------------------------------------------Version-11.0---------
  09-jan-95	M.P.Hildebrand	New _iomode bit set when lseek() done with
				positive offset, so that write() can pad
				with zero bytes if lseek() done past EOF.
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
  03-jan-96	Greg Bentz	remove stubs that are now aliases
  26-jul-96	M. Hildebrand	Use rtcheck.h support.
*/

#include "variety.h"
#include <stdio.h>
#include <unistd.h>
#include "iomode.h"
#include "tinyio.h"
#include "rtcheck.h"
#include "seterrno.h"

#ifdef __WINDOWS_386__
#include <dos.h>
#endif


_WCRTLINK long lseek( int handle, long offset, int origin )
{
    uint_32		pos;
    unsigned	    	_iomode;

    #ifdef __WINDOWS_386__
	union REGS regs;
	
	__handle_check( handle, -1 );

	/*** Set the _FILEEXT _iomode bit if positive offset ***/
	_iomode = __IOMode( handle );
	if( offset>0 && !(_iomode&_APPEND) )
	    __SetIOMode( handle, _iomode | _FILEEXT );

	regs.h.ah = DOS_LSEEK;
	regs.h.al = origin;
	regs.w.bx = handle;
	regs.w.cx = (offset >> 16) & 0xffff;
	regs.w.dx = offset & 0xffff;
	intdos( &regs, &regs );
	pos = (regs.w.dx << 16) | regs.w.ax;
	if( regs.w.cflag ) {
	    __set_errno_dos( regs.w.ax );
	    return( -1L );
	}
    #else
	unsigned short rc;
	
	__handle_check( handle, -1 );

	/*** Set the _FILEEXT _iomode bit if positive offset ***/
	_iomode = __IOMode( handle );
	if( offset>0 && !(_iomode&_APPEND) )
	    __SetIOMode( handle, _iomode | _FILEEXT );
	    
	rc = TinyLSeek( handle, offset, origin, (void _WCNEAR *) &pos );
	if( TINY_ERROR(rc) ) {
	    __set_errno_dos( TINY_INFO(rc) );
	    return( -1L );
	}
    #endif
    return( pos );
}
