/*
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %       Copyright (C) 1992, by WATCOM Systems Inc. All rights     %
  %       reserved. No part of this software may be reproduced      %
  %       in any form or by any means - graphic, electronic or      %
  %       mechanical, including photocopying, recording, taping     %
  %       or information storage and retrieval systems - except     %
  %       with the written permission of WATCOM Systems Inc.        %
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  05-mar-92     Craig Eisler	defined for NT
  28-mar-92	Craig Eisler	default windowing
  16-oct-92	Craig Eisler	use integer file ids instead of handles
  14-nov-92	F.W.Crigger	Make _AccessFList and _ReleaseFList indirect
				pointers to functions.
  09-dec-92	Craig Eisler	__NTAddSpecificFileHandle was not returning
  				once it added the handle to the array
  13-jan-93	Craig Eisler	added __NTAtMaxFiles
  29-sep-93	A.F.Scian	if 0, 1, or 2 are changed then notify NT
--------------------------------------------------------Version-11.0---------
  28-may-95	Greg Bentz	variety.h and dll linkage
  20-nov-95	Greg Bentz	don't allow negative handles
  12-mar-96	Greg Bentz	deal with non-existent stdio handles
  				improve creation of fake handles
  07-jan-98	Greg Bentz	clean up fake handles on shutdown
*/
#include "variety.h"
#include "osver.h"
#include "rtinit.h"
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <direct.h>
#include <malloc.h>
#include <windows.h>

extern	unsigned __NFiles;

#ifdef __SW_BM
extern void (*_AccessFList)( void );
extern void (*_ReleaseFList)( void );
#else
#define _AccessFList()
#define _ReleaseFList()
#endif

HANDLE *__FileHandleIDs = 0;
static int topFileID = 0;
HANDLE *__FakeFileHandles = 0;
static int topFakeFile = 0;

int __NTAtMaxFiles( void )
{
    int	i;

    if( topFileID < __NFiles ) {
	return( 0 );
    }

    for( i=0;i<topFileID;i++ ) {
	if( __FileHandleIDs[i] == NULL ) {
	    return( 0 );
	}
    }
    return( 1 );
}


int __NTAddFileHandle( HANDLE h )
{
    int	i;

    _AccessFList();
    for( i=0;i<topFileID;i++ ) {
	if( __FileHandleIDs[i] == NULL ) {
	    __FileHandleIDs[i] = h;
	    _ReleaseFList();
	    return( i );
	}
    }
    __FileHandleIDs = realloc( __FileHandleIDs, (topFileID+1) * sizeof( HANDLE ) );
    __FileHandleIDs[ topFileID ] = h;
    topFileID++;
    _ReleaseFList();
    return( topFileID - 1 );
}

void __NTAddSpecificFileHandle( HANDLE h, int hid )
{
    int	i;

    if( hid < 0 ) return;
    _AccessFList();
    switch( hid ) {
    case STDIN_FILENO:
	SetStdHandle( STD_INPUT_HANDLE, h );
	break;
    case STDOUT_FILENO:
	SetStdHandle( STD_OUTPUT_HANDLE, h );
	break;
    case STDERR_FILENO:
	SetStdHandle( STD_ERROR_HANDLE, h );
	break;
    }
    if( hid < topFileID ) {
	__FileHandleIDs[ hid ] = h;
	_ReleaseFList();
	return;
    }
    __FileHandleIDs = realloc( __FileHandleIDs, (hid+1) * sizeof( HANDLE ) );
    for( i=topFileID;i<hid;i++ ) {
	__FileHandleIDs[i] = NULL;
    }
    __FileHandleIDs[ hid ] = h;
    topFileID = hid+1;
    _ReleaseFList();
}

void __NTRemoveFileHandle( int hid )
{
    _AccessFList();
    if( hid > 0 && hid < topFileID ) {
	__FileHandleIDs[hid] = NULL;
    }
    _ReleaseFList();
}

void __NTInitFileHandles( void )
{
    HANDLE h;
    //topFileID = 0L;
    //__FileHandleIDs = NULL;
    h = GetStdHandle( STD_INPUT_HANDLE );
    if( h == 0 || h == INVALID_HANDLE_VALUE ) {
	h = (HANDLE)__NTGetFakeHandle();
    }
    __NTAddFileHandle( h );
    h = GetStdHandle( STD_OUTPUT_HANDLE );
    if( h == 0 || h == INVALID_HANDLE_VALUE ) {
	h = (HANDLE)__NTGetFakeHandle();
    }
    __NTAddFileHandle( h );
    h = GetStdHandle( STD_ERROR_HANDLE );
    if( h == 0 || h == INVALID_HANDLE_VALUE ) {
	h = (HANDLE)__NTGetFakeHandle();
    }
    __NTAddFileHandle( h );
}

int __NTGetFakeHandle( void )
{
    int h;
    _AccessFList();
    h = (int)CreateEvent( 0, 0, 0, 0 );
    if( !h ) {
	// win32s does not support event handles
	static DWORD fakeHandle = 0x80000000L;
	fakeHandle++;
	h = fakeHandle;
    } else {
	__FakeFileHandles = realloc( __FakeFileHandles, (topFakeFile+1) * sizeof( HANDLE ) );
	__FakeFileHandles[ topFakeFile ] = (HANDLE)h;
	topFakeFile++;
    }
    _ReleaseFList();
    return( h );
}

static void __NTFiniFileHandles( void )
{
    if( __FileHandleIDs ) {
	free( __FileHandleIDs );
	__FileHandleIDs = 0;
    }
    if( __FakeFileHandles ) {
	int i;
	for( i = 0 ; i < topFakeFile ; i++ ) {
	    CloseHandle( __FakeFileHandles[i] );
	}
	free( __FakeFileHandles );
	__FakeFileHandles = 0;
    }
}
AYI( __NTFiniFileHandles, INIT_PRIORITY_LIBRARY-1 )
