/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  27-dec-85	F.W.Crigger	Initial implementation
  12-nov-87	F.W.Crigger	Added permission parm
  10-jan-88	F.W.Crigger	Check for existing file before doing CREAT
  19-jan-88	F.W.Crigger	Added support for O_TRUNC
  19-jan-88	F.W.Crigger	Incorporated sopen into this module
-----------------------------------------------------Version 6.0---------
  30-mar-88	F.W.Crigger	added set_binary routine
  11-apr-88	A.F.Scian	used tinyio
  17-may-88	F.W.Crigger	Added call to _set_errno
  21-may-88	F.W.Crigger	Moved call to isatty() from ioalloc to here.
-----------------------------------------------------Version 6.5---------
  11-aug-88	F.W.Crigger	turn off _READ, _WRITE, _APPEND bits
  16-nov-88	F.W.Crigger	'creat' logic moved inline
  01-mar-90	F.W.Crigger	set_binary: don't set RAW flag for binary files
  16-mar-90	F.W.Crigger	__umaskval (file permission mask) added
  26-mar-90	F.W.Crigger	check for O_WRONLY or O_CREAT before trying to
				open an existing file (append command problem)
  07-apr-90	F.W.Crigger	open and sopen changed to variable # of args
  09-apr-90	A.F.Scian	added va_end() for strict ANSI compliance
  15-apr-90	F.W.Crigger	Use TinyAccess to see if file exists first
  17-apr-90	F.W.Crigger	Just try to open file if O_RDONLY
  19-apr-90	F.W.Crigger	make __set_binary a global routine
  20-apr-90	F.W.Crigger	moved call to isatty so all paths execute it
  12-sep-90	F.W.Crigger	Reordered logic so that TinyAccess is not called
				for devices.
-----------------------------------------------------Version 8.0---------
  11-oct-90	A.F.Scian	fixed open() creating a file when WRITE|APPEND
				without O_CREATE being set
  21-nov-90	A.F.Scian	fixed sopen() so that creating a file opens
				it with the correct share permissions
				- cleaned up code, also
  05-sep-91	J.B.Schueler	include O_NOINHERIT bit in open mode
  05-sep-91	J.B.Schueler	added support for O_EXCL flag
--------------------------------------------------Version 9.0----------
  16-dec-91	Craig Eisler	default windowing system (multiple windows)
--------------------------------------------------Version 9.0----------
  29-mar-92	Craig Eisler	mods for NT
  13-jul-92	B.J. Stecher	O_NOINHERIT screwed up setting of _iomode[]
  01-sep-92	F.W.Crigger	Turn off _BINARY bit
  21-oct-92	Craig Eisler	ANSI function names for default windowing
  22-nov-92	F.W.Crigger	change reference to _iomode array
  12-mar-93	B.J. Stecher	only do a TinyCreate for O_CREAT if the
  				TinyOpen error code was "file not found"
  19-oct-94	Greg Bentz	Use TinyCreateEx if DOS version >= 5
  				and not Flashtek or DOS/4G
  07-nov-94	Greg Bentz	backout above change do to inconsistent
  				availability accross dos boxes
--------------------------------------------------------Version-11.0---------
  05-jan-95	M.P.Hildebrand	Fix incorrect handling of umask() permissions
				mask _RWD_umaskval.
  03-mar-95	M. Hildebrand	Added _wopen and _wsopen.
  28-may-95	Greg Bentz	variety.h and dll linkage
  21-jun-95	T. Schiller	Added range checking to make certain that we
  				don't open too many files.
  28-jun-95     T. Schiller     Really fixed it this time.
  21-sep-95	M. Hildebrand	Added _u version, which calls wide version.
  03-jan-96	Greg Bentz	remove stubs that are now aliases
*/
#include "variety.h"
#include "widechar.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <share.h>
#include "dos.h"
#include "dosfunc.h"
#include "tinyio.h"
#include "rtdata.h"
#include "iomode.h"
#include "seterrno.h"
#include "defwin.h"
#include "extender.h"
#ifdef __WIDECHAR__
    #include <mbstring.h>
    #include "mbwcconv.h"
#endif

/* file attributes */

#define _A_RDONLY	0x01

extern unsigned __NFiles;


_WCRTLINK int __F_NAME(open,_wopen)( const CHAR_TYPE *name, int mode, ... )
    {
	int permission;
	va_list args;

	va_start( args, mode );
	permission = va_arg( args, int );
	va_end( args );
	return( __F_NAME(sopen,_wsopen)( name, mode, 0, permission ) );
    }


_WCRTLINK int __F_NAME(sopen,_wsopen)( const CHAR_TYPE *name, int mode, int shflag, ... )
{
    va_list		args;

    va_start( args, shflag );
    return( __F_NAME(_sopen,__wsopen)( name, mode, shflag, args ) );
}


static int __F_NAME(_sopen,__wsopen)( const CHAR_TYPE *name, int mode, int shflag, va_list args )
    {
	int	    rwmode;
	int	    handle;
	int	    attr;
	int	    permission;
	unsigned    io_mode;
	tiny_ret_t  rc;
	char	    dummy;
#ifdef __WIDECHAR__
	char	    mbName[MB_CUR_MAX*_MAX_PATH];   /* single-byte char */
#endif

	while( *name == ' ' ) ++name;
	handle = -1;

	/*** If necessary, convert the wide filename to multibyte form ***/
	#ifdef __WIDECHAR__
	    __filename_from_wide( mbName, name );
	#endif
							/* 05-sep-91 */
	rwmode = mode & ( O_RDONLY | O_WRONLY | O_RDWR | O_NOINHERIT );
	#ifdef __WIDECHAR__
	    rc = TinyOpen( mbName, rwmode | shflag );
	#else
	    rc = TinyOpen( name, rwmode | shflag );
	#endif
	if( TINY_OK( rc ) ) {
	    handle = TINY_INFO( rc );
	    if (handle >= __NFiles) {
		TinyClose( handle );
		__set_errno(EMFILE);
		return -1;
	    }
	}
					    /* 17-apr-90   05-sep-91 */
	if( (mode & (O_RDONLY | O_WRONLY | O_RDWR)) != O_RDONLY ) {
	    if( handle != -1 ) {
		if( ! isatty( handle ) ) {	/* if not a device */
#if 0
		    rc = TinyAccess( name, 0 ); /* check for existence */
		    if( TINY_ERROR( rc ) ) {	/* file does not exist */
			TinyClose( handle );	/* close whatever file we got */
			handle = -1;
		    } else if( mode & O_EXCL ) {    /* must not exist */
#else
	/*
	Don't need to do the access check, since the file was opened
	and therefore must exist (TinyOpen can't create a file).
	We don't want to do the check because there are classes of items
	in the file system namespace that are not devices, but the TinyAccess
	will fail on (e.g. named pipes).
	*/
		    /* must not exist if O_CREAT specified */
		    if( mode & O_EXCL && mode & O_CREAT ) {
#endif
			TinyClose( handle );
			__set_errno( EEXIST );
			return( -1 );
		    } else if( mode & O_TRUNC ) {   /* truncate file */
			rc = TinyWrite( handle, &dummy, 0 );
			if( TINY_ERROR( rc ) ) {
			    TinyClose( handle );
			    return( __set_errno_dos( TINY_INFO(rc) ) );
			}
		    }
		}
	    }
	}
	if( handle == -1 ) {			/* could not open */
	    if(( mode & O_CREAT ) == 0
	    	|| TINY_INFO( rc ) != TIO_FILE_NOT_FOUND ) {
		return( __set_errno_dos( TINY_INFO( rc ) ) );
	    }
	    /* creating the file */
	    permission = va_arg( args, int );
	    va_end( args );
	    if( permission == 0 ) permission = S_IWRITE | S_IREAD;
	    permission &= ~_RWD_umaskval;		/* 05-jan-95 */
	    attr = 0;
	    if(( permission & S_IWRITE) == 0 ) attr = _A_RDONLY;
	    #if 0
		/* remove this support because it is not consistently available */
		if( _RWD_osmajor >= 5
		    #ifdef __DOS_EXT__
			&& !_IsFlashTek() 
			&& !_IsRational()
		    #endif
		    ) {
		    /* this function is only available in version DOS 5 and up */
		    /* this new way was added to handle the case of creating a */
		    /* new file with read-only access, but with a writeable */
		    /* file handle */
		    #ifdef __WIDECHAR__
			rc = TinyCreateEx( mbName, rwmode|shflag, attr, TIO_OPEN );
		    #else
			rc = TinyCreateEx( name, rwmode|shflag, attr, TIO_OPEN );
		    #endif
		    if( TINY_ERROR( rc ) ) {
			return( __set_errno_dos( TINY_INFO( rc ) ) );
		    }
		    handle = TINY_INFO( rc );
		} else
	    #endif
	    {
		/* do it the old way */
		#ifdef __WIDECHAR__
		    rc = TinyCreate( mbName, attr );
		#else
		    rc = TinyCreate( name, attr );
		#endif
		if( TINY_ERROR( rc ) ) {
		    return( __set_errno_dos( TINY_INFO( rc ) ) );
		}

		handle = TINY_INFO( rc );
		if (handle >= __NFiles)
		{
			TinyClose(handle);
			__set_errno(EMFILE);
			return -1;
		}

		/* 21-nov-90 AFS: the file is created so now the file must be */
		/*		      opened with the correct share permissions */
		if( shflag != 0 ) {
		    rc = TinyClose( handle );
		    if( TINY_ERROR( rc ) ) {
			return( __set_errno_dos( TINY_INFO( rc ) ) );
		    }
		    #ifdef __WIDECHAR__
			rc = TinyOpen( mbName, rwmode | shflag );
		    #else
			rc = TinyOpen( name, rwmode | shflag );
		    #endif
		    if( TINY_ERROR( rc ) ) {
			return( __set_errno_dos( TINY_INFO( rc ) ) );
		    }
		    handle = TINY_INFO( rc );
		    /* handle does not equal -1 now */
		}
	    }
	}
	io_mode = __IOMode( handle );
	io_mode &= ~(_READ|_WRITE|_APPEND|_BINARY);	/* 11-aug-88 */
	if( isatty( handle ) )	io_mode |= _ISTTY;
	rwmode &= ~O_NOINHERIT;
	if( rwmode == O_RDWR )	io_mode |= _READ | _WRITE;
	if( rwmode == O_RDONLY) io_mode |= _READ;
	if( rwmode == O_WRONLY) io_mode |= _WRITE;
	if( mode & O_APPEND )	io_mode |= _APPEND;
	if( mode & (O_BINARY|O_TEXT) ) {
	    if( mode & O_BINARY ) {
	       io_mode |= _BINARY;
	    }
	} else {
	    if( _RWD_fmode == O_BINARY ) {
	       io_mode |= _BINARY;
	    }
	}
	__SetIOMode( handle, io_mode );
	#ifdef DEFAULT_WINDOWING
	    if( _WindowsNewWindow != 0 ) {
		#ifdef __WIDECHAR__
		    if( !wcscmp( name, L"con" ) ) {
			_WindowsNewWindow( NULL, handle, -1 );
		    }
		#else
		    if( !stricmp( name, "con" ) ) {
			_WindowsNewWindow( NULL, handle, -1 );
		    }
		#endif
	    }
	#endif

	return( handle );
    }


#ifndef __WIDECHAR__			/* compile one version only */
int __set_binary( int handle )
    {
	unsigned	io_mode;

	io_mode = __IOMode( handle );
	io_mode |= _BINARY;
	__SetIOMode( handle, io_mode );
	if( io_mode & _ISTTY ) {
	    tiny_ret_t rc;

	    rc = TinyGetDeviceInfo( handle );
	    if( TINY_ERROR( rc ) ) {
		return( __set_errno_dos( TINY_INFO( rc ) ) );
	    }
	    rc = TinySetDeviceInfo( handle, TINY_INFO(rc) | TIO_CTL_RAW );
	    if( TINY_ERROR( rc ) ) {
		return( __set_errno_dos( TINY_INFO( rc ) ) );
	    }
	}
	return( 0 );
    }
#endif
