/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  09-aug-89	John Dahms	defined for OS/2
  09-apr-90	A.F.Scian	made open & sopen "..." type functions
				for strict POSIX compliance
--------------------------------------------------Version 8.5----------
  09-jul-91     G. Coschi	support OS/2 2.0
--------------------------------------------------Version 9.0----------
  16-dec-91	C.G.Eisler	default windowing system (multiple windows)
  21-oct-92	Craig Eisler	ANSI function names for default windowing
  22-nov-92	F.W.Crigger	change reference to _iomode array
--------------------------------------------------------Version-11.0---------
  05-jan-95	M.P.Hildebrand	Fix incorrect handling of umask() permissions
  				mask _RWD_umaskval.
  10-jan-95	M.P.Hildebrand	The file permissions are now always initialized
  				when calling DosOpen.
  03-mar-95	M. Hildebrand	Added _wopen and _wsopen.
  28-may-95	Greg Bentz	variety.h and dll linkage
  21-jun-95	T. Schiller	Added range checking to make certain that we
  				don't open too many files.
  25-sep-95	M. Hildebrand	Added _u version, which calls wide version.
  03-jan-96	Greg Bentz	remove stubs that are now aliases
*/
#include "variety.h"
#include "widechar.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <share.h>
#include "dos.h"
#include "dosfunc.h"
#include <direct.h>
#include "fileacc.h"
#include <wos2.h>
#include <string.h>
#include "openmode.h"
#include "iomode.h"
#include "rtdata.h"
#include "seterrno.h"
#include "defwin.h"
#ifdef __WIDECHAR__
    #include <mbstring.h>
    #include "mbwcconv.h"
#endif

extern unsigned __NFiles;


static int _set_binary( int handle )
    {
	handle = handle;
	return( 0 );
    }


_WCRTLINK int __F_NAME(open,_wopen)( const CHAR_TYPE *name, int mode, ... )
    {
	int permission;
	va_list args;

	va_start( args, mode );
	permission = va_arg( args, int );
	va_end( args );
	return( __F_NAME(sopen,_wsopen)( name, mode, SH_COMPAT, permission ) );
    }


_WCRTLINK int __F_NAME(sopen,_wsopen)( const CHAR_TYPE *name, int mode, int shflag, ... )
{
    va_list		args;

    va_start( args, shflag );
    return( __F_NAME(_sopen,__wsopen)( name, mode, shflag, args ) );
}


static int __F_NAME(_sopen,__wsopen)( const CHAR_TYPE *name, int mode, int share, va_list args )
    {
	OS_UINT	    rwmode, error, actiontaken, fileattr, openflag, openmode;
	HFILE	    handle;
	int	    perm = S_IREAD | S_IWRITE;
	unsigned    iomode_flags;
#ifdef __WIDECHAR__
	char	    mbName[MB_CUR_MAX*_MAX_PATH];   /* single-byte char */
#endif

	while( *name == ' ' ) ++name;

	/*** If necessary, convert the wide filename to multibyte form ***/
	#ifdef __WIDECHAR__
	    __filename_from_wide( mbName, name );
	#endif
							/* 05-sep-91 */
	if( mode & O_CREAT ) {
	    perm = va_arg( args, int );
	    va_end( args );
	    if( mode & O_EXCL ) {
		openflag = OPENFLAG_FAIL_IF_EXISTS |
			   OPENFLAG_CREATE_IF_NOT_EXISTS;
	    } else if( mode & O_TRUNC ) {
		openflag = OPENFLAG_REPLACE_IF_EXISTS |
			   OPENFLAG_CREATE_IF_NOT_EXISTS;
	    } else {
		openflag = OPENFLAG_OPEN_IF_EXISTS |
			   OPENFLAG_CREATE_IF_NOT_EXISTS;
	    }
	} else if( mode & O_TRUNC ) {
	    openflag = OPENFLAG_REPLACE_IF_EXISTS;
	} else {
	    openflag = OPENFLAG_OPEN_IF_EXISTS;
	}
	rwmode = mode & OPENMODE_ACCESS_MASK;
#if defined(M_I86)
	if( rwmode == OPENMODE_ACCESS_WRONLY && !_RWD_osmode ) {
	    /* Can't open WRONLY file in bound application under DOS */
	    rwmode = OPENMODE_ACCESS_RDWR;
	}
#endif
	if( share == SH_COMPAT ) {
	    share = SH_DENYNO;
	}
	openmode = share+rwmode;

	perm &= ~_RWD_umaskval;			/* 05-jan-95 */
/*	if( ( perm & S_IREAD ) && !( perm & S_IWRITE ) )  */
	if( !(perm & S_IWRITE) ) {		/* 10-jan-95 */
	    fileattr = _A_RDONLY;
	} else {
	    fileattr = _A_NORMAL;
	}
	
	#ifdef __WIDECHAR__
	    error = DosOpen( (PSZ)mbName, &handle, &actiontaken, 0ul,
				fileattr, openflag, openmode, 0ul );
	#else
	    error = DosOpen( (PSZ)name, &handle, &actiontaken, 0ul,
				fileattr, openflag, openmode, 0ul );
	#endif
	if( error ) {
	    return( __set_errno_dos( error ) );
	}

	if (handle >= __NFiles)
	{
		DosClose( handle );
		__set_errno(ENOMEM);
		return -1;
	}

	if( rwmode == O_RDWR )  iomode_flags = _READ | _WRITE;
	if( rwmode == O_RDONLY) iomode_flags = _READ;
	if( rwmode == O_WRONLY) iomode_flags = _WRITE;
	if( mode & O_APPEND )   iomode_flags |= _APPEND;
	if( mode & (O_BINARY|O_TEXT) ) {
	    if( mode & O_BINARY )	  iomode_flags |= _BINARY;
	} else {
	    if( _RWD_fmode == O_BINARY )  iomode_flags |= _BINARY;
	}
	if( iomode_flags & _BINARY )  _set_binary( handle );
	__SetIOMode( handle, iomode_flags );
	_InitAccessH( handle );
	#ifdef DEFAULT_WINDOWING
	    if( _WindowsNewWindow != 0 ) {
		#ifdef __WIDECHAR__
		    if( ( _wcsicmp( name, L"con" ) == 0 ) ||
			( _wcsicmp( name, L"\\dev\\con" ) == 0 ) ) {
			_WindowsNewWindow( NULL, handle, -1 );
		    }
		#else
		    if( ( stricmp( name, "con" ) == 0 ) ||
			( stricmp( name, "\\dev\\con" ) == 0 ) ) {
			_WindowsNewWindow( NULL, handle, -1 );
		    }
		#endif
	    }
	#endif
	return( handle );
    }
