/*
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %       Copyright (C) 1992, by WATCOM Systems Inc. All rights     %
  %       reserved. No part of this software may be reproduced      %
  %       in any form or by any means - graphic, electronic or      %
  %       mechanical, including photocopying, recording, taping     %
  %       or information storage and retrieval systems - except     %
  %       with the written permission of WATCOM Systems Inc.        %
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  06-mar-92     Craig Eisler    defined for NT
  27-mar-92     Craig Eisler    default windowing
  21-oct-92     Craig Eisler    ANSI function names for default windowing
  22-nov-92     F.W.Crigger     change reference to _iomode array
  13-jan-93     Craig Eisler    check __NTAtMaxFiles to make sure
				we don't go past the end of _iomode
--------------------------------------------------------Version-11.0---------
  05-jan-95     M.P.Hildebrand  Fix incorrect handling of umask() permissions
				mask _RWD_umaskval.
  11-jan-95     M.P.Hildebrand  If O_CREAT and O_EXCL are specified and the
				file exists, the open call will fail.
				If O_CREAT is specified and the file exists,
				NT will no longer set the file attributes
				anyway (for POSIX compliance).
  31-jan-95	M.P.Hildebrand	Fixed a bug which caused open() to not
  				truncate an existing file opened with O_TRUNC.
  03-mar-95	M. Hildebrand	Added _wopen and _wsopen.
  28-may-95	Greg Bentz	variety.h and dll linkage
  21-jun-95	T. Schiller	Added range checking to make certain that we
  				don't open too many files.
  25-sep-95	Greg Bentz	fixup share mode handling
  25-sep-95	M. Hildebrand	Added _u version, which calls wide version.
  03-jan-96	Greg Bentz	remove stubs that are now aliases
  10-jun-96	M. Hildebrand	Win95 doesn't support CreateFileW.
  05-jul-96	M. Hildebrand	Use __lib_ versions of ...W Win32 functions.
  25-jul-96	M. Hildebrand	Handles are inheritable by default.
  16-jun-97	Greg Bentz	Tweak O_TRUNC handling for existing files
*/
#include "variety.h"
#include "widechar.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <share.h>
#include "fileacc.h"
#include "openmode.h"
#include "iomode.h"
#include "rtdata.h"
#include <windows.h>
#include "libwin32.h"
#include "ntex.h"
#include "seterrno.h"
#include "defwin.h"

extern unsigned __NFiles;


_WCRTLINK int __F_NAME(open,_wopen)( const CHAR_TYPE *name, int mode, ... )
{
    int         permission;
    va_list     args;

    va_start( args, mode );
    permission = va_arg( args, int );
    va_end( args );
    return( __F_NAME(sopen,_wsopen)( name, mode, SH_COMPAT, permission ) );
}


_WCRTLINK int __F_NAME(sopen,_wsopen)( const CHAR_TYPE *name, int mode, int shflag, ... )
{
    va_list		args;

    va_start( args, shflag );
    return( __F_NAME(_sopen,__wsopen)( name, mode, shflag, args ) );
}


static int __F_NAME(_sopen,__wsopen)( const CHAR_TYPE *name, int mode, int share, va_list args )
{
    DWORD		create_disp, exists_disp;
    DWORD		perm, fileattr;
    DWORD		desired_access, share_mode;
    SECURITY_ATTRIBUTES	security;
    HANDLE		handle;
    int			hid, rwmode;
    unsigned		iomode_flags;

    if( __NTAtMaxFiles() ) {
	__set_errno( EMFILE );
	return( -1 );
    }

    rwmode = mode & OPENMODE_ACCESS_MASK;
    __GetNTAccessAttr( rwmode, &desired_access, &perm );
    __GetNTShareAttr( share|rwmode, &share_mode );
    fileattr = FILE_ATTRIBUTE_NORMAL;

    security.nLength = sizeof( SECURITY_ATTRIBUTES );
    security.lpSecurityDescriptor = NULL;
    security.bInheritHandle = mode&O_NOINHERIT ? FALSE : TRUE;

    #ifdef __WIDECHAR__
	if( _WindowsNewWindow != 0 && !_wcsicmp( name, L"con" ) )
    #else
	if( _WindowsNewWindow != 0 && !stricmp( name, "con" ) )
    #endif
    {
	handle = (HANDLE) __NTGetFakeHandle();
	iomode_flags = _ISTTY;
	hid = _InitAccessH( (int) handle );
	_WindowsNewWindow( NULL, hid, -1 );
    } else {
	if( mode & O_CREAT ) {
	    perm = va_arg( args, int );
		va_end( args );
		perm &= ~_RWD_umaskval;             /* 05-jan-95 */
	    if( ( perm & S_IREAD ) && !( perm & S_IWRITE ) ) {
		fileattr = FILE_ATTRIBUTE_READONLY;
	    }
	    if( mode & O_EXCL ) {
		create_disp = CREATE_NEW;
		exists_disp = CREATE_NEW;
	    } else if( mode & O_TRUNC ) {
		create_disp = CREATE_ALWAYS;
		exists_disp = CREATE_NEW;
	    } else {
		create_disp = OPEN_ALWAYS;
		exists_disp = OPEN_EXISTING;
	    }
	} else if( mode & O_TRUNC ) {
	    exists_disp = TRUNCATE_EXISTING;
	} else {
	    exists_disp = OPEN_EXISTING;
	}

        /*** Open the file ***/
	#ifdef __WIDECHAR__
	    handle = __lib_CreateFileW( name, desired_access, share_mode,
					&security, exists_disp, fileattr,
					NULL );
	#else
	    handle = CreateFileA( name, desired_access, share_mode,
				  &security, exists_disp, fileattr, NULL );
	#endif
    	if( handle==(HANDLE)-1 ) {
	    if( mode&O_CREAT ) {
		#ifdef __WIDECHAR__
		    handle = __lib_CreateFileW( name, desired_access,
						share_mode, NULL, create_disp,
						fileattr, NULL );
		#else
		    handle = CreateFileA( name, desired_access,
					  share_mode, NULL, create_disp,
					  fileattr, NULL );
		#endif
	    }
	    if( handle == (HANDLE)-1 ) {
	        return( __set_errno_nt() );
	    }
	}
	
	iomode_flags = 0;

	hid = _InitAccessH( (int) handle );
	if (hid >= __NFiles)
	{
		CloseHandle(handle);
		__set_errno(ENOMEM);
		return -1;
	}
	else if (isatty(hid))
	{
	    iomode_flags = _ISTTY;
	}
    }
    
    if( rwmode == O_RDWR )       iomode_flags |= _READ | _WRITE;
    else if( rwmode == O_RDONLY) iomode_flags |= _READ;
    else if( rwmode == O_WRONLY) iomode_flags |= _WRITE;
    if( mode & O_APPEND )        iomode_flags |= _APPEND;
    if( mode & (O_BINARY|O_TEXT) ) {
	if( mode & O_BINARY )    iomode_flags |= _BINARY;
    } else {
	if( _RWD_fmode == O_BINARY ) iomode_flags |= _BINARY;
    }
    __SetIOMode( hid, iomode_flags );
    return( hid );
}
