/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %     Copyright (C) 1995, by Watcom International Inc.  All rights    %
   %     reserved.  No part of this software may be reproduced or        %
   %     used in any form or by any means - graphic, electronic or       %
   %     mechanical, including photocopying, recording, taping or        %
   %     information storage and retrieval systems - except with the     %
   %     written permission of Watcom International Inc.                 %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
   Date		By		Reason
   ----		--		------
--------------------------------------------------------Version-11.0---------
   17-oct-95	M. Hildebrand	Initial implementation.
   25-oct-95	M. Hildebrand	Added OS/2 support.
   25-oct-95	M. Hildebrand	Use _hdopen to create POSIX handles.
   30-oct-95	M. Hildebrand	Pipe handles weren't inheritable under NT.
*/

#include "variety.h"
#include <errno.h>
#include <fcntl.h>
#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#if defined(__NT__)
    #include <windows.h>
#elif defined(__OS2__)
    #include <wos2.h>
#endif
#include "fileacc.h"
#include "iomode.h"
#include "rtdata.h"
#include "seterrno.h"


_WCRTLINK int _pipe( int *phandles, unsigned psize, int textmode )
/****************************************************************/
{
#if defined(__NT__)
    HANDLE		hRead, hWrite;
    HANDLE		osHandle;
    BOOL		rc;
#elif defined(__OS2__) && defined(__386__)
    HFILE		hRead, hWrite;
    APIRET		rc;
#elif defined(__OS2__) && !defined(__386__)
    HFILE		hRead, hWrite;
    USHORT		rc;
#endif
    int			hReadPosix, hWritePosix;

    /*** Sanity check ***/
    if( textmode != 0  &&  textmode != _O_TEXT  &&  textmode != _O_BINARY ) {
	return( -1 );
    }

    /*** Create the pipes (note that psize==0 ==> use default size) ***/
    #if defined(__NT__)
	rc = CreatePipe( &hRead, &hWrite, NULL, psize );
	if( rc == FALSE ) {
	    __set_errno_nt();
	    return( -1 );
	}
    #elif defined(__OS2__)
	if( psize == 0 )  psize = 4096;
	#ifdef __386__
	    rc = DosCreatePipe( &hRead, &hWrite, psize );
	#else
	    rc = DosMakePipe( &hRead, &hWrite, psize );
	#endif
	if( rc != NO_ERROR ) {
	    __set_errno( ENOMEM );
	    return( -1 );
	}
    #endif

    /*** Make read handle inheritable ***/
    #ifdef __NT__
	rc = DuplicateHandle( GetCurrentProcess(), hRead, GetCurrentProcess(),
			      &osHandle, 0, TRUE, DUPLICATE_SAME_ACCESS );
	if( rc == FALSE ) {
	    CloseHandle( hRead );
	    CloseHandle( hWrite );
	    return( -1 );
	}
	CloseHandle( hRead );
	hRead = osHandle;
    #elif defined(__OS2__)
	/* Handle is inheritable by default */
    #endif

    /*** Make write handle inheritable ***/
    #ifdef __NT__
	rc = DuplicateHandle( GetCurrentProcess(), hWrite, GetCurrentProcess(),
			      &osHandle, 0, TRUE, DUPLICATE_SAME_ACCESS );
	if( rc == FALSE ) {
	    CloseHandle( hRead );
	    CloseHandle( hWrite );
	    return( -1 );
	}
	CloseHandle( hWrite );
	hWrite = osHandle;
    #elif defined(__OS2__)
	/* Handle is inheritable by default */
    #endif

    /*** Initialize the POSIX-level handles ***/
    hReadPosix = _hdopen( (int)hRead, textmode|_O_RDONLY );
    hWritePosix = _hdopen( (int)hWrite, textmode|_O_WRONLY );
    if( hReadPosix == -1  ||  hWritePosix == -1 ) {
	if( hReadPosix != -1 ) {
	    close( hReadPosix );
	} else {
	    #if defined(__NT__)
		CloseHandle( hRead );
	    #elif defined(__OS2__)
		DosClose( hRead );
	    #endif
	}
	if( hWritePosix != -1 ) {
	    close( hWritePosix );
	} else {
	    #if defined(__NT__)
		CloseHandle( hWrite );
	    #elif defined(__OS2__)
		DosClose( hWrite );
	    #endif
	}
	return( -1 );
    }

    /*** Store the new POSIX handles in return buffer ***/
    phandles[0] = hReadPosix;
    phandles[1] = hWritePosix;
    return( 0 );
}
