/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1992, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Note:
  	This file is used by the C++ iostream library.
	It supports the distinction between the various operating systems
	for the fstat() mode bits.
	This allows the iostream library to remain platform independent.

  Modified:	By:		Reason:
  ---------	---		-------
  22-mar-93	Greg Bentz	Initial implementation
--------------------------------------------------------Version-11.0---------
  28-may-95	Greg Bentz	variety.h and dll linkage
  07-nov-96	Greg Bentz	rewrite for correctness

*/
#include "variety.h"
#include "iomode.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

// these must be the same as is defined in iostream.h

#define __in          0x0001    // - open for input
#define __out         0x0002    // - open for output
#define __atend       0x0004    // - seek to end after opening
#define __append      0x0008    // - open for output, append to the end
#define __truncate    0x0010    // - discard contents after opening
#define __nocreate    0x0020    // - open only an existing file
#define __noreplace   0x0040    // - open only a new file
#define __text        0x0080    // - open as text file
#define __binary      0x0100    // - open as binary file

_WCRTLINK int __plusplus_fstat( int handle, int *pios_mode )
{
    int flags;
    int ios_mode = 0;

#if defined(__NETWARE__)
    handle=handle;
    flags=flags;
    // make a guess
    ios_mode |= __in | __out | __text;
#elif defined(__QNX__)
    if( (flags = fcntl( handle, F_GETFL )) == -1 ) {
	return( -1 );
    }
    if( flags & O_APPEND ) {
	ios_mode |= __append;
    }
    if( (flags & O_ACCMODE) == O_RDONLY ) {
	ios_mode |= __in;
    } else if( (flags & O_ACCMODE) == O_WRONLY ) {
	ios_mode |= __out;
    } else {
	ios_mode |= __in|__out;
    }
#else
    flags = __IOMode( handle );
    if( flags & _APPEND ) {
	ios_mode |= __append;
    }
    if( flags & _READ ) {
	ios_mode |= __in;
    }
    if( flags & _WRITE ) {
	ios_mode |= __out;
    }
    if( flags & _BINARY ) {
	ios_mode |= __binary;
    } else {
	ios_mode |= __text;
    }
#endif
    *pios_mode = ios_mode;
    return( 0 );
}
