/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  25-may-88     A.F.Scian       split from fgetc.c
--------------------------------------------------------Version 6.5---------
  04-jul-88     A.F.Scian       fixed bug that didn't return the DOS EOF
                                char in the compacted buffer
  17-aug-88     A.F.Scian       removed compaction for new buffered I/O
  04-sep-89     F.W.Crigger     Multi-Thread and OS/2 support
  23-oct-89     F.W.Crigger     Use TINY_LINFO instead of TINY_INFO for 386
  11-apr-91     G. Coschi       added special support for default windowing
                                when i/o to standard output
--------------------------------------------------Version 8.5----------
  09-jul-91     G. Coschi	support OS/2 2.0
  05-aug-91	C.G. Eisler	new default windowing support
--------------------------------------------------Version 9.0----------
  16-oct-91     G. Coschi	default windowing for OS/2
  15-dec-91	C.G.Eisler	multiple windowing
  28-jan-92	C.G.Eisler	fixed reads > 64k in 32-bit windows
--------------------------------------------------Version 9.x----------
  06-mar-92	C.G.Eisler	support NT
  16-oct-92	Craig Eisler	NT - use integer ids for file handles
  21-oct-92	Craig Eisler	ANSI function names for default windowing
--------------------------------------------------------Version-11.0---------
  28-may-95	Greg Bentz	variety.h and dll linkage
  26-oct-95	Greg Bentz	ensure release of semaphore with multi-thread
  				default windowing
  26-jul-96	M. Hildebrand	Use rtcheck.h support.
*/

#include "variety.h"
#if defined(__NT__)
#include <windows.h>
#include "ntex.h"
#elif defined(__OS2__)
    #include "tinyos2.h"
    #define INCL_DOSMEMMGR
    #include <wos2.h>
#else
    #include "tinyio.h"
#endif
#include "fileacc.h"
#include "rtcheck.h"
#include "seterrno.h"
#include "defwin.h"

#ifdef __WINDOWS_386__
#define MAXBUFF	0x8000
tiny_ret_t __TinyRead( int handle, char *buffer, unsigned len )
{
    unsigned	total=0,readamt;
    int		rc;

    while( len > 0 ) {

	if( len > MAXBUFF ) {
	    readamt = MAXBUFF;
	} else {
	    readamt = len;
	}
	rc = TinyRead( handle, buffer, readamt );
        if( rc < 0 ) return( rc );
	total += (unsigned) rc;
	if( (unsigned) rc != readamt ) return( total );

	len -= readamt;
	buffer += readamt;

    }
    return( total );
}
#endif



int __qread( int handle, char *buffer, unsigned len )
    {
#if defined( __NT__ )
	DWORD		amount_read;
        BOOL		rc;
	HANDLE		h;
#else
        tiny_ret_t rc;
#if defined(__WARP__)
	ULONG		amount_read;
#elif defined(__OS2_286__)
	USHORT		amount_read;
#else
	unsigned	amount_read;
#endif
#endif

	__handle_check( handle, -1 );
        _AccessFileH( handle );
	#ifdef DEFAULT_WINDOWING
	    if( _WindowsStdin != 0 ) {
		LPWDATA res;
    
		res = _WindowsIsWindowedHandle( handle );
		if( res ) {
		    int rt;
		    rt = _WindowsStdin( res, buffer, len );
		    _ReleaseFileH( handle );
		    return( rt );
		}
	    }
	#endif
#if defined(__NT__)
	h = __FileHandleIDs[ handle ];
	rc = ReadFile( h, buffer, len, &amount_read, NULL );
	if( !rc ) {
	    DWORD	err;
	    err = GetLastError();
	    __set_errno_dos( err );
	    if( err != ERROR_BROKEN_PIPE || amount_read != 0 ) {
		_ReleaseFileH( handle );
		return( -1 );
	    }
	}
#else
#if defined(__OS2__)
        rc = DosRead( handle, buffer, len, &amount_read );
#else
#ifdef __WINDOWS_386__
        rc = __TinyRead( handle, buffer, len );
#else
        rc = TinyRead( handle, buffer, len );
#endif
        amount_read = TINY_LINFO(rc);
#endif
        if( TINY_ERROR(rc) ) {
            _ReleaseFileH( handle );
            return( __set_errno_dos( TINY_INFO(rc) ) );
        }
#endif
        _ReleaseFileH( handle );
        return( amount_read );
    }
