/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1988, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  25-may-88	A.F.Scian	split from fflush.c
--------------------------------------------------------Version 6.5---------
  17-aug-88	A.F.Scian	removed compaction for new buffered I/O
  21-oct-88	F.W.Crigger	Add in code to seek to end-of-file for append
  04-sep-89	F.W.Crigger	Multi-Thread and OS/2 support
  23-oct-89	F.W.Crigger	Use TINY_LINFO instead of TINY_INFO for 386
  11-apr-91	G. Coschi	added special support for default windowing
				when i/o to standard output
--------------------------------------------------Version 8.5----------
  09-jul-91     G. Coschi	support OS/2 2.0
  05-aug-91	C.G. Eisler	new default windowing support
--------------------------------------------------Version 9.0----------
  09-oct-91     G. Coschi	default windowing for OS/2
  15-dec-91	C.G.Eisler	multiple windowing
  28-jan-92	C.G.Eisler	fixed writes > 64k for 32-bit windows
--------------------------------------------------Version 9.x----------
  06-mar-92	C.G.Eisler	support NT
  15-oct-92	Craig Eisler	use iomode array indices as handles
  21-oct-92	Craig Eisler	ANSI function names for default windowing
  22-nov-92	F.W.Crigger	change reference to _iomode array
--------------------------------------------------------Version-11.0---------
  20-mar-95	M. Hildebrand	Undefine CHAR_TYPE before including rtdata.h, so
				sigdefn.h won't complain.
  28-may-95	Greg Bentz	variety.h and dll linkage
  26-oct-95	Greg Bentz	ensure release of semaphore with multi-thread
  				default windowing
  13-mar-96	Greg Bentz	tweak typing inconsistencies
  26-jul-96	M. Hildebrand	Use rtcheck.h support.
*/
#include "variety.h"
#include <stdio.h>
#include <errno.h>
#if defined(__NT__)
    #include <windows.h>
    #include "ntex.h"
#elif defined(__OS2__)
    #include "tinyos2.h"
    #include <wos2.h>
#else
    #include "tinyio.h"
#endif
#include "fileacc.h"
#include "iomode.h"
#include "rtcheck.h"
#include "rtdata.h"
#include "seterrno.h"
#include "defwin.h"

#ifdef __WINDOWS_386__
#define MAXBUFF	0x8000
tiny_ret_t __TinyWrite( int handle, const void *buffer, unsigned len )
{
    unsigned	total=0,readamt;
    int		rc;

    while( len > 0 ) {

	if( len > MAXBUFF ) {
	    readamt = MAXBUFF;
	} else {
	    readamt = len;
	}
	rc = TinyWrite( handle, buffer, readamt );
        if( rc < 0 ) return( rc );
	total += (unsigned) rc;
	if( (unsigned) rc != readamt ) return( total );

	len -= readamt;
	buffer = ((const char *)buffer) + readamt;

    }
    return( total );
}
#endif

int __qwrite( int handle, const void *buffer, unsigned len )
    {
#if defined(__NT__)
	DWORD		len_written;
	HANDLE		h;
#else
	tiny_ret_t	rc;
#if defined(__WARP__)
	ULONG		len_written;
#elif defined(__OS2_286__)
	USHORT		len_written;
#else
	unsigned	len_written;
#endif
#endif

	__handle_check( handle, -1 );
#ifdef __NT__
	h = __FileHandleIDs[ handle ];
#endif
	_AccessFileH( handle );
	if( __IOMode( handle ) & _APPEND ) {
#ifdef __NT__
	    if( SetFilePointer( h, 0, NULL, FILE_END ) == -1 ) {
		_ReleaseFileH( handle );
		return( __set_errno_nt() );
	    }
#else
#ifdef __OS2__
	    {
	    unsigned long	dummy;
	    rc = DosChgFilePtr( handle, 0L, SEEK_END, &dummy );
	    }
#else
	    rc = TinySeek( handle, 0L, SEEK_END );
#endif
	    if( TINY_ERROR(rc) ) {
		_ReleaseFileH( handle );
		return( __set_errno_dos( TINY_INFO(rc) ) );
	    }
#endif
	}
	#ifdef DEFAULT_WINDOWING
	    if( _WindowsStdout != 0 ) {
		LPWDATA res;
    
		res = _WindowsIsWindowedHandle( handle );
		if( res ) {
		    int rt;
		    rt = _WindowsStdout( res, buffer, len );
		    _ReleaseFileH( handle );
		    return( rt );
		}
	    }
	#endif
#if defined(__NT__)
	if( !WriteFile( h, buffer, len, &len_written, NULL ) ) {
	    _ReleaseFileH( handle );
	    return( __set_errno_nt() );
	}
#else
#if defined( __OS2__ )
	rc = DosWrite( handle, (PVOID)buffer, len, &len_written );
#else
#ifdef __WINDOWS_386__
	rc = __TinyWrite( handle, buffer, len );
#else
	rc = TinyWrite( handle, buffer, len );
#endif
	len_written = TINY_LINFO(rc);
#endif
	if( TINY_ERROR(rc) ) {
	    _ReleaseFileH( handle );
	    return( __set_errno_dos( TINY_INFO(rc) ) );
	}
#endif
	if( len_written != len ) {
	    __set_errno( ENOSPC );
	}
	_ReleaseFileH( handle );
	return( len_written );
    }
