/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  27-dec-85	F.W.Crigger	Initial implementation
  19-jan-88	F.W.Crigger	check for bad handle
  19-jan-88	F.W.Crigger	strip CR's in O_TEXT mode
  03-feb-88	A.F.Scian	simplified handle check
				handle O_TEXT fully.
  04-feb-88	A.F.Scian	adjusted to return correct # bytes
  04-feb-88	F.W.Crigger	only do 1 read in O_TEXT mode if char device
--------------------------------------------------Version 6.0----------
  11-apr-88	F.W.Crigger	use tinyio
------------------------------------------------------------Version 6.5--------
  24-oct-88	F.W.Crigger	__NFiles introduced to replace _NFILES constant
------------------------------------------------------------Version 7.0--------
  23-feb-89	F.W.Crigger	return EACCES vs EBADF for permission error
  16-aug-89	John Dahms	Added multi-thread & OS/2 support
  23-oct-89	F.W.Crigger	Use TINY_LINFO instead of TINY_INFO for 386
------------------------------------------------------------Version 8.x--------
  24-feb-91	C.G.Eisler	Windows fix - if we haven't opened the file,
				then try using _lread to read the file.
  25-feb-91	C.G.Eisler	Added large file read for win386
  03-apr-91	C.G.Eisler	updated for WINDOWS_386 macro
  09-jul-91     G. Coschi	support OS/2 2.0
  05-aug-91	C.G. Eisler	default windowing support for Windows
--------------------------------------------------Version 9.0----------
  15-oct-91	G. Coschi	default windowing support for OS/2 PM
  15-dec-91	C.G.Eisler	multiple windowing
--------------------------------------------------Version 9.x----------
  06-mar-92	C.G.Eisler	support NT
  15-oct-92	Craig Eisler	use iomode array indices as handles
  21-oct-92	Craig Eisler	ANSI function names for default windowing
  22-nov-92	F.W.Crigger	change reference to _iomode array
  19-jan-93	Greg Bentz	handle ctrl-z when reading O_TEXT mode
--------------------------------------------------------Version-11.0---------
  20-mar-95	M. Hildebrand	Undefine CHAR_TYPE before including rtdata.h, so
				sigdefn.h won't complain.
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
  02-oct-95	T. Schiller	For NT a return code of ERROR_BROKEN_PIPE
				indicates that an anonymous pipe was closed.
				We should be returning EOF instead of an
				error.
  04-dec-95	T. Schiller	Under NT, if we encounter an
				ERROR_BROKEN_PIPE, should be returning the
				size of the partial buffer read rather than
				just 0.
  01-feb-96	T. Schiller	Return total_len instead of amount_read when
				we encounter an ERROR_BROKEN_PIPE.
  26-jul-96	M. Hildebrand	Use rtcheck.h support.
*/

#include "variety.h"
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>
#if defined(__NT__)
    #include <windows.h>
    #include "ntex.h"
#elif defined(__OS2__)
    #include "tinyos2.h"
    #define INCL_DOSMEMMGR
    #include <wos2.h>
#else
    #include "tinyio.h"
#endif
#include "fileacc.h"
#include "iomode.h"
#include "rtcheck.h"
#include "rtdata.h"
#include "seterrno.h"
#include "defwin.h"


#if defined(DEFAULT_WINDOWING) && defined( __WINDOWS__ )
    extern short _WCI86FAR pascal _lread( short, char _WCI86FAR *, short );
#endif


#if defined(__WINDOWS_386__)
    int __read
#else
    _WCRTLINK int read
#endif
		     ( int handle, char *buffer, unsigned len )
{
    unsigned read_len, total_len;
    unsigned reduce_idx, finish_idx;
    unsigned iomode_flags;
    #if defined(__NT__)
	DWORD	amount_read;
	BOOL	rc;
	HANDLE  h;
    #else
	tiny_ret_t rc;
	#if defined(__WARP__)
	    ULONG amount_read;
	#elif defined(__OS2_286__)
	    USHORT amount_read;
	#else
	    unsigned amount_read;
	#endif
    #endif
    #ifdef DEFAULT_WINDOWING
	LPWDATA res;
    #endif

    __handle_check( handle, -1 );
    #ifdef __NT__
	h = __FileHandleIDs[ handle ];
    #endif
    iomode_flags = __IOMode( handle );
    if( iomode_flags == 0 ) {
	#if defined( __WINDOWS__ ) || defined( __WINDOWS_386__ )
	    return( _lread( handle, buffer, len ) );
	#else
	    __set_errno( EBADF );
	    return( -1 );
	#endif
    }
    _AccessFileH( handle );
    if( !(iomode_flags & _READ) ) {
	__set_errno( EACCES );     /* changed from EBADF to EACCES 23-feb-89 */
	_ReleaseFileH( handle );
	return( -1 );
    }
    if( iomode_flags & _BINARY ) {	 /* if binary mode */
	#ifdef DEFAULT_WINDOWING
	    if( _WindowsStdin != 0 &&
		    (res = _WindowsIsWindowedHandle( handle )) != 0 ) {
		total_len = _WindowsStdin( res, buffer, len );
		rc = 0;
	    } else {
	#endif
	#if defined(__NT__)
		rc = ReadFile( h, buffer, len, &amount_read, NULL );
		total_len = amount_read;
		if( !rc ) {
		    _ReleaseFileH( handle );

		    if (GetLastError() == ERROR_BROKEN_PIPE)
			return total_len;

		    return( __set_errno_nt() );
		}
	#elif defined( __OS2__ )
		rc = DosRead( handle, buffer, len, &amount_read );
		total_len = amount_read;
	#else
		rc = TinyRead( handle, buffer, len );
		total_len = TINY_LINFO(rc);
	#endif
	#ifdef DEFAULT_WINDOWING
	    }
	#endif
	#if !defined(__NT__)
	    if( TINY_ERROR(rc) ) {
		_ReleaseFileH( handle );
		return( __set_errno_dos( TINY_INFO(rc) ) );
	    }
	#endif
    } else {
	total_len = 0;
	read_len = len;
	do {
	    #ifdef DEFAULT_WINDOWING
		if( _WindowsStdin != 0 &&
			(res = _WindowsIsWindowedHandle( handle )) != 0L ) {
		    amount_read = _WindowsStdin( res, buffer, read_len );
		    rc = 0;
		} else {
	    #endif
	    #if defined(__NT__)
		rc = ReadFile( h, buffer, read_len, &amount_read, NULL );
		if( !rc ) {
		    _ReleaseFileH( handle );

		    if (GetLastError() == ERROR_BROKEN_PIPE)
			return total_len;

		    return( __set_errno_nt() );
		}
	    #elif defined( __OS2__ )
		rc = DosRead( handle, buffer, read_len, &amount_read );
	    #else
		rc = TinyRead( handle, buffer, read_len );
		amount_read = TINY_LINFO(rc);
	    #endif
	    #ifdef DEFAULT_WINDOWING
	    }
	    #endif
	    #if !defined(__NT__)
		if( TINY_ERROR(rc) ) {
		    _ReleaseFileH( handle );
		    return( __set_errno_dos( TINY_INFO(rc) ) );
		}
	    #endif
	    if( amount_read == 0 ) {			/* EOF */
		break;
	    }
	    reduce_idx = 0;
	    finish_idx = reduce_idx;
	    for(; reduce_idx < amount_read; ++reduce_idx ) {
		if( buffer[ reduce_idx ] == 0x1a ) {	/* EOF */
		    lseek( handle,
			   ((long)reduce_idx - (long)amount_read)+1L,
			   SEEK_CUR );
		    total_len += finish_idx;
		    _ReleaseFileH( handle );
		    return( total_len );
		}
		if( buffer[ reduce_idx ] != '\r' ) {
		    buffer[ finish_idx++ ] = buffer[ reduce_idx ];
		}
	    }
	    total_len += finish_idx;
	    buffer += finish_idx;
	    read_len -= finish_idx;
	    if( iomode_flags & _ISTTY ) break;	/* 04-feb-88, FWC */
	} while( read_len != 0 );
    }
    _ReleaseFileH( handle );
    return( total_len );
}

#if defined( __WINDOWS_386__ )
#define MAXBUFF 0x8000
_WCRTLINK int read( int handle, char *buffer, unsigned len )
{
    unsigned	total=0,readamt;
    int		rc;

    __handle_check( handle, -1 );

    while( len > 0 ) {
	if( len > MAXBUFF ) {
	    readamt = MAXBUFF;
	} else {
	    readamt = len;
	}
	rc = __read( handle, buffer, readamt );
	if( rc < 0 ) return( rc );
	total += (unsigned) rc;
	if( rc != readamt ) return( total );

	len -= readamt;
	buffer += readamt;
    }
    return( total );

}
#endif

