/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  27-dec-85	F.W.Crigger	Initial implementation
  14-aug-87	J.U.Trabandt	set errno if error
  20-aug-87	F.W.Crigger	seek to end of file if open for append
  03-feb-88	A.F.Scian	simplified handle check
				handle O_TEXT fully.
  05-feb-88	F.W.Crigger	fix return value for text file, don't count
				the extra CR's written.
--------------------------------------------------Version 6.0----------
  11-apr-88	F.W.Crigger	use tinyio
  22-may-88	F.W.Crigger	use a buffer allocated from stack
------------------------------------------------------------Version 6.5--------
  24-oct-88	F.W.Crigger	__NFiles introduced to replace _NFILES constant
  12-nov-88	F.W.Crigger	Added code to check for ENOSPC
------------------------------------------------------------Version 7.0--------
  23-feb-89	F.W.Crigger	return EACCES vs EBADF for permission error
  16-aug-89	John Dahms	Added multi-thread & OS/2 support
  23-oct-89	F.W.Crigger	Use TINY_LINFO instead of TINY_INFO for 386
------------------------------------------------------------Version 8.x--------
  24-feb-91	C.G.Eisler	Windows fix - if we haven't opened the file,
				then try using _lwrite to write the file.
  25-feb-91	C.G.Eisler	Added large file write for win386
  03-apr-91	C.G.Eisler	adapted for new WINDOWS_386 macro
  22-apr-91	C.G.Eisler	Fixed bug having stddef.h occur after
				tinyio.h - tinyio.h redefines the fortran
				pragma for 32-bit windows.
  09-jul-91     G. Coschi	support OS/2 2.0
  05-aug-91	C.G. Eisler	default windowing support for Windows
--------------------------------------------------Version 9.0----------
  11-oct-91	G. Coschi	modifications for OS/2 PM (default windowing)
  15-dec-91	C.G.Eisler	multiple windowing
--------------------------------------------------Version 9.x----------
  06-mar-92	C.G.Eisler	support NT
  27-aug-92	G. Coschi	for 32-bit windows, write of length 0 should
  				truncate the file (like 16-bit windows & DOS)
  15-oct-92	Craig Eisler	use iomode array indices as handles
  21-oct-92	Craig Eisler	ANSI function names for default windowing
  22-nov-92	F.W.Crigger	change reference to _iomode array
--------------------------------------------------------Version-11.0---------
  09-jan-95	M.P.Hildebrand	For POSIX conformance, write() will now extend
  				file size by zero padding if lseek() has been
				used to seek past EOF.  (Does not apply when
				file opened with _APPEND.)
  20-mar-95	M. Hildebrand	Undefine CHAR_TYPE before including rtdata.h, so
				sigdefn.h won't complain.
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
  15-sep-95	Greg Bentz	ppc support
  13-mar-96	Greg Bentz	tweak typing inconsistencies
  26-jul-96	M. Hildebrand	Use rtcheck.h support.
*/
#include "variety.h"
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <malloc.h>
#include <errno.h>
#include <stddef.h>
#include <string.h>
#ifdef __NT__
    #include <windows.h>
    #include "ntex.h"
#elif defined(__OS2__)
    #include "tinyos2.h"
    #include <wos2.h>
#else
    #include "tinyio.h"
#endif
#include "fileacc.h"
#include "iomode.h"
#include "rtcheck.h"
#include "rtdata.h"
#include "seterrno.h"
#include "defwin.h"



#if defined( DEFAULT_WINDOWING ) && defined( __WINDOWS__ )
    extern int _WCI86FAR pascal _lwrite( short, const void _WCI86FAR *, short );
#endif

extern	void	__STKOVERFLOW();


static int zero_pad( int handle )			/* 09-jan-95 */
/*******************************/
{
    int			rc;
    long		curPos, eodPos;
    long		bytesToWrite;
    unsigned		writeAmt;
    char		zeroBuf[512];

    /*** Pad with zeros due to lseek() past EOF (POSIX) ***/
    curPos = lseek( handle, 0L, SEEK_CUR );	/* current offset */
    if( curPos == -1 )  return( -1 );
    eodPos = lseek( handle, 0L, SEEK_END );	/* end of data offset */
    if( eodPos == -1 )  return( -1 );

    if( curPos > eodPos ) {
	bytesToWrite = curPos - eodPos;		/* amount to pad by */

	__SetIOMode( handle, __IOMode(handle)&(~_FILEEXT) );
        if( bytesToWrite > 0 ) {		/* only write if needed */
	    memset( zeroBuf, 0x00, 512 );   	/* zero out a buffer */
	    do {    				/* loop until done */
		if( bytesToWrite > 512 )
		    writeAmt = 512;
		else
		    writeAmt = (unsigned)bytesToWrite;
		rc = write( handle, zeroBuf, writeAmt );
		if( rc<0 )  return( rc );
		bytesToWrite -= writeAmt;	/* more bytes written */
	    } while( bytesToWrite != 0 );
        }
    } else {
        curPos = lseek( handle, curPos, SEEK_SET );
	__SetIOMode( handle, __IOMode(handle)&(~_FILEEXT) );
        if( curPos == -1 )  return( -1 );
    }
    
    return( 0 );				/* return success code */
}


#if defined(__WINDOWS_386__)
    int __write
#else
    _WCRTLINK int write
#endif
			( int handle, const void *buffer, unsigned len )
    {
	char *buf;
	unsigned buf_size;
	unsigned iomode_flags;
#if defined(__NT__)
	DWORD	len_written,i,j;
	DWORD	rc;
	HANDLE	h;
#else
	tiny_ret_t rc;
#if defined(__WARP__)
	ULONG i,j,len_written;
#elif defined(__OS2_286__)
	USHORT i,j,len_written;
#else
	unsigned i,j,len_written;
#endif
#endif
#ifdef DEFAULT_WINDOWING
	LPWDATA res;
#endif

	__handle_check( handle, -1 );
#ifdef __NT__
	h = __FileHandleIDs[ handle ];
#endif
	iomode_flags = __IOMode( handle );
	if( iomode_flags == 0 ) {
	    
#if defined(__WINDOWS__) || defined(__WINDOWS_386__)
	    return( _lwrite( handle, buffer, len ) );
#else
	    __set_errno( EBADF );
	    return( -1 );
#endif
	}
	if( !(iomode_flags & _WRITE) ) {
	    __set_errno( EACCES );     /* changed from EBADF to EACCES 23-feb-89 */
	    return( -1 );
	}
	_AccessFileH( handle );
	if( iomode_flags & _APPEND ) {
#ifdef __NT__
	    if( SetFilePointer( h, 0, NULL, FILE_END ) == -1 ) {
		_ReleaseFileH( handle );
		return( __set_errno_nt() );
	    }
#else
#ifdef __OS2__
	    {
	    unsigned long	dummy;
	    rc = DosChgFilePtr( handle, 0L, SEEK_END, &dummy );
	    }
#else
	    rc = TinySeek( handle, 0L, SEEK_END );
#endif
	    if( TINY_ERROR(rc) ) {
		_ReleaseFileH( handle );
		return( __set_errno_dos( TINY_INFO(rc) ) );
	    }
#endif
	}

	/*** Pad the file with zeros if necessary ***/
	if( iomode_flags & _FILEEXT ) {
	    rc = zero_pad( handle );
	    if( rc != 0 ) {
		_ReleaseFileH( handle );
		return( -1 );
	    }
	}
	
	if( iomode_flags & _BINARY ) {	 /* if binary mode */
#ifdef DEFAULT_WINDOWING
	if( _WindowsStdout != 0 &&
		    (res = _WindowsIsWindowedHandle( handle )) != 0 ) {
	    len_written = _WindowsStdout( res, buffer, len );
	    rc = 0;
	} else {
#endif
#if defined(__NT__)
	if( !WriteFile( h, buffer, len, &len_written, NULL ) ) {
	    _ReleaseFileH( handle );
	    return( __set_errno_nt() );
	}
#elif defined(__OS2__)
	    rc = DosWrite( handle, (PVOID)buffer, len, &len_written );
#else
	    rc = TinyWrite( handle, buffer, len );
	    len_written = TINY_LINFO(rc);
#endif
#ifdef DEFAULT_WINDOWING
	}
#endif
#if !defined(__NT__)
	    if( TINY_ERROR(rc) ) {
		_ReleaseFileH( handle );
		return( __set_errno_dos( TINY_INFO(rc) ) );
	    }
#endif
	    if( len_written != len ) {
		__set_errno( ENOSPC );
		len = len_written;
	    }
	} else {
	    i = stackavail();
	    if( i < 0x00b0 ) {
		__STKOVERFLOW();	/* not enough stack space */
	    }
	    buf_size = 512;
	    if( i < (512 + 48) )  buf_size = 128;
	    #if defined(__AXP__) || defined(__PPC__)
		buf = alloca( buf_size );
	    #else
		buf = __alloca( buf_size );
	    #endif
	    i = 0;
	    j = 0;
	    len_written = 0;
	    while( i < len ) {
		if( ((const char*)buffer)[i] == '\n' ) {
		    buf[j] = '\r';
		    ++j;
		    if( j == buf_size ) {
#ifdef DEFAULT_WINDOWING
			if( _WindowsStdout != 0 && (res =
				_WindowsIsWindowedHandle( handle )) != 0 ) {
			    j = _WindowsStdout( res, buf, buf_size );
			    rc = 0;
			} else {
#endif
#ifdef __NT__
			    if( !WriteFile( h, buf, buf_size, &j, NULL ) ) {
				_ReleaseFileH( handle );
				return( __set_errno_nt() );
			    }
#elif defined(__OS2__)
			    rc = DosWrite( handle, buf, buf_size, &j );
#else
			    rc = TinyWrite( handle, buf, buf_size );
			    j = TINY_LINFO(rc);
#endif
#ifdef DEFAULT_WINDOWING
			}
#endif
#if !defined(__NT__)
			if( TINY_ERROR(rc) ) {
			    _ReleaseFileH( handle );
			    return( __set_errno_dos( TINY_INFO(rc) ) );
			}
#endif
			if( j != buf_size ) {
			    __set_errno( ENOSPC );
			    _ReleaseFileH( handle );
			    return( len_written + j );
			}
			len_written = i;
			j = 0;
		    }
		}
		buf[j] = ((const char*)buffer)[i];
		++i;
		++j;
		if( j == buf_size ) {
#ifdef DEFAULT_WINDOWING
		    if( _WindowsStdout != 0 && (res =
		    		_WindowsIsWindowedHandle( handle )) != 0 ) {
			j = _WindowsStdout( res, buf, buf_size );
			rc = 0;
		    } else {
#endif
#if defined(__NT__)
			if( !WriteFile( h, buf, buf_size, &j, NULL ) ) {
			    _ReleaseFileH( handle );
			    return( __set_errno_nt() );
			}
#elif defined(__OS2__)
			rc = DosWrite( handle, buf, buf_size, &j );
#else
			rc = TinyWrite( handle, buf, buf_size );
			j = TINY_LINFO(rc);
#endif
#ifdef DEFAULT_WINDOWING
		    }
#endif
#if !defined(__NT__)
		    if( TINY_ERROR(rc) ) {
			_ReleaseFileH( handle );
			return( __set_errno_dos( TINY_INFO(rc) ) );
		    }
#endif
		    if( j != buf_size ) {
			__set_errno( ENOSPC );
			_ReleaseFileH( handle );
			return( len_written + j );
		    }
		    len_written = i;
		    j = 0;
		}
	    }
	    if( j ) {
#ifdef DEFAULT_WINDOWING
		if( _WindowsStdout != 0 &&
			(res = _WindowsIsWindowedHandle( handle )) != 0 ) {
		    i = _WindowsStdout( res, buf, j );
		    rc = 0;
		} else {
#endif
#if defined(__NT__)
		    if( !WriteFile( h, buf, j, &i, NULL ) ) {
			_ReleaseFileH( handle );
			return( __set_errno_nt() );
		    }
#elif defined(__OS2__)
		    rc = DosWrite( handle, buf, j, &i );
#else
		    rc = TinyWrite( handle, buf, j );
		    i = TINY_LINFO(rc);
#endif
#ifdef DEFAULT_WINDOWING
		}
#endif
#if !defined(__NT__)
		if( TINY_ERROR(rc) ) {
		    _ReleaseFileH( handle );
		    return( __set_errno_dos( TINY_INFO(rc) ) );
		}
#endif
		if( i != j ) {
		    __set_errno( ENOSPC );
		    _ReleaseFileH( handle );
		    return( len_written + i );
		}
	    }
	}
	_ReleaseFileH( handle );
	return( len );
   }


#if defined(__WINDOWS_386__)
#define MAXBUFF 0x8000
_WCRTLINK int write( int handle, const void *buffer, unsigned len )
{
    unsigned	total=0,writeamt;
    int		rc;
    
    __handle_check( handle, -1 );

    // allow file to be truncated
    if( len == 0 ) return( __write( handle, buffer, 0 ) );

    while( len > 0 ) {
	if( len > MAXBUFF ) {
	    writeamt = MAXBUFF;
	} else {
	    writeamt = len;
	}
	rc = __write( handle, buffer, writeamt );
	if( rc < 0 ) return( rc );
	total += (unsigned) rc;
	if( rc != writeamt ) return( total );

	len -= writeamt;
	buffer = ((const char*)buffer) + writeamt;
    }
    return( total );

}
#endif

