/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1990, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  21-dec-87     A.F.Scian       stub for now
  10-jan-88     F.W.Crigger     initial implementation
  02-feb-88     F.W.Crigger     fixed problem when allocating new segment
  16-feb-88     John Dahms      if _nmalloc returns 0, return (void far *)0
                                since (void near *) 0 doesn't convert correctly
-----------------------------------------------------------Version 6.0-------
  21-mar-88     F.W.Crigger     make sure size is >= sizeof(farfrl) in _fmalloc
  03-apr-88     F.W.Crigger     implement new doubly linked free list.
  27-apr-88     F.W.Crigger     Use info returned by alloc_seg in BX when
                                DOS allocation requests fails to speed up rtn.
                                Try to extend previous segment before
                                allocating a new one.
-------------------------------------------------------Version 6.5----------
  09-jun-88     A.F.Scian       - improved coalesce routine
                                - added heap support
                                - used tinyio.h
  14-jun-88     A.F.Scian       when splitting blocks, return slice on the
                                bottom
  15-jul-88     F.W.Crigger     check for overflow past 64K on amount
-------------------------------------------------------Version 7.0----------
  21-mar-89     F.W.Crigger     Added __num_fallocs, __num_ffrees counters
                                for faster ffree function.
  28-mar-89     J.D.Graham      Added _nmemneed hook
  23-may-89     F.W.Crigger     check for 64K allocation in _getseg
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  20-dec-89     John Dahms      check _osmode is DosAllocSeg
  15-feb-90     J.B.Schueler    added QNX support
  04-apr-90     J.B.Schueler    Added QNX realloc support
-------------------------------------------------------Version 8.0----------
  28-dec-90     F.W.Crigger     split out from fmalloc.c
  26-mar-91	A.F.Scian	added 16-bit Windows support
  03-apr-91	C.G.Eisler	updated for WINDOWS_386 macro
  23-may-91	A.F.Scian	fixed to follow rules for Windows 3.1
  24-jul-91	A.F.Scian	made GlobalAlloc use MOVEABLE instead of FIXED
  21-aug-91	A.F.Scian	changed OS/2 DosAlloc to allocate private seg
  				so that OS/2 emulators like PharLap work
  29-oct-91	C.G.Eisler	added __win_alloc_flags
   3-may-92	B.J. Stecher	added __qnx_alloc_flags, moved
   				__win_alloc_flags to seperate file
  93/08/05	B.J. Stecher	added __heap_enabled check.
--------------------------------------------------------Version-11.0---------
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#if defined(__OS2__)
    #include <dos.h>
    #include "tinyos2.h"
    #define INCL_DOSMEMMGR
    #include <wos2.h>
#elif defined(__QNX__)
    #include <sys/types.h>
    #include <sys/seginfo.h>
    #include <i86.h>
#elif defined(__WINDOWS__) && defined(M_I86)
    #include <dos.h>
    #include "windows.h"
#else
    #include <dos.h>
    #include "tinyio.h"
#endif
#include "heap.h"

#if defined(__WINDOWS_286__)
extern unsigned long _WCNEAR __win_alloc_flags;
#elif defined(__QNX__)
extern unsigned __qnx_alloc_flags;
#endif

unsigned __AllocSeg( unsigned int amount )
    {
        unsigned n;             /* number of paragraphs desired   */
        short seg;
        struct heapstart _WCFAR *p;
        tag _WCFAR *last_tag;
#if !(defined(__QNX__) || defined(__WINDOWS_286__) )
        tiny_ret_t rc;
#endif

        if( !__heap_enabled ) return( 0 );
	/*               heapinfo + frl + frl,       end tags */
        if( amount > - (sizeof(struct heapstart) + TAG_SIZE*2 ) ) {
            return( 0 );
        }
	/*        heapinfo + frl,        allocated blk,  end tags */
        amount += sizeof(struct heapblk) + TAG_SIZE + TAG_SIZE * 2;
        if( amount < _amblksiz )  amount = _amblksiz;
        n = ( amount + 0x0f ) >> 4;
        if( n == 0 )  n = 0x1000;                       /* 23-may-89 */
#if defined(__OS2__)
        rc = DosAllocSeg( n << 4, (PSEL)&seg, 0 );	/* 21-aug-91 */
        if( rc != 0 ) return( 0 );
#elif defined(__QNX__)
        seg = qnx_segment_alloc_flags( ((long)n) << 4, __qnx_alloc_flags );
        if( seg == -1 ) return( 0 );
#elif defined(__WINDOWS_286__)
	{
	    HANDLE hmem;
	    LPSTR p;

	    hmem = GlobalAlloc( __win_alloc_flags, ((long)n) << 4 );
	    if( hmem == NULL ) return( 0 );
	    p = GlobalLock( hmem );
	    if( p == NULL ) {
		GlobalFree( hmem );
		return( 0 );
	    }
#if 0	    /* code generator can't handle this */
	    if( FP_OFF( p ) != 0 ) {	/* in case, Microsoft changes Windows */
		GlobalUnlock( hmem );	/* in post 3.1 versions */
		GlobalFree( hmem );
		return( 0 );
	    }
#endif
	    seg = FP_SEG( p );
	}
#else
        rc = TinyAllocBlock( n );
        if( TINY_ERROR( rc ) ) {
            return( 0 );
        }
        seg = TINY_INFO( rc );
#endif
        p = (struct heapstart _WCFAR *)MK_FP( seg, 0 );
        p->h.heaplen = n << 4;
        p->h.prevseg = 0;
        p->h.nextseg = 0;
        p->h.rover   = offsetof( struct heapstart, first );
        p->h.b4rover     = 0;
        p->h.numalloc = 0;
        p->h.numfree  = 1;
        p->h.freehead.len  = 0;
        p->h.freehead.prev = offsetof( struct heapstart, first );
        p->h.freehead.next = offsetof( struct heapstart, first );
        p->first.len  = p->h.heaplen - sizeof(struct heapblk) - TAG_SIZE*2;
        p->h.largest_blk = p->first.len;
        p->first.prev = offsetof( struct heapblk, freehead );
        p->first.next = offsetof( struct heapblk, freehead );
        last_tag = MK_FP( seg, p->h.heaplen - TAG_SIZE*2 );
        *last_tag = END_TAG;
        last_tag[1] = 0;        /* link to next piece of near heap */
        return( seg );          /* return allocated segment */
    }
