/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985,1986 by WATCOM Systems Inc. All rights %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  10-jun-88     A.F.Scian       defined
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  25-jan-90     F.W.Crigger     Set __heap_modified to 0 if heap is OK.
  31-dec-90     F.W.Crigger     New implementation
  26-mar-91	A.F.Scian	fixed multi-thread problem
--------------------------------------------------------Version-11.0---------
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
  14-jan-97	Greg Bentz	Add __fheapchk_current
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <malloc.h>
#include <i86.h>

#include "heap.h"
#include "heapacc.h"

farfrlptr __fheapchk_current;

static int checkFreeList( unsigned long *free_size )
{
    farfrlptr curr;
    unsigned short seg;
    struct heapblk _WCFAR *p;
    unsigned long total_size;

    total_size = 0;
    for( seg = __fheap; seg ;) {
	p = MK_FP( seg, 0 );
	__fheapchk_current = curr = MK_FP( seg, p->freehead.next );
	for(;;) {
	    if( FP_OFF(curr) == offsetof(struct heapblk, freehead) ) break;
	    total_size += curr->len;
	    __fheapchk_current = curr = MK_FP( seg, curr->next );
	}
	seg = p->nextseg;
    }
    *free_size = total_size;
    return( _HEAPOK );
}

static int checkFree( farfrlptr p )
{
    unsigned short seg;
    farfrlptr prev;
    farfrlptr next;
    farfrlptr prev_prev;
    farfrlptr next_next;

    __fheapchk_current = p;
    seg = FP_SEG( p );
    prev = MK_FP( seg, p->prev );
    next = MK_FP( seg, p->next );
    if( prev->next != FP_OFF(p) || next->prev != FP_OFF(p) ) {
	return( _HEAPBADNODE );
    }
    prev_prev = MK_FP( seg, prev->prev );
    next_next = MK_FP( seg, next->next );
    if( prev_prev->next != FP_OFF(prev) || next_next->prev != FP_OFF(next) ) {
	return( _HEAPBADNODE );
    }
    return( _HEAPOK );
}

#if defined(__BIG_DATA__)
_WCRTLINK int _heapchk( void )
{
    return( _fheapchk() );
}
#endif


_WCRTLINK int _fheapchk( void )
{
    struct _heapinfo hi;
    int heap_status;
    unsigned long free_size;

    _AccessFHeap();
    if( __fheap_clean ) {
	_ReleaseFHeap();
	return( _HEAPOK );
    }
    heap_status = checkFreeList( &free_size );
    if( heap_status != _HEAPOK ) {
	_ReleaseFHeap();
	return( heap_status );
    }
    hi._pentry = NULL;
    for(;;) {
	heap_status = __HeapWalk( &hi, __fheap, 0 );
	if( heap_status != _HEAPOK ) break;
	if( hi._useflag == _FREEENTRY ) {
	    heap_status = checkFree( hi._pentry );
	    if( heap_status != _HEAPOK ) break;
	    free_size -= hi._size;
	}
    }
    if( free_size != 0 ) {
	heap_status = _HEAPBADNODE;
    } else if( heap_status == _HEAPBADPTR ) {
	heap_status = _HEAPBADNODE;
    } else {
	if( heap_status == _HEAPEND ) {
	    heap_status = _HEAPOK;
	}
	if( heap_status == _HEAPOK ) {
	    __fheap_clean = 1;
	}
    }
    _ReleaseFHeap();
    return( heap_status );
}
