/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1987-1991, by WATCOM Systems Inc. All rights%
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  14-jun-88     A.F.Scian       defined
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  ------------------------------Version 8.0--------------------------
  23-mar-90     A.F.Scian       fixed problem with an empty heap
                                (p1=p1->next would be invalid because
                                 p1's heap may have been freed!)
  08-aug-90     J.B.Schueler    added support for QNX
  11-oct-90     A.F.Scian       fixed problem with shrinking an almost full seg
  29-jan-91	A.F.Scian	rewritten for new memory manager
  26-mar-91	A.F.Scian	added 16-bit Windows support
  03-apr-91	C.G.Eisler	updated for WINDOWS_386 macro
  23-may-91	A.F.Scian	fixed to follow rules for Windows 3.1
  ------------------------------Version 8.5--------------------------
  26-oct-91	F.W.Crigger	split big chunk out into __HeapMin
  ------------------------------Version 9.0--------------------------
  03-dec-92	F.W.Crigger	update __fheapRover if segment deleted
--------------------------------------------------------Version-11.0---------
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <stdlib.h>
#include <malloc.h>
#if defined(__QNX__)
    #include <sys/seginfo.h>
#elif defined(__OS2__)
    #include <dos.h>
    #define INCL_DOSMEMMGR
    #include <wos2.h>
    #include "tinyos2.h"
    #if defined(M_I86)
	#if defined(__BIG_DATA__)
	    #define MODIFIES ds es
	#else
	    #define MODIFIES es
	#endif
    #endif
#elif defined(__WINDOWS__) 
    #include <dos.h>
    #include "windows.h"
    #define MODIFIES es
#else
    #include <dos.h>
    #include "tinyio.h"
#endif
#include "heap.h"
#include "heapacc.h"


#if defined(__BIG_DATA__)

#if defined(__WINDOWS_286__) || defined(__OS2_286__)
    #pragma aux _heapmin modify [MODIFIES]
    #pragma aux _heapshrink modify [MODIFIES]
#endif

_WCRTLINK int _heapshrink( void )
    {
        return( _fheapshrink() );
    }
_WCRTLINK int _heapmin( void )
    {
        return( _fheapshrink() );
    }

#endif

#if defined(__WINDOWS_286__) || defined(__OS2_286__)
    #pragma aux _fheapmin modify [MODIFIES]
    #pragma aux _fheapshrink modify [MODIFIES]
#endif

_WCRTLINK int _fheapmin( void )
    {
        return( _fheapshrink() );
    }

_WCRTLINK int _fheapshrink( void )
{
    __segment		seg;
    __segment		heap_seg;
    __segment		prev_seg;
    int			heap_status;
    struct heapblk _WCFAR *heap;
    struct heapblk _WCFAR *next_heap;
    struct heapblk _WCFAR *prev_heap;

    heap_status = __HeapMin( __fheap, 0 );
    _AccessFHeap();
    for( seg = __fheap; seg; ) {
	heap = MK_FP( seg, 0 );
	/* we might free this segment so get the next one now */
	heap_seg = seg;
	seg = heap->nextseg;
	if( heap->numalloc == 0 ) {	/* empty heap */
	    /* unlink from heap list */
	    prev_seg = heap->prevseg;
	    if( seg != 0 ) {
		next_heap = MK_FP( seg, 0 );
		next_heap->prevseg = prev_seg;
	    }
	    if( prev_seg == 0 ) {
		__fheap = seg;
	    } else {
		prev_heap = MK_FP( prev_seg, 0 );
		prev_heap->nextseg = seg;
	    }
	    __fheapRover = __fheap;			/* 03-dec-92 */
	    __fheap_clean = 0;
	    heap_status = __FreeSeg( heap_seg );
	}
    }
    _ReleaseFHeap();
    return( heap_status );
}
