/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985,1986 by WATCOM Systems Inc. All rights %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  28-May-86     A.C.C.Wai       Initial implementation
   1-Mar-87     A.C.C.Wai       Changed to new free list direction
-------------------------------------------------------Version 6.0--------
  03-apr-88     F.W.Crigger     new doubly linked format of free storage.
  25-apr-88     F.W.Crigger     turn off allocated bit before coalescing
  29-apr-88     F.W.Crigger     check to see if pointer is part of DGroup
  03-may-88     F.W.Crigger     Forgot to mark second half as free before
                                calling free on it.
-------------------------------------------------------Version 6.5--------
  15-jul-88     F.W.Crigger     check for overflow past 64K on size
                                don't free old stg if can't grow the stg.
  04-aug-88     A.F.Scian       split into _nrealloc & _frealloc
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
-------------------------------------------------------Version 8.0--------
  29-dec-90     F.W.Crigger     New implementation based on _nexpand & _fexpand
  08-feb-91     F.W.Crigger     Check for NULL from _nexpand
--------------------------------------------------------Version-11.0---------
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <malloc.h>
#include <string.h>
#include "heap.h"

#if defined(__386__)
#define MOVSW   0x66 0xa5
#define _DI     edi
#define _SI     esi
#define _CX     ecx
#else
#define MOVSW   0xa5
#define _DI     di
#define _SI     si
#define _CX     cx
#endif


extern void _WCNEAR *_mymemcpy( void _WCFAR *, void _WCFAR *, size_t );
#pragma aux _mymemcpy = \
        0x1e            /* push ds */ \
        0x8e 0xda       /* mov ds,dx */ \
        0xd1 0xe9       /* shr cx,1 */ \
        0xf3 MOVSW      /* rep movsw */ \
        0x11 0xc9       /* adc cx,cx */ \
        0xf3 0xa4       /* rep movsb */ \
        0x1f            /* pop ds */ \
        parm caller     [es _DI] [dx _SI] [_CX] \
        value           [_SI] \
        modify exact    [_SI _DI _CX];


#if defined(__BIG_DATA__)

_WCRTLINK void *realloc( void *stg, size_t amount )
    {
        return( _frealloc( stg, amount ) );
    }

#endif


_WCRTLINK void _WCFAR *_frealloc( void _WCFAR *stg, size_t req_size )
    {
        size_t    old_size;
        void _WCFAR *p;

        if( stg == NULL ) {
            return( _fmalloc( req_size ) );
        }
        if( req_size == 0 ) {
            _ffree( stg );
            return( NULL );
        }
        old_size = _fmsize( stg );
        if( FP_SEG(stg) == _DGroup() ) {
            p = stg;
            if( _nexpand( (void _WCNEAR *) stg, req_size ) == NULL ) p = NULL;
        } else {
            p = _fexpand( stg, req_size );
        }
        if( p == NULL ) {       /* couldn't be expanded inline */
            p = _fmalloc( req_size );
            if( p != NULL ) {
                _mymemcpy( p, stg, old_size );
                _ffree( stg );
            } else {
                if( FP_SEG(stg) == _DGroup() ) {
                    _nexpand( (void _WCNEAR *) stg, old_size );
                } else {
                    _fexpand( stg, old_size );
                }
            }
        }
        return( p );
    }
