/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1991, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  14-jun-88     A.F.Scian       defined
  ------------------------------Version 8.0--------------------------
  08-aug-90     J.B.Schueler    added support for QNX
  26-mar-91	A.F.Scian	added 16-bit Windows support
  03-apr-91	C.G.Eisler	updated for WINDOWS_386 macro
  23-may-91	A.F.Scian	fixed to follow rules for Windows 3.1
  ------------------------------Version 8.5--------------------------
  28-oct-91	F.W.Crigger	split out from _fheapmin
  29-apr-94	Greg Bentz	added __FreeWinSeg()
--------------------------------------------------------Version-11.0---------
  27-mar-95	Greg Bentz	use __DoFreeSeg() for OS/2 1.x too
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  16-may-95     T. Schiller     modified __DoFreeSeg() to work with
                                multi-segment huge allocations under Win-16.
  16-may-95     T. Schiller     Added reference to crwd.h.
  28-may-95	Greg Bentz	dll linkage
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include "rtdata.h"
#include "crwd.h"
#if defined(__QNX__)
    #include <sys/seginfo.h>
#elif defined(__OS2__)
    #include <dos.h>
    #define INCL_DOSMEMMGR
    #include <wos2.h>
    #include "tinyos2.h"
    #if defined(__BIG_DATA__)
	#define MODIFIES ds es
    #else
	#define MODIFIES es
    #endif
#elif defined(__WINDOWS_286__)
    #include <dos.h>
    #include "windows.h"
    #define MODIFIES es
#else
    #include <dos.h>
    #include "tinyio.h"
#endif
#include "heap.h"
#include "seterrno.h"

#if defined(__WINDOWS_286__) || defined(__OS2__)

#pragma aux DoFreeSeg modify [MODIFIES]
static int DoFreeSeg( __segment seg )
{
    #if defined(__WINDOWS_286__)
	HANDLE hmem;
    
	hmem = GlobalHandle( seg );
	if( hmem == NULL ) {
	    return( -1 );
	}
	GlobalUnlock( hmem );
	if( GlobalFree( hmem ) == hmem ) {
	    return( -1 );
	}
	return( 0 );
    #else
	return( DosFreeSeg( seg ) );
    #endif
}

extern int tricky_free_seg( int, int );
#if defined(__OS2__) && defined(__BIG_DATA__)
#pragma aux tricky_free_seg = \
	"mov	cx,es"		\
	"cmp	cx,ax"		\
	"jl     L1"		\
	"cmp	cx,bx"		\
	"jg     L1"		\
	"xor	cx,cx"		\
	"L1:"			\
	"mov	dx,ds"		\
	"cmp	dx,ax"		\
	"jl     L2"		\
	"cmp	dx,bx"		\
	"jg     L2"		\
	"xor	dx,dx"		\
	"L2:"			\
	"call	DoFreeSeg"	\
	"mov	es,cx"		\
	"mov	ds,dx"		\
	parm [ax] [bx] value [ax] modify [cx dx]
#else
#pragma aux tricky_free_seg = \
	"mov	cx,es"		\
	"cmp	cx,ax"		\
	"jl     L1"		\
	"cmp	cx,bx"		\
	"jg     L1"		\
	"xor	cx,cx"		\
	"L1:"			\
	"call	DoFreeSeg"	\
	"mov	es,cx"		\
	parm [ax] [bx] value [ax] modify [cx]
#endif
	
#pragma aux __DoFreeSeg modify [MODIFIES]
int __DoFreeSeg(__segment first)
{
	__segment     last;
	long          segments;
	unsigned long size;
#ifdef __OS2__
	DosSizeSeg(first, &size);
#else
	size = GlobalSize(first);
#endif
	segments = (size + 65535L) >> 16;
	last = first + (__segment)((segments - 1L) << _RWD_HShift);

	return (tricky_free_seg(first, last));
}
#endif


#if defined(__WINDOWS_286__) || defined(__OS2__)
#pragma aux __FreeSeg modify [MODIFIES]
#endif
int __FreeSeg( __segment seg )
{
    #if defined(__QNX__)
	if( qnx_segment_free( seg ) == -1 ) {
	    return( -1 );
	}
    #elif defined(__WINDOWS_286__)
	if( __DoFreeSeg( seg ) ) {
	    return( -1 );
	}
    #else
	tiny_ret_t rc;

	#if defined(__OS2__)
	    rc = __DoFreeSeg( seg );
	#else
	    rc = TinyFreeBlock( seg );
	#endif
	if( TINY_ERROR( rc ) ) {
	    return( __set_errno_dos( TINY_INFO( rc ) ) );
	}
    #endif
    return( 0 );
}
