/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985,1992 by WATCOM Systems Inc. All rights %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  18-Feb-87     A.C.C.Wai       From older malloc.c, but allocation goes up.
  03-Mar-87     A.C.C.Wai       Added heap expansion scheme
  10-jan-88     F.W.Crigger     Adapted for different memory models
  09-mar-88     F.W.Crigger     added #include for <stdlib.h> to get proper
                                prototype for sbrk to fix bug in large
                                data models.
--------------------------------------------------------Version 6.0--------
  03-apr-88     F.W.Crigger     use the new brk() function.
                                free list made doubly linked.
                                and low order bit indicates allocated or not.
  12-apr-88     F.W.Crigger     _nfree checks to make sure block not already
                                indicated as free.
  08-may-88     F.W.Crigger     split away from malloc.c
-------------------------------------------------------Version 6.5----------
  09-jun-88     A.F.Scian       - improved coalesce routine
                                - added heap support
  14-jun-88     A.F.Scian       when splitting blocks, return slice on the
                                bottom
  15-jul-88     F.W.Crigger     check for overflow past 64K on amount
  03-aug-88     A.F.Scian       removed large data restrictions on DS=64k
--------------------------------------------------------Version 7.0--------
  06-mar-89     F.W.Crigger     Make int size independent for 386
  20-mar-89     F.W.Crigger     Update __num_allocs and __num_frees variables
  28-mar-89     J.D.Graham      Added _nmemneed hook
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  18-oct-89     F.W.Crigger     _expand (account for free block at end of heap)
  02-dec-89     F.W.Crigger     __DynRover added to speed up memory allocations
  25-jan-90     F.W.Crigger     start lookup at start of free list if amount
                                being allocated is < 16
  26-jan-90     F.W.Crigger     keep track of largest block before __DynRover
  07-feb-90     A.F.Scian       changes in _nmalloc & _expand adjusted
                                expansion of DGROUP if a free block at the
                                end of the heap could be used, of course,
                                _nheapgrow depended on this and ceased working
                                MORAL: any changes to DGROUP expansion
                                must verify that _nheapgrow still works!!!
                                fix: added __LastFree function
  20-dec-90     J.B.Schueler    brk changed to __brk
  28-dec-90     F.W.Crigger     increment __num_allocs in __ExpandDGROUP()
  22-feb-91     A.F.Scian       tweaked __ExpandDGROUP()
                                - adjusted extra amounts added to allocation
                                - cannot assume 'amount' is even
                                - if size of last free block is equal to
                                  expansion amount problems arose
  18-mar-91     A.F.Scian       tweaked __ExpandDGROUP()
                                - changed +1 & ~1 code to use sizeof( tag )
                                  for the 386 libraries
  21-mar-91     C.G.Eisler      removed extender for 386 windows
  26-mar-91     A.F.Scian       added 16-bit Windows support
                                - fixed bug if a new heap block was too small
  26-mar-91     C.G.Eisler      removed extender for 386 windows (AGAIN!)
  02-apr-91	A.F.Scian	added 32-bit Rational support
  03-apr-91	C.G.Eisler	updated for WINDOWS_386 macro
  15-apr-91	A.F.Scian	added 32-bit Intel CodeBuilder support
  21-may-91	C.G.Eisler	fixed minimum block allocation
  14-jun-91	A.F.Scian	adjusted insertion of heap blocks so that
  				heap blocks could be inserted before previous
				heap blocks
  08-jul-91	G. Coschi	added OS/2 2.0 support
  19-jul-91	A.F.Scian	added allocation from DOS memory for Rational
  07-aug-91	John Dahms	don't ask for more than 64K DOS memory under
  				in a DPMI environment. It ignore the hi word.
  02-mar-92	A.F.Scian	fixed check for 64k so that exactly 64k is OK
--------------------------------------------------------Version 9.x--------
  06-mar-92	Craig Eisler	NT support
  30-jun-92	B.J. Stecher 	used __DOS_EXT__ macro
  21-jul-92	B.J. Stecher	split out from NMALLOC.C
--------------------------------------------------------Version 9.5--------
  05-aug-93	B.J. Stecher	added __heap_enabled check.
  17-sep-93	F.W.Crigger	Added support for TNT DOS Extender v6.0
  08-nov-93	F.W.Crigger	added better support for DPMI environments
  09-nov-93	F.W.Crigger	heaps were not in proper ascending order
  15-nov-93	F.W.Crigger	Added __ExpandDPMIBlock for Rational
  11-feb-94	Greg Bentz	support Rationals non-zero based data
  				pages that allow the segment to grow
  19-feb-94	F.W.Crigger	__LastFree and __ExpandDGROUP:
				add check for __X32VM
  24-feb-94	Kevin Hui	__ExpandDGroup() now calls __CreateNewNHeap()
  				for systems that cannot grow the near heap by
				slicing. Near heap now uses mini-heaps data
				structure. Modified the DPMI functions to work
				for this new data structure. Modified the
				pragma for SegmentLimit.
  10-may-94	Greg Bentz	consolidate code to round up to nearest 4k
  				boundary and check for overflow of size
  24-may-94	Greg Bentz	add version checking to lsl instruction
  				because inline assembler is different between
				version 9.5 and 10.0
  15-aug-94	Greg Bentz	fixes in __ReallocDPMIBlocks()
  				- improve check for containing only one block
				- ensure that spare free block uses even size
				- fix computation for new required size
--------------------------------------------------------Version-11.0---------
  22-feb-95	Greg Bentz	alpha support
  06-mar-95	Greg Bentz	bugs in MiniHeapRover code
  06-mar-95	Greg Bentz	miniheaps may not be in order
  14-mar-95	M. Hildebrand	Reference __LargestSizeB4MiniHeapRover, not
  				__LargestSizeB4Rover.
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
  11-jun-95	F.W.Crigger	__CreateNewNHeap - for OSI subtract extra tag
  08-aug-95	T. Schiller	modified to use VirtualAlloc() instead of
				LocalAlloc()
  07-sep-95	Greg Bentz	__unlink needs to update free rover
*/

#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <stdlib.h>
#include <malloc.h>
#include "heapacc.h"
#include "heap.h"
#include <errno.h>
#ifdef _M_IX86
 #include <i86.h>
#endif
#if defined(__DOS_EXT__)
 #include "extender.h"
 #include "tinyio.h"
#endif
#if defined(__CALL21__)
 #include "tinyio.h"
#endif
#if defined(__WINDOWS_286__) || defined(__NT__)
 #include "windows.h"
#endif
#if defined(__OS2__)
 #include <wos2.h>
#endif
#if defined(__WINDOWS_386__)
 extern void * pascal DPMIAlloc(unsigned long);
#endif

static frlptr __LinkUpNewMHeap( mheapptr );

#if defined(__DOS_EXT__)
extern	int			SegmentLimit();
#if __WATCOMC__ > 950
#pragma aux SegmentLimit	= \
	"xor	eax,eax"	\
	"mov	ax,ds"		\
	"lsl	eax,ax"		\
	"inc	eax"		\
	value			[eax] \
	modify exact		[eax];
#else
#pragma aux SegmentLimit	= \
	"xor	eax,eax"	\
	"mov	ax,ds"		\
	"lsl	eax,eax"	\
	"inc	eax"		\
	value			[eax] \
	modify exact		[eax];
#endif

static void __unlink( mheapptr miniheapptr )
{
    mheapptr		prev_link;
    mheapptr		next_link;

    if( __nheapbeg == miniheapptr ) {
	__nheapbeg = miniheapptr->next;
    }
    if( miniheapptr == __MiniHeapRover ) {
	__MiniHeapRover = miniheapptr->prev;
	if( __MiniHeapRover == NULL ) {
	    __MiniHeapRover = __nheapbeg;
	    __LargestSizeB4MiniHeapRover = 0;
	}
    }
    if( miniheapptr == __MiniHeapFreeRover ) {
	__MiniHeapFreeRover = 0;
    }
    prev_link = miniheapptr->prev;
    next_link = miniheapptr->next;
    if( prev_link != NULL )  prev_link->next = next_link;
    if( next_link != NULL )  next_link->prev = prev_link;
}

void __FreeDPMIBlocks()
{
    mheapptr		mhp;
    struct dpmi_hdr	*dpmi;

    mhp = __nheapbeg;
    while( mhp != NULL ) {
	// see if the last free entry has the full size of
	// the DPMI block ( - overhead).  If it is then we can give this
	// DPMI block back to the DPMI host.
	if( (mhp->freehead.prev)->len + sizeof(struct miniheapblkp) ==
	    mhp->len ) {
	    mheapptr	pnext;

	    dpmi = ((struct dpmi_hdr *)mhp) - 1;
	    pnext = mhp->next;
	    __unlink( mhp );
	    mhp = pnext;
	    if( dpmi->dos_seg_value == 0 ) {	// if DPMI block
		TinyDPMIFree( dpmi->dpmi_handle );
	    } else {				// else DOS block below 1MB
		TinyFreeBlock( dpmi->dos_seg_value );
	    }
	} else {
	    mhp = mhp->next;
	}
    }
}

void *__ReAllocDPMIBlock( frlptr p1, unsigned req_size )
{
    mheapptr		mhp;
    struct dpmi_hdr	*dpmi;
    struct dpmi_hdr	*prev_dpmi;
    unsigned		size;
    frlptr		flp, flp2;

    if( !__heap_enabled ) return( 0 );
    __FreeDPMIBlocks();
    prev_dpmi = NULL;
    for( mhp = __nheapbeg; mhp; mhp = mhp->next ) {
	if( ((PTR)mhp + sizeof(struct miniheapblkp) == (PTR)p1)
	  && (mhp->numalloc == 1) ) {
	    // The mini-heap contains only this memblk
	    __unlink( mhp );
	    dpmi = ((struct dpmi_hdr *)mhp) - 1;
	    if( dpmi->dos_seg_value != 0 ) return( NULL );
	    size = mhp->len + sizeof(struct dpmi_hdr) + TAG_SIZE;
	    size += ( req_size - (p1->len-TAG_SIZE) );
	    size += 0x0fff;
	    size &= ~0x0fff;
	    prev_dpmi = dpmi;
	    dpmi = TinyDPMIRealloc( dpmi, size );
	    if( dpmi == NULL ) {
		dpmi = prev_dpmi;
		return( NULL );		// indicate resize failed
	    }
	    dpmi->dos_seg_value = 0;
	    mhp = (mheapptr)( dpmi + 1 );
	    mhp->len = size - sizeof(struct dpmi_hdr) - TAG_SIZE;
	    flp = __LinkUpNewMHeap( mhp );
	    mhp->numalloc = 1;

	    // round up to even number
	    req_size = (req_size + 1) & ~1;
	    size = flp->len - req_size;
	    if( size >= FRL_SIZE ) {	// Enough to spare a free block
		flp->len = req_size | 1;// adjust size and set allocated bit
		// Make up a free block at the end
		flp2 = (frlptr)((PTR)flp + req_size);
		flp2->len = size | 1;
		++mhp->numalloc;
		mhp->largest_blk = 0;
		_nfree( (PTR)flp2 + TAG_SIZE );
	    } else {
		flp->len |= 1; // set allocated bit
	    }
	    return( flp );
	}
    }
    return( NULL );
}
#endif

static frlptr __LinkUpNewMHeap( mheapptr p1 ) // originally __AddNewHeap()
{
    mheapptr	p2;
    mheapptr	p2_prev;
    tag		*last_tag;
    unsigned	amount;

    /* insert into ordered heap list (14-jun-91 AFS) */
    /* logic wasn't inserting heaps in proper ascending order */
    /* (09-nov-93 Fred) */
    p2_prev = NULL;
    for( p2 = __nheapbeg; p2 != NULL; p2 = p2->next ) {
	if( p1 < p2 ) break;
	p2_prev = p2;
    }
    /* ascending order should be: p2_prev < p1 < p2  */
    /* except for special cases when p2_prev and/or p2 are NULL */
    p1->prev = p2_prev;
    p1->next = p2;
    if( p2_prev != NULL ) {
	p2_prev->next = p1;
    } else {		/* add p1 to beginning of heap */
	__nheapbeg = p1;
    }
    if( p2 != NULL ) {
	/* insert before 'p2' (list is non-empty) */
	p2->prev = p1;
    }
    amount = p1->len - sizeof( struct miniheapblkp );
    /* Fill out the new miniheap descriptor */
    p1->freehead.len = 0;
    p1->freehead.prev = &p1->freehead;
    p1->freehead.next = &p1->freehead;
    p1->rover = &p1->freehead;
    p1->b4rover = 0;
    p1->numalloc = 0;
    p1->numfree  = 0;
    p1++;
    ((frlptr)p1)->len = amount;
    /* fix up end of heap links */
    last_tag = (tag *) ( (PTR)p1 + amount );
    *last_tag = END_TAG;
    return( (frlptr) p1 );
}

#if ! ( defined(__WINDOWS_286__) || defined(__WARP__) || \
        defined(__NT__) || defined(__WINDOWS_386__) )
size_t __LastFree( void )    /* used by nheapgrow to know about adjustment */
    {
        frlptr p1;
        unsigned brk_value;

        if( __nheapbeg == NULL ) {      /* no heap? can't have free blocks */
            return( 0 );
        }
        p1 = __nheapbeg->freehead.prev;        /* point to last free block */
        brk_value = (unsigned)((PTR)p1 + p1->len + TAG_SIZE );
	#if defined(__DOS_EXT__)
	if( _IsPharLap() && !__X32VM) _curbrk = SegmentLimit(); /*19-feb-94*/
	#endif
        if( brk_value == _curbrk ) {   /* if last free block is at the end */
            return( p1->len );
        }
        return( 0 );
    }
#endif

#if ! defined(__CALL21__)
#if defined(__DOS_EXT__)
static void *RationalAlloc( size_t size )
{
    struct dpmi_hdr	*dpmi;
    mheapptr		mhp;
    tiny_ret_t		save_DOS_block;
    tiny_ret_t		DOS_block;

    __FreeDPMIBlocks();
    /* size is a multiple of 4k */
    dpmi = TinyDPMIAlloc( size );
    if( dpmi != NULL ) {
	mhp = (mheapptr)( dpmi + 1 );
	mhp->len = size - sizeof( struct dpmi_hdr );
	dpmi->dos_seg_value = 0;	// indicate DPMI block
	return( (void *)mhp );
    }
    if( __minreal & 0xfff00000 ) {
	/* checks for users that want >1M real memory saved */
	__minreal = 0xfffff;
    }
    if( size > 0x00010000 ) {
	/* cannot allocate more than 64k from DOS real memory */
	return( NULL );
    }
    save_DOS_block = TinyAllocBlock(( __minreal >> 4 ) | 1 );
    if( TINY_OK( save_DOS_block ) ) {
	DOS_block = TinyAllocBlock( size >> 4 );
	TinyFreeBlock( save_DOS_block );
	if( TINY_OK( DOS_block ) ) {
	    dpmi = (struct dpmi_hdr *) TinyDPMIBase( DOS_block );
	    dpmi->dos_seg_value = DOS_block;
	    mhp = (mheapptr)( dpmi + 1 );
	    mhp->len = size - sizeof( struct dpmi_hdr );
	    return( (void *)mhp );
	}
    }
    return( NULL );
}
#endif
#endif

#if defined(__WARP__) || defined(__NT__) || defined(__CALL21__) || \
    defined(__WINDOWS_386__) || defined(__WINDOWS_286__) || \
    defined(__DOS_EXT__)
static int __CreateNewNHeap( unsigned amount )
{
        mheapptr	p1;
	frlptr		flp;
        unsigned	brk_value;
#if defined(__WARP__)
	PBYTE	p;
#endif
        if( !__heap_enabled ) return( 0 );
        if( _curbrk == ~1u ) return( 0 );
	if( __AdjustAmount( &amount ) == 0 ) return( 0 );
#if defined(__WARP__)
	if( DosAllocMem( &p, amount, PAG_COMMIT|PAG_READ|PAG_WRITE ) ) {
	    return( 0 );
	}
	brk_value = (unsigned)p;
#elif defined(__NT__)
        brk_value = (unsigned) VirtualAlloc(NULL, amount, MEM_COMMIT,
					    PAGE_EXECUTE_READWRITE);
        //brk_value = (unsigned) LocalAlloc( LMEM_FIXED, amount );
        if( brk_value == 0 ) {
            return( 0 );
        }
#elif defined(__WINDOWS_386__)
        brk_value = (unsigned) DPMIAlloc( amount );
        if( brk_value == 0 ) {
            return( 0 );
        }
#elif defined(__WINDOWS_286__)
        brk_value = (unsigned) LocalAlloc( LMEM_FIXED, amount );
        if( brk_value == 0 ) {
            return( 0 );
        }
#elif defined(__CALL21__)
	{
	    tag _WCNEAR *tmp_tag;

	    tmp_tag = (tag _WCNEAR *)TinyMemAlloc( amount );
	    if( tmp_tag == NULL ) {
		return( 0 );
	    }
	    /* make sure it will not look like the end of a heap */
	    tmp_tag[0] = ! END_TAG;
	    brk_value = (unsigned) &tmp_tag[2];
	    amount -= 2 * TAG_SIZE; // 11-jun-95, subtract extra tag
	}
#elif defined(__DOS_EXT__)
	// if( _IsRationalZeroBase() || _IsCodeBuilder() ) {
	{
	    tag		*tmp_tag;

	    if( _IsRational() ) {
		tmp_tag = RationalAlloc( amount );
		if( tmp_tag ) amount = *tmp_tag;
	    } else {	/* CodeBuilder */
		tmp_tag = TinyCBAlloc( amount );
		amount -= TAG_SIZE;
	    }
	    if( tmp_tag == NULL ) {
		return( 0 );
	    }
	    brk_value = (unsigned) tmp_tag;
	}
	// Pharlap, RSI/non-zero can never call this function
#endif
        if( amount - TAG_SIZE > amount ) {
            return( 0 );
        } else {
            amount -= TAG_SIZE;
        }
	if( amount < sizeof( struct miniheapblkp ) + sizeof( frl ) ) {
	    /* there isn't enough for a heap block (struct miniheapblkp) and
	       one free block (frl) */
	    return( 0 );
	}
	/* we've got a new heap block */
	p1 = (mheapptr) brk_value;
	p1->len = amount;
	// Now link it up
	flp = __LinkUpNewMHeap( p1 );
	amount = flp->len;
        /* build a block for _nfree() */
        flp->len = amount | 1;
        ++p1->numalloc;				/* 28-dec-90 */
        p1->largest_blk = 0;
        _nfree( (PTR)flp + TAG_SIZE );
        return( 1 );
}
#endif

int __ExpandDGROUP( unsigned amount )
    {
    #if defined(__WARP__) || defined(__NT__) || defined(__CALL21__) || \
	defined(__WINDOWS_386__) || defined(__WINDOWS_286__)
	// first try to free any available storage
	_nheapshrink();
	return( __CreateNewNHeap( amount ) );
    #else
    	mheapptr	p1;
	frlptr		flp;
        unsigned brk_value;
        tag	*last_tag;
        unsigned new_brk_value;
        void _WCNEAR *brk_ret;

	#if defined(__DOS_EXT__)
 	if( ( _IsRationalZeroBase() || _IsCodeBuilder() ) ) {
	    return( __CreateNewNHeap( amount ) );	// Won't slice either
 	}
	// Rational non-zero based system should go through.
	#endif
        if( !__heap_enabled ) return( 0 );
        if( _curbrk == ~1u ) return( 0 );
	if( __AdjustAmount( &amount ) == 0 ) return( 0 );
	#if defined(__DOS_EXT__)
	if( _IsPharLap() && !__X32VM ) {	/* 19-feb-94 */
	    _curbrk = SegmentLimit();
	}
	#endif
        new_brk_value = amount + _curbrk;
        if( new_brk_value < _curbrk ) {
            new_brk_value = ~1u;
        }
        brk_ret = __brk( new_brk_value );
        if( brk_ret == (void _WCNEAR *) -1 ) {
            return( 0 );
        }
        brk_value = (unsigned) brk_ret;
        if( brk_value >  /*0xfff8*/ ~7u ) {
            return( 0 );
        }
        if( new_brk_value <= brk_value ) {
            return( 0 );
        }
        amount = new_brk_value - brk_value;
        if( amount - TAG_SIZE > amount ) {
            return( 0 );
        } else {
            amount -= TAG_SIZE;
        }
	for( p1 = __nheapbeg; p1 != NULL; p1 = p1->next ) {
	    if( p1->next == NULL ) break;
	    if( (unsigned)p1 <= brk_value &&
	        ((unsigned)p1)+p1->len+TAG_SIZE >= brk_value ) break;
	}
        if( (p1 != NULL) &&
	((brk_value - TAG_SIZE) == (unsigned)( (PTR)p1 + p1->len) ) ) {
		/* we are extending the previous heap block (slicing) */
		/* nb. account for the end-of-heap tag */
		brk_value -= TAG_SIZE;
		amount += TAG_SIZE;
		flp = (frlptr) brk_value;
		/* adjust current entry in heap list */
		p1->len += amount;
		/* fix up end of heap links */
		last_tag = (tag *) ( (PTR)flp + amount );
		last_tag[0] = END_TAG;
        } else {
            if( amount < sizeof( struct miniheapblkp ) + sizeof( frl ) ) {
            /*  there isn't enough for a heap block (struct miniheapblkp) and
		one free block (frl) */
                return( 0 );
            }
	    // Initializing the near heap if __nheapbeg == NULL,
	    // otherwise, a new mini-heap is getting linked up
	    p1 = (mheapptr) brk_value;
	    p1->len = amount;
	    flp = __LinkUpNewMHeap( p1 );
	    amount = flp->len;
 	}
        /* build a block for _nfree() */
        flp->len = amount | 1;
        ++p1->numalloc;				/* 28-dec-90 */
        p1->largest_blk = ~0;    /* set to largest value to be safe */
        _nfree( (PTR)flp + TAG_SIZE );
        return( 1 );
    #endif
    }

static int __AdjustAmount( unsigned *amount )
    {
	unsigned amt;

	amt = *amount;
        amt = ( amt + ROUND_SIZE) & ~ROUND_SIZE;
        if( amt == 0 ) return( 0 );
#if ! ( defined(__WINDOWS_286__) || defined(__WARP__)|| defined(__NT__)  ||defined(__WINDOWS_386__))
#if defined(__DOS_EXT__)
	if( _IsRationalZeroBase() || _IsCodeBuilder() ) {
	    // Allocating extra to identify the dpmi block
 	    amt += sizeof(struct dpmi_hdr);
	} else
#endif
	    amt -= __LastFree();         /* adjust for last free block */
#endif
        /* amount is even here */
        /*
          extra amounts        (22-feb-91 AFS)

           (1) adding a new heap needs:
               frl                    free block req'd for _nmalloc request
	       			      (frl is the MINIMUM because the block
				      may be freed)
               tag                    end of miniheap descriptor
               struct miniheapblkp    start of miniheap descriptor
           (2) extending heap needs:
               tag               free block req'd for _nmalloc request
        */
	*amount = amt;
        amt += ((TAG_SIZE)+sizeof(frl)+sizeof(struct miniheapblkp));
        if( amt < *amount ) return( 0 );
        if( amt < _amblksiz ) {
            /*
              _amblksiz may not be even so round down to an even number
              nb. pathological case: where _amblksiz == 0xffff, we don't
                                     want the usual round up to even
            */
            amt = _amblksiz & ~1u;
        }
	#if defined(__WARP__) || defined(__NT__) || defined(__CALL21__) || \
	    defined(__WINDOWS_386__) || defined(__DOS_EXT__)
	    /* make sure amount is a multiple of 4k */
	    *amount = amt;
	    amt += 0x0fff;
	    if( amt < *amount ) return( 0 );
	    amt &= ~0x0fff;
	#endif
	*amount = amt;
	return( *amount != 0 );
    }
