/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  30-oct-87     F.W.Crigger     Initial implementation
  15-mar-90     F.W.Crigger     check for len too large
  13-mar-91     F.W.Crigger     Rewrite so that it works in real or prot mode
  26-mar-91     A.F.Scian       integrated Windows support
  23-may-91	A.F.Scian	fixed Windows support to follow strict Windows
  				rules
  24-jul-91	A.F.Scian	made GlobalAlloc use MOVEABLE instead of FIXED
  28-sep-93	A.F.Scian	return NULL if final size > 64k and size isn't
  				a power of 2
  29-apr-94	Greg Bentz	added __FreeWinSeg()
  28-mar-95	Greg Bentz	changed __FreeWinSeg() to __FreeSeg()
--------------------------------------------------------Version-11.0---------
  28-may-95	Greg Bentz	variety.h and dll linkage
*/

#include "variety.h"
#include <dos.h>
#include <string.h>
#if defined(__WINDOWS__)
  #include "windows.h"
#endif
#include "heap.h"

extern  long _dosalloc(unsigned);
#pragma aux     _dosalloc = 0xb4 0x48   /* mov ah,48h */        \
                            0xcd 0x21   /* int 21h    */        \
                            0x19 0xd2   /* sbb dx,dx  */        \
                        parm caller [bx] value [dx ax] modify [dx ax];

extern  void    _dosfree(void _WCHUGE *);
#pragma aux     _dosfree  = 0xb4 0x49   /* mov ah,49h */        \
                            0xcd 0x21   /* int 21h    */        \
                        parm caller [ax es] modify [es];

#define HUGE_NULL	((void _WCHUGE *)NULL)

static int only_one_bit( size_t x )
{
    if( x == 0 ) {
	return 0;
    }
    /* turns off lowest 1 bit and leaves all other bits on */
    if(( x & ( x - 1 )) != 0 ) {
	return 0;
    }
    /* only one bit was on! */
    return 1;
}

_WCRTLINK void _WCHUGE * halloc( unsigned long n, unsigned size )
    {
#if defined(__WINDOWS__)
	HANDLE hmem;
	LPSTR p;
        
        n *= size;
        if( n == 0 ) return( HUGE_NULL );
	if( n > 65536 && ! only_one_bit( size ) ) return( HUGE_NULL );
        hmem = GlobalAlloc( GMEM_MOVEABLE|GMEM_ZEROINIT, n );
	if( hmem == NULL ) return( HUGE_NULL );
	p = GlobalLock( hmem );
        return( p );
#else
        long seg;
        unsigned int paras;
        unsigned int len;
        char _WCHUGE *hp;

        n *= size;
        if( n == 0  || n >= 0x100000 ) return( HUGE_NULL );
	if( n > 65536 && ! only_one_bit( size ) ) return( HUGE_NULL );
        paras = (n + 15) >> 4;
        seg = _dosalloc( paras );
        if( seg < 0 ) return( HUGE_NULL );  /* allocation failed */
        hp = (char _WCHUGE *)MK_FP( (unsigned short)seg, 0 );
        for(;;) {
            len = 0x8000;
            if( paras < 0x0800 )  len = paras << 4;
            _fmemset( hp, 0, len );
            if( paras < 0x0800 )  break;
            hp = hp + len;
            paras -= 0x0800;
        }
        return( (void _WCHUGE *)MK_FP( (unsigned short)seg, 0 ) );
#endif
    }

#if defined(__WINDOWS__)
  #pragma aux hfree modify [es]
#endif
_WCRTLINK void hfree( void _WCHUGE *ptr )
    {
#if defined(__WINDOWS__)
	__FreeSeg( FP_SEG( ptr ) );
#else
        _dosfree( ptr );
#endif
    }
