/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  09-aug-89     John Dahms      defined for OS/2
  20-dec-89     John Dahms      check _osmode is DosAllocSeg
  30-apr-91	F.W.Crigger	check to see if last segment is valid
  13-jun-91	F.W.Crigger	if alloc fails, try again with flags = 0
  				Phar Lap 286 Extender requires flags = 0
  21-aug-91	A.F.Scian	simplified 13-jun-91 to always use flags = 0
  28-sep-93	A.F.Scian	return NULL if final size > 64k and size isn't
  				a power of 2
  27-mar-95	Greg Bentz	use __freeseg()
--------------------------------------------------------Version-11.0---------
  28-may-95	Greg Bentz	variety.h and dll linkage
*/

#include "variety.h"
#include "dos.h"
#include "seterrno.h"
#define INCL_DOSMEMMGR
#include <wos2.h>
#include <stdlib.h>
#include "heap.h"


void _os2zero64k( unsigned ax, unsigned cx, unsigned es, unsigned di );
#pragma aux _os2zero64k = 0xf3 0xab /* rep stosw */     \
        parm caller [ ax ] [ cx ] [ es ] [ di ]                     \
        modify [ cx di ];

void _os2zero_rest( unsigned ax, unsigned cx, unsigned es, unsigned di );
#pragma aux _os2zero_rest = 0xf3 0xaa /* rep stosb */   \
        parm caller [ ax ] [ cx ] [ es ] [ di ]                     \
        modify [ cx di ];


static int only_one_bit( size_t x )
{
    if( x == 0 ) {
	return 0;
    }
    /* turns off lowest 1 bit and leaves all other bits on */
    if(( x & ( x - 1 )) != 0 ) {
	return 0;
    }
    /* only one bit was on! */
    return 1;
}

_WCRTLINK void _WCHUGE * halloc( unsigned long n, unsigned size )
{
    unsigned long len;
    USHORT      error, tseg;
    SEL         seg;
    USHORT      number_segments, remaining_bytes;
    USHORT      increment;

    len = n * size;
    if( len == 0 ) return( (void _WCHUGE *)0 );
    if( n > 65536 && ! only_one_bit( size ) ) return( (void _WCHUGE *)0 );
    error = DosGetHugeShift( &increment );
    if( error ) {
        __set_errno_dos( error );
        return( ( void _WCHUGE *)0 );
    }
    number_segments = len >> 16;
    remaining_bytes = len & 0xffff;
    error = DosAllocHuge( number_segments, remaining_bytes, &seg, 0, 0 );
    if( error ) {
        __set_errno_dos( error );
        return( (void _WCHUGE *)0 );  /* allocation failed */
    }
    tseg = seg;
    increment = 1 << increment;
    while( number_segments-- ) {
        _os2zero64k( 0, 0x8000, tseg, 0 );
        tseg += increment;
    }
    if( remaining_bytes != 0 ) {			/* 30-apr-91 */
        _os2zero_rest( 0, remaining_bytes, tseg, 0 );
    }
    return( (void _WCHUGE *)((unsigned long )seg << 16) );
}

#pragma aux hfree modify [es]
_WCRTLINK void hfree( void _WCHUGE *ptr )
    {
	__FreeSeg( FP_SEG( ptr ) );
    }
