/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985,1988 by WATCOM Systems Inc. All rights %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  22-jul-88     A.F.Scian       defined
  03-aug-88     A.F.Scian       removed large data restrictions on DS=64k
  16-dec-88     F.W.Crigger     Fixed calculation of expand. (it was too big)
-------------------------------------------------------------Version 7.0-----
  04-apr-89     F.W.Crigger     added "- sizeof(frl)" in call to _nmalloc
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  07-feb-90     A.F.Scian       changes in _nmalloc & _expand adjusted
                                expansion of DGROUP if a free block at the
                                end of the heap could be used, of course,
                                _nheapgrow depended on this and ceased working
                                fix: use __LastFree call to get adjustment
                                also fixed calculation of expand
  04-apr-90     J.B.Schueler    added support for QNX
  26-mar-91	A.F.Scian	added 16-bit Windows support
  03-apr-91	C.G.Eisler	updated for WINDOWS_386 macro
  23-apr-91	F.W.Crigger	Check _osmode for DOS/16M
  01-mar-94	Kevin Hui	Got rid of __nheapblk; use __nheapbeg
--------------------------------------------------------Version-11.0---------
  22-feb-95	Greg Bentz	alpha support
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
  15-sep-95	Greg Bentz	ppc support
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <stdlib.h>
#include <malloc.h>
#if defined(__QNX__)
    #include <sys/seginfo.h>
#elif defined(__OS2__)
    #include "tinyos2.h"
    #define INCL_DOSMEMMGR
    #include <wos2.h>
#elif defined(_M_IX86)
    #include "tinyio.h"
#endif
#include "heap.h"
#include "heapacc.h"

#if defined(M_I86)

_WCRTLINK void _fheapgrow( void )
{
    /* multiple heaps are used so growing the far heaps is not necessary */
}

#endif


#if defined(__SMALL_DATA__)

_WCRTLINK void _heapgrow( void )
    {
        _nheapgrow();
    }

#else

_WCRTLINK void _heapgrow( void )
    {
        _fheapgrow();
    }

#endif


_WCRTLINK void _nheapgrow( void )
    {
#if defined(__WINDOWS_286__) || defined(__386__) || defined(__AXP__) || defined(__PPC__)
	_nfree( _nmalloc( 1 ) );	/* get something into the heap */
#else
        unsigned max_paras;
        unsigned curr_paras;
        unsigned diff_paras;
        unsigned expand;

        _AccessNHeap();
        /* calculate # pages which always has extra slack space (ie. 0x10) */
        curr_paras = (( _curbrk + 0x10 ) & ~0x0f ) >> 4;
        if( curr_paras == 0 ) {
            /* we're already at 64k */
            _ReleaseNHeap();
            return;
        }
#if defined(__QNX__)
        if( qnx_segment_realloc( _DGroup(), 65536L ) == -1 ) {
            _ReleaseNHeap();
            return;
        }
        max_paras = PARAS_IN_64K;
#elif defined(__OS2__)
        if( DosReallocSeg( 0, _DGroup() ) )  {
            _ReleaseNHeap();
            return;
        }
        max_paras = PARAS_IN_64K;
#else
	if( _osmode != DOS_MODE ) {			/* 23-apr-91 */
            max_paras = PARAS_IN_64K;
	} else {
	    max_paras = TinyMaxSet( _psp );
	    /* subtract off code size */
	    max_paras -= _DGroup() - _psp;
	    if( max_paras > PARAS_IN_64K ) {
		max_paras = PARAS_IN_64K;
	    }
	}
#endif
        if( max_paras <= curr_paras ) {
            /* '<' -> something is wrong, '==' -> can't change size */
            _ReleaseNHeap();
            return;
        }
        diff_paras = max_paras - curr_paras;
        expand = (( diff_paras + 1 ) << 4 ) - ( _curbrk & 0x0f );
        expand += __LastFree(); /* compensate for _expand's adjustment */
        _ReleaseNHeap();
        _nfree( _nmalloc( expand - ( sizeof( size_t ) + sizeof(frl) ) ) );
#endif
    }
