/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1991, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  14-jun-88     A.F.Scian       defined
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  ------------------------------Version 8.0--------------------------
  23-mar-90     A.F.Scian       fixed problem with an empty heap
                                (p1=p1->next would be invalid because
                                 p1's heap may have been freed!)
  08-aug-90     J.B.Schueler    added support for QNX
  11-oct-90     A.F.Scian       fixed problem with shrinking an almost full seg
  29-jan-91	A.F.Scian	rewritten for new memory manager
  26-mar-91	A.F.Scian	added 16-bit Windows support
  03-apr-91	C.G.Eisler	updated for WINDOWS_386 macro
  23-may-91	A.F.Scian	fixed to follow rules for Windows 3.1
--------------------------------------------------------Version-11.0---------
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <stdlib.h>
#include <malloc.h>
#if defined(__QNX__)
    #include <sys/seginfo.h>
#elif defined(__OS2__)
    #include <dos.h>
    #define INCL_DOSMEMMGR
    #include <wos2.h>
    #include "tinyos2.h"
#elif defined(__WINDOWS__) 
    #include <dos.h>
    #include "windows.h"
#else
    #include <dos.h>
    #include "tinyio.h"
#endif
#include "heap.h"
#include "heapacc.h"
#include "seterrno.h"


int __HeapMin( __segment seg, unsigned one_seg )
{
    tag			last_len;
    tag			adjust_len;
    tag			new_heap_len;
    __segment		heap_seg;
    struct heapblk _WCFAR *heap;
    farfrlptr		last_free;
    farfrlptr		end_tag;
#if !( defined(__QNX__) || defined(__WINDOWS__) )
    tiny_ret_t		rc;
#endif

    _AccessFHeap();
    while( seg != 0 ) {
	heap = MK_FP( seg, 0 );
	/* we might free this segment so get the next one now */
	heap_seg = seg;
	seg = heap->nextseg;
	if( heap->numfree == 0 ) {	/* full heap */
	    if( one_seg != 0 )  break;
	    continue;
	}
	if( heap->numalloc == 0 ) {	/* empty heap */
	    continue;
	}
	/* verify the last block is free */
	last_free = MK_FP( heap_seg, heap->freehead.prev );
	if(( last_free->len & 1 ) != 0 ) continue;

	/* verify the last block is just before the end of the heap */
	last_len = last_free->len;
	end_tag = (farfrlptr)(((FARPTR) last_free) + last_len );
	if( end_tag->len != END_TAG ) continue;

	/* adjust sizes so the last free block stays in the heap */
	if( last_len <= FRL_SIZE ) continue;

	new_heap_len = heap->heaplen - ( last_len - FRL_SIZE );
	new_heap_len = ( new_heap_len + 0x0f ) & ~0x0f;
	adjust_len = heap->heaplen - new_heap_len;
	if( adjust_len == 0 ) continue;

#if defined(__QNX__)
	if( qnx_segment_realloc( heap_seg, new_heap_len ) == -1 ) {
	    _ReleaseFHeap();
	    return( -1 );
	}
#elif defined(__WINDOWS__)
	{
	    HANDLE hmem;

	    hmem = GlobalHandle( heap_seg );
	    if( hmem == NULL ) {
		_ReleaseFHeap();
		return( -1 );
	    }
	    if( GlobalReAlloc( hmem, new_heap_len, GMEM_ZEROINIT ) == NULL){
		_ReleaseFHeap();
		return( -1 );
	    }
	}
#else
  #if defined(__OS2__)
	rc = DosReallocSeg( new_heap_len, heap_seg );
  #else
	if( new_heap_len != 0 ) {
	    rc = TinySetBlock( new_heap_len >> 4, heap_seg );
	} else {
	    rc = TinySetBlock( PARAS_IN_64K, heap_seg );
	}
  #endif
	if( TINY_ERROR( rc ) ) {
	    _ReleaseFHeap();
	    return( __set_errno_dos( TINY_INFO( rc ) ) );
	}
#endif

	/* make the changes to the heap structure */
	heap->heaplen = new_heap_len;
	last_free->len -= adjust_len;
	end_tag = (farfrlptr)(((FARPTR) last_free) + last_free->len );
	end_tag->len = END_TAG;
	end_tag->prev = 0;
    }
    _ReleaseFHeap();
    return( _HEAPOK );
}
