/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985-1991 by WATCOM Systems Inc. All rights %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  10-jun-88     A.F.Scian       defined
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  31-dec-90     F.W.Crigger     New implementation for new memory manager
  26-oct-91	F.W.Crigger	Modified for use by _bheapwalk
  28-nov-91	A.F.Scian	removed an infinite loop problem
  10-jan-92	A.F.Scian	added verifyHeapList
  01-mar-94	Kevin Hui	fixed parameter type for __HeapWalk()
--------------------------------------------------------Version-11.0---------
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <malloc.h>
#include <i86.h>
#include "heap.h"

static int verifyHeapList( unsigned seg )
{
    /* make sure list of heaps is a doubly-linked NULL terminated list */
    struct heapblk _WCFAR *prev_heap;
    struct heapblk _WCFAR *next_heap;
    struct heapblk _WCFAR *curr;
    unsigned next_seg;
    unsigned prev_seg;

    curr = MK_FP( seg, 0 );
    /* check previous heaps end in NULL */
    for(;;) {
	prev_seg = curr->prevseg;
	if( prev_seg == seg ) {
	    return( _HEAPBADBEGIN );
	}
	prev_heap = MK_FP( prev_seg, 0 );
	if( prev_heap == NULL ) break;
	if( prev_heap->nextseg != FP_SEG( curr ) ) {
	    return( _HEAPBADBEGIN );
	}
	curr = prev_heap;
    }
    /* check next heaps end in NULL */
    for(;;) {
	next_seg = curr->nextseg;
	if( next_seg == seg ) {
	    return( _HEAPBADBEGIN );
	}
	next_heap = MK_FP( next_seg, 0 );
	if( next_heap == NULL ) break;
	if( next_heap->prevseg != FP_SEG( curr ) ) {
	    return( _HEAPBADBEGIN );
	}
	curr = next_heap;
    }
    return( _HEAPOK );
}

int __HeapWalk( struct _heapinfo *entry, __segment seg, unsigned all_segs )
    {
        struct heapblk _WCFAR *prev_heap;
        struct heapblk _WCFAR *next_heap;
        struct heapblk _WCFAR *pheap;
        farfrlptr p;
        farfrlptr q;

        if( seg == 0 )  return( _HEAPEMPTY );
        p = entry->_pentry;
        if( p != NULL ) {
	    seg = FP_SEG(p);
	} else if( all_segs == 0 ) {
	    /* we are starting a multi-heap walk */
	    if( verifyHeapList( seg ) != _HEAPOK ) {
		return( _HEAPBADBEGIN );
	    }
	}
        for(;;) {
            pheap = MK_FP( seg, 0 );
	    prev_heap = MK_FP( pheap->prevseg, 0 );
	    next_heap = MK_FP( pheap->nextseg, 0 );
	    if( prev_heap != NULL ) {
		if( prev_heap->nextseg != seg || prev_heap == next_heap ) {
		    return( _HEAPBADBEGIN );
		}
	    }
	    if( next_heap != NULL ) {
		if( next_heap->prevseg != seg ) {
		    return( _HEAPBADBEGIN );
		}
	    }
            if( p == NULL ) {
                if( pheap->freehead.len != 0 )  return( _HEAPBADBEGIN );
                p = MK_FP( FP_SEG(pheap), sizeof(struct heapblk) );
            } else {    /* advance to next entry */
                q = (farfrlptr)((FARPTR)p + (p->len & ~1));
		if( q <= p )  return( _HEAPBADNODE );
		p = q;
		if( pheap->heaplen != 0 && (tag)p > pheap->heaplen ) {
		    return( _HEAPBADNODE );
		}
            }
            if( p->len != END_TAG ) break;
            seg = pheap->nextseg;
            if( seg == 0  ||  all_segs != 0 ) {
                entry->_useflag = _USEDENTRY;
                entry->_size    = 0;
                entry->_pentry  = NULL;
                return( _HEAPEND );
            }
            p = NULL;
        }
        entry->_pentry  = p;
        entry->_useflag = _FREEENTRY;
        entry->_size    = p->len & ~1;
        if( p->len & 1 ) {
            entry->_useflag = _USEDENTRY;
        }
        return( _HEAPOK );
    }
