/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985-1990 by WATCOM Systems Inc. All rights %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  28-May-86     A.C.C.Wai       Initial implementation
   1-Mar-87     A.C.C.Wai       Changed to new free list direction
-------------------------------------------------------Version 6.0--------
  03-apr-88     F.W.Crigger     new doubly linked format of free storage.
  25-apr-88     F.W.Crigger     turn off allocated bit before coalescing
  29-apr-88     F.W.Crigger     check to see if pointer is part of DGroup
  03-may-88     F.W.Crigger     Forgot to mark second half as free before
                                calling free on it.
-------------------------------------------------------Version 6.5--------
  15-jul-88     F.W.Crigger     check for overflow past 64K on size
                                don't free old stg if can't grow the stg.
  04-aug-88     A.F.Scian       split into _nrealloc & _frealloc
-------------------------------------------------------Version 7.0--------
  14-apr-89     F.W.Crigger     Added __386__
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  27-apr-90     A.F.Scian       enlarging current allocation always allocated
                                a new block and copied; changed to check for
                                free blocks afterwards that can be used to
                                enlarge the allocation
-------------------------------------------------------Version 8.0--------
  27-dec-90     F.W.Crigger     split out from _nrealloc
  08-jul-91	D.J.Gaudet	_ffree of partial block left over from a shrink
  				was never malloced so we have to fake free out
  07-nov-93	F.W.Crigger	don't try to expand block if DPMI environment
  15-nov-93	F.W.Crigger	Added call to __ExpandDPMIBlock for Rational
  14-feb-94	Kevin Hui	Extracted the common code in _bexpand() and
  				_nexpand() to form __HeapManager_expand().
				Removed __ExpandDPMIBlock() and applied the
				idea in _nrealloc(). Added __IsCtsNHeap to
				avoid useless calls to __ExpandDGroup.
--------------------------------------------------------Version-11.0---------
  22-feb-95	Greg Bentz	alpha support
  06-mar-95	Greg Bentz	miniheaps may not be in order
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
  15-sep-95	Greg Bentz	ppc support
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <malloc.h>
#include "heap.h"
#include "heapacc.h"
#if defined(__DOS_EXT__)
 #include "extender.h"
 extern void *__ExpandDPMIBlock(frlptr,unsigned);
#endif

#if defined(__SMALL_DATA__)

_WCRTLINK void *_expand( void *stg, size_t amount )
    {
        return( _nexpand( stg, amount ) );
    }

#endif

#if defined(__AXP__) || defined(__PPC__)
    #define _SEGMENT int
#else
    #define _SEGMENT __segment
#endif

int __HeapManager_expand( _SEGMENT seg,
			  unsigned offset,
			  size_t req_size,
			  size_t *growth_size )
{
    #if defined(M_I86)
	typedef struct freelistp __based(seg) *fptr;
	typedef char __based(void) *cptr;

	struct miniheapblkp __based(seg) *hblk;
    #else
	typedef struct freelistp _WCNEAR *fptr;
	typedef char _WCNEAR *cptr;

	mheapptr hblk;
    #endif
    fptr	p1;
    fptr	p2;
    fptr	pnext;
    fptr	pprev;
    size_t	new_size;
    size_t	old_size;
    size_t	free_size;

    /* round (new_size + tag) to multiple of pointer size */
    new_size = (req_size + TAG_SIZE + ROUND_SIZE) & ~ROUND_SIZE;
    if( new_size < req_size ) new_size = ~0; //go for max
    if( new_size < FRL_SIZE ) {
	new_size = FRL_SIZE;
    }
    p1 = (fptr) ((cptr)offset - TAG_SIZE);
    old_size = p1->len & ~1;
    if( new_size > old_size ) {
	/* enlarging the current allocation */
	p2 = (fptr) ((cptr)p1 + old_size);
	*growth_size = new_size - old_size;
	for(;;) {
	    free_size = p2->len;
	    if( p2->len == END_TAG ) {
		return( __HM_TRYGROW );
	    } else if( free_size & 1 ) { /* next piece is allocated */
		break;
	    } else {
		pnext = p2->next;
		pprev = p2->prev;
		
		if( seg == _DGroup() ) { // near heap
		    for( hblk = __nheapbeg; hblk->next; hblk = hblk->next ) {
			if( (fptr)hblk <= (fptr)offset &&
			    (fptr)((PTR)hblk+hblk->len) > (fptr)offset ) break;
		    }
		}
		#if defined(M_I86) 
		    else {	// Based heap
			hblk = 0;
		    }
		#endif
		
		if( hblk->rover == p2 ) { /* 09-feb-91 */
		    hblk->rover = p2->prev;
		}
		if( free_size < *growth_size  ||
		    free_size - *growth_size < FRL_SIZE ) {
		    /* unlink small free block */
		    pprev->next = pnext;
		    pnext->prev = pprev;
		    p1->len += free_size;
		    hblk->numfree--;
		    __fheap_clean = 0;
		    if( free_size >= *growth_size ) {
			return( __HM_SUCCESS );
		    }
		    *growth_size -= free_size;
		    p2 = (fptr) ((cptr)p2 + free_size);
		} else {
		    p2 = (fptr) ((cptr)p2 + *growth_size);
		    p2->len = free_size - *growth_size;
		    p2->prev = pprev;
		    p2->next = pnext;
		    pprev->next = p2;
		    pnext->prev = p2;
		    p1->len += *growth_size;
		    __fheap_clean = 0;
		    return( __HM_SUCCESS );
		}
	    }
	}
	/* no suitable free blocks behind, have to move block */
	return( __HM_FAIL );
    } else {
	/* shrinking the current allocation */
	if( old_size - new_size >= FRL_SIZE ) {
	    /* block big enough to split */
	    p1->len = new_size | 1;
	    p1 = (fptr) ((cptr)p1 + new_size);
	    p1->len = (old_size - new_size) | 1;
	    if( seg == _DGroup() ) { // near heap
		for( hblk = __nheapbeg; hblk->next; hblk = hblk->next ) {
		    if( (fptr)hblk <= (fptr)offset &&
			(fptr)((PTR)hblk+hblk->len) > (fptr)offset ) break;
		}
	    }
	    #if defined(M_I86)
		else 	// Based heap
		    hblk = 0;
	    #endif
	    /* _bfree will decrement 'numalloc' 08-jul-91 */
	    hblk->numalloc++;
	    #if defined(M_I86)
		_bfree( seg, (cptr)p1 + TAG_SIZE );
		/* free the top portion */
	    #else
		_nfree( (cptr)p1 + TAG_SIZE );
	    #endif
	}
    }
    return( __HM_SUCCESS );
}


_WCRTLINK void _WCNEAR *_nexpand( void _WCNEAR *stg, size_t req_size )
    {
        struct {
            unsigned expanded : 1;
        } flags;
	int retval;
	size_t growth_size;

	flags.expanded = 0;
	_AccessNHeap();
	for( ;; ) {
	    retval = __HeapManager_expand( _DGroup(),
					   (unsigned) stg,
					   req_size,
					   &growth_size );
	    if( retval == __HM_SUCCESS ) {
		_ReleaseNHeap();
		return( stg );
	    }
	    if( retval == __HM_FAIL || !__IsCtsNHeap() ) break;
	    if( retval == __HM_TRYGROW ) {
		if( flags.expanded ) break;
		if( __ExpandDGROUP( growth_size ) == 0 ) {
		    break;
		}
		flags.expanded = 1;
	    }
	}
	_ReleaseNHeap();
	return( NULL );
    }
