/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985,1986 by WATCOM Systems Inc. All rights %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  10-jun-88     A.F.Scian       defined
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  30-jan-91	A.F.Scian	defined for new memory manager
  26-mar-91	A.F.Scian	fixed problem with multiple heap blocks
  24-feb-94	Kevin Hui	Adapted for new near heap structure
--------------------------------------------------------Version-11.0---------
  22-feb-95	Greg Bentz	alpha support
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
  14-jan-97	Greg Bentz	Add __nheapchk_current, beef up checks
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <malloc.h>
#include "heap.h"
#include "heapacc.h"

frlptr __nheapchk_current;

static int checkFreeList( size_t *free_size )
{
    frlptr p;
    frlptr end;
    size_t new_size;
    size_t free_list_size = 0;
    mheapptr mhp;

    for( mhp = __nheapbeg; mhp != NULL; mhp = mhp ->next ) {
	/* check that the free list is a doubly linked ring */
	__nheapchk_current = p = mhp->freehead.next;
	/* make sure we start off on the right track */
	if( (p->prev == NULL) ||
	    (p->prev < &(mhp->freehead)) ||
	    (((PTR)p->prev) > (((PTR)mhp)+mhp->len)) ) {
	    return( _HEAPBADNODE );
	}
	if( p->prev->next != p ) {
	    return( _HEAPBADNODE );
	}
	end = p;
	do {
	    /* loop invariant: p->prev->next == p */
	    /* are we still in a ring if we move to p->next? */
	    /* nb. this check is sufficient to ensure that we will
	       never cycle 		*/
	    if( (p->next == NULL) ||
		(p->next < &(mhp->freehead)) ||
		(((PTR)p->next) > (((PTR)mhp)+mhp->len)) ) {
		return( _HEAPBADNODE );
	    }
	    if( p->next->prev != p ) {
		return( _HEAPBADNODE );
	    }
	    /* is entry allocated? */
	    if( p->len & 1 ) {
		return( _HEAPBADNODE );
	    }
	    new_size = free_list_size + p->len;
	    if( new_size < free_list_size ) {
		/* this is a case where we do not know where memory
		   is corrupted		*/
		return( _HEAPBADNODE );
	    }
	    free_list_size = new_size;
	    __nheapchk_current = p = p->next;
	} while( p != end );
    }
    *free_size = free_list_size;
    return( _HEAPOK );
}

static int checkFree( frlptr p )
{
    frlptr next;
    frlptr prev;
    frlptr next_next;
    frlptr prev_prev;

    __nheapchk_current = p;
    if( p->len & 1 ) {
	return( _HEAPBADNODE );
    }
    next = p->next;
    prev = p->prev;
    if( next->prev != p || prev->next != p ) {
	return( _HEAPBADNODE );
    }
    next_next = next->next;
    prev_prev = prev->prev;
    if( next_next->prev != next || prev_prev->next != prev ) {
	return( _HEAPBADNODE );
    }
    if( next_next->prev != next || prev_prev->next != prev ) {
	return( _HEAPBADNODE );
    }
    return( _HEAPOK );
}

#if defined(__SMALL_DATA__)
_WCRTLINK int _heapchk( void )
{
    return( _nheapchk() );
}
#endif

_WCRTLINK int _nheapchk( void )
{
    struct _heapinfo hi;
    int heap_status;
    size_t free_size;

    _AccessNHeap();
    if( __nheap_clean ) {
	_ReleaseNHeap();
	return( _HEAPOK );
    }
    heap_status = checkFreeList( &free_size );
    if( heap_status != _HEAPOK ) {
	_ReleaseNHeap();
	return( heap_status );
    }
    hi._pentry = NULL;
    for(;;) {
	heap_status = __NHeapWalk( &hi, __nheapbeg );
	if( heap_status != _HEAPOK ) break;
	if( hi._useflag == _FREEENTRY ) {
	    heap_status = checkFree( (frlptr) hi._pentry );
	    if( heap_status != _HEAPOK ) break;
	    free_size -= hi._size;
	}
    }
    if( free_size != 0 ) {
	heap_status = _HEAPBADNODE;
    } else if( heap_status == _HEAPBADPTR ) {
	heap_status = _HEAPBADNODE;
    } else {
	if( heap_status == _HEAPEND ) {
	    heap_status = _HEAPOK;
	}
	if( heap_status == _HEAPOK ) {
	    __nheap_clean = 1;
	}
    }
    _ReleaseNHeap();
    return( heap_status );
}
