/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985,1986 by WATCOM Systems Inc. All rights %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  18-Feb-87     A.C.C.Wai       From older malloc.c, but allocation goes up.
  03-Mar-87     A.C.C.Wai       Added heap expansion scheme
  10-jan-88     F.W.Crigger     Adapted for different memory models
  09-mar-88     F.W.Crigger     added #include for <stdlib.h> to get proper
                                prototype for sbrk to fix bug in large
                                data models.
--------------------------------------------------------Version 6.0--------
  03-apr-88     F.W.Crigger     use the new brk() function.
                                free list made doubly linked.
                                and low order bit indicates allocated or not.
  12-apr-88     F.W.Crigger     _nfree checks to make sure block not already
                                indicated as free.
  08-may-88     F.W.Crigger     split away from malloc.c
-------------------------------------------------------Version 6.5----------
  09-jun-88     A.F.Scian       - improved coalesce routine
                                - added heap support
  14-jun-88     A.F.Scian       when splitting blocks, return slice on the
                                bottom
  15-jul-88     F.W.Crigger     check for overflow past 64K on amount
  03-aug-88     A.F.Scian       removed large data restrictions on DS=64k
--------------------------------------------------------Version 7.0--------
  06-mar-89     F.W.Crigger     Make int size independent for 386
  20-mar-89     F.W.Crigger     Update __num_allocs and __num_frees variables
  28-mar-89     J.D.Graham      Added _nmemneed hook
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  18-oct-89     F.W.Crigger     _expand (account for free block at end of heap)
  02-dec-89     F.W.Crigger     __DynRover added to speed up memory allocations
  25-jan-90     F.W.Crigger     start lookup at start of free list if amount
                                being allocated is < 16
  26-jan-90     F.W.Crigger     keep track of largest block before __DynRover
  07-feb-90     A.F.Scian       changes in _nmalloc & _expand adjusted
                                expansion of DGROUP if a free block at the
                                end of the heap could be used, of course,
                                _nheapgrow depended on this and ceased working
                                MORAL: any changes to DGROUP expansion
                                must verify that _nheapgrow still works!!!
                                fix: added __LastFree function
  20-dec-90     J.B.Schueler    brk changed to __brk
  28-dec-90     F.W.Crigger     increment __num_allocs in __ExpandDGROUP()
  22-feb-91     A.F.Scian       tweaked __ExpandDGROUP()
                                - adjusted extra amounts added to allocation
                                - cannot assume 'amount' is even
                                - if size of last free block is equal to
                                  expansion amount problems arose
  18-mar-91     A.F.Scian       tweaked __ExpandDGROUP()
                                - changed +1 & ~1 code to use sizeof( tag )
                                  for the 386 libraries
  21-mar-91     C.G.Eisler      removed extender for 386 windows
  26-mar-91     A.F.Scian       added 16-bit Windows support
                                - fixed bug if a new heap block was too small
  26-mar-91     C.G.Eisler      removed extender for 386 windows (AGAIN!)
  02-apr-91	A.F.Scian	added 32-bit Rational support
  03-apr-91	C.G.Eisler	updated for WINDOWS_386 macro
  15-apr-91	A.F.Scian	added 32-bit Intel CodeBuilder support
  21-may-91	C.G.Eisler	fixed minimum block allocation
  14-jun-91	A.F.Scian	adjusted insertion of heap blocks so that
  				heap blocks could be inserted before previous
				heap blocks
  08-jul-91	G. Coschi	added OS/2 2.0 support
  19-jul-91	A.F.Scian	added allocation from DOS memory for Rational
  07-aug-91	John Dahms	don't ask for more than 64K DOS memory under
  				in a DPMI environment. It ignore the hi word.
  02-mar-92	A.F.Scian	fixed check for 64k so that exactly 64k is OK
--------------------------------------------------------Version 9.x--------
  06-mar-92	Craig Eisler	NT support
  30-jun-92	B.J. Stecher 	used __DOS_EXT__ macro
  21-jul-92	B.J. Stecher	split __ExpandDGroup out into GROWNEAR.C
  25-jan-93	A.F.Scian	added __outside_CLIB attributes to __nmemneed
  24-feb-94	Kevin Hui	Adapted for new mini-heap data structure
--------------------------------------------------------Version-11.0---------
  06-mar-95	Greg Bentz	bugs in MiniHeapRover code
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <stdlib.h>
#include <malloc.h>
#include "extfunc.h"
#include "heapacc.h"
#include "heap.h"

#if defined(_M_IX86)
    #pragma aux (__outside_CLIB) __nmemneed;
#endif

mheapptr _WCNEAR __nheapbeg = NULL;

struct miniheapblkp _WCNEAR *__MiniHeapRover = NULL;
unsigned int   __LargestSizeB4MiniHeapRover = 0;
 


#if defined(__SMALL_DATA__)

_WCRTLINK void *malloc( size_t amount )
{
    return( _nmalloc( amount ) );
}

#endif


_WCRTLINK void _WCNEAR *_nmalloc( size_t amt )
{
    unsigned        largest;
    unsigned        size;
    unsigned        ptr;
    unsigned char   expanded;
    mheapptr	    miniheap_ptr;

    if( (amt == 0) || (amt > -sizeof(struct heapblk)) ) {
	return( (void _WCNEAR *)NULL );
    }

    // Try to determine which miniheap to begin allocating from.
    // first, round up the amount
    size = (amt + TAG_SIZE + ROUND_SIZE) & ~ROUND_SIZE;
    if( size < FRL_SIZE ) {
	size = FRL_SIZE;
    }

    _AccessNHeap();
    ptr = NULL;
    expanded = 0;
    for(;;) {
	if( size > __LargestSizeB4MiniHeapRover ) {
	    miniheap_ptr = __MiniHeapRover;
	    if( miniheap_ptr == NULL ) {
		__LargestSizeB4MiniHeapRover = 0;
		miniheap_ptr = __nheapbeg;
	    }
	} else {
	    __LargestSizeB4MiniHeapRover = 0;	// force value to be updated
	    miniheap_ptr = __nheapbeg;
	}
	for(;;) {
	    if( miniheap_ptr == NULL ) {
		break;
	    }
	    __MiniHeapRover = miniheap_ptr;
	    largest = miniheap_ptr->largest_blk;
	    if( largest >= amt ) {
		ptr = __MemAllocator( amt, _DGroup(), (unsigned)miniheap_ptr );
		if( ptr != NULL ) {
		    goto lbl_release_heap;
		}
	    }
	    if( largest > __LargestSizeB4MiniHeapRover ) {
		__LargestSizeB4MiniHeapRover = largest;
	    }
	    miniheap_ptr = miniheap_ptr->next;
	}
	if( expanded || !__ExpandDGROUP( amt ) ) {
	    if( !__nmemneed( amt ) ) {
		break;
	    }
	    expanded = 0;
	} else {
	    expanded = 1;
	}
    }
lbl_release_heap:
    __nheap_clean = 0;
    _ReleaseNHeap();
    return( (void _WCNEAR *)ptr );
}
