/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985-1990 by WATCOM Systems Inc. All rights %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  28-May-86     A.C.C.Wai       Initial implementation
   1-Mar-87     A.C.C.Wai       Changed to new free list direction
-------------------------------------------------------Version 6.0--------
  03-apr-88     F.W.Crigger     new doubly linked format of free storage.
  25-apr-88     F.W.Crigger     turn off allocated bit before coalescing
  29-apr-88     F.W.Crigger     check to see if pointer is part of DGroup
  03-may-88     F.W.Crigger     Forgot to mark second half as free before
                                calling free on it.
-------------------------------------------------------Version 6.5--------
  15-jul-88     F.W.Crigger     check for overflow past 64K on size
                                don't free old stg if can't grow the stg.
  04-aug-88     A.F.Scian       split into _nrealloc & _frealloc
-------------------------------------------------------Version 7.0--------
  14-apr-89     F.W.Crigger     Added __386__
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  27-apr-90     A.F.Scian       enlarging current allocation always allocated
                                a new block and copied; changed to check for
                                free blocks afterwards that can be used to
                                enlarge the allocation
-------------------------------------------------------Version 8.0--------
  27-dec-90     F.W.Crigger     _nexpand split out from this module
--------------------------------------------------Version 8.5----------
  16-jul-91     G. Coschi	called _inline_memcpy instead of using a pragma
  13-mar-93	B.J. Stecher	use #pragma intrinsic for above
  24-feb-94	Kevin Hui	Added __ReAllocDPMIBlock for Rational Systems
--------------------------------------------------------Version-11.0---------
  22-feb-95	Greg Bentz	alpha support
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stddef.h>
#include <malloc.h>
#include <string.h>
#include "heap.h"
#if defined(__DOS_EXT__)
 #include "extender.h"
#endif

#if defined(__SMALL_DATA__)

_WCRTLINK void *realloc( void *stg, size_t amount )
    {
        return( _nrealloc( stg, amount ) );
    }

#endif

#if defined(__DOS_EXT__)
 void *__ReAllocDPMIBlock( frlptr, unsigned );
#endif

#if __WATCOMC__ < 950
 #define memcpy(p1,p2,n)	_inline_memcpy(p1,p2,n)
 void *memcpy( void *__s1, const void *__s2, size_t __n );
#else
 #pragma intrinsic(memcpy)
#endif

_WCRTLINK void _WCNEAR *_nrealloc( void _WCI86NEAR *stg, size_t req_size )
    {
        void _WCNEAR *p;
        size_t     old_size;

        if( stg == NULL ) {
            return( _nmalloc( req_size ) );
        }
        if( req_size == 0 ) {
            _nfree( stg );
            return( (void _WCNEAR *) NULL );
        }
        old_size = _nmsize( stg );
        p = _nexpand( stg, req_size );  /* try to expand it in place */
        if( p == NULL ) {               /* if couldn't be expanded in place */
	    #if defined(__DOS_EXT__)
	    if( _IsRational() ) {
		frlptr	flp, newflp;
		
		flp = (frlptr) ((PTR)stg - TAG_SIZE);
		newflp = __ReAllocDPMIBlock( flp, req_size + TAG_SIZE );
		if( newflp ) {
		    return( (void _WCNEAR *)((PTR)newflp + TAG_SIZE) );
		}
	    }
	    #endif
            p = _nmalloc( req_size );   /* - allocate a new block */
            if( p != NULL ) {           /* - if we got one */
                memcpy( p, stg, old_size );  /* copy it */
                _nfree( stg );                  /* and free old one */
            } else {
                _nexpand( stg, old_size );      /* reset back to old size */
            }
        }
        return( p );
    }
