/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  31-jul-87     A.F.Scian       defined
  21-nov-87     F.W.Crigger     _amblksiz variable introduced
-----------------------------------------------------------Version 6.0-------
  03-apr-88     F.W.Crigger     allow signed input argument. Also added brk().
  17-apr-88     F.W.Crigger     make sure value is not less than _STACKTOP.
  03-aug-88     A.F.Scian       removed large data restrictions on DS=64k
  16-aug-89     John Dahms      Added multi-thread & OS/2 support
  04-apr-90     J.B.Schueler    Added QNX support
-----------------------------------------------------------Version 8.0-------
  20-dec-90     F.W.Crigger     brk changed to __brk
  26-mar-91     A.F.Scian       added 16-bit Windows support
  03-apr-91     C.G.Eisler      updated for WINDOWS_386 macro
  24-apr-91     F.W.Crigger     Check _osmode for DOS/16M
  16-may-91	A.F.Scian	fixed sbrk for 16-bit Windows
--------------------------------------------------------Version-11.0---------
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  28-may-95	Greg Bentz	dll linkage
*/
#include "dll.h"	// needs to be first
#include "variety.h"
#include <stdlib.h>
#include <errno.h>
#if defined(__OS2__)
    #include "tinyos2.h"
    #define INCL_DOSMEMMGR
    #include <wos2.h>
#elif defined(__QNX__)
    #include <sys/types.h>
    #include <sys/seginfo.h>
    #include <unistd.h>
    #if defined(__386__)
	extern int _brk(void *);
    #endif
#elif defined(__WINDOWS__)
    #include "windows.h"
#endif
#include "heapacc.h"
#include "heap.h"

extern  unsigned                _STACKTOP;

#if !( defined(__OS2__) || defined(__QNX__))

extern  unsigned short SS_Reg( void );
#pragma aux SS_Reg              = \
        0x8c 0xd0               /* mov ax,ss */ \
        parm caller \
        value                   [ax];

extern  int SetBlock( unsigned short, size_t );
#pragma aux SetBlock            = \
        0xb4 0x4a               /* mov ah,04ah */ \
        0xcd 0x21               /* int 021h */ \
        0x1b 0xc0               /* sbb ax,ax */ \
        parm caller             [es] [bx] \
        value                   [ax];

#endif

#if defined(__QNX__) && defined(__386__)
_WCRTLINK void *sbrk( int increment ) {
    return( __brk( _curbrk + increment ) );
}

_WCRTLINK int brk( void *endds ) {
    return( __brk( (unsigned) endds ) == (void *)-1 ? -1 : 0 );
}
#else
_WCRTLINK void _WCNEAR *sbrk( int increment ) {
    #if defined(__WINDOWS_286__)
        HANDLE h;

	if( increment > 0 ) {
	    h = LocalAlloc( LMEM_FIXED, increment );
	    if( h == NULL ) {
		errno = ENOMEM;
		h = -1;
	    }
	} else {
	    errno = EINVAL;
	    h = -1;
	}
        return( (void _WCNEAR *) h );
    #else
        return( __brk( _curbrk + increment ) );
    #endif
}
#endif

#if ! defined(__WINDOWS_286__)

_WCRTLINK void _WCNEAR *__brk( unsigned brk_value )
    {
        unsigned old_brk_value;
        unsigned seg_size;
        unsigned segment;

        if( brk_value < _STACKTOP ) {
            errno = ENOMEM;
            return( (void _WCNEAR *) -1 );
        }
        seg_size = ( brk_value + 0x0f ) >> 4;
        if( seg_size == 0 ) {
            seg_size = 0x1000;
        }
        /* try setting the block of memory */
        _AccessNHeap();
        segment = _DGroup();
#if defined(__OS2__)
        if( DosReallocSeg( seg_size << 4, segment ) != 0 ) {
#elif defined(__QNX__) && defined(__386__)
	if( _brk((void *)(seg_size << 4)) == -1 ) {
#elif defined(__QNX__)
        if( qnx_segment_realloc( segment,((unsigned long)seg_size) << 4) == -1){
#else
        if( _osmode == DOS_MODE ) {                     /* 24-apr-91 */
            seg_size += SS_Reg() - _psp;/* add in code size (in paragraphs) */
            segment = _psp;
        }
        if( SetBlock( segment, seg_size ) != 0 ) {
#endif
            errno = ENOMEM;
            _ReleaseNHeap();
            return( (void _WCNEAR *) -1 );
        }

        old_brk_value = _curbrk;        /* return old value of _curbrk */
        _curbrk = brk_value;            /* set new break value */

        _ReleaseNHeap();
        return( (void _WCNEAR *) old_brk_value );
    }

#if defined(__QNX__)
/*
 * This is used by the QNX/386 shared memory functions to tell the
 * memory manager that the break value has changed. That way things don't
 * get screwed up next time we grow the data segment.
 */
void __setcbrk( unsigned offset )
{
    _curbrk = offset;
}
#endif
#endif
