/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %     Copyright (C) 1996, by Watcom International Inc.  All rights    %
   %     reserved.  No part of this software may be reproduced or        %
   %     used in any form or by any means - graphic, electronic or       %
   %     mechanical, including photocopying, recording, taping or        %
   %     information storage and retrieval systems - except with the     %
   %     written permission of Watcom International Inc.                 %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   Date         By              Reason
   ----         --              ------
--------------------------------------------------------Version-11.0---------
   03-jul-96	M. Hildebrand	Initial implementation.
*/

#include "variety.h"
#include <mbstring.h>
#include <windows.h>
#include "liballoc.h"
#include "libwin32.h"
#include "osver.h"


/*
 * The Win32 CreateProcessW actually has a few other parameters, but the
 * C library doesn't really use them.  Instead of bloating the executables
 * with conversion code that won't ever be executed, always supply the
 * values that would be passed in anyway.
 */

BOOL __lib_CreateProcessW( LPWSTR lpCommandLine,
			   BOOL bInheritHandles,
			   LPVOID lpEnvironment,
			   LPPROCESS_INFORMATION lpProcessInformation )
/*********************************************************************/
{
    if( WIN32_IS_NT ) {					/* NT */
	STARTUPINFOW	startupInfo;

	memset( &startupInfo, 0, sizeof( startupInfo ) );
	startupInfo.wShowWindow = SW_NORMAL;
	return( CreateProcessW( NULL, lpCommandLine, NULL, NULL,
				bInheritHandles, CREATE_UNICODE_ENVIRONMENT,
				lpEnvironment, NULL, &startupInfo,
				lpProcessInformation ) );
    } else {						/* Win95 or Win32s */
	char *		mbCommandLine;
	char *		mbEnvironment;
	BOOL		osrc;
	STARTUPINFOA	mbStartupInfo;
	size_t		cvt;
	size_t		len;
	size_t		size;
	wchar_t *	wp;
	char *		p;

	/*** Convert lpCommandLine to MBCS ***/
	len = wcslen( lpCommandLine ) * MB_CUR_MAX + 1;
	mbCommandLine = lib_malloc( len );
	if( mbCommandLine == NULL ) {
	    return( FALSE );
	}
	cvt = wcstombs( mbCommandLine, lpCommandLine, len );
	if( cvt == (size_t)-1 ) {
	    lib_free( mbCommandLine );
	    return( FALSE );
	}

	/*** Determine the size of a MBCS version of the environment ***/
	size = 0;
	wp = lpEnvironment;
	while( *wp != L'\0' ) {
	    len = wcslen( wp ) + 1;
	    size += len;
	    wp += len;
	}
	size *= MB_CUR_MAX;
	size++;			/* for terminating null byte */

	/*** Create a MBCS version of the environment ***/
	mbEnvironment = lib_malloc( size );	/* allocate some room */
	if( mbEnvironment == NULL ) {
	    lib_free( mbCommandLine );
	    return( FALSE );
	}
	wp = lpEnvironment;			/* convert the strings */
	p = mbEnvironment;
	while( *wp != L'\0' ) {
	    cvt = wcstombs( p, wp, size );
	    if( cvt != (size_t)-1 ) {
		p += cvt;
		*p = '\0';
		p++;
		size -= cvt + 1;
	    } else {
		/* just ignore this string */
	    }
	    wp += wcslen( wp ) + 1;
	}
	*p = '\0';			/* terminate with empty string */

	/*** Call the OS ***/
	memset( &mbStartupInfo, 0, sizeof( mbStartupInfo ) );
	mbStartupInfo.wShowWindow = SW_NORMAL;
	osrc = CreateProcessA( NULL, mbCommandLine, NULL, NULL,
			       bInheritHandles, 0, mbEnvironment, NULL,
			       &mbStartupInfo, lpProcessInformation );
	lib_free( mbCommandLine );
	lib_free( mbEnvironment );

	return( osrc );
    }
}
