/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %     Copyright (C) 1996, by Watcom International Inc.  All rights    %
   %     reserved.  No part of this software may be reproduced or        %
   %     used in any form or by any means - graphic, electronic or       %
   %     mechanical, including photocopying, recording, taping or        %
   %     information storage and retrieval systems - except with the     %
   %     written permission of Watcom International Inc.                 %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   Date         By              Reason
   ----         --              ------
--------------------------------------------------------Version-11.0---------
   03-jul-96	M. Hildebrand	Initial implementation.
*/

#include "variety.h"
#include <mbstring.h>
#include <windows.h>
#include "liballoc.h"
#include "libwin32.h"
#include "osver.h"


DWORD __lib_GetFullPathNameW( LPCWSTR lpFileName, DWORD nBufferLength,
			      LPWSTR lpBuffer, LPWSTR *lpFilePart )
/********************************************************************/
{
    if( WIN32_IS_NT ) {					/* NT */
	return( GetFullPathNameW( lpFileName, nBufferLength, lpBuffer, lpFilePart ) );
    } else {						/* Win95 or Win32s */
	char *		mbFileName;
	char *		mbBuffer;
	char *		mbFilePart;
	char *		p;
	size_t		len;
	DWORD		osrc;
	size_t		cvt;

	/*** Allocate some memory ***/
	len = wcslen( lpFileName ) * MB_CUR_MAX + 1;
	mbFileName = lib_malloc( len );
	if( mbFileName == NULL ) {
	    return( 0 );
	}
	mbBuffer = lib_malloc( _MAX_PATH * MB_CUR_MAX );
	if( mbBuffer == NULL ) {
	    lib_free( mbFileName );
	    return( 0 );
	}

	/*** Prepare to call the OS ***/
	cvt = wcstombs( mbFileName, lpFileName, len );
	if( cvt == (size_t)-1 ) {
	    lib_free( mbFileName );
	    lib_free( mbBuffer );
	    return( 0 );
	}

	/*** Call the OS ***/
	osrc = GetFullPathNameA( mbFileName, _MAX_PATH*MB_CUR_MAX,
				 mbBuffer, &mbFilePart );
	lib_free( mbFileName );
	if( osrc == 0 ) {
	    lib_free( mbBuffer );
	    return( 0 );
	}

	/*** If the buffer is too small, return required length ***/
	len = _mbslen( mbBuffer ) + 1;
	if( len > nBufferLength ) {
	    lib_free( mbBuffer );
	    return( len );
	}

	/*** Convert returned pathname ***/
	cvt = mbstowcs( lpBuffer, mbBuffer, nBufferLength );
	if( cvt == (size_t)-1 ) {
	    lib_free( mbBuffer );
	    return( 0 );
	}

	/*** Convert file part pointer ***/
	p = mbBuffer;
	for( len=0; p<mbFilePart; len++ ) {
	    p = _mbsinc( p );
	}
	*lpFilePart = lpBuffer + len;

	/*** Clean up and go home ***/
	lib_free( mbBuffer );
	return( wcslen( lpBuffer ) );	/* return string length */
    }
}
