/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %     Copyright (C) 1996, by Watcom International Inc.  All rights    %
   %     reserved.  No part of this software may be reproduced or        %
   %     used in any form or by any means - graphic, electronic or       %
   %     mechanical, including photocopying, recording, taping or        %
   %     information storage and retrieval systems - except with the     %
   %     written permission of Watcom International Inc.                 %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   Date         By              Reason
   ----         --              ------
--------------------------------------------------------Version-11.0---------
   02-jul-96	M. Hildebrand	Initial implementation.
   09-jul-96	M. Hildebrand	Call MultiByteToWideChar directly.
   11-jul-96	M. Hildebrand	Export __lib_GetModuleFileNameW.
*/

#include "variety.h"
#include <mbstring.h>
#include <stdlib.h>
#include <windows.h>
#include "liballoc.h"
#include "libwin32.h"


/*
 * Since the multibyte stuff isn't initialized when this module is called
 * from the startup code, call MultiByteToWideChar directly.  Since we tell
 * the OS to use the system default code page during conversion, this
 * module will have limited use outside the startup code... which at present
 * is the only place it is used.
 *
 * Note that we can't use osver.h either, since _osbuild and _osmajor are
 * also uninitialized at this point.
 *
 * This function is exported (_WCRTLINK'ed) so that the startup code works
 * when using the DLL version of the library.
 */

_WCRTLINK DWORD __lib_GetModuleFileNameW( HINSTANCE hModule,
					  LPWSTR lpFilename,
					  DWORD nSize )
/**********************************************************/
{
    if( HIWORD( GetVersion() ) < 0x8000 ) {		/* NT */
	return( GetModuleFileNameW( hModule, lpFilename, nSize ) );
    } else {						/* Win95 or Win32s */
	char *		mbFileName;
	DWORD		osrc;
	size_t		cvt;

	/*** Allocate some memory ***/
	mbFileName = lib_malloc( _MAX_PATH * MB_CUR_MAX );
	if( mbFileName == NULL ) {
	    return( 0 );
	}

	/*** Call the OS ***/
	osrc = GetModuleFileNameA( hModule, mbFileName, _MAX_PATH*MB_CUR_MAX );
	if( osrc == 0 ) {
	    lib_free( mbFileName );
	    return( 0 );
	}

	/*** Convert returned pathname ***/
	cvt = MultiByteToWideChar( CP_OEMCP, MB_PRECOMPOSED, mbFileName,
				   -1, lpFilename, nSize );
	lib_free( mbFileName );
	if( cvt == 0 )  return( 0 );
	lpFilename[nSize-1] = L'\0';	/* ensure null-terminated */
	
	/*** Clean up and go home ***/
	return( wcslen( lpFilename ) );	/* return string length */
    }
}
