;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;%	 Copyright (C) 1990 by WATCOM Systems Inc. All rights	   %
;%	 reserved. No part of this software may be reproduced	   %
;%	 in any form or by any means - graphic, electronic or	   %
;%	 mechanical, including photocopying, recording, taping	   %
;%	 or information storage and retrieval systems - except	   %
;%	 with the written permission of WATCOM Systems Inc.	   %
;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;
; FPE387	: connect/disconnect 80x87 interrupt handler
;
; Modified:	By:		Reason:
; ---------	---		------
; 27-jul-89	F.W.Crigger	initial implementation
; 05-jan-90	F.W.Crigger	save and restore real & protected-mode vectors
; 22-jan-90	F.W.Crigger	only hook into interrupts if an 8087 is present
; 16-apr-90	F.W.Crigger	Mask out status bits using mask from control wd
; 21-jul-90	F.W.Crigger	set IDT with interrupt gate instead of task gate
; 08-aug90	F.W.Crigger	Add support for OS/386 from Ergo
; 30-may-91	F.W.Crigger	change 'mov reg,offset var' to 'lea reg,var'
;				to avoid incorrect segment fixups
; 03-jul-91	J.B.Schueler	Moved FPEHandler to FPEINTH.ASM
; 22-jul-91	John Dahms	added support for DOS4G from Rational/WATCOM.
; 03-jun-92	D. Wellstood	moved code to connect/disconnect the handlers
;				for Pharlap,DOS4G,Ergo,Intel to separate files
; 04-aug-92	D.J.Gaudet	made PMSeg, PMAddr, RMAddr public
; 07-mar-95	M. Hildebrand	Deal with different versions of WASM.
; 12-mar-95	Greg Bentz	remove _ from "C" symbols

.8087
.386p

include struct.inc
include mdef.inc
include extender.inc

	if __WASM__ ge 100
	    xref	"C",__Enable_FPE	; enable exceptions
	else
	    xref	<"C",__Enable_FPE>	; enable exceptions
	endif
	xref	__Phar_hook_init_	; setup interupts for Pharlap
	xref	__Phar_hook_fini_	; restore interupts for Pharlap
	xref	__DOS4G_hook_init_	; setup interupts for Rational
	xref	__DOS4G_hook_fini_	; restore interupts for Rational
	xref	__Ergo_hook_init_	; setup interupts for Ergo
	xref	__Ergo_hook_fini_	; restore interupts for Ergo
	xref	__Intel_hook_init_
	xref	__Intel_hook_fini_

	modstart	fpe387

	datasegment

	extrn	__8087		: byte
	extrn	"C",_Extender	: byte

	public	"C",__PMSeg
	public	"C",__PMAddr
	public	"C",__RMAddr
__PMSeg		dw	0
__PMAddr	dd	0
__RMAddr	dd	0

	enddata

	if __WASM__ ge 100
	    xdefp	"C",__Init_FPE_handler
	else
	    xdefp	<"C",__Init_FPE_handler>
	endif
defp	__Init_FPE_handler
	_guess				; guess initialization required
	  cmp	word ptr __PMSeg,0	; - quit if already initialized
	  _quif ne			; - ...
	  cmp	byte ptr __8087,0	; - quit if no 80x87 present
	  _quif e			; - ...
	_admit				; admit: already initialized
	   ret				; - return
	_endguess			; endguess
	push	EAX			; save registers
	push	EBX			; ...
	push	ECX			; ...
	push	EDX			; ...
	push	ES			; ...
	push	DS			; ...
	_guess				; guess: Ergo OS/386
	  cmp	byte ptr _Extender,X_ERGO ; - quit if not OS/386 from Ergo
	  _quif ne			; - ...
	  call __Ergo_hook_init_	; - setup interupt vector
	_admit				; guess: Rational DOS/4G
	  cmp	byte ptr _Extender,X_RATIONAL ; - quit if not DOS/4G
	  _quif ne			; - ...
	  call __DOS4G_hook_init_	; - setup interupt vector
	_admit				; guess: Intel Code Builder
	  cmp	byte ptr _Extender,X_INTEL ; - quit if not Code Builder
	  _quif ne			; - ...
	  call __Intel_hook_init_	; - setup interupt vector
	_admit				; admit: Pharlap
	  call __Phar_hook_init_	; - setup interupt vector
	_endguess			; endguess
	pop	DS			; restore DS
	call	__Enable_FPE		; enable exceptions
	pop	ES			; restore registers
	pop	EDX			; ...
	pop	ECX			; ...
	pop	EBX			; ...
	pop	EAX			; ...
	ret				; return
endproc __Init_FPE_handler


	if __WASM__ ge 100
	    xdefp	"C",__Fini_FPE_handler
	else
	    xdefp	<"C",__Fini_FPE_handler>
	endif
defp	__Fini_FPE_handler
	cmp	word ptr __PMSeg,0	; if not initialized
	_if	e			; - then
	  ret				; - return
	_endif				; endif
	push	EAX			; save registers
	push	EBX			; ...
	push	EDX			; ...
	push	DS			; ...
	sub	ESP,4			; allocate space for control word
	fstcw	word ptr [ESP]		; get control word
	fwait				; ...
	mov	byte ptr [ESP],7Fh	; disable exception interrupts
	fldcw	word ptr [ESP]		; ...
	fwait				; ...
	add	ESP,4			; remove temporary
	_guess				; guess: Ergo OS/386
	  cmp	byte ptr _Extender,X_ERGO ; quit if not OS/386 from Ergo
	  _quif ne			; - ...
	  call __Ergo_hook_fini_	; - restore int vector
	_admit				; guess: Rational DOS/4G
	  cmp byte ptr _Extender,X_RATIONAL ; - quit if not DOS/4G
	  _quif ne			; - ...
	  call __DOS4G_hook_fini_	; - restore int vector
	_admit				; guess: Intel Code Builder
	  cmp byte ptr _Extender,X_INTEL ; quit if not Code Builder
	  _quif ne			; - ...
	  call __Intel_hook_fini_	; - restore int vector
	_admit				; admit: Pharlap
	  call __Phar_hook_fini_	; - restore int vector
	_endguess			; endguess
	pop	DS			; restore registers
	pop	EDX			; ...
	pop	EBX			; ...
	pop	EAX			; ...
	ret
endproc __Fini_FPE_handler

	endmod
	end
