/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  16-mar-87     F.W.Crigger     Initial implementation
---------------------------------------------------------Version 6.0--------
  29-apr-88     F.W.Crigger     movebwd case fixed up.
                                Only do movebwd if buffers overlap.
  09-may-88     F.W.Crigger     Don't use movebwd pragma in HUGE memory model
  27-jun-88     A.F.Scian       optimized movebwd to use movsw
  11-apr-91     A.F.Scian       fixed for FLAT model
---------------------------------------------------------Version 9.0--------
  11-Jun-92	D. Wellstood	Unicode support
  27-sep-92	F.W.Crigger	movebwd pragma added for 386
--------------------------------------------------------Version-11.0---------
  22-feb-95	Greg Bentz	alpha support
  02-mar-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  30-may-95	Greg Bentz	dll linkage
  15-sep-95	Greg Bentz	ppc support
  28-may-96	M. Hildebrand	Return value was wrong for AXP & PPC.
  30-may-96	M. Hildebrand	Removed wide character version.
*/
#include "variety.h"
#include <stddef.h>
#include <string.h>

#if defined(__386__)
extern  void    movefwd( char _WCFAR *dst, const char _WCNEAR *src, unsigned len);
#pragma aux     movefwd =  \
        0x06            /* push es   */\
        0x8e 0xc2       /* mov es,dx */\
	0x51		/* push ecx  */\
        0xc1 0xe9 0x02  /* shr ecx,2 */\
        0xf3 0xa5       /* rep movsd */\
	0x59		/* pop ecx   */\
        0x83 0xe1 0x03  /* and ecx,3 */\
        0xf3 0xa4       /* rep movsb */\
        0x07            /* pop es    */\
        parm [dx edi] [esi] [ecx] \
        modify exact [edi esi ecx];
extern  void    movebwd( char _WCFAR *dst, const char _WCNEAR *src, unsigned len);
#pragma aux     movebwd =  \
        0x06            /* push es */\
        0x8e 0xc2       /* mov es,dx */\
        0xfd            /* std */\
        0x4e            /* dec esi */\
        0x4f            /* dec edi */\
        0xd1 0xe9       /* shr ecx,1 */\
        0x66 0xf3 0xa5  /* rep movsw */\
        0x11 0xc9       /* adc ecx,ecx */\
        0x46            /* inc esi */\
        0x47            /* inc edi */\
        0x66 0xf3 0xa4  /* rep movsb */\
        0x07            /* pop es */\
        0xfc            /* cld */\
        parm [dx edi] [esi] [ecx] \
        modify exact [edi esi ecx];

#elif defined(M_I86) && defined(__SMALL_DATA__)
extern  void    movebwd( char _WCFAR *dst, const char _WCNEAR *src, unsigned len);
#pragma aux     movebwd =  \
        0xfd            /* std */\
        0x4e            /* dec si */\
        0x4f            /* dec di */\
        0xd1 0xe9       /* shr cx,1 */\
        0xf3 0xa5       /* rep movsw */\
        0x11 0xc9       /* adc cx,cx */\
        0x46            /* inc si */\
        0x47            /* inc di */\
        0xf3 0xa4       /* rep movsb */\
        0xfc            /* cld */\
        parm [es di] [si] [cx] \
        modify exact [di si cx];

extern  void    movefwd( char _WCFAR *dst, const char _WCNEAR *src, unsigned len);
#pragma aux     movefwd =  \
        0xd1 0xe9       /* shr cx,1 */\
        0xf3 0xa5       /* rep movsw */\
        0x11 0xc9       /* adc cx,cx */\
        0xf3 0xa4       /* rep movsb */\
        parm [es di] [si] [cx] \
        modify exact [di si cx];
#elif defined(M_I86) && defined(__BIG_DATA__)
extern  void    movebwd( char _WCFAR *dst, const char _WCFAR *src, unsigned len);
#pragma aux     movebwd =  \
        0x1e            /* push ds */ \
        0x8e 0xda       /* mov ds,dx */ \
        0xfd            /* std */\
        0x4e            /* dec si */\
        0x4f            /* dec di */\
        0xd1 0xe9       /* shr cx,1 */\
        0xf3 0xa5       /* rep movsw */\
        0x11 0xc9       /* adc cx,cx */\
        0x46            /* inc si */\
        0x47            /* inc di */\
        0xf3 0xa4       /* rep movsb */\
        0xfc            /* cld */\
        0x1f            /* pop ds */ \
        parm [es di] [dx si] [cx] \
        modify exact [di si cx];

extern  void    movefwd( char _WCFAR *dst, const char _WCFAR *src, unsigned len);
#pragma aux     movefwd =  \
        0x1e            /* push ds */ \
        0x8e 0xda       /* mov ds,dx */ \
        0xd1 0xe9       /* shr cx,1 */\
        0xf3 0xa5       /* rep movsw */\
        0x11 0xc9       /* adc cx,cx */\
        0xf3 0xa4       /* rep movsb */\
        0x1f            /* pop ds */ \
        parm [es di] [dx si] [cx] \
        modify exact [di si cx];
#elif defined(__AXP__) || defined(__PPC__)
// no pragma for these ones
#else
#error unrecognized platform
#endif


_WCRTLINK void *memmove( char *to, const char *from, size_t len )
    {
#if defined(__HUGE__) || defined(__AXP__) || defined(__PPC__)
	char *		toStart = to;
#endif

        if( from == to ) {
            return( to );
        }
        if( from < to  &&  from + len > to ) {  /* if buffers are overlapped*/
#if defined(__HUGE__) || defined(__AXP__) || defined(__PPC__)
            to += len;
            from += len;
            while( len != 0 ) {
                *--to = *--from;
                len--;
            }
#else
            movebwd(( to + len ) - 1, ( from + len ) - 1, len );
#endif
        } else {
#if defined(__AXP__) || defined(__PPC__)
            while( len != 0 ) {
                *to++ = *from++;
                len--;
            }
#else
            movefwd( to, from, len );
#endif
        }

#if defined(__HUGE__) || defined(__AXP__) || defined(__PPC__)
        return( toStart );
#else
        return( to );
#endif
    }
