/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %     Copyright (C) 1996, by Watcom International Inc.  All rights    %
   %     reserved.  No part of this software may be reproduced or        %
   %     used in any form or by any means - graphic, electronic or       %
   %     mechanical, including photocopying, recording, taping or        %
   %     information storage and retrieval systems - except with the     %
   %     written permission of Watcom International Inc.                 %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
   Date         By              Reason
   ----         --              ------
--------------------------------------------------------Version-11.0---------
   16-may-96	M. Hildebrand	Initial implementation.
*/

#include <string.h>
#include "riscstr.h"


void *memcpy( void *dest, const void *src, size_t n )
/***************************************************/
{
    void *		destStart = dest;
    UINT *		dwSrc;
    UINT *		dwDest;
    UINT		dword, tmpdword;
    int			shl, shr;
    int			offset = OFFSET(dest);
    char *		destP = (char*)dest;
    char *		srcP = (char*)src;

    if( n <= 0 )  return( destStart );

    /*** Copy any unaligned bytes at the start ***/
    while( offset != 0  &&  n >= 1 ) {
	*destP++ = *srcP++;
	offset = (offset+1) & 0x03;	/* offset = (offset+1) % 4 */
	n--;
    }

    /*** Initialize locals ***/
    dwSrc = ROUND(srcP);
    dwDest = (UINT*)destP;			/* already rounded */
    shr = OFFSET(srcP) * 8;
    shl = INT_SIZE - shr;
    if( shr != 0 ) {
	dword = *dwSrc++;
    }

    /*** Copy in aligned 4-byte groups ***/
    for( ;; ) {
	/*** Form the dword containing the next four bytes ***/
	if( shr == 0 ) {
	    tmpdword = *dwSrc++;
	} else {
	    tmpdword = dword >> shr;
	    dword = *dwSrc++;
	    tmpdword |= (dword << shl);
	}
	if( n < BYTES_PER_WORD )  break;

        /*** Copy the bytes ***/
	*dwDest++ = tmpdword;
	n -= BYTES_PER_WORD;
    }

    /*** Copy any remaining unaligned bytes at the end ***/
    if( n == 1 ) {				/* copy 1 byte */
	dword = *dwDest & ~(BYTE1);
	tmpdword &= (BYTE1);
	*dwDest = dword | tmpdword;
    } else if( n == 2 ) {			/* copy 2 bytes */
	dword = *dwDest & ~(BYTE1|BYTE2);
	tmpdword &= (BYTE1|BYTE2);
	*dwDest = dword | tmpdword;
    } else if( n == 3 ) {			/* copy 3 bytes */
	dword = *dwDest & ~(BYTE1|BYTE2|BYTE3);
	tmpdword &= (BYTE1|BYTE2|BYTE3);
	*dwDest = dword | tmpdword;
    }

    return( destStart );
}
