/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1991, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  06-mar-91	A.F.Scian	defined
--------------------------------------------------Version 8.5----------
  09-jul-91	G. Coschi	support OS/2 2.0
--------------------------------------------------Version 9.x----------
  11-dec-92	Craig Eisler	support NT
  14-jan-93	F.W.Crigger	Only ignore ".\" if following a "\"
  23-feb-93	Ashish Gupta	support Netware
  19-jul-93	J.B.Schueler	changed PATH_MAX refs to _MAX_PATH
				changed MAX_PATH ref to _MAX_PATH
				(Note: MS defines MAX_PATH in TOOLHELP.H
				and documents MAX_PATH in _fullpath)
  29-sep-93	A.F.Scian	work around Windows NT bug for fullpath("con")
  01-mar-94	F.W.Crigger	Adapted for OSI library
--------------------------------------------------------Version-11.0---------
  30-may-95	Greg Bentz	variety.h and dll linkage
*/
#include "variety.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <ctype.h>
#include <direct.h>
#include "tinyio.h"
#include "ostype.h"
#include "liballoc.h"
#include "rtdata.h"
#include "seterrno.h"

extern __GetFullPathName( char *buff, const char *path, size_t size );
#pragma aux __GetFullPathName = \
	"mov	AH,60h"	\
	_INT_21		\
	parm [ebx] [edx] [ecx];

#define _WILL_FIT( c )	if(( (c) + 1 ) > size ) {	\
			    __set_errno( ERANGE );	\
			    return( NULL );		\
			}				\
			size -= (c);

#define _IS_SLASH( c )	(( (c) == '/' ) || ( (c) == '\\' ))


_WCRTLINK char *_fullpath( char *buff, const char *path, size_t size )
{
    const char	*p;
    char	*q;
    size_t	len;
    tiny_ret_t	rc;
    unsigned	path_drive_idx;
    char	curr_dir[_MAX_PATH];

    if( buff == NULL ) {
	size = _MAX_PATH;
	buff = lib_malloc( size );
	if( buff == NULL ) {
	    __set_errno( ENOMEM );
	    return( NULL );
	}
    }
    if( path == NULL || path[0] == '\0' ) {
	return( getcwd( buff, size ) );
    }
    if( __OS == OS_OS2  ||  __OS == OS_NT ) {
	__GetFullPathName( buff, path, size );
    } else {
	p = path;
	q = buff;
	_WILL_FIT( 2 );
	if( isalpha( p[0] ) && p[1] == ':' ) {
	    path_drive_idx = ( tolower( p[0] ) - 'a' ) + 1;
	    q[0] = p[0];
	    q[1] = p[1];
	    p += 2;
	} else {
	    path_drive_idx = TinyGetCurrDrive() + 1;
	    q[0] = 'A' + ( path_drive_idx - 1 );
	    q[1] = ':';
	}
	q += 2;
	if( ! _IS_SLASH( p[0] ) ) {
	    rc = TinyGetCWDir( curr_dir, path_drive_idx );
	    if( TINY_ERROR( rc ) ) {
		__set_errno( ENOENT );
		return( NULL );
	    }
	    len = strlen( curr_dir );
	    if( curr_dir[0] != '\\' ) {
		_WILL_FIT( 1 );
		*(q++) = '\\';
	    }
	    _WILL_FIT( len );
	    strcpy( q, curr_dir );
	    q += len;
	    if( q[-1] != '\\' ) {
		_WILL_FIT( 1 );
		*(q++) = '\\';
	    }
	    for(;;) {
		if( p[0] == '\0' ) break;
		if( p[0] != '.' ) {
		    _WILL_FIT( 1 );
		    *(q++) = *(p++);
		    continue;
		}
		++p;
		if( _IS_SLASH( p[0] ) ) {
		    /* ignore "./" in directory specs */
		    if( ! _IS_SLASH( q[-1] ) ) {	/* 14-jan-93 */
			*q++ = '\\';
		    }
		    ++p;
		    continue;
		}
		if( p[0] == '\0' ) break;
		if( p[0] == '.' && _IS_SLASH( p[1] ) ) {
		    /* go up a directory for a "../" */
		    p += 2;
		    if( ! _IS_SLASH( q[-1] ) ) {
			return( NULL );
		    }
		    q -= 2;
		    for(;;) {
			if( q < buff ) {
			    return( NULL );
			}
			if( _IS_SLASH( *q ) ) break;
			if( *q == ':' ) {
			    ++q;
			    *q = '\\';
			    break;
			}
			--q;
		    }
		    ++q;
		    *q = '\0';
		    continue;
		}
		_WILL_FIT( 1 );
		*(q++) = '.';
	    }
	    *q = '\0';
	} else {
	    len = strlen( p );
	    _WILL_FIT( len );
	    strcpy( q, p );
	}
	/* force to all backslashes */
	for( q = buff; *q; ++q ) {
	    if( *q == '/' ) {
		*q = '\\';
	    }
	}
    }
    return( buff );
}
