/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1988, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  02-mar-88	F.W.Crigger	Initial implementation
  04-jun-91	J.B.Schueler	Added PenPoint support (same as QNX)
  10-jul-92	D. Wellstood	Unicode Support
  08-feb-93	A. Gupta	Netware support
  19-feb-93	J.B.Schueler	Fixed processing for PENPOINT, QNX, NETWARE, DOS
				fixed /dir1/dir2 .c -> /dir1/dir2/.c for QNX
				fixed \dir1\dir2 .c -> \dir1\dir2\.c for PP
				fixed /foo/ \bar.c  -> /foo/bar.c for DOS, NET
				fixed \foo /bar.c   -> \foo\bar.c for DOS, NET
  10-mar-93	J.B.Schueler	Fixed c: \foo\	    -> c:\foo\ for DOS, NET
				fixed //node /foo/  -> //node/foo/ for QNX
				fixed \\node \foo/  -> \\node\foo\ for PP
  23-apr-93	J.B.Schueler	Moved insertion of NUL char in output buffer
				to fix strcpy( buffer, "C:\\FOO" );
				       _makepath( buffer, "C:", "", "a", ".c" );
  25-nov-93	F.W.Crigger	made pickup routine static
--------------------------------------------------------Version-11.0---------
  02-mar-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  24-mar-95	M. Hildebrand	Added multibyte character support.
  04-may-95     T. Schiller     Fixed overrun but that occured when you
                                passed an empty string in for `fname'.
  30-may-95	Greg Bentz	variety.h and dll linkage
  18-sep-95	M. Hildebrand	Removed __mbvtop.
  21-sep-95	M. Hildebrand	Added _u version, which calls wide version.
  18-oct-95	M. Hildebrand	Added _wmakepath.
  27-nov-95	Greg Bentz	fixup const in function definition
  29-nov-95	M. Hildebrand	Use SBCS code for QNX _makepath.
  13-dec-95	M. Hildebrand	Changed _mbptov to _mbsnextc.
  03-jan-96	Greg Bentz	remove stubs that are now aliases
  01-feb-96	Greg Bentz	don't strip PC from QNX fname
  06-jan-97	T. Schiller	Add support for NT-style UNC paths
*/

#include "variety.h"
#include "widechar.h"
#include <stdlib.h>
#include <string.h>
#ifndef __WIDECHAR__
    #include <mbstring.h>
#endif

#undef _makepath

#if defined(__PENPOINT__)
  #define PC '\\'
#elif defined(__QNX__)
  #define PC '/'
#else	/* DOS, OS/2, Windows, Netware */
  #define PC '\\'
  #define ALT_PC '/'
#endif


#if defined(__PENPOINT__) || defined(__QNX__)

/* create full PENPOINT/QNX path name from the components */

_WCRTLINK void __F_NAME(_makepath,_wmakepath)(
	CHAR_TYPE	    *path,
	const CHAR_TYPE  *node,
	const CHAR_TYPE  *dir,
	const CHAR_TYPE  *fname,
	const CHAR_TYPE  *ext )
{
#if !defined(__WIDECHAR__) && !defined(__QNX__)
    char    *pathstart = path;
#endif
    *path = '\0';

    if( node != NULL ) {
	if( *node != '\0' ) {
	    __F_NAME(strcpy,wcscpy)( path, node );
	    path = __F_NAME(strchr,wcschr)( path, NULLCHAR );

	    /* if node did not end in '/' then put in a provisional one */
	    #ifdef __WIDECHAR__
		if( path[-1] == PC )
		    path--;
		else
		    *path = PC;
	    #else
		#ifdef __QNX__
		    if( path[-1] == PC ) {
			path--;
		    } else {
			*path = PC;
		    }
		#else
		    if( *(_mbsdec(pathstart,path)) == PC )
			path--;
		    else
			*path = PC;
		#endif
	    #endif
	}
    }
    if( dir != NULL ) {
	if( *dir != '\0' ) {
	    /*	if dir does not start with a '/' and we had a node then
		    stick in a separator
	    */
	    if( (*dir != PC) && (*path == PC) ) path++;

	    __F_NAME(strcpy,wcscpy)( path, dir );
	    path = __F_NAME(strchr,wcschr)( path, NULLCHAR );

	    /* if dir did not end in '/' then put in a provisional one */
	    #ifdef __WIDECHAR__
		if( path[-1] == PC )
		    path--;
		else
		    *path = PC;
	    #else
		#ifdef __QNX__
		    if( path[-1] == PC ) {
			path--;
		    } else {
			*path = PC;
		    }
		#else
		    if( *(_mbsdec(pathstart,path)) == PC )
			path--;
		    else
			*path = PC;
		#endif
	    #endif
	}
    }

    if( fname != NULL ) {
	if( (*fname != PC) && (*path == PC) ) path++;

	__F_NAME(strcpy,wcscpy)( path, fname );
	path = __F_NAME(strchr,wcschr)( path, NULLCHAR );

    } else {
	if( *path == PC ) path++;
    }
    if( ext != NULL ) {
	if( *ext != '\0' ) {
	    if( *ext != '.' )  *path++ = '.';
	    __F_NAME(strcpy,wcscpy)( path, ext );
	    path = __F_NAME(strchr,wcschr)( path, NULLCHAR );
	}
    }
    *path = '\0';
}

#elif defined( __NETWARE__ )

/*
    For silly two choice DOS path characters / and \,
    we want to return a consistent path character.
*/

static char pickup( char c, char *pc_of_choice )
{
    if( c == PC || c == ALT_PC ) {
	if( *pc_of_choice == '\0' ) *pc_of_choice = c;
	c = *pc_of_choice;
    }
    return( c );
}

_WCRTLINK extern void _makepath( char *path, const char *volume,
		const char *dir, const char *fname, const char *ext )
{
    char first_pc = '\0';

    if( volume != NULL ) {
	if( *volume != '\0' ) {
	    do {
		*path++ = *volume++;
	    } while( *volume != '\0' );
	    if( path[ -1 ] != ':' ) {
		*path++ = ':';
	    }
	}
    }
    *path = '\0';
    if( dir != NULL ) {
	if( *dir != '\0' ) {
	    do {
		*path++ = pickup( *dir++, &first_pc );
	    } while( *dir != '\0' );
	    /* if no path separator was specified then pick a default */
	    if( first_pc == '\0' ) first_pc = PC;
	    /* if dir did not end in path sep then put in a provisional one */
	    if( path[-1] == first_pc ) {
		path--;
	    } else {
		*path = first_pc;
	    }
	}
    }
    /* if no path separator was specified thus far then pick a default */
    if( first_pc == '\0' ) first_pc = PC;
    if( fname != NULL ) {
	if( (pickup( *fname, &first_pc ) != first_pc)
	    && (*path == first_pc) ) path++;
	while( *fname != '\0' ) *path++ = pickup( *fname++, &first_pc );
    } else {
	if( *path == first_pc ) path++;
    }
    if( ext != NULL ) {
	if( *ext != '\0' ) {
	    if( *ext != '.' )  *path++ = '.';
	    while( *ext != '\0' ) *path++ = *ext++;
	}
    }
    *path = '\0';
}

#else

/*
    For silly two choice DOS path characters / and \,
    we want to return a consistent path character.
*/

static unsigned pickup( unsigned c, unsigned *pc_of_choice )
{
    if( c == PC || c == ALT_PC ) {
	if( *pc_of_choice == '\0' ) *pc_of_choice = c;
	c = *pc_of_choice;
    }
    return( c );
}

/* create full MS-DOS path name from the components */

_WCRTLINK void __F_NAME(_makepath,_wmakepath)( CHAR_TYPE *path, const CHAR_TYPE *drive,
		const CHAR_TYPE *dir, const CHAR_TYPE *fname, const CHAR_TYPE *ext )
{
    unsigned		first_pc = '\0';
#ifndef __WIDECHAR__
    char *		pathstart = path;
    unsigned		ch;
#endif

    if( drive != NULL ) {
	if( *drive != '\0' ) {
	    if ((*drive == '\\') && (drive[1] == '\\')) {
		__F_NAME(strcpy, wcscpy)(path, drive);
		path += __F_NAME(strlen, wcslen)(drive);
	    } else {
		*path++ = *drive;				/* OK for MBCS */
		*path++ = ':';
	    }
	}
    }
    *path = '\0';
    if( dir != NULL ) {
	if( *dir != '\0' ) {
	    do {
		#ifdef __WIDECHAR__
		    *path++ = pickup( *dir++, &first_pc );
		#else
		    #ifdef __QNX__
			*path++ = pickup( *dir++, &first_pc );
		    #else
			ch = pickup( _mbsnextc(dir), &first_pc );
			_mbvtop( ch, path );
			path[_mbclen(path)] = '\0';
			path = _mbsinc( path );
			dir = _mbsinc( dir );
		    #endif
		#endif
	    } while( *dir != '\0' );
	    /* if no path separator was specified then pick a default */
	    if( first_pc == '\0' ) first_pc = PC;
	    /* if dir did not end in '/' then put in a provisional one */
	    #ifdef __WIDECHAR__
		if( path[-1] == first_pc )
		    path--;
		else
		    *path = first_pc;
	    #else
		#ifdef __QNX__
		    if( path[-1] == PC ) {
			path--;
		    } else {
			*path = first_pc;
		    }
		#else
		    if( *(_mbsdec(pathstart,path)) == first_pc )
			path--;
		    else
			*path = first_pc;
		#endif
	    #endif
	}
    }

    /* if no path separator was specified thus far then pick a default */
    if( first_pc == '\0' ) first_pc = PC;
    if( fname != NULL ) {
	#ifdef __WIDECHAR__
	    if( pickup(*fname,&first_pc) != first_pc  &&  *path == first_pc )
	    	path++;
	#else
	    #ifdef __QNX__
		if( pickup(*fname,&first_pc) != first_pc  &&  *path == first_pc )
		    path++;
	    #else
		ch = _mbsnextc( fname );
		if( pickup(ch,&first_pc) != first_pc  &&  *path == first_pc )
		    path++;
	    #endif
	#endif

	while (*fname != '\0')
	{
	//do {
	    #ifdef __WIDECHAR__
	        *path++ = pickup( *fname++, &first_pc );
	    #else
		#ifdef __QNX__
		    *path++ = pickup( *fname++, &first_pc );
		#else
		    ch = pickup( _mbsnextc(fname), &first_pc );
		    _mbvtop( ch, path );
		    path[_mbclen(path)] = '\0';
		    path = _mbsinc( path );
		    fname = _mbsinc( fname );
		#endif
	    #endif
	} //while( *fname != '\0' );
    } else {
	if( *path == first_pc ) path++;
    }
    if( ext != NULL ) {
	if( *ext != '\0' ) {
	    if( *ext != '.' )  *path++ = '.';
	    while( *ext != '\0' ) *path++ = *ext++;	/* OK for MBCS */
	}
    }
    *path = '\0';
}
#endif
