/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1992, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  10-nov-92	F.W.Crigger	defined
  20-dec-92	F.W.Crigger	_ThreadExitRtn defined
  20-jan-93	A.Gupta		extended for Netware
  17-feb-93	F.W.Crigger	use modulus of handle to map all possible
				file handles onto limited # of semaphores
				Close file semaphores in __FiniSema4s
  04-mar-93	A.Gupta		modified Netware code - now allocates a
                                thread data struct for use when _threadid
				returns 0
  18-jun-93	John Dahms  	all library allocations use lib_xxx routines
  04-may-94	Greg Bentz	improve allocation failure checking
--------------------------------------------------------Version-11.0---------
  18-jan-95	Greg Bentz	support QNX multi-threading
  22-feb-95	Greg Bentz	alpha support
  30-may-95	Greg Bentz	variety.h and dll linkage
  25-aug-95     T. Schiller     Added code to preserve NT last error code
                                when retrieving thread information.
  15-sep-95	Greg Bentz	ppc support
  21-sep-95	Alex Brodsky	Fixed syntax error so that it can be compiled.
  13-nov-95	Greg Bentz	change WIN32 support to use critical sections
  				ensured that at shutdown, thread storage is
				released before heap semaphores are closed
  23-apr-96	Greg Bentz	Win32 CloseSemaphore no longer deletes the
  				critical section.  So we don't want to
				mark our semaphore object as not being intialized.
				This shows up as a memory leak.
*/
#include "variety.h"
#include "stacklow.h"
#include <stdio.h>
#include <stdlib.h>
#include "liballoc.h"

#if defined( __RUNTIME_CHECKS__ ) && defined( _M_IX86 )
  //#include <assert.h>
  extern void BreakPoint( void );
  #pragma aux BreakPoint = "int 3" \
  			   "nop";
  #define assert( __x ) if( !(__x) ) BreakPoint();
#endif

#include "thread.h"
#include "trdlist.h"
#include "mthread.h"
#include "rtdata.h"
#include "rtinit.h"
#include "exitwmsg.h"
#include "osver.h"

#if defined( __QNX__ )
  #include "semaqnx.h"
  #include <sys/magic.h>
  extern thread_data *__QNXAddThread( thread_data *tdata );
#endif

extern	void		__FiniThreadProcessing(void);
#if defined( __386__ ) || defined( __AXP__ ) || defined( __PPC__ )
  extern void		(*_AccessFileH)(int);
  extern void		(*_ReleaseFileH)(int);
  extern int		(*_InitAccessH)(int);
  extern void		(*_FiniAccessH)(int);
  extern void		(*_AccessIOB)(void);
  extern void		(*_ReleaseIOB)(void);
  extern void		(*_AccessTDList)(void);
  extern void		(*_ReleaseTDList)(void);
  #if !defined( __NETWARE__ )
    extern void		(*_AccessNHeap)(void);
    extern void		(*_AccessFHeap)(void);
    extern void		(*_ReleaseNHeap)(void);
    extern void		(*_ReleaseFHeap)(void);
  #endif
  #if defined( __NT__ )
    extern void		(*_AccessFList)(int);
    extern void		(*_ReleaseFList)(int);
    extern void		(*_ThreadExitRtn)(void);
    static semaphore_object FListSemaphore;
  #endif
#endif

#if defined( __386__ ) || defined( __AXP__ ) || defined( __PPC__ )
    void static nullSema4Rtn( semaphore_object *p ) { p = p; }
    _WCRTLINK void (*__AccessSema4)( semaphore_object *) = &nullSema4Rtn;
    _WCRTLINK void (*__ReleaseSema4)( semaphore_object *) = &nullSema4Rtn;
    _WCRTLINK void (*__CloseSema4)( semaphore_object *) = &nullSema4Rtn;
#endif

extern	int		__Sema4Fini;		// in finalizer segment
#ifdef _M_IX86
 #pragma aux 		__Sema4Fini "_*";
#endif
extern	unsigned	__MaxThreads;
extern	thread_data	*__FirstThreadData;
extern  void		**__ThreadIDs;

#define	MAX_SEMAPHORE	16

static semaphore_object FileSemaphores[ MAX_SEMAPHORE ];

#if !defined( __NETWARE__ )
  static semaphore_object FHeapSemaphore;
  static semaphore_object NHeapSemaphore;
#endif

static semaphore_object IOBSemaphore;

#if defined( __386__ ) || defined( __AXP__ ) || defined( __PPC__ )
  static semaphore_object InitSemaphore;
  static semaphore_object TDListSemaphore;
#endif

#if defined( __NT__ )
#define MAX_CRITICAL_SECTION 64
static CRITICAL_SECTION critsect_cache[MAX_CRITICAL_SECTION];
static int critsect_next;
static CRITICAL_SECTION **critsect_vector;
static int critsect_vectornext;

CRITICAL_SECTION *__NTGetCriticalSection( void ) {
    CRITICAL_SECTION *ptr;

    if( critsect_next < MAX_CRITICAL_SECTION ) {
	ptr = &(critsect_cache[critsect_next]);
	critsect_next++;
    } else {
	ptr = lib_calloc( 1, sizeof( *ptr ) );
	if( ptr == NULL ) {
	    __fatal_runtime_error(
		"Unable to allocate semaphore data\r\n", 1 );
	} 
	critsect_vector = lib_realloc( critsect_vector,
		(critsect_vectornext+1)*sizeof(CRITICAL_SECTION*));
	if( critsect_vector == NULL ) {
	    __fatal_runtime_error(
		"Unable to allocate semaphore data\r\n", 1 );
	}
	critsect_vector[critsect_vectornext] = ptr;
	critsect_vectornext++;
    }
    InitializeCriticalSection( ptr );
    return( ptr );
}
__NTDeleteCriticalSection( void ) {
    int i;
    for( i = 0 ; i < critsect_next ; i++ ) {
	DeleteCriticalSection( &(critsect_cache[i]) );
    }
}
__NTFreeCriticalSection( void ) {
    int i;
    for( i = 0 ; i < critsect_vectornext ; i++ ) {
	DeleteCriticalSection( critsect_vector[i] );
	lib_free( critsect_vector[i] );
    }
    if( critsect_vector ) {
	lib_free( critsect_vector );
    }
}
#endif

_WCRTLINK void __CloseSemaphore( semaphore_object *obj )
{
    #if defined( __RUNTIME_CHECKS__ ) && defined( _M_IX86 )
	// 0 is ok
	// 1 is ok
        assert( obj->count < 2 );
    #endif
    #if defined( __386__ ) || defined( __AXP__ ) || defined( __PPC__ )
	if( obj->initialized != 0 ) {
	    #if defined( __NETWARE__ )
		obj->semaphore = 0;
	    #elif defined( __NT__ )
	    #elif defined( __QNX__ )
		__qsem_destroy( &obj->semaphore );
	    #else
		DosCloseMutexSem( obj->semaphore );
	    #endif
	}
    #else
	if( obj->count > 0 ) {
	    DosSemClear( &obj->semaphore );
	}
    #endif
    #if !defined( __NT__ )
	obj->initialized = 0;
    #endif
    obj->owner = 0;
    obj->count = 0;
}

_WCRTLINK void __AccessSemaphore( semaphore_object *obj )
{
    TID tid;

    tid = GetCurrentThreadId();
    #if defined( __NETWARE__ )
	if( tid == 0 ) return;
    #endif
    if( obj->owner != tid ) {
	#if defined( __386__ ) || defined( __AXP__ ) || defined( __PPC__ )
	    #if !defined( __NETWARE__ )
		if( obj->initialized == 0 ) {
		    #if defined( __RUNTIME_CHECKS__ ) && defined( _M_IX86 )
			assert( obj != &InitSemaphore );
		    #endif
		    __AccessSemaphore( &InitSemaphore );
		    if( obj->initialized == 0 ) {
			#if defined( __NT__ )
			    obj->semaphore = __NTGetCriticalSection();
			#elif defined( __QNX__ )
			    __qsem_init( &obj->semaphore, 1, 1 );
			#else
			    DosCreateMutexSem( NULL, &obj->semaphore, 0, FALSE );
			#endif
			obj->initialized = 1;
		    }
		    __ReleaseSemaphore( &InitSemaphore );
			
		}
	    #endif
	    #if defined( __NETWARE__ )
		while( obj->semaphore != 0 ) ThreadSwitch();
		obj->semaphore = 1;
		obj->initialized = 1;
	    #elif defined( __NT__ )
		EnterCriticalSection( obj->semaphore );
	    #elif defined( __QNX__ )
		__qsem_wait( &obj->semaphore );
	    #else
		DosRequestMutexSem( obj->semaphore, SEM_INDEFINITE_WAIT );
	    #endif
	#else
	    DosSemRequest( &obj->semaphore, -1L );
	#endif
	obj->owner = tid;
    }
    obj->count++;
}

_WCRTLINK void __ReleaseSemaphore( semaphore_object *obj )
{
    #if defined( __NETWARE__ )
	if( GetCurrentThreadId() == 0 ) return;
    #endif
    if( obj->count > 0 ) {
	if( --obj->count == 0 ) {
	    obj->owner = 0;
	    #if defined( __386__ ) || defined( __AXP__ ) || defined( __PPC__ )
		#if defined( __NETWARE__ )
		    obj->semaphore = 0;
		#elif defined( __NT__ )
		    LeaveCriticalSection( obj->semaphore );
		#elif defined( __QNX__ )
		    __qsem_post( &obj->semaphore );
		#else
		    DosReleaseMutexSem( obj->semaphore );
		#endif
	    #else
		DosSemClear( &obj->semaphore );
	    #endif
	}
    } else {
	#if defined( __RUNTIME_CHECKS__ ) && defined( _M_IX86 )
	    assert( obj->count == 0 );
	#endif
    }
}

void	__AccessIOB()
/*******************/
{
    __AccessSemaphore( &IOBSemaphore );
}

void	__ReleaseIOB()
/********************/
{
    __ReleaseSemaphore( &IOBSemaphore );
}


void __AccessFileH( int handle )
/******************************/
{
    __AccessSemaphore( &FileSemaphores[ (unsigned)handle % MAX_SEMAPHORE ] );
}


void __ReleaseFileH( int handle )
/*******************************/
{
    __ReleaseSemaphore( &FileSemaphores[ (unsigned)handle % MAX_SEMAPHORE ] );
}


int	__InitAccessH( int handle )
/*********************************/
{
    #if defined( __NT__ )
	handle = __NTAddFileHandle( handle );
    #endif
    return( handle );
}


void	__FiniAccessH( int handle )
/*********************************/
{
    __CloseSemaphore( &FileSemaphores[ (unsigned)handle % MAX_SEMAPHORE ] );
    #if defined( __NT__ )
	__NTRemoveFileHandle( handle );
    #endif
}


#if !defined( __NETWARE__ )
void	__AccessNHeap()
/*********************/
{
    __AccessSemaphore( &NHeapSemaphore );
}

void	__ReleaseNHeap()
/**********************/
{
    __ReleaseSemaphore( &NHeapSemaphore );
}

void	__AccessFHeap()
/*********************/
{
    __AccessSemaphore( &FHeapSemaphore );
}

void	__ReleaseFHeap()
/**********************/
{
    __ReleaseSemaphore( &FHeapSemaphore );
}
#endif

#if defined( __386__ ) || defined( __AXP__ ) || defined( __PPC__ )

void	__AccessTDList()
/*********************/
{
    __AccessSemaphore( &TDListSemaphore );
}

void	__ReleaseTDList()
/**********************/
{
    __ReleaseSemaphore( &TDListSemaphore );
}

#if defined( __NT__ )
void	__AccessFList()
/*********************/
{
    __AccessSemaphore( &FListSemaphore );
}

void	__ReleaseFList()
/**********************/
{
    __ReleaseSemaphore( &FListSemaphore );
}
#endif
#endif

_WCRTLINK void *__MultipleThread()
{
    #if defined( __NT__ )
	/*
	 * Preserve old error code -- important because this code can get
	 * called from _STK.
	 */
	DWORD old = GetLastError();

	thread_data *tdata;
	tdata = (thread_data *)TlsGetValue( __TlsIndex );
	if( tdata == NULL ) {
	    tdata = __GetThreadData();
	} else if( tdata->__resize ) {
	    tdata = __ReallocThreadData();
	}
	SetLastError(old);
	return( tdata );
    #elif defined( __WARP__ )
	// 32 bit OS/2
	TID tid;
	thread_data *tdata = NULL;
	tid = GetCurrentThreadId();
	if( tid <= __MaxThreads ) {
	    tdata = __ThreadData[tid].data;
	}
	if( tdata == NULL ) {
	    tdata = __GetThreadData();
	} else if( tdata->__resize ) {
	    tdata = __ReallocThreadData();
	}
	return( tdata );
    #elif defined( __OS2_286__ )
	// 16 bit OS/2
	return( __ThreadData[GetCurrentThreadId()] );
    #elif defined( __QNX__ )
	void *tdata;
	__getmagicvar( &tdata, _m_thread_data );
	if( tdata == NULL ) {
	    tdata = __QNXAddThread( tdata );
	}
	return( tdata );
    #else
	return( __ThreadData[GetCurrentThreadId()].data );
    #endif
}

#if defined( __386__ ) || defined( __AXP__ ) || defined( __PPC__ )

thread_data *__AllocInitThreadData( thread_data *tdata )
/******************************************************/
{
    if( tdata == NULL ) {
	tdata = lib_calloc( 1, __ThreadDataSize );
	if( tdata != NULL ) {
	    tdata->__allocated = 1;
	    tdata->__data_size = __ThreadDataSize;
	}
    }
    __InitThreadData( tdata );
    return( tdata );
}

#if defined( __NT__ )

BOOL __NTThreadInit( void )
/*************************/
{
    if( __TlsIndex == NO_INDEX ) {
	__TlsIndex = TlsAlloc();
	// avoid Win32s bug (0-2 are assumed to be reserved by Win32s)
	if( WIN32_IS_WIN32S ) {
	    while( (__TlsIndex != NO_INDEX) && (__TlsIndex <= 2) ) {
		__TlsIndex = TlsAlloc();
	    }
	}
    }
    if( __TlsIndex == NO_INDEX ) {
	return( FALSE );
    }
    return( TRUE );
}


void __NTThreadFini( void )
/*************************/
{
    if( __TlsIndex != NO_INDEX ) {
	TlsFree( __TlsIndex );
	__TlsIndex = NO_INDEX;
    }
}


BOOL __NTAddThread( thread_data *tdata )
/**************************************/
{
    if( __TlsIndex == NO_INDEX ) {
	return( FALSE );
    }
    
    tdata = __AllocInitThreadData( tdata );
    if( tdata == NULL ) {
	return( FALSE );
    }
    if( !__AddThreadData( tdata->thread_id, tdata ) ) {
	lib_free( tdata );
	return( FALSE );
    }
    TlsSetValue( __TlsIndex, tdata );

    return( TRUE );
}


void __NTRemoveThread( int close_handle )
/***************************************/
{
    thread_data	*tdata;
    HANDLE	thread_handle;

    if( __TlsIndex != NO_INDEX ) {
	tdata = TlsGetValue( __TlsIndex );
	#if defined( __RUNTIME_CHECKS__ ) && defined( _M_IX86 )
	    if( tdata == (thread_data *)2 ) return;
	#else
	    if( tdata == NULL ) return;
	#endif
	thread_handle = tdata->thread_handle;
	__RemoveThreadData( tdata->thread_id );
	#if defined( __RUNTIME_CHECKS__ ) && defined( _M_IX86 )
	    TlsSetValue( __TlsIndex, (void*)2 );
	#else
	    TlsSetValue( __TlsIndex, NULL );
	#endif
	if( thread_handle != 0 && close_handle ) {
	    CloseHandle( thread_handle );
	}
    }
}


static void __ThreadExit()
/************************/
{
    __NTRemoveThread( TRUE );
    __NTThreadFini();
}

#elif defined( __OS2__ )

int __OS2AddThread( TID tid, thread_data *tdata )
/***********************************************/
{
    tdata = __AllocInitThreadData( tdata );
    if( tdata == NULL ) return( 0 );
    if( tid <= __MaxThreads ) {
	if( __initthread( tdata ) ) {
	    lib_free( tdata );
	    return( 0 );
	} else {
	    __ThreadData[tid].data = tdata;
	    __ThreadData[tid].allocated_entry = tdata->__allocated;
	}
    } else {
	if( !__AddThreadData( tid, tdata ) ) {
	    // unable to setup storage
	    lib_free( tdata );
	    return( 0 );
	}
    }
    return( 1 );
}

void __OS2RemoveThread( void )
/****************************/
{
    TID tid;
    tid = *_threadid;
    if( tid <= __MaxThreads ) {
	if( __ThreadData[tid].allocated_entry ) {
	    lib_free( __ThreadData[tid].data );
	}
	__ThreadData[tid].data = NULL;
    } else {
	__RemoveThreadData( tid );
    }
}

#elif defined( __QNX__ )

thread_data *__QNXAddThread( thread_data *tdata )
/***********************************************/
{
    void *tmp;
    tdata = __AllocInitThreadData( tdata );
    // if tdata is NULL it doesn't matter what we do with it
    tmp = (void *)tdata;
    __setmagicvar( &tmp, _m_thread_data );
    return( tdata );
}

void __QNXRemoveThread( void )
/****************************/
{
    void *tmp;
    thread_data *tdata;

    __getmagicvar( &tmp, _m_thread_data );
    if( tmp != NULL ) {
	tdata = tmp;
	if( tdata->__allocated ) {
	    lib_free( tdata );
	}
	tmp = 0;
	__setmagicvar( &tmp, _m_thread_data );
    }
}

#endif
 
void __InitMultipleThread()
/*************************/
{
    _AccessFileH   = &__AccessFileH;
    _ReleaseFileH  = &__ReleaseFileH;
    _InitAccessH   = &__InitAccessH;
    _FiniAccessH   = &__FiniAccessH;
    _AccessIOB     = &__AccessIOB;
    _ReleaseIOB    = &__ReleaseIOB;
    _AccessTDList  = &__AccessTDList;
    _ReleaseTDList = &__ReleaseTDList;
    __AccessSema4  = &__AccessSemaphore;
    __ReleaseSema4 = &__ReleaseSemaphore;
    __CloseSema4   = &__CloseSemaphore;
    #if !defined( __NETWARE__ )
	_AccessNHeap  = &__AccessNHeap;
	_AccessFHeap  = &__AccessFHeap;
	_ReleaseNHeap = &__ReleaseNHeap;
	_ReleaseFHeap = &__ReleaseFHeap;
    #endif
    #if defined( __NETWARE__ )
    {
    /* __ThreadData[ 0 ] is used whenever GetThreadID() returns a pointer
       not in our __ThreadIDs list - ie. whenever it returns NULL, a
       pointer to a thread we didn't create, or an invalid pointer */
	void *ptr;
	ptr = lib_calloc( 1, __ThreadDataSize );
	if( ptr == NULL ) {
	    __fatal_runtime_error(
		"Unable to allocate thread-specific data\r\n", 1 );
	}
	__ThreadData[ 0 ].data = ptr;
	__ThreadData[ 0 ].allocated_entry = 1;
	__ThreadData[ 0 ].data->__allocated = 1;
	__ThreadData[ 0 ].data->__randnext = 1;
	__ThreadData[ 0 ].data->__data_size = __ThreadDataSize;
	if( __initthread( ptr ) ) {
	    lib_free( ptr );
	    __fatal_runtime_error(
		"Unable to initialize thread-specific data\r\n", 1 );
	}
	ptr = lib_calloc( 1, __ThreadDataSize );
	if( ptr == NULL ) {
	    __fatal_runtime_error(
		"Unable to allocate thread-specific data\r\n", 1 );
	}
	__FirstThreadData = ptr;
	__FirstThreadData->__allocated = 1;
	__FirstThreadData->__randnext = 1;
	__FirstThreadData->__data_size = __ThreadDataSize;
    	__ThreadData[ 1 ].data = __FirstThreadData;
    	__ThreadData[ 1 ].allocated_entry = __FirstThreadData->__allocated;
	__ThreadIDs[ 1 ] = GetThreadID();
	if( __initthread( ptr ) ) {
	    lib_free( ptr );
	    __fatal_runtime_error(
		"Unable to initialize thread-specific data\r\n", 1 );
	}
    }
    #elif defined( __NT__ )
	InitSemaphore.semaphore = __NTGetCriticalSection();
	InitSemaphore.initialized = 1;
	_AccessFList  = &__AccessFList;
	_ReleaseFList = &__ReleaseFList;
	_ThreadExitRtn= &__ThreadExit;
	// note that __AddThreadData uses the InitSemaphore
	__AddThreadData( __FirstThreadData->thread_id, __FirstThreadData );
	TlsSetValue( __TlsIndex, __FirstThreadData );
    #elif defined( __QNX__ )
	__qsem_init( &InitSemaphore.semaphore, 1, 1 );
	InitSemaphore.initialized = 1;
	// first thread data already in magic memory
    #else
	DosCreateMutexSem( NULL, &InitSemaphore.semaphore, 0, FALSE );
	InitSemaphore.initialized = 1;
	__ThreadData[1].data = __FirstThreadData;
	__ThreadData[1].allocated_entry = __FirstThreadData->__allocated;
    #endif
    __GetThreadPtr  = &__MultipleThread;
}
#endif

void static __FiniSema4s()		// called from finalizer
/************************/
{
    int		i;

    _CloseSemaphore( &IOBSemaphore );
    for( i = 0; i < MAX_SEMAPHORE; i++ ) {		/* 17-feb-93 */
	_CloseSemaphore( &FileSemaphores[ i ] );
    }
    #if defined( __NT__ )
	_CloseSemaphore( &FListSemaphore );
	__NTFreeCriticalSection();
    #endif
    #if !defined( __QNX__ )
	__FiniThreadProcessing();
    #endif
    #if !defined( __NETWARE__ )
	_heapshrink();
	_CloseSemaphore( &NHeapSemaphore );
	_CloseSemaphore( &FHeapSemaphore );
    #endif
    #if defined( __386__ ) || defined( __AXP__ ) || defined( __PPC__ )
	_CloseSemaphore( &TDListSemaphore );
        _CloseSemaphore( &InitSemaphore );
	#if defined( __NT__ )
	    __NTDeleteCriticalSection();
	    __NTThreadFini();
	#endif
    #endif
}

AYI( __FiniSema4s, INIT_PRIORITY_RUNTIME )
