/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1989, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  19-jan-95	Greg Bentz	defined
--------------------------------------------------------Version-11.0---------
  07-mar-95	M. Hildebrand	Include semaphor.h, not semaphore.h.
  30-may-95	Greg Bentz	variety.h and dll linkage
*/

#include "variety.h"
#include <semaphor.h>
#include <string.h>
#include <signal.h>
#include <float.h>
#include <unistd.h>
#include <errno.h>
#include <process.h>

#include "liballoc.h"
#include "osthread.h"
#include "stacklow.h"
#include "rtdata.h"
#include "extfunc.h"
#include "mthread.h"
#include "seterrno.h"
#include "semaqnx.h"

extern	void		__InitMultipleThread(void);
extern	void		_endthread( void );
extern	thread_data	*__QNXAddThread( thread_data *tdata );
extern	void		__QNXRemoveThread( void );
extern  void    	__qnx_exit( int __status );

typedef struct thread_args {
    thread_fn	*rtn;
    void	*argument;
    void	*stack_bottom;
    sem_t	event;
} thread_args;


static int begin_thread_helper( void *ptr )
/*****************************************/
{
    thread_args			*td;
    thread_fn			*rtn;
    void			*arg;
    thread_data 		*tdata;
    
    td = ptr;
    rtn = td->rtn;
    arg = td->argument;

    tdata = __alloca( __ThreadDataSize );
    memset( tdata, 0, __ThreadDataSize );
    // tdata->__allocated = 0;
    tdata->__data_size = __ThreadDataSize;
    __QNXAddThread( tdata );
    _STACKLOW = (unsigned)td->stack_bottom;

    __qsem_post( &td->event );
    _fpreset();
    (*rtn)( arg );
    _endthread();
    return( 0 );
}

int __CBeginThread( thread_fn *start_addr, void *stack_bottom,
		    unsigned stack_size, void *arglist )
/******************************************************/
{
    pid_t	pid;
    thread_args	td;
    int		rc;
	
    if( stack_bottom == NULL ) {
	if( stack_size == 0 ) {
	    stack_size = 1024*4;
	}
	stack_bottom = lib_calloc( stack_size, 1 );
	if( stack_bottom == NULL ) {
	    __set_errno( ENOMEM );
	    return( -1 );
	}
    }
    if( __GetThreadPtr != &__MultipleThread ) {
	__InitMultipleThread();
    }
    td.rtn = start_addr;
    td.argument = arglist;
    td.stack_bottom = stack_bottom;
    rc = __qsem_init( &td.event, 1, 0 );
    if( rc == -1 ) return( -1 );
    pid = tfork( stack_bottom, stack_size, begin_thread_helper, &td, 0 );
    if( pid != -1 ) {
	/*
	   suspend parent thread so that it can't call _beginthread() again
	   before new thread extracts data from "td" (no problem if new
	   thread calls _beginthread() since it has its own stack)
	*/
	__qsem_wait( &td.event );
    }
    __qsem_destroy( &td.event );
    return( pid );
}

void __CEndThread( void )
/***********************/
{
    __QNXRemoveThread();
    __qnx_exit( 0 );
}
