/*
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %       Copyright (C) 1992, by WATCOM Systems Inc. All rights     %
  %       reserved. No part of this software may be reproduced      %
  %       in any form or by any means - graphic, electronic or      %
  %       mechanical, including photocopying, recording, taping     %
  %       or information storage and retrieval systems - except     %
  %       with the written permission of WATCOM Systems Inc.        %
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  06-mar-92     Craig Eisler	defined for NT
  24-sep-92	A.F.Scian	added __outside_CLIB pragma to fn ptrs
  21-oct-92	Craig Eisler	use thread local storage
  07-nov-92	F.W.Crigger	cleaned up thread support
  11-nov-92	A.F.Scian	fixed call to memset (last 2 parms were
  				flipped around)
  23-nov-92     Craig Eisler	exception handling
  16-dec-92     G. Coschi	in begin_thread_helper(), make sure
				TlsSetValue() is called before thread-specific
				data is accessed (i.e. _STACKLOW)
  04-feb-93     Craig Eisler	added thread_id to thread-specific data
  04-feb-93     G. Coschi	initialized thread_id in NTAddThread() for DLLs
  29-jul-93	F.W.Crigger	introduced __BeginThread, __EndThread vars so
				FORTRAN can change it to point to a FORTRAN
				specific _beginthread, _endthread functions
  16-aug-94	Greg Bentz	change alloca() to __alloca()
  27-aug-94	Greg Bentz	set stacklow to 0 in child thread and remove
  				call to GetThreadContext() because it causes
				a seg fault in KERNEL32.DLL under Chicago
				stacklow==0 is ok since NT grows the stack
				note that the child is now also started in
				a non-suspended state
--------------------------------------------------------Version-11.0---------
  30-may-95	Greg Bentz	variety.h and dll linkage
  22-jun-95	T. Schiller	modified _CBeginThread() to match the
  				Microsoft implementation
  15-sep-95	Greg Bentz	ppc support
  22-may-97	Greg Bentz	sequencing issue with setting the thread handle
*/
#include "variety.h"
#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <malloc.h>
#include "liballoc.h"
#include "ntex.h"
#include "stacklow.h"
#include "sigtab.h"
#include "thread.h"
#include "trdlist.h"
#include "mthread.h"
#include "rtdata.h"
#include "extfunc.h"

extern	void		__InitMultipleThread( void );

extern	DWORD		__TlsIndex;
extern	int		__Is_DLL;

extern	void		(*__sig_init_rtn)(void);
extern	void		(*__sig_fini_rtn)(void);

extern	void		_endthread(void);

typedef struct thread_args {
    thread_fn	*rtn;
    void	*argument;
    HANDLE	parent;
    HANDLE	event;
    HANDLE	*thread_handle;
} thread_args;

static __cdecl begin_thread_helper( thread_args *td ) 
/***************************************************/
{
    thread_fn		*rtn;
    void		*arg;
    REGISTRATION_RECORD	rr;
    thread_data		*tdata;

    rtn = td->rtn;
    arg = td->argument;

    if( !__Is_DLL ) {					/* 15-feb-93 */
	#if defined(__AXP__) || defined(__PPC__)
	    tdata = (thread_data *)alloca( __ThreadDataSize );
	#else
	    tdata = (thread_data *)__alloca( __ThreadDataSize );
	#endif
	memset( tdata, 0, __ThreadDataSize );
	// tdata->__allocated = 0;
	tdata->__data_size = __ThreadDataSize;
	if( !__NTAddThread( tdata ) ) {
	    return;
	}
    }
    td->thread_handle = &(__THREADDATAPTR->thread_handle);
    SetEvent( td->event );

    __NewExceptionHandler( &rr );
    __sig_init_rtn();
    (*rtn)( arg );
    _endthread();
}

unsigned long __CBeginThread(thread_fn *start_addr, unsigned stack_size,
		     void *arglist)
{

    DWORD	tid;
    thread_args	td;
    HANDLE	th;
    char	evname[32];
	
    if( __TlsIndex == NO_INDEX ) {
	if( __NTThreadInit() == FALSE )  return NULL;
	__InitMultipleThread();
    }
    //stack_bottom = stack_bottom;
    
    stack_size = __Align4K( stack_size );
    
    td.rtn = start_addr;
    td.argument = arglist;
    td.parent = GetCurrentThread();
    td.thread_handle = 0;
    strcpy( evname,"__bgnthd" );
    itoa( GetCurrentThreadId(), &evname[8], 16 );
    td.event = CreateEvent( NULL, FALSE, FALSE, evname );
    
    th = CreateThread( NULL, stack_size, (LPVOID) begin_thread_helper,
		(LPVOID) &td, 0, &tid );
    if( th == 0 ) {
	tid = -1;
    } else {
	WaitForSingleObject( td.event, -1 );
	if( td.thread_handle ) {
	    (*td.thread_handle) = th;
	} else {
	    CloseHandle( th );
	}
    }
    CloseHandle( td.event );
    return( (unsigned long)th );
}

void __CEndThread( void )
{
    __sig_fini_rtn();
    __DoneExceptionHandler();
    if( ! __Is_DLL ) {
	__NTRemoveThread( TRUE );
    }
    ExitThread( 0 );
}
