/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1991, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  91/12/03	G. Coschi	initial implementation
  92/01/02	G. Coschi	added functional interface for maximum
  				# of threads (required for FORTRAN)
  92/10/21	Greg Bentz	modified declaration of __ThreadData to
  				keep C++ compiler happy
  92/11/07	F.W.Crigger	Make __ThreadData a pointer to an array of
				pointers to thread_data
  93/01/29	A. Gupta	added __ThreadIDs structure for Netware
  93/03/04	A. Gupta	needed to free __ThreadData[ 0 ] which is
                                now always allocated (under Netware only)
  18-jun-93	John Dahms  	all library allocations use lib_xxx routines
  04-may-94	Greg Bentz	improve allocation failure checking
--------------------------------------------------------Version-11.0---------
  22-feb-95	Greg Bentz	alpha support
  30-may-95	Greg Bentz	variety.h and dll linkage
  15-sep-95	Greg Bentz	ppc support
*/

#include "variety.h"
#include <stdlib.h>
#include <string.h>
#include "thread.h"
#include "trdlist.h"
#include "liballoc.h"

#ifdef __NETWARE__
void			**__ThreadIDs;
#endif

#if defined(__386__) || defined(__AXP__) || defined(__PPC__)
    #if !defined(__NT__) && !defined(__QNX__)
	thread_data_vector	*__ThreadData;
    #endif
#else
    thread_data		**__ThreadData;
#endif


#if !defined(__NT__) && !defined(__QNX__)
void *__InitThreadProcessing()
/****************************/
{
// Thread structures must be initialized to 0 so that if it's for a DLL
// _STACKLOW is 0.

    // allocate 1 more than we have to since thread 1 will
    // actually access 2nd element of __ThreadData
    
    __MaxThreads = __GetMaxThreads();
    __ThreadData = lib_calloc( (__MaxThreads + 1), sizeof( *__ThreadData ) );
    #ifdef __NETWARE__
	if( __ThreadData != NULL ) {
	    __ThreadIDs = lib_calloc( __MaxThreads + 1, sizeof( int ) );
	    if( __ThreadIDs == NULL ) {
		lib_free( __ThreadData );
	    }
	    return( __ThreadIDs );
	}
    #endif
    return( __ThreadData );
}
#endif


#if defined(__OS2_286__)
// only needed by 16 bit OS/2
void __SetupThreadProcessing( int i ) {
/*************************************/
    thread_data	*tdata;

    tdata = calloc( 1, sizeof( thread_data ) );
    tdata->__allocated = TRUE;
    tdata->__randnext = 1;
    __ThreadData[i] = tdata;
}
#endif


void __FiniThreadProcessing()
/***************************/
{

    #ifdef __NETWARE__
	if( __ThreadIDs != NULL ) {
	    lib_free( __ThreadIDs );
	}
    #endif
    #if !defined(__NT__) && !defined(__QNX__)
	if( __ThreadData != NULL ) {
	    unsigned	i;
	    thread_data	*tdata;
	    #ifdef __NETWARE__
		lib_free( __ThreadData[ 0 ].data ); /* for Netware, this is always allocated */
	    #endif
	    for( i = 1 ; i <= __MaxThreads; i++ ) {
		#if defined(__386__) || defined(__AXP__) || defined(__PPC__)
		    tdata = __ThreadData[i].data;
		    if( tdata != NULL ) {
			if( __ThreadData[i].allocated_entry ) lib_free( tdata );
		    }
		#else
		    tdata = __ThreadData[i];
		    if( tdata != NULL ) {
			if( tdata->__allocated ) lib_free( tdata );
		    }
		#endif
	    }
	    lib_free( __ThreadData );
	}
    #endif
    #if !defined(__NETWARE__) && (defined(__386__) || defined(__AXP__) || defined(__PPC__) )
	__FreeThreadDataList();
    #endif
}
