/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  xx-sep-87	P.J.Bumbulis	Initial implementation
---------------------------------------------------------Version 6.5-------
  03-aug-88	A.F.Scian	fixed null check of pointer from nmalloc
  12-aug-88	F.W.Crigger	fix alignment of p to segment boundary in
				large data models
---------------------------------------------------------Version 7.0-------
  24-apr-89	F.W.Crigger	Don't have to put on segment boundary for 386
  18-mar-91	F.W.Crigger	Change _cenvarg to __cenvarg
---------------------------------------------------------Version 8.5-------
  10-feb-92	A.F.Scian	fixed problem with extra ' ' in front of args
---------------------------------------------------------Version 9.0-------
  02-jun-92	F.W.Crigger	calculation for cmd tail length was wrong
				for empty command tail
  21-jul-92	D.J.Gaudet	handle long command lines under OS/2 and NT
  07-oct-92	F.W.Crigger	Added envstrings parm
  18-jun-93	John Dahms  	all library allocations use lib_xxx routines
--------------------------------------------------------Version-11.0---------
  22-feb-95	Greg Bentz	alpha support
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  30-may-95	Greg Bentz	dll linkage
  15-sep-95	Greg Bentz	ppc support
  04-oct-95	M. Hildebrand	Added wide char versions for NT.
*/
#undef __INLINE_FUNCTIONS__
#include "dll.h"	// needs to be first
#include "variety.h"
#include "widechar.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "liballoc.h"
#include <dos.h>
#include "msdos.h"
#include "rtdata.h"
#include "seterrno.h"

#ifdef __WIDECHAR__
extern _WCRTLINK void	__create_wide_environment( void );
#endif


static CHAR_TYPE *stpcpy( CHAR_TYPE *dest, const CHAR_TYPE *src ) {

    while( *dest = *src ) {
	++dest;
	++src;
    }
    return( dest );
}


int __F_NAME(__cenvarg,__wcenvarg)( argv, envp, envptr, envstrings, envseg, cmdline_len, exec )
/*
 *  Build environment and command line for new process.  Length of environment
 *  (in bytes) is returned on success.	-1 is returned on failure.
 */
    CHAR_TYPE		*argv[];	/* i: arguments for new process */
    CHAR_TYPE		*envp[];	/* i: env strings for new process */
    CHAR_TYPE		**envptr;	/* o: allocated memory for env */
    CHAR_TYPE		**envstrings;	/* o: pointer to environment strings */
    unsigned		*envseg;	/* o: start of env (on para boundary) */
    size_t		*cmdline_len;	/* o: size required to hold cmd line */
    int 		exec;		/* i: TRUE if for exec */
{
    unsigned		length;
    unsigned		oamblksiz;
    CHAR_TYPE		*p;
    CHAR_TYPE		**envv;
    unsigned		len;
    CHAR_TYPE		*arg;

    if( envp == NULL ){
	#ifdef __WIDECHAR__
	    if( _RWD_wenviron == NULL )  __create_wide_environment();
	    envp = _RWD_wenviron;
	#else
	    envp = _RWD_environ;
	#endif
    }
    length = 0;
    if( envp != NULL ){
	for( envv = envp; *envv != NULL; ++envv ){
	    length += __F_NAME(strlen,wcslen)( *envv ) + 1;
	}
    }
    ++length; /* trailing \0 for env */
    if( exec ){
	/* store argv[0] at 2 bytes past end of env */
	length += 2 + __F_NAME(strlen,wcslen)( argv[0] ) + 1;
    }
    length += 15;	/* so we can start on a paragraph boundary */

    oamblksiz = _RWD_amblksiz;
    _RWD_amblksiz = 16; /* force allocation in 16 byte increments */
    p = lib_nmalloc( length*sizeof(CHAR_TYPE) );
    if( (void _WCNEAR *)p == NULL ){		/* 03-aug-88 */
	p = lib_malloc( length*sizeof(CHAR_TYPE) );
	if( p == NULL ){
	    __set_errno( ENOMEM );
	    __set_doserrno( E_nomem );
	    _RWD_amblksiz = oamblksiz;
	    return( -1 );
	}
    }
    _RWD_amblksiz = oamblksiz;
    *envptr = p;
#if defined(__386__) || defined(__AXP__) || defined(__OS2__) || defined(__PPC__)
    *envseg = 0;
#else
#if defined(__SMALL_DATA__)
    p = (char *) (((unsigned) p + 15) & 0xfff0);
#else		/* large data models */ 	/* 12-aug-88 */
    p = MK_FP( FP_SEG(p), (( FP_OFF(p) + 15) & 0xfff0) );
#endif
    {
	CHAR_TYPE _WCFAR *temp;

	temp = p;
	*envseg = FP_SEG( temp ) + FP_OFF( temp )/16;
    }
#endif
    *envstrings = p;		/* save ptr to env strings. 07-oct-92 */
    if( envp != NULL ){
	for( envv = envp; *envv != NULL; ++envv ){
	    p = stpcpy( p, *envv ) + 1;
	}
    }
    *p++ = '\0';
    if( exec ) {
	__F_NAME(strcpy,wcscpy)( p + 2, argv[0] );
    }

    len = 0;
    if( argv[0] != NULL ){
	for( ++argv; (arg = *argv) != NULL; ++argv ){
	    if( len != 0 ) ++len;	/* plus 1 for blank separator */
	    len += __F_NAME(strlen,wcslen)( arg );
	}
    }
#if defined( __OS2__ ) || defined( __NT__ )
    len += _MAX_PATH2 + 1;
#else
    if( len > 126 ){
	__set_errno( E2BIG );
	__set_doserrno( E_badenv );
	lib_free( *envptr );
	return( -1 );
    }
    len = _MAX_PATH;	/* always use _MAX_PATH chars for DOS */
#endif
    *cmdline_len = len;

    return( length/16 );
}


void __F_NAME(__ccmdline,__wccmdline)( CHAR_TYPE *path, CHAR_TYPE *argv[], CHAR_TYPE *buffer,
				       int just_args )
{
/*
 * path is the name of the program we are about to spawn
 * argv is the argument array
 * buffer points to memory large enough to hold the resulting command line
 * just_args means build a command line with arguments only
 *
 * Each OS wants the command line to look a little bit different.
 */

    CHAR_TYPE *p;

    if( just_args ) {
	p = buffer;
    } else {
#if defined( __OS2__ )
	path = path;
	/* OS/2 wants:  argv[0] '\0' arguments '\0' */
	p = stpcpy( buffer, argv[0] ) + 1;
#elif defined( __NT__ )
	/* NT wants: path ' ' arguments '\0' */
	p = stpcpy( buffer, path );
	*p++ = ' ';
#else
	path = path;
	/* DOS wants: len_byte arguments '\r' */
	p = buffer + 1;
#endif
    }
    if( argv[0] != NULL ){
	++argv;
	if( *argv != NULL ) {
	    for(;;) {
		p = stpcpy( p, *argv );
		++argv;
		if( *argv == NULL ) break;
		*p++ = ' '; /* put the space between args */
	    }
	}
    }
#if defined( __OS2__ ) || defined( __NT__ )
    *p = 0;
#else
    if( just_args ) {
	*p = 0;
    } else {
	*p = '\r';
	buffer[0] = ( p - buffer ) - 1;
    }
#endif
}
