/*
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %       Copyright (C) 1992, by WATCOM Systems Inc. All rights     %
  %       reserved. No part of this software may be reproduced      %
  %       in any form or by any means - graphic, electronic or      %
  %       mechanical, including photocopying, recording, taping     %
  %       or information storage and retrieval systems - except     %
  %       with the written permission of WATCOM Systems Inc.        %
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  05-mar-92     Craig Eisler	defined for NT
  21-jul-92	D.J.Gaudet	handle long command lines
  16-feb-94	Greg Bentz	make P_WAIT work for WIN32s
  		David Wellstood	^
--------------------------------------------------------Version-11.0---------
  22-feb-95	Greg Bentz	alpha support
  30-may-95	Greg Bentz	variety.h and dll linkage
  04-aug-95	Greg Bentz	NT error code checking
  15-sep-95	Greg Bentz	ppc support
  04-oct-95	M. Hildebrand	Added _wdospawn for NT.
  19-oct-95	M. Hildebrand	Returned process ID is now the HANDLE of
				the spawned process.
  28-nov-95	Greg Bentz	fix spawn of .exe from Netware Server with
  				long filename support by checking for extra
				error return code
  13-dec-95	M. Hildebrand	Return PID, not HANDLE, when mode==P_NOWAITO.
  28-mar-96	Greg Bentz	fix check for failure on dup handle create
  18-jun-96	Greg Bentz	make P_WAIT work for windows 95
  24-jun-96	M. Hildebrand	Win95 doesn't support CreateProcessW.
  05-jul-96	M. Hildebrand	Use __lib_ versions of ...W Win32 functions.
*/
#include "variety.h"
#include "widechar.h"
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include <string.h>
#include <stddef.h>
#include <malloc.h>
#include <process.h>
#include <windows.h>
#include "libwin32.h"
#include "osver.h"
#include "rtdata.h"
#include "seterrno.h"

extern  void	__F_NAME(__ccmdline,__wccmdline)( CHAR_TYPE *, CHAR_TYPE **, CHAR_TYPE *, int );

#ifdef _M_IX86
 #pragma aux	_dospawn "_*" parm caller [];
#endif
int __F_NAME(_dospawn,_wdospawn)( int mode, CHAR_TYPE *pgmname, CHAR_TYPE *cmdline,
				  CHAR_TYPE *envp, CHAR_TYPE *argv[] )
{
    STARTUPINFO		sinfo;
    PROCESS_INFORMATION	pinfo;
    DWORD		rc;
    BOOL		osrc;

    __F_NAME(__ccmdline,__wccmdline)( pgmname, argv, cmdline, 0 );

    memset( &sinfo, 0, sizeof( sinfo ) );
    sinfo.wShowWindow = SW_NORMAL;
    
    // When passing in Unicode environments, the OS may not know which code
    // page to use when translating to MBCS in spawned program's startup
    // code.  Result: Possible corruption of Unicode environment variables.
    #ifdef __WIDECHAR__
	#if defined(__AXP__) || defined(__PPC__)
	    osrc = CreateProcessW( NULL, cmdline, NULL, NULL, TRUE,
				   CREATE_UNICODE_ENVIRONMENT, envp,
				   NULL, &sinfo, &pinfo );
	#else
	    osrc = __lib_CreateProcessW( cmdline, TRUE, envp, &pinfo );
	#endif
    #else
	osrc = CreateProcessA( NULL, cmdline, NULL, NULL, TRUE, 0,
			       envp, NULL, &sinfo, &pinfo );
    #endif

    if( osrc == FALSE ) {
	DWORD err;
	err = GetLastError();
	if( (err == ERROR_ACCESS_DENIED)
	 || (err == ERROR_BAD_EXE_FORMAT)
	 || (err == ERROR_BAD_PATHNAME) ) {
	    err = ERROR_FILE_NOT_FOUND;
	}
	return( __set_errno_dos( err ) );
    }
    if( mode == P_WAIT ) {
	if( WIN32_IS_WIN32S ) {
	    // this is WIN32s
	    Sleep( 1000 );
	    rc = STILL_ACTIVE;
	    while( rc == STILL_ACTIVE ) {
		Sleep( 100 );
		if( !GetExitCodeProcess( pinfo.hProcess, &rc ) ) {
		    return( __set_errno_nt() );
		}
	    }
	} else {
	    // this is WIN32 or Windows95
	    if( WaitForSingleObject( pinfo.hProcess, -1 ) == 0 ) {
		GetExitCodeProcess( pinfo.hProcess, &rc );
	    } else {
		rc = __set_errno_nt();
	    }
	}
	CloseHandle( pinfo.hProcess );
    } else if( mode == P_NOWAITO ) {
	CloseHandle( pinfo.hProcess );
	rc = pinfo.dwProcessId;
    } else {
	/*
	 * Since pinfo.dwProcessId is only valid until the process
	 * terminates, cwait will be unable to use it to get the return
	 * code of the process if it's already ended before cwait is
	 * called.  Instead, use pinfo.hProcess as the process ID, since
	 * it is valid even after the process ends.  If cwait is called,
	 * it can do the CloseHandle; if it's not called, we'll just have
	 * an extra handle sitting about, which will be destroyed when
	 * the program terminates.
	 *		-- M.Hildebrand, 19-oct-95
	 */
	HANDLE		osHandle;
	BOOL		osRc;

	/*** Make a non-inheritable handle; use old one on failure ***/
	osRc = DuplicateHandle( GetCurrentProcess(), pinfo.hProcess,
				GetCurrentProcess(), &osHandle, 0,
				FALSE, DUPLICATE_SAME_ACCESS );
	if( !osRc )  osHandle = pinfo.hProcess;
	rc = (int)osHandle;
    }
    CloseHandle( pinfo.hThread );
    return( rc );
}
