/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  17-jul-87	F.W.Crigger	Initial implementation
  09-nov-87	F.W.Crigger	Reset signal to SIG_DFL after being raised
----------------------------------------------------------Version 6.0-------
  17-apr-88	F.W.Crigger	abort now prints "ABNORMAL TERMINATION"
  19-may-88	F.W.Crigger	Added support for catching int 0x23
  23-may-88	F.W.Crigger	Reset _int23_exit to point to null exit rtn
  24-may-88	F.W.Crigger	Introduce _child variable
----------------------------------------------------------Version 6.5-------
  16-aug-88	F.W.Crigger	__abort variable introduced
----------------------------------------------------------Version 7.0-------
  25-may-89	John Dahms	moved _raise to separate file
  12-jul-89	F.W.Crigger	__setvect for catching CTRL-C under Phar Lap
  13-jul-89	F.W.Crigger	__setvect modified to handle A.I.Architects
  20-jul-89	F.W.Crigger	__FPE_handler variable introduced
  31-jul-89	F.W.Crigger	call to __Init_FPE_handler() added
  05-apr-90	F.W.Crigger	Added parm to __sigfpe_handler
  05-apr-90	J.B.Schueler	added sig number range < 1 check
  11-may-90	F.W.Crigger	grab_fpe indirectly thru __Hook_FPE_Signal
  21-may-90	F.W.Crigger	Added SIGBREAK, SIGUSR1, SIGUSR2, SIGUSR3
  11-jun-90	F.W.Crigger	Added __FAR__ to __sigfpe_handler for 8086
  12-jul-90	F.W.Crigger	signal - Only grab interrupts if func=user rtn
  04-feb-91	C.G.Eisler	updated for Windows 386
  03-apr-91	C.G.Eisler	updated for WINDOWS_386 macro
  15-jan-92	G. Coschi	added SIGIDIVZ and SIGIOVFL
  04-feb-92	J.B.Schueler	Changed __FAR__ to _FAR_ for __sigfpe_handler
  24-sep-92	A.F.Scian	added __outside_CLIB pragma to fn ptrs
  17-dec-92	G. Coschi	argument to __sigfpe_handler() always passed
  				in register from exception handler, even if
				using stack calling convention
  93-mar-05	A. Gupta	Netware support - can't trap fp exceptions
  				under Netware
  26-Mar-93	A.F.Scian	if SIGINT is set back to SIG_DFL, restore int23
--------------------------------------------------------Version-11.0---------
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  11-may-95	T. Schiller	Updated default signal states.
  30-may-95	Greg Bentz	dll linkage
  01-jun-95	T. Schiller	Modifications to make FPE's work under Win-16.
  20-jun-95	T. Schiller	Fixed bug in which we didn't install the
  				fpe handler when targeting 16-bit MS-DOS.
  15-sep-95	Greg Bentz	ppc support
*/

#include "variety.h"
#include <stdio.h>
#include <signal.h>
#include <dos.h>
#include <errno.h>
#include <float.h>
#include "rtdata.h"
#include "sigtab.h"
#include "extfunc.h"
#include "seterrno.h"

typedef void sig_func();
#ifdef _M_IX86
    #pragma aux (__outside_CLIB) sig_func;
#endif

#ifndef __WINDOWS_386__
#ifndef __NETWARE__
    extern	void	__grab_int23();
    extern	void	__restore_int23();
    extern	void	__grab_int_ctrl_break();
    extern	void	__restore_int_ctrl_break();
    extern	void	__grab_fpe();
#endif

extern	void	__terminate();
#endif

extern	void	(*__abort)();

#define __SIGLAST	SIGIOVFL

static void (* _HUGEDATA SignalTable[])(int) = {
	SIG_IGN,	/* unused  */
	SIG_DFL,	/* SIGABRT */
	SIG_IGN,	/* SIGFPE  */
	SIG_DFL,	/* SIGILL  */
	SIG_DFL,	/* SIGINT  */
	SIG_DFL,	/* SIGSEGV */
	SIG_DFL,	/* SIGTERM */
	SIG_DFL,	/* SIGBREAK */
	SIG_IGN,	/* SIGUSR1 */
	SIG_IGN,	/* SIGUSR2 */
	SIG_IGN,	/* SIGUSR3 */
	SIG_DFL,	/* SIGIDIVZ */
	SIG_DFL 	/* SIGIOVFL */
};

void __sigabort()
    {
	raise( SIGABRT );
    }

#ifndef __WINDOWS_386__
/* arguments always passed in registers, even if stack calling convention */

#if defined( __386__ )
#pragma aux __sigfpe_handler parm [eax]
#else
#pragma aux __sigfpe_handler parm [ax]
#endif

#if defined(__WINDOWS__)

// called from emulator callback
void __based(__segname("_TEXT")) __raise_fpe(void) {
    raise(SIGFPE);
}

extern __far _fpmath;

unsigned int win87em_get_sw(void);
#pragma aux win87em_get_sw = \
	"push	bx"					\
	"mov	bx, 8h"					\
	"call	far _fpmath"				\
	"pop	bx"					\
	value [ax]
#endif

_WCRTLINK void _WCI86FAR __sigfpe_handler( int fpe_type )
    {
	sig_func     *func;

	#if defined(__WINDOWS__)
	    unsigned int  sw;
	    sw = win87em_get_sw();

	    if (sw & EM_INVALID) {
		fpe_type = FPE_INVALID;
	    } else if (sw & EM_DENORMAL) {
		fpe_type = FPE_DENORMAL;
	    } else if (sw & EM_ZERODIVIDE) {
		fpe_type = FPE_ZERODIVIDE;
	    } else if (sw & EM_OVERFLOW) {
		fpe_type = FPE_OVERFLOW;
	    } else if (sw & EM_UNDERFLOW) {
		fpe_type = FPE_UNDERFLOW;
	    } else if (sw & EM_INEXACT) {
		fpe_type = FPE_INEXACT;
	    }
	#endif

	func = (sig_func *) SignalTable[ SIGFPE ];
	if( func != SIG_IGN  &&  func != SIG_DFL  &&  func != SIG_ERR ) {
	    SignalTable[ SIGFPE ] = SIG_DFL;	  /* 09-nov-87 FWC */
	    (*func)( SIGFPE, fpe_type );	/* so we can pass 2'nd parm */
	}
    }
#endif

_WCRTLINK void (*signal( int sig, void (*func)(int) ))(int)
    {
	void (*prev_func)(int);

	if( (sig < 1) || (sig > __SIGLAST) ) {
	    __set_errno( EINVAL );
	    return( SIG_ERR );
	}
	__abort = __sigabort;		/* change the abort rtn address */
#if !defined( __WINDOWS_386__ ) && !defined( __NETWARE__ )
	if( sig == SIGINT ) {
	    if( func == SIG_DFL ) {
		__restore_int23();
	    } else if( func != SIG_ERR ) {
		__grab_int23();
	    }
	} else if( sig == SIGBREAK ) {
	    if( func == SIG_DFL ) {
		__restore_int_ctrl_break();
	    } else if( func != SIG_ERR ) {
		__grab_int_ctrl_break();
	    }
	} else if( sig == SIGFPE ) {
	    if( func == SIG_DFL ) // Turn exceptions on
	    {
		/*
		 * Mask all exceptions.
		 */
		_control87(~0, 0x002F);
	    }
	    else if( func != SIG_ERR ) // Turn exceptions off
	    {
		/*
		 * Unmask the default exceptions.
		 */
		_control87(0, EM_INVALID | EM_ZERODIVIDE | EM_OVERFLOW
				| EM_UNDERFLOW);
#ifndef __WINDOWS__
		__grab_fpe();
#endif
	    }
	}
#endif
	prev_func = _RWD_sigtab[ sig ];
	_RWD_sigtab[ sig ] = func;
	return( prev_func );
    }


_WCRTLINK int raise(int sig)
    {
	sig_func *func;

	func = _RWD_sigtab[ sig ];
	switch( sig ) {
#if !defined( __WINDOWS_386__ ) && !defined( __NETWARE__ )
	case SIGFPE:
	    __sigfpe_handler( FPE_EXPLICITGEN );
	    break;
	case SIGABRT:
	    if( func == SIG_DFL ) {
		__terminate();
	    }
	case SIGINT:
	    if( func != SIG_IGN  &&  func != SIG_DFL  &&  func != SIG_ERR ) {
		_RWD_sigtab[ sig ] = SIG_DFL;	   /* 09-nov-87 FWC */
		__restore_int23();
		(*func)(sig);
	    }
	    break;
	case SIGBREAK:
	    if( func != SIG_IGN  &&  func != SIG_DFL  &&  func != SIG_ERR ) {
		_RWD_sigtab[ sig ] = SIG_DFL;	   /* 09-nov-87 FWC */
		__restore_int_ctrl_break();
		(*func)(sig);
	    }
	    break;
#else
	case SIGABRT:
	case SIGFPE:
	case SIGINT:
	case SIGBREAK:
#endif
	case SIGILL:
	case SIGSEGV:
	case SIGTERM:
	case SIGUSR1:
	case SIGUSR2:
	case SIGUSR3:
	case SIGIDIVZ:
	case SIGIOVFL:
	    if( func != SIG_IGN  &&  func != SIG_DFL  &&  func != SIG_ERR ) {
		_RWD_sigtab[ sig ] = SIG_DFL;	   /* 09-nov-87 FWC */
		(*func)(sig);
	    }
	    break;
	default:
	    return( -1 );
	}
	return( 0 );
    }
