/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  17-jul-87	F.W.Crigger	Initial implementation
  09-nov-87	F.W.Crigger	Reset signal to SIG_DFL after being raised
----------------------------------------------------------Version 6.0-------
  17-apr-88	F.W.Crigger	abort now prints "ABNORMAL TERMINATION"
  19-may-88	F.W.Crigger	Added support for catching int 0x23
  23-may-88	F.W.Crigger	Reset _int23_exit to point to null exit rtn
  24-may-88	F.W.Crigger	Introduce _child variable
----------------------------------------------------------Version 6.5-------
  16-aug-88	F.W.Crigger	__abort variable introduced
----------------------------------------------------------Version 7.0-------
  25-may-89	John Dahms	moved _raise to separate file
  12-jul-89	F.W.Crigger	__setvect for catching CTRL-C under Phar Lap
  13-jul-89	F.W.Crigger	__setvect modified to handle A.I.Architects
  20-jul-89	F.W.Crigger	__FPE_handler variable introduced
  31-jul-89	F.W.Crigger	call to __Init_FPE_handler() added
  05-apr-90	F.W.Crigger	Added parm to __sigfpe_handler
  05-apr-90	J.B.Schueler	added sig number range < 1 check
  11-may-90	F.W.Crigger	grab_fpe indirectly thru __Hook_FPE_Signal
  21-may-90	F.W.Crigger	Added SIGBREAK, SIGUSR1, SIGUSR2, SIGUSR3
  11-jun-90	F.W.Crigger	Added __FAR__ to __sigfpe_handler for 8086
  12-jul-90	F.W.Crigger	signal - Only grab interrupts if func=user rtn
  02-aug-90	F.W.Crigger	Adapted for OS/2
  15-jan-92	G. Coschi	added SIGIDIVZ and SIGIOVFL
  24-sep-92	A.F.Scian	added __outside_CLIB pragma to fn ptrs
  30-mar-94	Greg Bentz	SIG_IGN means trap the signal and then
  				ignore it
--------------------------------------------------------Version-11.0---------
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  11-may-95	T. Schiller	Updated default signal states.
  30-may-95	Greg Bentz	dll linkage
*/

#include "variety.h"
#include <stdio.h>
#include <signal.h>
#include <dos.h>
#include <errno.h>
#include <float.h>
#define INCL_DOSSIGNALS
#include <wos2.h>
#include "rtdata.h"
#include "sigtab.h"
#include "extfunc.h"
#include "seterrno.h"

typedef void sig_func();
#if defined(_M_IX86)
    #pragma aux (__outside_CLIB) sig_func;
#endif

extern	void	__grab_fpe();
extern	void	__terminate();

extern	void	(*__abort)();
extern	void	__null_int23_exit();
extern	void	(*__int23_exit)();


static struct sigtab SignalTable[] = {
	{ SIG_IGN, NULL, 0, 0 },		/* unused  */
	{ SIG_DFL, NULL, 0, 0 },		/* SIGABRT */
	{ SIG_IGN, NULL, 0, 0 },		/* SIGFPE  */
	{ SIG_DFL, NULL, 0, 0 },		/* SIGILL  */
	{ SIG_DFL, NULL, 0, SIG_CTRLC },	/* SIGINT  */
	{ SIG_DFL, NULL, 0, 0 },		/* SIGSEGV */
	{ SIG_DFL, NULL, 0, SIG_KILLPROCESS },	/* SIGTERM */
	{ SIG_DFL, NULL, 0, SIG_CTRLBREAK },	/* SIGBREAK */
	{ SIG_IGN, NULL, 0, SIG_PFLG_A },	/* SIGUSR1 */
	{ SIG_IGN, NULL, 0, SIG_PFLG_B },	/* SIGUSR2 */
	{ SIG_IGN, NULL, 0, SIG_PFLG_C },	/* SIGUSR3 */
	{ SIG_DFL, NULL, 0, 0 },		/* SIGIDIVZ */
	{ SIG_DFL, NULL, 0, 0 }			/* SIGIOVFL */
};

void __sigabort()
    {
	raise( SIGABRT );
    }


_WCRTLINK void _WCI86FAR __sigfpe_handler( int fpe_type )
    {
	sig_func *func;

	func = _RWD_sigtab[ SIGFPE ].func;
	if( func != SIG_IGN  &&  func != SIG_DFL  &&  func != SIG_ERR ) {
	    _RWD_sigtab[ SIGFPE ].func = SIG_DFL;      /* 09-nov-87 FWC */
	    (*func)( SIGFPE, fpe_type );	/* so we can pass 2'nd parm */
	}
    }


static _WCFAR pascal break_handler( USHORT sigarg, USHORT signum )
    {
	int sig;

	sigarg = sigarg;
	for( sig = 1; sig <= __SIGLAST; sig++ ) {
	    if( _RWD_sigtab[sig].os_sig_code == signum ) {
		raise( sig );
		break;
	    }
	}
    }


static void restore_handler()
    {
	int sig;

	for( sig = 1; sig <= __SIGLAST; sig++ ) {
	    if( _RWD_sigtab[sig].os_func != NULL ) {
		    DosSetSigHandler( _RWD_sigtab[sig].os_func,
				      &_RWD_sigtab[sig].os_func,
				      &_RWD_sigtab[sig].prev_action,
				      _RWD_sigtab[sig].prev_action,
				      _RWD_sigtab[sig].os_sig_code );
	    }
	}
	__int23_exit = __null_int23_exit;
    }


_WCRTLINK void (*signal( int sig, void (*func)(int) ))(int)
    {
	void (*prev_func)(int);

	if( (sig < 1) || (sig > __SIGLAST) ) {
	    __set_errno( EINVAL );
	    return( SIG_ERR );
	}
	__abort = __sigabort;		/* change the abort rtn address */
	if( func != SIG_DFL  &&  func != SIG_ERR ) {
	    if( _RWD_sigtab[sig].os_sig_code != 0 ) {
		if( _RWD_sigtab[sig].os_func == NULL ) {
		    DosSetSigHandler( break_handler,
				      &_RWD_sigtab[sig].os_func,
				      &_RWD_sigtab[sig].prev_action,
				      2,
				      _RWD_sigtab[sig].os_sig_code );
		    __int23_exit = restore_handler;
		}
	    } else if( sig == SIGFPE ) {
		__grab_fpe();
	    }
	}
	prev_func = _RWD_sigtab[ sig ].func;
	_RWD_sigtab[ sig ].func = func;
	return( prev_func );
    }

_WCRTLINK int raise(int sig)
    {
	sig_func *func;

	func = _RWD_sigtab[ sig ].func;
	switch( sig ) {
	case SIGFPE:
	    __sigfpe_handler( FPE_EXPLICITGEN );
	    break;
	case SIGABRT:
	    if( func == SIG_DFL ) {
		__terminate();
	    }
	case SIGILL:
	case SIGINT:
	case SIGSEGV:
	case SIGTERM:
	case SIGBREAK:
	case SIGUSR1:
	case SIGUSR2:
	case SIGUSR3:
	case SIGIDIVZ:
	case SIGIOVFL:
	    if( func != SIG_IGN  &&  func != SIG_DFL  &&  func != SIG_ERR ) {
		_RWD_sigtab[ sig ].func = SIG_DFL;	/* 09-nov-87 FWC */
		(*func)(sig);
	    }
	    break;
	default:
	    return( -1 );
	}
	return( 0 );
    }
