/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1991, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  91/07/16	G. Coschi	initial implementation
  92/03/16	G. Coschi	added __ExceptionHandled flag so user can
  				inform signal processor whether he actually
				handled the exception or not
  92/08/20	G. Coschi	indicate integer overflow if fist(p) is
  				executed and value is too large to represent
				as a integer
  23-nov-92	Craig Eisler	split out FPU specific stuff to fpusig.h
  93/04/16	G. Coschi	OS/2 2.x doesn't distinguish between Ctrl/C
  				and Ctrl/Break
  93/11/01	Greg Bentz	make sure that Ctrl-C and Ctrl-Break
  				handled the same way during exception
  94/02/23	Greg Bentz	call DosAcknowledgeSignalException() when
  		G. Coschi	a signal exception is received so we can
				process subsequent signal exceptions.
  94/03/30	Greg Bentz	SIG_IGN means trap the signal and then
  				ignore it
--------------------------------------------------------Version-11.0---------
  95/05/11  	T. Schiller 	Updated default signal states.
  30-may-95	Greg Bentz	variety.h and dll linkage
  15-sep-95	Greg Bentz	ppc support
  17-oct-95	Greg Bentz	shutdown code may not be called by primary
  				thread, make sure that deal with possiblility
				of a null exception handler chain
  18-mar-96	Greg Bentz	don't invoke a null signal handler
  05-nov-96	Greg Bentz	fixup SIG_IGN, SIG_DFL, SIG_ERR handling
*/

#include "variety.h"
#include <stdio.h>
#include <dos.h>
#include <errno.h>
#include <float.h>
#define INCL_DOSEXCEPTIONS
#include <wos2.h>
#include "osthread.h"
#include "rtdata.h"
#include "errorno.h"
#include "sigtab.h"
#include "extfunc.h"
#include "fpusig.h"
#include "seterrno.h"
#include "rtinit.h"

typedef void sig_func();
#ifdef _M_IX86
    #pragma aux (__outside_CLIB) sig_func;
#endif

extern	void		__terminate();

extern	void		(*__abort)();
extern	void		__null_int23_exit();
extern	void		(*__int23_exit)();

unsigned	char	__ExceptionHandled;

#define	XCPT_FPE	-1	/* trap all floating-point exceptions */

sigtab	SignalTable[] = {
	{ SIG_IGN, 0 },					/* unused  */
	{ SIG_DFL, 0 },					/* SIGABRT */
	{ SIG_IGN, XCPT_FPE },				/* SIGFPE  */
	{ SIG_DFL, XCPT_ILLEGAL_INSTRUCTION },		/* SIGILL  */
	{ SIG_DFL, XCPT_SIGNAL },			/* SIGINT  */
	{ SIG_DFL, XCPT_ACCESS_VIOLATION },		/* SIGSEGV */
	{ SIG_DFL, XCPT_SIGNAL },			/* SIGTERM */
	{ SIG_DFL, XCPT_SIGNAL },			/* SIGBREAK */
	{ SIG_IGN, 0 },					/* SIGUSR1 */
	{ SIG_IGN, 0 },					/* SIGUSR2 */
	{ SIG_IGN, 0 },					/* SIGUSR3 */
	{ SIG_DFL, XCPT_INTEGER_DIVIDE_BY_ZERO },	/* SIGIDIVZ */
	{ SIG_DFL, XCPT_INTEGER_OVERFLOW }		/* SIGIOVFL */
};

void	__sigabort() {
/********************/
    
    raise( SIGABRT );
}


_WCRTLINK int	__sigfpe_handler( int fpe ) {
/***********************************/
	
    sig_func *func;

    func = _RWD_sigtab[ SIGFPE ].func;
    if( (func != SIG_IGN) && (func != SIG_DFL) && (func != SIG_ERR) ) {
	_RWD_sigtab[ SIGFPE ].func = SIG_DFL;
	(*func)( SIGFPE, fpe );
	return( 0 );
    }
    return( -1 );
}

static	ULONG	__syscall xcpt_handler( PEXCEPTIONREPORTRECORD pxcpt,
                                  PEXCEPTIONREGISTRATIONRECORD registration,
                                                PCONTEXTRECORD context,
                                                         PVOID unknown ) {
/************************************************************************/

    int			sig;
    int			fpe;
    unsigned char	*ip;
    unsigned short	tw;
    status_word		sw;

    registration = registration;
    unknown = unknown;
    
    if( ( pxcpt->ExceptionNum >= XCPT_FLOAT_DENORMAL_OPERAND ) &&
	( pxcpt->ExceptionNum <= XCPT_FLOAT_UNDERFLOW ) ) {
	switch( pxcpt->ExceptionNum ) {
	case XCPT_FLOAT_DENORMAL_OPERAND :
	    fpe = FPE_DENORMAL;
	    break;
	case XCPT_FLOAT_DIVIDE_BY_ZERO :
	    fpe = FPE_ZERODIVIDE;
	    break;
	case XCPT_FLOAT_INEXACT_RESULT :
	    fpe = FPE_INEXACT;
	    break;
	case XCPT_FLOAT_INVALID_OPERATION :
	    fpe = FPE_INVALID;
	    ip = (unsigned char *)context->ctx_env[3];
	    if( *(unsigned short *)ip == 0xfad9 ) {
	        // exception caused by "fsqrt" instruction
	        fpe = FPE_SQRTNEG;
	    } else if( *(unsigned short *)ip == 0xf1d9 ) {
	        // exception caused by "fyl2x" instruction
	        fpe = FPE_LOGERR;
	    } else if( *(unsigned short *)ip == 0xf8d9 ) {
	        // exception caused by "fprem" instruction
	        fpe = FPE_MODERR;
	    } else {
		if( ( ip[0] == 0xdb ) || ( ip[0] == 0xdf ) ) {
		    if( ( ip[1] & 0x30 ) == 0x10 ) {
			// exception caused by "fist(p)" instruction
			fpe = FPE_IOVERFLOW;
		    }
		}
		if( !(ip[0] & 0x01) ) {
		    if( (ip[1] & 0x30) == 0x30 ) {
			// it's a "fdiv" or "fidiv" instruction
			tw = context->ctx_env[2] & 0x0000ffff;
			sw.sw = context->ctx_env[1] & 0x0000ffff;
			if( ((tw >> (sw.b.st << 1)) & 0x01) == 0x01 ) {
			    fpe = FPE_ZERODIVIDE;
			}
		    }
		}
	    }
	    break;
	case XCPT_FLOAT_OVERFLOW :
	    fpe = FPE_OVERFLOW;
	    break;
	case XCPT_FLOAT_STACK_CHECK :
	    if( context->ctx_env[1] & SW_C1 ) {
		fpe = FPE_STACKOVERFLOW;
	    } else {
		fpe = FPE_STACKUNDERFLOW;
	    }
	    break;
	case XCPT_FLOAT_UNDERFLOW :
	    fpe = FPE_UNDERFLOW;
	    break;
	}
	_fpreset();
	__ExceptionHandled = 1;
	if( ( __sigfpe_handler( fpe ) == 0 ) && ( __ExceptionHandled ) ) {
	    context->ctx_env[1] &= ~( SW_BUSY | SW_XCPT_FLAGS | SW_IREQ );
	    return( XCPT_CONTINUE_EXECUTION );
	}
    } else {
	if( pxcpt->ExceptionNum == XCPT_SIGNAL ) {
	    DosAcknowledgeSignalException( pxcpt->ExceptionInfo[0] );
	}
	for( sig = 1; sig <= __SIGLAST; sig++ ) {
	    if( pxcpt->ExceptionNum == _RWD_sigtab[sig].os_sig_code ) {
		if( sig == SIGINT || sig == SIGBREAK ) {
		    if( pxcpt->ExceptionInfo[0] != XCPT_SIGNAL_INTR &&
		        pxcpt->ExceptionInfo[0] != XCPT_SIGNAL_BREAK ) {
			continue;
		    }
		}
		if( (_RWD_sigtab[sig].func == SIG_IGN) ) {
		    return( XCPT_CONTINUE_EXECUTION );
		}
		if( (_RWD_sigtab[sig].func == SIG_DFL) ||
		    (_RWD_sigtab[sig].func == SIG_ERR) ) {
		    return( XCPT_CONTINUE_SEARCH );
		}
		__ExceptionHandled = 1;
		raise( sig );
		if( __ExceptionHandled ) {
		    return( XCPT_CONTINUE_EXECUTION );
		} else {
		    return( XCPT_CONTINUE_SEARCH );
		}
	    }
	}
    }
    return( XCPT_CONTINUE_SEARCH );
}


static	void	restore_handler() {
/*********************************/

    __SigFini();
    __int23_exit = __null_int23_exit;
}


_WCRTLINK void (*signal( int sig, void (*func)(int) ))( int ) {
/************************************************************/

    void	(*prev_func)(int);
    ULONG	nesting;

    if( ( sig < 1 ) || ( sig > __SIGLAST ) ) {
	__set_errno( EINVAL );
	return( SIG_ERR );
    }
    __abort = __sigabort;		/* change the abort rtn address */
    if( (func != SIG_DFL) && (func != SIG_ERR) ) {
	if( _RWD_sigtab[sig].os_sig_code != 0 ) {
	    if( sig == SIGFPE ) {
		/* enable all interrupts, except precision exception */
		/* - precision exceptions are very common */
	        _control87( 0, ( MCW_EM & ~EM_PRECISION ) | 0x80 );
	    }
	    if( sig == SIGBREAK ) {
		// map SIGBREAK to SIGINT since OS/2 reports XCPT_SIGNAL_INTR
		// when Ctrl/Break is pressed
		sig = SIGINT;
	    }
	    if( __XCPTHANDLER->prev_structure == NULL ) {
		DosSetExceptionHandler( __XCPTHANDLER );
		__int23_exit = restore_handler;
	    }
	    if( _RWD_sigtab[sig].os_sig_code == XCPT_SIGNAL ) {
		DosSetSignalExceptionFocus( SIG_SETFOCUS, &nesting );
	    }
	}
    } else {
	if( _RWD_sigtab[sig].os_sig_code == XCPT_SIGNAL ) {
	    APIRET rc;
	    do {
		rc = DosSetSignalExceptionFocus( SIG_UNSETFOCUS, &nesting );
	    } while( rc == NO_ERROR && nesting > 0 );
	}
    }
    prev_func = _RWD_sigtab[ sig ].func;
    _RWD_sigtab[ sig ].func = func;
    return( prev_func );
}


_WCRTLINK int raise( int sig ) {
/*****************************/

    sig_func *func;

    func = _RWD_sigtab[ sig ].func;
    switch( sig ) {
    case SIGFPE:
	__sigfpe_handler( FPE_EXPLICITGEN );
	break;
    case SIGABRT:
	if( func == SIG_DFL ) {
	    __terminate();
	}
    case SIGILL:
    case SIGINT:
    case SIGSEGV:
    case SIGTERM:
    case SIGBREAK:
    case SIGUSR1:
    case SIGUSR2:
    case SIGUSR3:
    case SIGIDIVZ:
    case SIGIOVFL:
	if( (func != SIG_IGN) && (func != SIG_DFL) && (func != SIG_ERR) ) {
	    _RWD_sigtab[ sig ].func = SIG_DFL;
	    if( func ) {
		(*func)( sig );
	    }
	}
	break;
    default:
	return( -1 );
    }
    return( 0 );
}


void	__SigInit() {
/*******************/

#if defined( __SW_BM )
    int		i;

    for( i = 1; i <= __SIGLAST; ++i ) {
        _RWD_sigtab[ i ] = SignalTable[ i ];
    }
#endif
    __XCPTHANDLER->prev_structure = NULL;
    __XCPTHANDLER->ExceptionHandler = &xcpt_handler;
}


void	__SigFini() {
/*******************/

#if defined( __SW_BM )
    ULONG		nesting;
    APIRET		rc;
    __EXCEPTION_RECORD  *rr;

    rr = __XCPTHANDLER;
    if( rr && rr->prev_structure ) {
	do {
	    rc = DosSetSignalExceptionFocus( SIG_UNSETFOCUS, &nesting );
	} while( rc == NO_ERROR && nesting > 0 );
	DosUnsetExceptionHandler( rr );
    }
#endif
}

extern	void	(*__sig_init_rtn)(void);
extern	void	(*__sig_fini_rtn)(void);

static void __SetSigInit( void ) {
    __sig_init_rtn = &__SigInit;
    __sig_fini_rtn = &__SigFini;
}

AXI( __SetSigInit, INIT_PRIORITY_LIBRARY )
