/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1990, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  22-mar-90	J.B.Schueler	Adapted for PQNX
  13-feb-91	J.B.Schueler	Changed last arg to qnx_spawn to reference ctfd
  05-jan-95	B.J. Stecher	support for P_NOWAITO
--------------------------------------------------------Version-11.0---------
  30-may-95	Greg Bentz	variety.h and dll linkage
  17-jan-96	B.J. Stecher	handle EINTR while doing waitpid
  22-apr-96	Greg Bentz	tweak typing inconsistencies
  28-may-96	B.J. Stecher	handle SIGCHLD being set to SIG_IGN
*/

#undef __INLINE_FUNCTIONS__
#include "variety.h"
#include <sys/types.h>
#include <errno.h>
#include <process.h>
#include <sys/qnx_glob.h>
#include <sys/wait.h>
#include <stddef.h>
#include <signal.h>

#define SPAWN	0
#define EXEC	1

/* P_WAIT	= 0 wait for spawned task to complete			      */
/* P_NOWAIT	= 1 resume execution immediately after spawning task	      */
/* P_OVERLAY	= 2 replace current task with new task, equivalent to exec... */
/* P_NOWAITO	= 3 resume execution immediately, no result code 	      */

int __p_overlay = 2; /* defined here because of a trick in DOS version  */


_WCRTLINK int (spawnve)( mode, path, argv, envp )
    int		    mode;	/* wait, nowait or overlay(==exec) */
    const char	    *path;	/* path name of file to be executed */
    const char *const argv[];	/* array of pointers to arguments	*/
    const char *const envp[];	/* array of pointers to environment strings */
{
    pid_t 		pid;
    int 		err;
    int			status;
    struct sigaction	old;
    struct sigaction	new;

    old.sa_handler = SIG_DFL;
    if( mode == P_WAIT ) {
	sigaction( SIGCHLD, &old, NULL );
	if( old.sa_handler == SIG_IGN ) {
	    new = old;
	    new.sa_handler = SIG_DFL;
	    sigaction( SIGCHLD, NULL, &new );
	}
    }
    err = pid = qnx_spawn( mode == P_OVERLAY ? EXEC : SPAWN,
		    qnx_spawn_options.msgbuf,
		    (nid_t) qnx_spawn_options.node,
		    qnx_spawn_options.priority,
		    qnx_spawn_options.sched_algo,
		    qnx_spawn_options.flags
		    	| ( mode == P_NOWAITO ? _SPAWN_NOZOMBIE : 0),
		    path, (char **)argv, (char **)envp,
		    qnx_spawn_options.iov,
		    qnx_spawn_options.ctfd );
    /* EXEC's don't return, only SPAWN does */
    if( err != -1 ) {
	/* if P_WAIT return invoked task's status otherwise P_NOWAIT so
	   return pid and let user do the wait */
	if( mode == P_WAIT ) {
	    do {
		err = waitpid( pid, &status, 0 );
	    } while( err == -1 && errno == EINTR );
	    if( err == pid ) err = status;
	}
    }
    if( old.sa_handler == SIG_IGN ) {
	sigaction( SIGCHLD, NULL, &old );
    }
    return( err );
}
