/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1990, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  01-nov-87     F.W.Crigger     Initial implementation
---------------------------------------------------------Version 8.0-------
  22-mar-90     J.B.Schueler    Adapted for PQNX
  19-apr-90     J.B.Schueler    Use PATH only if name does not contain '/'
   3-may-90     J.T.Laughlin    Bug fix: local directory had a prefixed '/'
   9-jul-90     J.T.Laughlin    add support for __last_path
  29-oct-90     J.B.Schueler    Library routines cannot set errno to 0
                                so save and restore errno around attempts
  05-mar-91     J.B.Schueler    Changed loop exit condition from != ENOENT
                                to the list of conditions shown below
  15-may-91     B.J.Stecher     reversed sense of above change so that the
                                search only continues for errors we understand
--------------------------------------------------------Version-11.0---------
  30-may-95	Greg Bentz	variety.h and dll linkage
  22-apr-96	Greg Bentz	tweak typing inconsistencies
*/
#undef __INLINE_FUNCTIONS__
#include "variety.h"
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <process.h>

/*
 * if program name empty then quit
 * if program name contains '/' or PATH not defined
 *    attempt to spawn requested task
 * otherwise
 *    attempt to spawn task from one of the paths listed in the PATH
 *    environment variable (set PATH=/bin:/usr/jack/bin:/etc/bin).
 */

static char *__last_path;


_WCRTLINK int (spawnvpe)( mode, path, argv, envp )
    int		    mode;	/* wait, nowait or overlay(==exec) */
    const char	    *path;	/* path name of file to be executed */
    const char *const argv[];	/* array of pointers to arguments	*/
    const char *const envp[];	/* array of pointers to environment strings */
{
    register char *p;
    register char *p2;
    register int retval, err;
    auto char buffer[_POSIX_PATH_MAX];
    int trailler=0;

    __last_path = "";
    if( *path == '\0' ) {
        errno = ENOENT;
        return( -1 );
    }
    p = (char *) getenv( "PATH" );
    for( p2 = (char *)path; *p2 != '\0'; p2++ ) {   /* POSIX check for / in file name */
        if( *p2 == '/' ) break;
    }
    if( p == NULL || *p2 == '/' ) return( spawnve( mode, path, argv, envp ) );
    err = errno;
    for( retval = -1; ; ) {
        if( *p == '\0' ) break;
        for( __last_path = p, p2 = buffer; *p && *p != ':';  )
            *p2++ = *p++;
        if( p2 > buffer && p2[-1] != '/' )
            *p2++ = '/';
        strcpy( p2, path );
        retval = spawnve( mode, buffer, argv, envp );
        if( retval != -1 ) break;
        if( !(errno == ENOENT || errno == EACCES || errno == ENOTDIR) ) break;
        if( *p == '\0' ) break;
/*
 * Search current directory once if PATH has a trailling ':'
 */
        if( trailler ) break;
        if( *++p == '\0' ) {
            --p;
            trailler++;
        }
        errno = err;
    }
    return( retval );
}
