/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1987,1989 by WATCOM Systems Inc. All rights %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  20-mar-87     F.W.Crigger     Initial implementation
-------------------------------------------------------------Version 7.0-----
  27-apr-89     F.W.Crigger     Fixed problem with segment wrap around
                                Changed variables to be indices instead of
                                pointers to optimize large data models
                                Added a wordswap routine for faster swapping
  17-jun-89     F.W.Crigger     rewritten, keep pivot value at base, swap
                                middle value with pivot so already sorted
                                arrays don't take forever.
  07-nov-89     F.W.Crigger     optimization when all values are equal
                                do the swap if the values are <= instead of <
  18-jul-90     F.W.Crigger     optimize case where all elements at the end of
                                a sequence are equal.
  23-nov-90     D.J.Gaudet      wrote pragmas for inline_swap for extra speed
  13-dec-90     John Dahms      set up es in 386 compiler
  24-sep-92	A.F.Scian	fixed up call to compare routine to use pragma
  10-jan-94	Kevin Hui	Improved pivot-picking, optimized partition
  				routine. Speed doubles now. Also added a swap
				macro to decide on the use of inline assembly
				or just simple word exchange. Modified
				MAXDEPTH.
--------------------------------------------------------Version-11.0---------
  22-feb-95	Greg Bentz	alpha support
  21-apr-95	Greg Bentz	variety.h for near/far and composite macros
  30-may-95	Greg Bentz	dll linkage
  13-mar-96	Greg Bentz	tweak typing inconsistencies
  17-jun-96	Greg Bentz	handle 16bit boundary conditions
  18-jun-96	M. Hildebrand	AXP __unaligned optimizations.  In the
				process, split main logic to qsortrtn.c
				to support code reuse.
*/
#include "variety.h"
#undef __INLINE_FUNCTIONS__
#include <stdio.h>
#include <stdlib.h>
#include "extfunc.h"

typedef int qcomp( const void *, const void * );
#if defined(_M_IX86)
    #pragma aux (__outside_CLIB) qcomp;
#endif


/* Function to find the median value */
static char *med3( char *a, char *b, char *c, qcomp cmp )
{
    if( cmp( a, b ) > 0 ) {
	if( cmp( a, c ) > 0 ) {
	    if( cmp( b, c ) > 0 ) {
		return( b );
	    } else {
		return( c );
	    }
	} else {
	    return( a );
	}
    } else {
	if( cmp( a, c ) >= 0 ) {
	    return( a );
	} else {
	    if( cmp( b, c ) > 0 ) {
		return( c );
	    } else {
		return( b );
	    }
	}
    }
}


#ifdef __AXP__

    #define FUNCTION_LINKAGE	static
    #define FUNCTION_NAME	aligned_qsort
    #define PTRATTR
    #define MED3		med3
    #define BYTESWAP		aligned_byteswap
    #include "qsortrtn.c"

    #undef FUNCTION_NAME
    #undef PTRATTR
    #undef MED3
    #undef BYTESWAP
    #define FUNCTION_NAME	unaligned_qsort
    #define PTRATTR		__unaligned
    #define BYTESWAP		unaligned_byteswap
    #define MED3(a,b,c,f)	med3( (char*)(a), (char*)(b), (char*)(c), (f) )
    #include "qsortrtn.c"

    _WCRTLINK void qsort( void *in_base, size_t n, size_t size,
			  int (*compar)(const void *, const void *) )
    {
	/*
	 * If size%4 is 0 and the input base pointer is on a word boundary,
	 * call the aligned version of qsort.  Otherwise, call the version
	 * that knows its pointers will be unaligned, so all the fixups can
	 * be done by qsort rather than by expensive OS exceptions.
	 */
	if( (size&3) == 0  &&  (((unsigned)in_base)&3) == 0 ) {
	    aligned_qsort( in_base, n, size, compar );
	} else {
	    unaligned_qsort( (__unaligned void*)in_base, n, size, compar );
	}
    }

#else

    #define FUNCTION_LINKAGE	_WCRTLINK
    #define FUNCTION_NAME	qsort
    #define PTRATTR
    #define MED3		med3
    #define BYTESWAP		byteswap
    #include "qsortrtn.c"

#endif
