;========================================================================
;==	Name:		STK					       ==
;==	Operation:	Check for stack overflow
;==	Inputs: 	EAX	amount of stack space required
;==	Volatile:	none					       ==
;==	Date:		By:		Reason: 		       ==
;==	15-sep-87	F.W.Crigger	Initial implementation	       ==
;==-------------------------------------------------Version 6.0----------
;==	15-apr-88	F.W.Crigger	ref _STACKLOW thru DS instead SS
;==	22-may-88	F.W.Crigger	Added __STKOVERFLOW label     ==
;==	07-nov-89	F.W.Crigger	Don't trap when SS != DS       ==
;==-------------------------------------------------Version 8.5----------
;==	10-sep-91	F.W.Crigger	Added _init_stk and SS_seg     ==
;==	23-sep-91	C.G.Eisler      use __exit_with_msg__          ==
;==	18-oct-91	C.G.Eisler	chg ? to 1 dup(?) in _BSS (MASM bug)
;==					change '$' to '\0'
;==	29-nov-91	John Dahms	added __GRO & __CHK
;==	11-nov-93	Greg Bentz	rationalize exit_with_msg      ==
;==	28-jan-94	Greg Bentz	change to __fatal_runtime_error==
;==	07-mar-95	M. Hildebrand	Deal with different versions of WASM.
;==	12-mar-95	Greg Bentz	remove _ from "C" symbols      ==
;==	28-nov-95	Greg Bentz	wasm export support	       ==
;========================================================================
include mdef.inc
include struct.inc
include exitwmsg.inc

	modstart	stk

_DATA	segment dword public 'DATA'
	extrn	"C",_STACKLOW : dword
SS_seg	dw	0
msg	db	"Stack Overflow!", 0dh, 0ah, 0
_DATA	ends

include xinit.inc

	xinit	_init_stk,DEF_PRIORITY

	assume	ds:DGROUP

	xdefp	    __STK
	xdefp	    __CHK
	xdefp	    __GRO
	if __WASM__ ge 100
	    xdefp  "C",__STKOVERFLOW
	else
	    xdefp  <"C",__STKOVERFLOW>
	endif
	

	defp	_init_stk
	mov	SS_seg,SS	; save original SS value
	ret			; return
	endproc _init_stk

	
	defpe	__CHK
	xchg	eax,4[esp]
	call	__STK
	mov	eax,4[esp]
	ret	4
	endproc __CHK
	
       
	defpe	__GRO
	ret	4
	endproc	__GRO
	
       
	defpe	__STK
	_guess			; guess: no overflow
	  cmp	eax,esp 	; - quit if user asking for too much
	  _quif ae		; - . . .
	  sub	eax,esp 	; - calculate new low point
	  neg	eax		; - calc what new SP would be
	  cmp	eax,_STACKLOW	; - quit if too much
	  _quif be		; - . . .
	  ret			; - return
	_endguess		; endguess

;	We could have reached this point because we are in an interrupt
;	routine with different SS:ESP values from our DGROUP.

	_guess			; guess: SS != DS	07-nov-89
	  mov	ax,ss		; - get ss
	  cmp	ax,SS_seg	; - see if SS has been changed
	  _quif e		; - quit if same
	  ret			; - return (running with a different stack)
	_endguess		; endguess


__STKOVERFLOW:
ifdef __STACK__
	push	1
	push	offset msg
else
	mov	eax,offset msg
	mov	edx,1
endif
	call	__fatal_runtime_error
	endproc __STK

	endmod
	end
