;========================================================================
;==     Name:           STK                                            ==
;==     Operation:      Check for stack overflow
;==     Inputs:         AX      amount of stack space required
;==     Volatile:       none                                           ==
;==     Date:           By:             Reason:                        ==
;==     15-sep-87       F.W.Crigger     Initial implementation         ==
;==-------------------------------------------------Version 6.0----------
;==     15-apr-88       F.W.Crigger     ref __STACKLOW thru DS instead SS
;==     22-may-88       F.W.Crigger     Added __STKOVERFLOW_ label     ==
;==     11-jul-88       M.Glaum         Added exit_with_msg call for   ==
;==                                         portability                ==
;==     03-dec-91       G. Coschi	unlimited threads	       ==
;==	25-aug-93	Greg Bentz	remove use of stklowP	       ==
;==	11-nov-93	Greg Bentz	rationalize exit_with_msg      ==
;==	28-jan-94	Greg Bentz	change to __fatal_runtime_error==
;==	07-mar-95	M. Hildebrand	Deal with different versions of WASM.
;==	12-mar-95	Greg Bentz	remove _ from "C" symbols      ==
;==	28-nov-95	Greg Bentz	wasm export support	       ==
;========================================================================
include mdef.inc
include struct.inc
include exitwmsg.inc

        .286C

ifdef __MT__
	extrn	__threadid	    : dword
	extrn	"C",__ThreadData    : word
else
	extrn	"C",_STACKLOW	    : word
endif

        modstart        stk

CONST   segment word public 'DATA'
msg     db      "Stack Overflow!", 0dh, 0ah
CONST   ends

_DATA   segment word public 'DATA'
_DATA   ends

        DGROUP  group   CONST, _DATA
        assume  ds:DGROUP

        xdefp   __STK
	if __WASM__ ge 100
	    xdefp  "C",__STKOVERFLOW
	else
	    xdefp  <"C",__STKOVERFLOW>
	endif

dgroupp dw      DGROUP

        defpe   __STK
        _guess                  	; guess: no overflow
        cmp     ax,sp           	; - quit if user asking for too much
        _quif   ae              	; - . . .
        sub     ax,sp           	; - calculate new low point
        neg     ax              	; - calc what new SP would be
ifdef __MT__
        push    ds              	; - save registers
        push    si              	; - ...
        mov     ds,cs:dgroupp   	; - . . .
        lds     si,__threadid   	; - get thread id
        mov     si,[si]			; - ...
        shl     si,1			; - turn into index
	shl	si,1			; - ...
        mov     ds,cs:dgroupp   	; - . . .
	add	si,__ThreadData		; - get pointer to thread data
	mov	ds,__ThreadData+2	; - ...
	lds	si,[si]			; - ...
        cmp     ax,[si]			; - quit if too much
        pop     si              	; - restore registers
        pop     ds              	; - ...
else
    if _MODEL and (_BIG_DATA or _HUGE_DATA)
        push    ds              	; - save ds
        mov     ds,cs:dgroupp   	; - load ds from DGROUP
    endif
        cmp     ax,ds:_STACKLOW		; - quit if too much
    if _MODEL and (_BIG_DATA or _HUGE_DATA)
        pop     ds              	; - restore ds
    endif
endif
        _quif   be              	; - . . .
        ret                   		; - return
        _endguess               	; endguess

__STKOVERFLOW:
        add     sp, 100H        	; get enough of the stack
        mov     dx,cs:dgroupp   	; set stk overflow msg
        mov     ax, offset DGROUP:msg
        mov     bx, 01h         	; return error
        call    __fatal_runtime_error 	; arrrggg....
        endproc __STK
        endmod
        end
