;=========================================================================
;==     Name:           STK						==
;==     Operation:      Check for stack overflow for OS/2 2.0		==
;==     Inputs:         EAX	amount of stack space required		==
;==     Volatile:       none						==
;==     Date:           By:             Reason:				==
;==     21-nov-91       G. Coschi	initial implementation		==
;==     29-nov-91       John Dahms	Added __GRO and __CHK		==
;==	11-nov-92	F.W.Crigger	thread data done differently	==
;==	25-aug-93	Greg Bentz	remove use of stklowP		==
;==	11-nov-93	Greg Bentz	rationalize exit_with_msg       ==
;==	28-jan-94	Greg Bentz	change to __fatal_runtime_error==
;==	07-mar-95	M. Hildebrand	Deal with different versions of WASM.
;==	12-mar-95	Greg Bentz	remove _ from "C" symbols       ==
;==	28-nov-95	Greg Bentz	wasm export support		==
;=========================================================================
include mdef.inc
include struct.inc
include exitwmsg.inc

	modstart	stk

_DATA	segment dword public 'DATA'
ifdef __MT__
	extrn	"C",__GetThreadPtr : dword
else
	extrn	"C",_STACKLOW : dword
endif
msg	db	"Stack Overflow!", 0dh, 0ah, 0
_DATA	ends

include xinit.inc

	xinit	_init_stk,DEF_PRIORITY


	assume	ds:DGROUP

	xdefp	__STK
	xdefp	__CHK
	xdefp	__GRO
	if __WASM__ ge 100
	    xdefp  "C",__STKOVERFLOW
	else
	    xdefp  <"C",__STKOVERFLOW>
	endif


	defp	_init_stk
	ret				; return
	endproc _init_stk


	defpe	__CHK			; new style stack check
	xchg	eax,4[esp]		; get parm in eax
	call	__STK			; call stack checker
	mov 	eax,4[esp]		; restore eax
	ret	4
	endproc	__CHK
	
       
	defpe	__STK
	push	eax			; save parm for __GRO routine
	_guess				; guess: no overflow
	  cmp	eax,esp			; - quit if user asking for too much
	  _quif ae			; - ...
	  sub	eax,esp			; - calculate new low point
	  neg	eax			; - calc what new SP would be
ifdef __MT__
	  push	esi			; - save registers
	  push	eax			; - save eax
	  call	__GetThreadPtr		; - get thread data address
	  mov	esi,eax			; - ...
	  pop	eax			; - restore new ESP value
	  cmp	eax,[esi]		; - quit if too much
	  pop	esi			; - restore registers
else
	  cmp	eax,_STACKLOW		; - quit if too much
endif
	  _quif   be			; - ...
	  call  __GRO            	; - return
	  ret
        _endguess               	; endguess

__STKOVERFLOW:
	pop	eax			; pop the stack
ifdef __STACK__
	push	1			; ...
	push	offset msg		; print the error message
else
	mov	eax,offset msg		; print the error message
	mov	edx,1			; ...
endif
	call	__fatal_runtime_error	; ...
	endproc __STK


	defpe	__GRO			; touch each 4K page for sparse stack
	push	eax			; save regs
	push	ebx			; ...
	mov	eax,12[esp]		; get size to grow by
	mov	ebx,-4			; initialiaze index
lup:					; do {
	mov	[esp+ebx],ebx		; - touch that stack page
	sub	ebx,1000H		; - decrement by 4k
	sub	eax,1000H		; - decrement by 4k
	jg	lup			; } while stack left to go
	pop	ebx			; restore regs
	pop	eax			; ...
	ret	4			; return to caller
	endproc	__GRO			; 
	
        endmod
        end
