;========================================================================
;==     Name:           STK                                            ==
;==     Operation:      Check for stack overflow
;==     Inputs:         AX      amount of stack space required
;==     Volatile:       none                                           ==
;==     Date:           By:             Reason:                        ==
;==     15-sep-87       F.W.Crigger     Initial implementation         ==
;==-------------------------------------------------Version 6.0----------
;==     15-apr-88       F.W.Crigger     ref __STACKLOW thru DS instead SS
;==	22-may-88	F.W.Crigger	Added __STKOVERFLOW label      ==
;==     11-jul-88       M.Glaum         Added exit_with_msg call for   ==
;==                                         portability                ==
;==     20-sep-92	B.J. Stecher	print out allocating address   ==
;==	11-nov-93	Greg Bentz	rationalize exit_with_msg      ==
;==	28-jan-94	Greg Bentz	change to __fatal_runtime_error==
;==	07-mar-95	M. Hildebrand	Deal with different versions of WASM.
;==	12-mar-95	Greg Bentz	remove _ from "C" symbols      ==
;==	28-nov-95	Greg Bentz	wasm export support	       ==
;========================================================================
include mdef.inc
include struct.inc
include exitwmsg.inc

        .286C

        extrn   "C",_STACKLOW      : word

        modstart        stk

CONST   segment word public 'DATA'
CONST   ends

_DATA   segment word public 'DATA'
_DATA   ends

        DGROUP  group   CONST, _DATA
        assume  ds:DGROUP

dgroupp dw      DGROUP

        xdefp   __STK
	if __WASM__ ge 100
	    xdefp  "C",__STKOVERFLOW
	else
	    xdefp  <"C",__STKOVERFLOW>
	endif

msg     db      "Stack Overflow at "
msg_end label byte
hextab  db	"0123456789ABCDEF"

        defpe   __STK
        _guess                  ; guess: no overflow
          cmp   ax,sp           ; - quit if user asking for too much
          _quif ae              ; - . . .
          sub   ax,sp           ; - calculate new low point
          neg   ax              ; - calc what new SP would be
if _MODEL and (_BIG_DATA or _HUGE_DATA)
          push  ds              ; - save ds
          mov   ds,cs:dgroupp   ; - load ds from DGROUP
endif
	  cmp	ax,ds:_STACKLOW ; - quit if too much
if _MODEL and (_BIG_DATA or _HUGE_DATA)
          pop   ds              ; - restore ds
endif
          _quif be              ; - . . .
          ret                   ; - return
        _endguess               ; endguess

;* modified stack overflow code to print the return address of the
;* caller.

__STKOVERFLOW:

	pop	ax
if _MODEL and _BIG_CODE
	pop	dx
else
	mov	dx,cs
endif				; dx:ax points at retaddr
	add	sp,0100h	; cream last 256 bytes of stack
	mov	bp,sp

	push	ax		; push return address, segment first
	push	dx		; to make it easier to print it


	push	ss		; setup string instrs to copy
	pop	es		; exit message onto stack
	cld
	mov	di,bp		

	mov	si,offset msg
	mov	cx,offset msg_end
	sub	cx,si
	push	cs
	pop	ds
	rep	movsb

	pop	bx		; get segment
	call	_putw
	mov	al,':'
	stosb
	pop	bx		; get offset
	call	_putw
	mov	al,0dh
	stosb
	mov	al,0ah
	stosb
	mov	al,0
	stosb
	mov	ds,cs:dgroupp
	mov	dx,ss		; ds:ax  points at
	mov	ax,bp		; exit code in bx
	mov	bx,1
	call	__fatal_runtime_error

	endproc __STK

_putw proc near
	mov	dx,bx		; save value
	mov	cl,12		; setup shift count
_lup:
	mov	bx,dx		; get value
	shr	bx,cl		; put in bottom 4 bits.
	and	bx,0fh
	mov	al,cs:hextab[bx]
	stosb
	sub	cl,4
	jns	_lup
	ret
_putw	endp

	endmod

	end
