;=========================================================================
;==     Name:           STK						==
;==     Operation:      Check for stack overflow for QNX/386		==
;==     Inputs:         EAX	amount of stack space required		==
;==     Volatile:       none						==
;==     Date:           By:             Reason:				==
;==     21-nov-91       G. Coschi	initial implementation		==
;==     29-nov-91       John Dahms	Added __GRO and __CHK		==
;==     20-sep-92	B.J. Stecher	print out allocating address	==
;==	11-nov-93	Greg Bentz	rationalize exit_with_msg       ==
;==	28-jan-94	Greg Bentz	change to __fatal_runtime_error ==
;==	18-jan-95	Greg Bentz	support multi-threading		==
;==	07-mar-95	M. Hildebrand	Deal with different versions of WASM.
;==	12-mar-95	Greg Bentz	remove _ from "C" symbols	==
;==	28-nov-95	Greg Bentz	wasm export support		==
;=========================================================================
include mdef.inc
include struct.inc
include exitwmsg.inc

ifdef __MT__
	extrn	"C",__GetThreadPtr : dword
else
	extrn	"C",_STACKLOW : dword
endif
	extrn	"C",__qnx_exit: near

	modstart	stk

_DATA	segment dword public 'DATA'
_DATA	ends

        DGROUP  group   _DATA

	assume	ds:DGROUP

	xdefp	__STK
	xdefp	__CHK
	xdefp	__GRO
	if __WASM__ ge 100
	    xdefp  "C",__STKOVERFLOW
	else
	    xdefp  <"C",__STKOVERFLOW>
	endif


msg     db      "Stack Overflow at "
msg_end label byte
hextab  db	"0123456789ABCDEF"


	defpe	__GRO
	ret	4
	endproc	__GRO
	
	defpe	__STK
	xchg	eax,[esp]		; exchange parm with return addr
	push	eax			; push return addr
	; fall into __CHK
	endproc	__STK

	defpe	__CHK
	push	eax
	mov	eax,8[esp]
	_guess			; guess: no overflow
	  cmp	eax,esp 	; - quit if user asking for too much
	  _quif ae		; - . . .
	  sub	eax,esp 	; - calculate new low point
	  neg	eax		; - calc what new SP would be
ifdef __MT__
	  push	esi		; - save registers
	  push	eax		; - save eax
	  call	__GetThreadPtr	; - get thread data address
	  mov	esi,eax		; - ...
	  pop	eax		; - restore new ESP value
	  cmp	eax,[esi]	; - quit if too much
	  pop	esi		; - restore registers
else
	  cmp	eax,_STACKLOW	; - quit if too much
endif
	  _quif be		; - . . .
	  pop	eax		; - restore EAX
	  ret   4		; - return
	_endguess		; endguess

	endproc __CHK

;* modified stack overflow code to print the return address of the
;* caller.

__STKOVERFLOW:

	pop	eax		; throw away saved eax value
	pop	eax		; get return addr
	xor	edx,edx
	mov	dx,cs
	add	esp,0100h	; cream last 256 bytes of stack
	mov	ebp,esp

	push	eax		; push return address, segment first
	push	edx		; to make it easier to print it


	push	ss		; setup string instrs to copy
	pop	es		; exit message onto stack
	cld
	mov	edi,ebp		

	mov	esi,offset msg
	mov	ecx,offset msg_end
	sub	ecx,esi
	push	cs
	pop	ds
	rep	movsb

	pop	ebx		; get segment
	call	_putw
	mov	al,':'
	stosb
	mov	ebx,[esp]	; get offset
	shr	ebx,16		; print high word
	call	_putw
	pop	ebx		; get offset
	call	_putw		; print low word
	mov	al,0dh
	stosb
	mov	al,0ah
	stosb
	mov	al,0
	stosb
	push	ss
	pop	ds
ifdef __STACK__
	push	1
	push	ebp
else
	mov	eax,ebp
	mov	edx,1
endif
	call	__fatal_runtime_error
	jmp	__qnx_exit


_putw proc near
	mov	edx,ebx		; save value
	mov	cl,12		; setup shift count
_lup:
	mov	ebx,edx		; get value
	shr	ebx,cl		; put in bottom 4 bits.
	and	ebx,0fh
	mov	al,cs:hextab[ebx]
	stosb
	sub	cl,4
	jns	_lup
	ret
_putw	endp

	
        endmod
        end
