/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1991, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  06-mar-92     Craig Eisler    defined for NT
  10-apr-92	Craig Eisler	tweaked for dlls
  27-jun-92	B.J. Stecher	split out __exit_with_msg
  28-sep-92	G. Coschi	fixed problems with DLLs
  21-oct-92	Craig Eisler	use thread local storage
  08-nov-92	F.W.Crigger	cleaned up thread support
  16-nov-92	Craig Eisler	exception handling
  16-dec-92	G. Coschi	__DoneExceptionHandler() must be called before
  				__FiniRtns()
  20-dec-92	F.W.Crigger	_ThreadExitRtn defined
  04-feb-93	Craig Eisler	added thread_id to thread data
  05-feb-93	Craig Eisler	__FirstThreadData must be assigned in
  				__NTInit, not __NTMainInit
  15-feb-93	F.W.Crigger	Added __Is_DLL variable (for _beginthread)
  18-jun-93	John Dahms  	all library allocations use lib_xxx routines
  01-oct-93	A.F.Scian	- must initialize the file handles translation
  				  array quite early because some __InitRtns
				  may need to do some I/O
				- __exit() must be compiled as aborts;
				  #include "rtdata.h" fixes this
  25-aug-95     T. Schiller     Moved the ASTACKPTR code here from cmain().
--------------------------------------------------------Version-11.0---------
  17-feb-95	Greg Bentz	add _osbuild
  22-feb-95	Greg Bentz	alpha support
  13-sep-95	T. Schiller	Ported NT stack check fix from 10.5.
  15-sep-95	Greg Bentz	ppc support
  16-oct-95	Greg Bentz	handle embedded whitespace in filenames
  30-oct-95	M. Hildebrand	Force call to __ParsePosixHandleStr.
  03-nov-95	T. Schiller	Modified __sig_init/fini behaviour so the
				signal routines don't get dragged in unless
				they are actually needed
  02-may-96	Greg Bentz	wide character startup
  07-jun-96	T. Schiller	call __NewExceptionHandler() for axp
  11-jun-96	T. Schiller	also call __DoneExceptionHandler() for axp
				even though it doesn't do anything
  05-jul-96	M. Hildebrand	Use __lib_ versions of ...W Win32 functions.
  16-jul-96	T. Schiller	Need to run the init routines before calling
				__sig_init_rtn()
  97/11/14	Greg Bentz	make sure __exit runs the right number
  				of FiniRtns() in a parent exe image
*/

#include "variety.h"
#include <windows.h>

#include <stdio.h>
#include <io.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#ifdef _M_IX86
 #include <i86.h>
#endif

#include "strdup.h"
#include "liballoc.h"
#include "libwin32.h"
#include "stacklow.h"
#include "sigtab.h"
#include "ntex.h"
#include "rtdata.h"
#include "initfini.h"
#include "rtinit.h"

DWORD __TlsIndex = NO_INDEX;

_WCRTLINK int *__threadid( void )
{
    return( (int *) &(__THREADDATAPTR->thread_id) );
}

thread_data		*__FirstThreadData;
thread_data		*__AllocInitThreadData( thread_data * );

static void *__SingleThread()
{
    return( __FirstThreadData );
}

static void __NullAccessRtn( int handle )
{
    handle = handle;
}

static void __NullExitRtn() {}

static void __NullAccIOBRtn(void) {}
static void __NullAccHeapRtn(void) {}
static void __NullAccTDListRtn(void) {}

static int __DefInitAccessH( int handle ) 
{
    return( __NTAddFileHandle( handle ) );
}
static void __DefFiniAccessH( int hid )
{
    __NTRemoveFileHandle( hid );
}

_WCRTLINK void 	*(*__GetThreadPtr)() = &__SingleThread;
void	(*_AccessFileH)(int)	 = &__NullAccessRtn;
void	(*_ReleaseFileH)(int)	 = &__NullAccessRtn;
int	(*_InitAccessH)(int)	 = &__DefInitAccessH;
void	(*_FiniAccessH)(int)	 = &__DefFiniAccessH;
void	(*_AccessIOB)(void)	 = &__NullAccIOBRtn;
void	(*_ReleaseIOB)(void)	 = &__NullAccIOBRtn;
void	(*_AccessNHeap)(void)	 = &__NullAccHeapRtn;
void	(*_AccessFHeap)(void)	 = &__NullAccHeapRtn;
void	(*_ReleaseNHeap)(void)	 = &__NullAccHeapRtn;
void	(*_ReleaseFHeap)(void)	 = &__NullAccHeapRtn;
void	(*_AccessTDList)(void)	 = &__NullAccTDListRtn;
void	(*_ReleaseTDList)(void)  = &__NullAccTDListRtn;
void	(*_AccessFList)(void)	 = &__NullAccIOBRtn;
void	(*_ReleaseFList)(void)	 = &__NullAccIOBRtn;
void	(*_ThreadExitRtn)(void)	 = &__NullExitRtn;

void __sig_null_rtn(void) {}
_WCRTLINK void	(*__sig_init_rtn)(void) = __sig_null_rtn;
_WCRTLINK void	(*__sig_fini_rtn)(void) = __sig_null_rtn;

#ifdef _M_IX86
 #pragma aux _end "*"
#endif
extern	char		_end;

extern	char 		*_Envptr;

_WCRTLINK extern char	*_LpCmdLine;	/* pointer to command line */
_WCRTLINK extern char	*_LpPgmName;	/* pointer to program name */
          extern char	*_LpDllName;	/* pointer to dll name */
_WCRTLINK extern wchar_t *_LpwCmdLine;	/* pointer to wide command line */
_WCRTLINK extern wchar_t *_LpwPgmName;	/* pointer to wide program name */
          extern wchar_t *_LpwDllName;	/* pointer to wide dll name */

int			__Is_DLL;	/* TRUE => DLL, else not a DLL */
static char		*_cmd_ptr;
static wchar_t		*_wcmd_ptr;

extern void __NTInitFileHandles( void );

int __NTInit( int is_dll, thread_data *tdata, HANDLE hdll )
{
    DWORD	ver;
    WORD	os_ver;

    __Is_DLL = is_dll;					/* 15-feb-93 */
    __FirstThreadData = __AllocInitThreadData( tdata );
    if( __FirstThreadData == NULL ) {
	if( is_dll ) {
	    return( FALSE );
	}
	ExitProcess( 1 );
    }

    __NTInitFileHandles();

    _Envptr = GetEnvironmentStrings();

    /*
     * Force reference to environ so that __setenvp is linked in; hence,
     * __ParsePosixHandleStr will be called.
     */
    environ = NULL;

    ver = GetVersion();
    os_ver = LOWORD( ver );
    _RWD_osmajor = os_ver & 0xff;
    _RWD_osminor = (os_ver >> 8) & 0xff;
    _RWD_osbuild = HIWORD( ver );
    _RWD_osver = _RWD_osbuild;
    _RWD_winmajor = _RWD_osmajor;
    _RWD_winminor = _RWD_osminor;
    _RWD_winver = (_RWD_winmajor << 8) | _RWD_winminor;

    {
	static char	fn[_MAX_PATH];
	GetModuleFileNameA( NULL, fn, sizeof( fn ) );
	_LpPgmName = fn;
    }
    {
	static wchar_t	wfn[_MAX_PATH];
	__lib_GetModuleFileNameW( NULL, wfn, sizeof( wfn ) );
	_LpwPgmName = wfn;
    }

    {
	char	*cmd;
	_cmd_ptr = cmd = __clib_strdup( GetCommandLineA() );
	if( *cmd == '"' ) {
	    cmd++;
	    while( *cmd != '"' && *cmd != 0 ) {
		cmd++;
	    }
	    if( *cmd ) cmd++;
	} else {
	    while( !isspace( *cmd ) && *cmd != 0 ) {
		cmd++;
	    }
	}
	while( isspace( *cmd ) ) {
	    cmd++;
	}
	_LpCmdLine = cmd;
    }
    {
	wchar_t	*wcmd;
	wcmd = GetCommandLineW();	/* Win95 supports GetCommandLineW */
	if( wcmd ) {
	    _wcmd_ptr = wcmd = __clib_wcsdup( wcmd );
	    if( *wcmd == '"' ) {
		wcmd++;
		while( *wcmd != '"' && *wcmd != 0 ) {
		    wcmd++;
		}
		if( *wcmd ) wcmd++;
	    } else {
		while( !isspace( *wcmd ) && *wcmd != 0 ) {
		    wcmd++;
		}
	    }
	    while( isspace( *wcmd ) ) {
		wcmd++;
	    }
	} else {
	    wcmd = L"";
	}
	_LpwCmdLine = wcmd;
    }

    if( is_dll ) {
	{
	    static char    fn[_MAX_PATH];
	    GetModuleFileNameA( hdll, fn, sizeof( fn ) );
	    _LpDllName = fn;
	}
	{
	    static wchar_t wfn[_MAX_PATH];
	    __lib_GetModuleFileNameW( hdll, wfn, sizeof( wfn ) );
	    _LpwDllName = wfn;
	}
    }

    return( TRUE );
}

void __NTMainInit( REGISTRATION_RECORD *rr, thread_data *tdata )
{
    __NTInit( FALSE, tdata, GetModuleHandle(NULL) );
    __init_stack_limits( &_STACKLOW, &_STACKTOP );
    __NewExceptionHandler( rr );
    __InitRtns( INIT_PRIORITY_LIBRARY+1 );
    __sig_init_rtn();
    __InitRtns( 255 );
}

_WCRTLINK void (*__process_fini)(unsigned,unsigned) = 0;

_WCRTLINK void __exit( unsigned ret_code )
{
    if( __Is_DLL ) {
	if( __process_fini != 0 ) {
	    (*__process_fini)( 0, FINI_PRIORITY_EXIT-1 );
	}
    } else {
	__DoneExceptionHandler();
	__FiniRtns( 0, FINI_PRIORITY_EXIT-1 );
	(*_ThreadExitRtn)();
    }
    ExitProcess( ret_code );
}

static void __ntfini( void )
{
    if( _cmd_ptr ) {
	lib_free( _cmd_ptr );
	_cmd_ptr = NULL;
    }
    if( _wcmd_ptr ) {
	lib_free( _wcmd_ptr );
	_wcmd_ptr = NULL;
    }
}
AYI( __ntfini, INIT_PRIORITY_LIBRARY )
