/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  26-may-89	John Dahms	split out of fopen.c (and added __freefp)
  06-sep-89	F.W.Crigger	Added calls to _InitAccess and _FiniAccess
  26-oct-90	J.B.Schueler	First free file varies from DOS to QNX
   2-dec-90	B.J. Stecher	for QNX, alloc any number of FILE structures
  10-sep-91	J.B.Schueler	Added handle parm to allocfp to match PenPoint
				version (it is not used by DOS, OS/2 or QNX)
  10-sep-91	B.J. Stecher	quit loop when end of list encountered (QNX)
  01-jun-92	F.W.Crigger	Always start looking at 0 instead of 3 or 5
  02-jan-93	B.J. Stecher	all systems dynamically alloc FILE structs
  05-jan-93	B.J. Stecher	moved PENPOINT stuff here as well
  18-jun-93	John Dahms  	all library allocations use lib_xxx routines
--------------------------------------------------------Version-11.0---------
  13-mar-95	M. Hildebrand	Renamed stream_link to __stream_link, added
  				some new fields to it, and modified __iobuf.
  29-may-95	Greg Bentz	variety.h and dll linkage
0001
*/

#include "variety.h"
#if defined(__PENPOINT__)
#include <clsmgr.h>
#include <ostypes.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "liballoc.h"
#include "fileacc.h"
#include "rtdata.h"
#include "seterrno.h"

#ifdef __PENPOINT__
#define KEEP_FLAGS (_READ|_WRITE|_DYNAMIC|_SHRMEM)
#else
#define KEEP_FLAGS (_READ|_WRITE|_DYNAMIC)
#endif

FILE *__allocfp( int handle )
{
    FILE *		end;
    FILE *		fp;
    __stream_link *	link;
    unsigned		flags;

    handle = handle;
    _AccessIOB();
    /* Try and take one off the recently closed list */
    link = _RWD_cstream;
    if( link != NULL ) {
	_RWD_cstream = link->next;
	fp = link->stream;
	flags = (fp->_flag & KEEP_FLAGS) | (_READ | _WRITE);
	goto got_one;
    }
    /* See if there is a static FILE structure available. */
#if defined(__PENPOINT__)
    end = &_RWD_iob[NUM_STD_STREAMS];
#else
    end = &_RWD_iob[_NFILES];
#endif
    for( fp = _RWD_iob; fp < end; ++fp ) {
	if( (fp->_flag & (_READ | _WRITE)) == 0 ) {
	    link = lib_malloc( sizeof( __stream_link ) );
	    if( link == NULL ) goto no_mem;
	    flags = _READ | _WRITE;
	    goto got_one;
	}
    }
    /* Allocate a new dynamic structure */
    flags = _DYNAMIC | _READ | _WRITE;
#if defined(__PENPOINT__)
    if( handle > NUM_STD_STREAMS &&
	stsOK == ObjectCall(msgCan, (OBJECT)handle, (P_ARGS)objCapCall)) {
	flags |= _SHRMEM;
	link = _smalloc( sizeof( __stream_link ) + sizeof( FILE ) );
    } else {
	link = lib_malloc( sizeof( __stream_link ) + sizeof( FILE ) );
    }
#else
    link = lib_malloc( sizeof( __stream_link ) + sizeof( FILE ) );
#endif
    if( link == NULL ) goto no_mem;
    fp = (FILE *)(link + 1);
got_one:
    memset( fp, 0, sizeof( *fp ) );
    fp->_flag = flags;
    link->stream = fp;
    #ifndef __NETWARE__
	link->stream->_link = link;	/* point back to link structure */
    #endif
    link->next = _RWD_ostream;
    _RWD_ostream = link;
    _ReleaseIOB();
    return( fp );
no_mem:
    __set_errno( ENOMEM );
    _ReleaseIOB();
    return( NULL ); 	/* no free slots */
}

/*
    NOTE: This routine does not actually free the link/FILE structures.
    That is because code assumes that it can fclose the file and then
    freopen is a short time later. The __purgefp routine can be called
    to actually release the storage.
*/

void __freefp( FILE * fp )
{
    __stream_link **	owner;
    __stream_link *	link;

    owner = &_RWD_ostream;
    for( ;; ) {
	link = *owner;
	if( link == NULL ) return;
	if( link->stream == fp ) break;
	owner = &link->next;
    }
    fp->_flag |= _READ | _WRITE;
    (*owner) = link->next;
    link->next = _RWD_cstream;
    _RWD_cstream = link;
}


void __purgefp()
{
    __stream_link *	next;

    while( _RWD_cstream != NULL ) {
	next = _RWD_cstream->next;
	lib_free( _RWD_cstream );
	_RWD_cstream = next;
    }
}
