/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  27-dec-85     F.W.Crigger     Initial implementation
  06-Jun-86     A.C.C.Wai       added fflush()
  27-may-87     J.Trabandt      moved ffush() to its own file
  25-nov-87     F.W.Crigger     Check for file already closed
-------------------------------------------------------Version 6.0--------
  02-may-88     F.W.Crigger     Only call fflush if buffer is dirty.
-------------------------------------------------------Version 7.0--------
  27-may-89     John Dahms      got rid of _tmpfname array
  30-may-89     John Dahms      Added multi-thread support
  26-jul-89     John Dahms      Added FILE pointer validation
  02-aug-90     F.W.Crigger     Moved logic for __remove_tmpfile inline
  02-apr-92     Craig Eisler    fixed broken QNX version
  12-nov-92     F.W.Crigger     call __mktmpfile to get tmp filename to delete
  13-dec-92     B.J. Stecher    POSIX conformance
  14-dec-92     J.B.Schueler    Do not include process.h for PenPoint, QNX
  28-dec-92	B.J. Stecher	call __flush for better modularity
  29-dec-92	B.J. Stecher	call lseek for better modularity
  19-jan-93	A. Gupta	defined getpid() as GetThreadID() for
  				Netware since GetThreadID() returns a pointer
				to a thread struct (and will be unique)
  05-feb-93	A. Gupta	moved _AccessFile() call from fclose() to
                                __doclose() to avoid semaphore problems when
				an invalid pointer is passed to freopen()
  25-apr-93	B.J. Stecher	added a parm do __doclose() so that a stream
  				could be shutdown without closing the
				underlying handle
  18-jun-93	John Dahms  	all library allocations use lib_xxx routines
--------------------------------------------------------Version-11.0---------
  13-mar-95	M. Hildebrand	Renamed stream_link to __stream_link, added
  				some new fields to it, and modified __iobuf.
  06-mar-95	Greg Bentz	use more pid bits for 32bit platforms in
  				temp file names
  12-may-95	T. Schiller	Modified MkTmpFile() to include the path to
				the tmpfile.
  29-may-95	Greg Bentz	variety.h and dll linkage
  15-sep-95	Greg Bentz	ppc support
*/
#include "variety.h"
#include <stdio.h>
#include <unistd.h>
#if !(defined(__PENPOINT__)  ||  defined(__QNX__))
#include <process.h>
#endif
#include <errno.h>
#include "liballoc.h"
#include <string.h>
#include "fileacc.h"
#include "tmpfname.h"
#include "rtdata.h"

#ifdef __NETWARE__
#define getpid()	GetThreadID()
extern int		GetThreadID( void );
#endif

extern  void    __freefp( FILE *fp );
extern  int	__flush( FILE * );
extern  int	__close( int );

_WCRTLINK int fclose( FILE *fp )
{
    __stream_link *	link;

    _AccessIOB();
    link = _RWD_ostream;
    for( ;; ) {
        if( link == NULL ) {
	    _ReleaseIOB();
	    return( -1 );     /* file not open */
	}
        if( link->stream == fp ) break;
        link = link->next;
    }
    _ReleaseIOB();
    return( __shutdown_stream( fp, 1 ) );
}

int __shutdown_stream( FILE *fp, int close_handle )
{
    int		ret;

    ret = __doclose( fp, close_handle );
    __freefp( fp );
    return( ret );
}

#ifndef __QNX__
static char __hex( int num )
{
    num += '0';
    if( num > '9' )  num += 'a' - '0' - 10;
    return( num );
}

extern char *__GetTmpPath(void);

void __MkTmpFile( char *buf, int num )
{
    unsigned pid;
    unsigned i;
	char     *ptr;

    pid = getpid();
    #if defined(__386__) || defined(__AXP__) || defined(__PPC__)
	// try to use more of the 32bit pid bits
	pid |= pid >> 16;
    #endif

	strcpy(buf, __GetTmpPath());
	ptr = &buf[strlen(buf)];
    ptr[0] = 't';
    for( i = 4; i != 0; i-- ) {
        ptr[i] = __hex( pid & 0x000F );
        pid = pid >> 4;
    }
    ptr[5] = '_';
    ptr[6] = __hex( (num >> 4) & 0x000F );
    ptr[7] = __hex( num & 0x000F );
    ptr[8] = '.';
    ptr[9] = 't';
    ptr[10] = 'm';
    ptr[11] = 'p';
    ptr[12] = '\0';
}
#endif

int __doclose( FILE *fp, int close_handle )
{
    int 		ret;
    long int 		offset;

    if( fp->_flag == 0 ) {
        return( -1 );                       /* file already closed */
    }
    ret = 0;
    if( fp->_flag & _DIRTY ) {
        ret = __flush( fp );
    }
    _AccessFile( fp );
/*
 *      02-nov-92 G.Turcotte  Syncronize buffer pointer with the file pointer
 *                        IEEE Std 1003.1-1988 B.8.2.3.2
 */
    if ((offset = ftell(fp)) != -1L) {
	/* match position with file descriptors */
        lseek(fileno(fp), offset, SEEK_SET);
    }

    if( close_handle ) {
	#if defined(__QNX__) || defined(__NETWARE__) || defined(__PENPOINT__)
	    // we don't get to implement the close function on these systems
	    ret |= close( fileno( fp ) );
	#else
	    ret |= __close( fileno( fp ) );
	#endif
    }
    if( fp->_flag & _BIGBUF ) {     /* if we allocated the buffer */
	lib_free( _FP_BASE(fp) );
	_FP_BASE(fp) = NULL;
    }
#ifndef __QNX__
    /* this never happens under QNX */
    if( fp->_flag & _TMPFIL ) {     /* if this is a temporary file */
        char        name[PATH_MAX + _TMPFNAME_LENGTH + 1]; /* 02-aug-90 */

        __MkTmpFile( name, _FP_TMPFCHAR(fp) );
        remove( name );
    }
#endif
    _ReleaseFile( fp );
    /* don't make the semaphore go away if the handle is not really closed */
    /* since FILE* semaphores are really semaphores on the handle number */
    if( close_handle ) {
	_FiniAccess( fp );
    }
#ifdef __NETWARE__
    __FileHandles[ fp->_handle ] = NULLHANDLE;
#endif
    return( ret );
}
