/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  --------	--		------
  89-05-26	John Dahms	split out of fopen.c
  14-aug-89	John Dahms	only need to make sure that the R/W permissions
				requested are a subset of the permissions
				available for the given file handle.
  19-apr-90	F.W.Crigger	check for handle == -1
  25-jul-90	J.B.Schueler	Added QNX conditional for unistd
  17-aug-90	F.W.Crigger	Some && should have been & in error checks
   2-dec-90	B.J. Stecher	preserve _DYNAMIC bit for QNX
   5-dec-90	J.B.Schueler	Merged QNX version with DOS version
  31-may-91	J.B.Schueler	Removed check for tty (done in ioalloc)
  31-may-91	J.B.Schueler	Set buffer size to zero (set in ioalloc)
  31-may-91	J.B.Schueler	Added chktty call so it matches fopen code
				(though I think it's unnecessary in both cases)
  23-jul-91	B.J. Stecher	had to set fp->_handle to handle before the
				__chktty call since that routine uses the field
  15-aug-91	J.B.Schueler	Added PENPOINT support, no TEXT/BINARY
  30-aug-91	J.B.Schueler	Added handle parm to allocfp (used to determine				if shared memory should be used for files)
  12-Jun-92	D. Wellstood	Unicode support
   3-jan-93	B.J. Stecher	all systems use linked FILE structs
  19-jan-93	A. Gupta	Modified for Netware
  21-dec-93	Greg Bentz	make sure "a" append access seeks to end
--------------------------------------------------------Version-11.0---------
  23-feb-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  14-mar-95	M. Hildebrand	Modified to reflect changes in __iobuf.
  29-may-95	Greg Bentz	variety.h and dll linkage
  21-sep-95	M. Hildebrand	Added _u version, which calls wide version.
  17-oct-95	M. Hildebrand	Now knows about pipe streams.
  03-jan-96	Greg Bentz	remove stubs that are now aliases
  27-jun-96	M. Hildebrand	Support commit / no commit flag.
*/
#include "variety.h"
#include "widechar.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>
#include "fileacc.h"
#include "iomode.h"
#include "rtdata.h"
#include "seterrno.h"

extern FILE	*__allocfp( int handle );
extern	void	__freefp(FILE *);
#ifdef __NETWARE__
    extern int	__F_NAME(__open_flags,__wopen_flags)( const CHAR_TYPE * );
#else
    extern int	__F_NAME(__open_flags,__wopen_flags)( const CHAR_TYPE *, int * );
#endif
extern	void	__chktty( FILE *fp );


_WCRTLINK FILE *__F_NAME(fdopen,_wfdopen)( int handle, const CHAR_TYPE *access_mode )
    {
	unsigned	flags;
	FILE *		fp;
#ifndef __NETWARE__
	int		extflags;
#endif

	if( handle == -1 ) {
	    __set_errno( EBADF ); 	    /* 5-dec-90 */
	    return( NULL );		    /* 19-apr-90 */
	}
	#ifdef __NETWARE__
	    flags = __F_NAME(__open_flags,__wopen_flags)( access_mode );
	#else
	    flags = __F_NAME(__open_flags,__wopen_flags)( access_mode,
							  &extflags );
	#endif
	if( flags == 0 ) return( NULL );

#if !defined(__PENPOINT__)
	/* make sure the handle has the same text/binary mode */
	if( __iomode( handle, flags ) == -1 ) {
	    return( NULL );
	}
#endif
	fp = __allocfp( handle );		/* JBS 30-aug-91 */
	if( fp ) {
	    fp->_flag &= ~(_READ | _WRITE); /* 2-dec-90 */
	    fp->_flag |= flags;
	    fp->_cnt = 0;
	    _FP_BASE(fp) = NULL;
	    fp->_bufsize = 0;			/* was BUFSIZ JBS 91/05/31 */
	    #ifndef __NETWARE__
		_FP_ORIENTATION(fp) = _NOT_ORIENTED; /* initial orientation */
		_FP_EXTFLAGS(fp) = extflags;
	    #endif
	    #if defined(__NT__) || defined(__OS2__)
		_FP_PIPEDATA(fp).isPipe = 0;	/* not a pipe */
	    #endif
#if defined(__NETWARE__)
	    {
		extern void __NewFileHandle( int *handle );

		__NewFileHandle( &( fp->_handle ) );
		if( fp->_handle == -1 ) {
		    __freefp( fp );
		    return( NULL );
		}
	        __FileHandles[ fp->_handle ] = handle;
	    }
#else
	    fileno( fp ) = handle;		/* BJS 91-07-23 */
#endif
	    _InitAccess( fp );
	    if( __F_NAME(tolower,towlower)( *access_mode ) == 'a' ) {
		fseek( fp, 0L, SEEK_END );
	    }
	    __chktty( fp );			/* JBS 31-may-91 */
#if !defined(__PENPOINT__) && !defined(__QNX__) && !defined(__NETWARE__)
	    __SetIOMode( handle, flags );
#endif
	}
	return( fp );
    }

#if !defined(__PENPOINT__)

static int __iomode( int handle, int amode )
{
    int flags;
    int __errno;

#if defined(__NETWARE__)
    handle=handle;amode=amode;flags=flags;
    __errno = 0;
#elif defined(__QNX__)
    if( (flags = fcntl( handle, F_GETFL )) == -1 ) {
	return( -1 );
    }

    __errno = EOK;
    if( (flags & O_APPEND) && !(amode & _APPEND) ) {
	__errno = EACCES;
    }
    if( (flags & O_ACCMODE) == O_RDONLY ) {
	if( amode & _WRITE ) __errno = EACCES;
    } else if( (flags & O_ACCMODE) == O_WRONLY ) {
	if( amode & _READ ) __errno = EACCES;
    }
#else
    /* make sure the handle has the same text/binary mode */
    flags = __IOMode( handle );
    __errno = 0;
    if( ( amode ^ flags ) & (_BINARY|_APPEND) ) __errno = EACCES;
    if( ( amode & _READ )  && !( flags & _READ ) ) __errno = EACCES;
    if( ( amode & _WRITE ) && !( flags & _WRITE ) ) __errno = EACCES;
#endif
    if( __errno == EACCES ) {
	__set_errno( __errno );
	return( -1 );
    }
    return(0);
}

#endif
