/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  27-dec-85	F.W.Crigger	Initial implementation from Kernighan & Ritchie
  28-jan-88	F.W.Crigger	check for DOS EOF character in text mode
---------------------------------------------------------Version 6.0-------
  30-mar-88	F.W.Crigger	Only turn on _ERR bit if not open for _READ
				in _fillbuf()
  02-may-88	F.W.Crigger	set fp->_cnt to 0 at start of _fillbuf
  03-may-88	F.W.Crigger	_fillbuf merged into fgetc.
  22-may-88	F.W.Crigger	Use special _read rtn instead of read()
  23-may-88	A.F.Scian	optimized loop in _read rtn
  25-may-88	A.F.Scian	moved _read to separate file for Express
---------------------------------------------------------Version 6.5-------
  17-aug-88	A.F.Scian	new buffered I/O scheme used
---------------------------------------------------------Version 7.0-------
  25-may-89	John Dahms	change _qread() to read() for portability
  30-may-89	John Dahms	Added multi-thread support
  26-jul-89	John Dahms	Added FILE pointer validation
---------------------------------------------------------------------------
  31-aug-89	F.W.Crigger	__fill_buffer rtn added to avoid duplicate code
  04-sep-89	F.W.Crigger	change read() back to __qread() for TEXT I/O
  07-oct-89	F.W.Crigger	Removed call to ReleaseFile from __fill_buffer
  09-mar-90	J.B.Schueler	Added QNX support, no TEXT/BINARY or EOF_CHAR
				for QNX
  10-mar-90	F.W.Crigger	Turn off _UNGET flag in __fill_buffer
				Only read 1 character when mode is _IONBF
  16-jul-90	F.W.Crigger	Ignore _EOF flag on TTY devices
  20-aug-90	F.W.Crigger	__fill_buffer - call __flushall if TTY device
  22-aug-90	J.B.Schueler	no io.h for QNX - use unistd.h instead
  31-may-91	J.B.Schueler	DOS qread buffers stdin - use getche instead
  29-aug-91	J.B.Schueler	Revamped fgetc, removed error check before
				attempting read
  29-aug-91	J.B.Schueler	Changed __fill_buffer to always return length
  29-aug-91	J.B.Schueler	Added _filbuf to always return char (can be
				used from macro implementation of getc on
				non-DOS systems)
  11-sep-91	J.B.Schueler	Predecrement count rather than post decrement
  13-dec-91	A.F.Scian	fixed 31-may-91's assumption that file handle
				0 is always the console device
  28-jan-92	J.B.Schueler	Renamed _filbuf to __filbuf
  13-feb-92	B.J. Stecher	changed _ERR to _SFERR (IBM used it too)
  12-Jun-92	D. Wellstood	Unicode support
  24-Aug-92 	D. Wellstood	fixed the unicode version of __filbuf so
  				that the return value was not truncated into
				an ASCII character
  13-dec-92	B.J. Stecher	POSIX conformance
  19-jan-93	A. Gupta	Modified for Netware
  19-feb-93	A. Gupta	changed from __qread to setmode and read
  				(for Netware only)
  14-apr-93	John Dahms	Don't call setmode for netware
--------------------------------------------------------Version-11.0---------
  23-feb-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  09-mar-95	M. Hildebrand	Implemented Normative Addendum:  Changed to
  				reflect stream orientation (wide/byte/none).
  29-may-95	Greg Bentz	variety.h and dll linkage
  19-sep-95	M. Hildebrand	fgetwc reads wide chars in binary mode and
				MBCS in text mode.
  04-dec-95	M. Hildebrand	Changed _mbislead to _ismbblead.
  22-dec-95	M. Hildebrand	Automatically resync if the buffer is _DIRTY.
  05-jun-96	M. Hildebrand	Fixed multithreading issues.
*/

#include "variety.h"
#include "widechar.h"
#include <stdio.h>
#if defined(__PENPOINT__)  ||  defined(__QNX__)
#include <unistd.h>
#else
#include <conio.h>
#include <io.h>
#endif
#include "fileacc.h"
#include <errno.h>
#include "rtdata.h"
#include "seterrno.h"
#ifdef __WIDECHAR__
    #include <mbstring.h>
    #include <wchar.h>
#endif

extern	void	__ioalloc( FILE * );
extern	int	__flushall( int );

#if defined(__PENPOINT__)
 #define __qread( h, b, l ) (read( h, b, l ) )
#elif defined(__QNX__)
 extern int  __qread( int handle, char *buffer, unsigned len );
#elif defined(NETWARE)
 #define DOS_EOF_CHAR	0x1a
 #define __qread( h, b, l ) read( h, b, l )
#else
 #define DOS_EOF_CHAR	0x1a
 extern int  __qread( int handle, char *buffer, unsigned len );
#endif

#if defined( __WIDECHAR__ )
 #define CHARMASK	0xffff
#else
 #define CHARMASK	0xff
#endif


#ifndef __WIDECHAR__

_WCRTLINK int fgetc( FILE *fp )
{
    int c;

    _ValidFile( fp, EOF );
    _AccessFile( fp );

    /*** Deal with stream orientation ***/
    #ifndef __NETWARE__
	if( _FP_ORIENTATION(fp) != _BYTE_ORIENTED ) {
	    if( _FP_ORIENTATION(fp) == _NOT_ORIENTED ) {
		_FP_ORIENTATION(fp) = _BYTE_ORIENTED;
	    } else {
		_ReleaseFile( fp );
	        return( EOF );		    /* error return */
	    }
	}
    #endif

    if(( fp->_flag & _READ ) == 0 ) {
	__set_errno( EBADF );
	fp->_flag |= _SFERR;
	c = EOF;
    } else {
#if 0
	/*** If the buffer is _DIRTY, resync it before reading ***/
	if( fp->_flag & (_WRITE|_UNGET) ) {
	    if( fp->_flag & _DIRTY ) {
		fseek( fp, 0L, SEEK_CUR );
	    }
	}
#endif

	fp->_cnt--;
	// it is important that this remain a relative comparison
	// to ensure that the getc() macro works properly
	if( fp->_cnt < 0 ) {
	    c = __F_NAME(__filbuf,__wfilbuf)(fp);
	} else {
	    c = *(char*)fp->_ptr;
	    fp->_ptr++;
	}
    }
    #if !defined(__PENPOINT__) && !defined(__QNX__)
	if( ! (fp->_flag & _BINARY) ) {
	    if( c == '\r' ) {
		fp->_cnt--;
		// it is important that this remain a relative comparison
		// to ensure that the getc() macro works properly
		if( fp->_cnt < 0 ) {
		    c = __F_NAME(__filbuf,__wfilbuf)(fp);
		} else {
		    c = *(CHAR_TYPE*)fp->_ptr & CHARMASK;
		    fp->_ptr += CHARSIZE;
		}
	    }
	    if( c == DOS_EOF_CHAR ) {
		fp->_flag |= _EOF;
		c = EOF;
	    }
	}
    #endif
    _ReleaseFile( fp );
    return( c );
}


#else


static int __read_wide_char( FILE *fp, wchar_t *wc )
/**************************************************/
{
    if( fp->_flag & _BINARY ) {
	/*** Read a wide character ***/
	return( fread( wc, sizeof(wchar_t), 1, fp ) );
    } else {
	char		mbc[MB_CUR_MAX];
	wchar_t		wcTemp;
	int		rc;

	/*** Read the multibyte character ***/
	if( !fread( &mbc[0], 1, 1, fp ) )  return( 0 );
	if( _ismbblead( mbc[0] ) ) {
	    if( !fread( &mbc[1], 1, 1, fp ) )  return( 0 );
	}

	/*** Convert it to wide form ***/
	rc = mbtowc( &wcTemp, mbc, MB_CUR_MAX );
	if( rc >= 0 ) {
	    *wc = wcTemp;
	    return( 1 );
	} else {
	    __set_errno( EILSEQ );
	    return( 0 );
	}
    }
}


_WCRTLINK wint_t fgetwc( FILE *fp )
{
    wchar_t		c;

    _ValidFile( fp, WEOF );
    _AccessFile( fp );

    /*** Deal with stream orientation ***/
    #ifndef __NETWARE__
	if( _FP_ORIENTATION(fp) != _WIDE_ORIENTED ) {
	    if( _FP_ORIENTATION(fp) == _NOT_ORIENTED ) {
		_FP_ORIENTATION(fp) = _WIDE_ORIENTED;
	    } else {
		_ReleaseFile( fp );
		return( WEOF );		    /* error return */
	    }
	}
    #endif

    /*** Read the character ***/
    if( !__read_wide_char( fp, &c ) ) {
	_ReleaseFile( fp );
	return( WEOF );
    }
    if( !(fp->_flag & _BINARY)  &&  c == L'\r' ) {
	if( !__read_wide_char( fp, &c ) ) {
	    _ReleaseFile( fp );
	    return( WEOF );
	}
    }

    _ReleaseFile( fp );
    return( (wint_t)c );
}

#endif


int __F_NAME(__filbuf,__wfilbuf)( FILE *fp )
{
    if( __F_NAME(__fill_buffer,__wfill_buffer)( fp ) == 0 ) {
	return( EOF );
    } else {
	fp->_cnt -= CHARSIZE;
	fp->_ptr += CHARSIZE;
	return( *(CHAR_TYPE*)(fp->_ptr - CHARSIZE) & CHARMASK );
    }
}

int __F_NAME(__fill_buffer,__wfill_buffer)( FILE *fp )
{
    if( _FP_BASE(fp) == NULL ) {
	__ioalloc( fp );
    }
    if( fp->_flag & _ISTTY ) {			    /* 20-aug-90 */
	if( fp->_flag & (_IONBF | _IOLBF) ) {
	    __flushall( _ISTTY );	    /* flush all TTY output */
	}
    }
    fp->_flag &= ~_UNGET;			    /* 10-mar-90 */
    fp->_ptr = _FP_BASE(fp);
#if defined(__PENPOINT__)
    fp->_cnt = __qread( fileno( fp ), fp->_ptr,
			(fp->_flag & _IONBF) ? CHARSIZE : fp->_bufsize );
#elif defined(__QNX__)
    fp->_cnt = __qread( fileno( fp ), fp->_ptr,
			(fp->_flag & _IONBF) ? CHARSIZE : fp->_bufsize );
#else
    if(( fp->_flag & (_IONBF|_ISTTY)) == (_IONBF|_ISTTY) &&
#ifndef __NETWARE__
       ( fileno( fp ) == STDIN_FILENO )) {
#else
       ( fileno( fp ) == STDIN_HANDLE )) {
#endif
	int c;			    /* JBS 31-may-91 */

	fp->_cnt = 0;
	c = getche();
	if( c != EOF ) {
	    *(CHAR_TYPE*)fp->_ptr = c;
	    fp->_cnt = CHARSIZE;
	}
    } else {
#ifndef __NETWARE__
	fp->_cnt = __qread( fileno( fp ), fp->_ptr,
			    (fp->_flag & _IONBF) ? CHARSIZE : fp->_bufsize );
#else
	fp->_cnt = read( fileno( fp ), fp->_ptr,
			     (fp->_flag & _IONBF) ? CHARSIZE : fp->_bufsize );
#endif
    }
#endif
    if( fp->_cnt <= 0 ) {
	if( fp->_cnt == 0 ) {
	    fp->_flag |= _EOF;
	} else {
	    fp->_flag |= _SFERR;
	    fp->_cnt = 0;
	}
    }
    return( fp->_cnt );
}
