/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  27-may-87	J.Trabandt	split from fclose; proper file positioning added
  14-aug-87	F.W.Crigger	comment out setting of errno; reset buffer
  09-oct-87	F.W.Crigger	change fseek parm from -fp->_cnt to 0L
  09-nov-87	F.W.Crigger	check for NULL fp value
  05-feb-88	F.W.Crigger	change fseek to use SEEK_SET instead of SEEK_CUR
----------------------------------------------------------Version 6.0--------
  02-may-88	F.W.Crigger	No need to do fseek(...) if no read been done
----------------------------------------------------------Version 6.5--------
  17-aug-88	A.F.Scian	new buffered I/O scheme used
  12-nov-88	F.W.Crigger	added "errno = ENOSPC"
----------------------------------------------------------Version 7.0--------
  26-may-89	John Dahms	use posix instead of tinyio for portability
  30-may-89	John Dahms	Added multi-thread support
  24-jun-89	F.W.Crigger	save flags and turn on _BINARY for the write
  29-jun-89	John Dahms	Call setmode instead of using _iomode directly
				for portability (NETWARE 386)
  26-jul-89	John Dahms	Added FILE pointer validation
  10-aug-89	John Dahms	DOS didn't mind if you did a seek of zero bytes
				on a device handle. OS2 does!
  02-aug-90	F.W.Crigger	Call __qwrite instead of setmode and write
  02-jan-91	B.J. Stecher	report error if file not open for writing
  13-feb-92	B.J. Stecher	changed _ERR to _SFERR (IBM used it too)
  13-dec-92	B.J. Stecher	POSIX conformance
  28-dec-92	B.J. Stecher	split out from fflush
  19-jan-93	A. Gupta	Modified for Netware
  19-feb-93	A. Gupta	changed from __qwrite to setmode and write
  				(for Netware only)
   2-apr-93	B.J. Stecher	more posix conformance
   9-apr-93	B.J. Stecher	handle _fflush in input streams.
  14-apr-93	John Dahms	Don't call setmode for netware.
  07-oct-94	B.J. Stecher	handle partial writes from __qwrite (POSIX)
--------------------------------------------------------Version-11.0---------
  14-mar-95	M. Hildebrand	Modified to reflect changes in __iobuf.
  29-may-95	Greg Bentz	variety.h and dll linkage
  27-jun-96	M. Hildebrand	Support commit / no commit flag.

  NOTE:

	here is how the buffers look upon entry to fflush
	*************************************************

( _flag & _DIRTY ) == 0 	( after fgetc calls )

			 _cnt			    0
_base			 _ptr
+---------------------------------------------------+
|			  |/////////////////////////|
|			  |/////////////////////////|
+---------------------------------------------------+
						    ^DOS

( _flag & _DIRTY ) != 0 	( after fputc calls )

0			 _cnt
_base			 _ptr
+---------------------------------------------------+
|/////////////////////////|			    |
|/////////////////////////|			    |
+---------------------------------------------------+
^DOS

	here is how the buffers look upon exit of fflush
	************************************************

_cnt=0
_ptr=_base
+---------------------------------------------------+
|						    |
|						    |
+---------------------------------------------------+
^DOS

*/
#include "variety.h"
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include "fileacc.h"
#include "rtdata.h"
#include "seterrno.h"
#include "qwrite.h"

_WCRTLINK int __flush( FILE *fp )
{
    int		len;
    long	offset;
    int		ret;
    char	*ptr;
    unsigned	amount;

    ret = 0;
    _AccessFile( fp );
    if( fp->_flag & _DIRTY ) {
	fp->_flag &= ~_DIRTY;
	if( (fp->_flag & _WRITE) && (_FP_BASE(fp) != NULL) ) {
	    ptr = _FP_BASE(fp);
	    amount = fp->_cnt;
	    while( amount != 0 && ret == 0 ) {
		len = __qwrite( fileno( fp ), ptr, amount );/* 02-aug-90 */
		if( len == -1 ) {
		    fp->_flag |= _SFERR;
		    ret = EOF;
		}
		#if !defined(__QNX__)
		    else if( len == 0 ) {
			__set_errno( ENOSPC );		     /* 12-nov-88 */
			fp->_flag |= _SFERR;
			ret = EOF;
		    }
		#endif
		ptr += len;
		amount -= len;
	    }
	}
    } else if( _FP_BASE(fp) != NULL ) {		/* not dirty */
	/* fseek( fp, ftell(fp), SEEK_SET ); */
	fp->_flag &= ~_EOF;
	if( !(fp->_flag & _ISTTY) ) {
	    offset = fp->_cnt;
	    if( offset != 0 ) { /* 10-aug-89 JD */
		offset = lseek( fileno( fp ), -offset, SEEK_CUR );
	    }
	    if( offset == -1 ) {
		fp->_flag |= _SFERR;
		ret = EOF;
	    }
	}
    }
    fp->_ptr = _FP_BASE(fp);   /* reset ptr to start of buffer */
    fp->_cnt = 0;
    #ifndef __NETWARE__
	if( ret == 0  &&  ( _FP_EXTFLAGS(fp) & _COMMIT ) ) {
	    if( fsync( fileno( fp ) ) == -1 )  ret = EOF;
	}
    #endif
    _ReleaseFile( fp );
    return( ret );
}
