/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights	    %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping	    %
 *%	  or information storage and retrieval systems - except	    %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  27-dec-85	F.W.Crigger	Initial implementation
  15-mar-87			Added errno codes
  17-mar-87			Added freopen
  09-jun-87	F.W.Crigger	Fixed access_mode logic
  19-aug-87	J.U.Trabandt	... (again)
  21-oct-87	F.W.Crigger	start allocating iob's at 5.
  03-dec-87	F.W.Crigger	Added fdopen
  04-jan-88	F.W.Crigger	change PMODE from 0 to S_IREAD|S_IWRITE
-----------------------------------------------------Version 6.0---------
  30-mar-88	F.W.Crigger	Removed calls to creat, let open do it.
				So that binary is handled properly.
  01-may-88	F.W.Crigger	set O_APPEND flag before calling open.
				Don't set O_TRUNC if open for "append"
  12-may-88	F.W.Crigger	Check _fmode if neither 't' or 'b' specified
				Set one of O_TEXT or O_BINARY
  21-may-88	F.W.Crigger	Copy _ISTTY flag from _iomode to fp->_flag
-----------------------------------------------------Version 6.5---------
  04-aug-88	A.F.Scian	optimized & cleaned up
  24-oct-88	F.W.Crigger	__NFiles introduced to replace _NFILES constant
-----------------------------------------------------Version 7.0---------
  26-may-89	John Dahms	Split out fdopen() to fdopen.c
				Changed _iomode to isatty() call
				Don't set _iomode - its handled by open()
  26-may-89	John Dahms	split out __allocfp()
  28-may-89	John Dahms	always use O_BINARY so that we can call
				read() & write() in a POSIX environment (NW386)
  30-may-89	John Dahms	Added multi-thread support
  26-jul-89	John Dahms	Added FILE pointer validation
  04-sep-89	F.W.Crigger	see 28-may-89: use O_TEXT for text I/O
  20-sep-89	John Dahms	freopen returns NULL if the open fails!
-----------------------------------------------------Version 8.0---------
  29-mar-90	J.B.Schueler	for QNX, define a different PMODE
  29-mar-90	J.B.Schueler	add "const" to 1st parm to doopen
  06-apr-90	John Dahms	__freefp has to be called for NETWARE. Just
				setting fp->flag to NULL isn't good enough!
  28-aug-90	J.B.Schueler	Replaced isatty check with call to
				check for tty default buffering
  29-oct-90	John Dahms	don't call InitAccess in doopen when mode = 'a'
				unless fp->handle != -1
   2-dec-90	B.J. Stecher	preserve _DYNAMIC bit for QNX
  31-may-91	J.B.Schueler	Set buffer size to zero (set in ioalloc)
  13-aug-91	J.B.Schueler	Added PENPOINT support, no TEXT/BINARY
  30-aug-91	J.B.Schueler	Added handle parm to allocfp (used to determine
				if shared memory should be used for files)
-----------------------------------------------------Version 9.0---------
  12-Jun-92	D. Wellstood	Unicode support
  23-jul-92	F.W.Crigger	Added _fsopen function
  24-jul-92	F.W.Crigger	set _flag to 0 in freopen
  28-jul-92	J.B.Schueler	added sopen stub function for PenPoint
  13-dec-92	B.J. Stecher	POSIX conformance
   3-jan-93	B.J. Stecher	all systems use linked FILE structs
  19-jan-93	A. Gupta	Modified for Netware
  05-feb-93	A. Gupta	put _ReleaseFile() in freopen() to avoid
  				semaphore problems if freopen() is called
				with an invalid pointer
   9-apr-93	B.J. Stecher	allow freopen to work with recently closed
   				files.
  14-apr-93	John Dahms	make sure sopen() always passed O_BINARY for
  				the Netware version.
  29-sep-94	Greg Bentz	validate file open access mode
--------------------------------------------------------Version-11.0---------
  23-feb-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  09-mar-95	M. Hildebrand	Implemented Normative Addendum:  Changed to
  				reflect stream orientation (wide/byte/none).
  13-mar-95	M. Hildebrand	Renamed stream_link to __stream_link, added
  				some new fields to it, and modified __iobuf.
  29-may-95	Greg Bentz	variety.h and dll linkage
  21-sep-95	M. Hildebrand	Added _u version, which calls wide version.
  17-oct-95	M. Hildebrand	Now knows about pipe streams.
  27-nov-95	M. Hildebrand	Changed _w_fsopen to _wfsopen.
  03-jan-96	Greg Bentz	remove stubs that are now aliases
  14-jun-96	M. Hildebrand	Use OPENMODE_DENY_COMPAT in fopen, not 0.
  27-jun-96	M. Hildebrand	Support commit / no commit flag.
  08-jul-96	M. Hildebrand	Improved _commit handling.
*/
#include "variety.h"
#include "widechar.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "fileacc.h"
#include "fmode.h"
#include "openmode.h"
#include "rtdata.h"
#include "seterrno.h"
#include "defwin.h"

#if defined(__PENPOINT__)  ||  defined(__QNX__)
 #define PMODE ( S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH )
#else
 #define PMODE ( S_IREAD | S_IWRITE )
#endif

#ifdef __NETWARE__
 extern unsigned		__NFiles;
#endif


#if defined(__PENPOINT__)
static int __F_NAME(sopen,_wsopen)( const CHAR_TYPE *path, int oflag,
				    int share, ... )
{
    va_list  ap;
    int	     mode;

    va_start( ap, share );
    mode = va_arg( ap, int );
    va_end( ap );
    return( open( path, oflag, mode ) );
}
#endif

extern	FILE	*__allocfp( int handle );
extern	void	__freefp(FILE *);
extern	void	__chktty( FILE *fp );
extern	int	__doclose( FILE *fp, int );


#ifdef __NETWARE__
    int __F_NAME(__open_flags,__wopen_flags)( const CHAR_TYPE *modestr )
#else
    int __F_NAME(__open_flags,__wopen_flags)( const CHAR_TYPE *modestr,
					      int *extflags )
#endif
{
    int			flags;
    int			alive = 1;
    int			gotplus = 0;
    int			gottextbin = 0;
#ifndef __NETWARE__
    int			gotcommit = 0;
#endif

    flags = 0;
    #ifndef __NETWARE__
	if( extflags != NULL ) {
	    if( _commode == _COMMIT ) {
		*extflags = _COMMIT;
	    } else {
		*extflags = 0;
	    }
	}
    #endif

    /*
     * The first character in modestr must be 'r', 'w', or 'a'.
     */
    switch( *modestr ) {
	case 'r':
	    flags |= _READ;
	    break;
	case 'w':
	    flags |= _WRITE;
	    break;
	case 'a':
	    flags |= _WRITE | _APPEND;
	    break;
	default:
	    __set_errno( EINVAL );
	    return( 0 );
    }
    modestr++;

    /*
     * Next we might have, in any order, some additional mode modifier
     * characters:
     *	    1.	A '+' character.
     *	    2.	Either a 't' or a 'b'.
     *	    3.	Either a 'c' or a 'n'.  (Not available for Netware.)
     * For MS compatability, scanning stops when any of the three groups
     * is encountered twice; e.g., "wct+b$&!" is valid and will result in
     * a text, not binary, stream.  Also for MS compatability, scanning
     * stops at any unrecognized character, without causing failure.
     */
    while( *modestr != NULLCHAR  &&  alive ) {
	switch( *modestr ) {
	    case '+':
		if( gotplus ) {
		    alive = 0;
		} else {
		    flags |= _READ | _WRITE;
		    gotplus = 1;
		}
		break;
	    case 't':
		if( gottextbin ) {
		    alive = 0;
		} else {
		    gottextbin = 1;
		}
		break;
	    case 'b':
		if( gottextbin ) {
		    alive = 0;
		} else {
		    #if !defined(__PENPOINT__) && !defined(__QNX__)
			flags |= _BINARY;
		    #endif
		    gottextbin = 1;
		}
		break;
#ifndef __NETWARE__
	    case 'c':
		if( gotcommit ) {
		    alive = 0;
		} else {
		    *extflags |= _COMMIT;
		    gotcommit = 1;
		}
		break;
	    case 'n':
		if( gotcommit ) {
		    alive = 0;
		} else {
		    *extflags &= ~_COMMIT;
		    gotcommit = 1;
		}
		break;
#endif
	    default:
		break;
	}
	modestr++;
    }

    /*
     * Handle defaults for any unspecified options.
     */
#if !defined(__PENPOINT__) && !defined(__QNX__)
    if( !gottextbin ) {
	if( _RWD_fmode == O_BINARY )  flags |= _BINARY;
    }
#endif

    return( flags );
}


#ifdef __NETWARE__
void __NewFileHandle( int *handle )
{
    for( *handle = 0; *handle < __NFiles; ++*handle ) {
	 if( __FileHandles[ *handle ] == NULLHANDLE ) {
	     return;
	 }
    }
    _grow_handles( __NFiles + 10 );
    for( *handle = 0; *handle < __NFiles; ++*handle ) {
	 if( __FileHandles[ *handle ] == NULLHANDLE ) {
	     return;
	 }
    }
    __set_errno( ENOMEM );
    *handle = -1;
}
#endif

static FILE *__F_NAME(__doopen,__wdoopen)( const CHAR_TYPE *name,
		       CHAR_TYPE    mode,
		       int	    file_flags,
#ifndef __NETWARE__
		       int	    extflags,
#endif
		       int	    shflag,	/* sharing flag */
		       FILE *	    fp )
    {
	int open_mode;

	SetupTGCSandNCS( RETURN_ARG( FILE *, 0 ) );	/* for NW386 */
	fp->_flag &= ~(_READ | _WRITE);
	fp->_flag |= file_flags;

	/* we need to mode character to indicate if the original */
	/* intention is to open for read or for write */
	mode = __F_NAME(tolower,towlower)( mode );
	if( mode == 'r' ) {
	    open_mode = O_RDONLY;
	    if( file_flags & _WRITE ) {		/* if "r+" mode */
		open_mode = O_RDWR;
	    }
	    #if !defined(__PENPOINT__) && !defined(__QNX__)
		if( file_flags & _BINARY ) {
		    open_mode |= O_BINARY;
		} else {
		    open_mode |= O_TEXT;
		}
	    #endif
	    #ifndef __NETWARE__
		 fp->_handle = __F_NAME(sopen,_wsopen)( name, open_mode,
		 					shflag, 0 );
	    #else
		 open_mode |= O_BINARY;
		 open_mode &= ~O_TEXT;
		 __NewFileHandle( &( fp->_handle ) );
		 if( fp->_handle == -1 ) {
		    __freefp( fp );
		    return( NULL );
		 }
		 __FileHandles[ fp->_handle ] =
			__F_NAME(sopen,_wsopen)( name, open_mode, shflag, 0 );
	    #endif
	} else {	/* mode == 'w' || mode == 'a' */
	    if( file_flags & _READ ) {		/* if "a+" or "w+" mode */
		open_mode = O_RDWR | O_CREAT;
	    } else {
		open_mode = O_WRONLY | O_CREAT;
	    }
	    if( file_flags & _APPEND ) {
		open_mode |= O_APPEND;
	    } else {			/* mode == 'w' */
		open_mode |= O_TRUNC;
	    }
	    #if !defined(__PENPOINT__) && !defined(__QNX__)
		if( file_flags & _BINARY ) {
		    open_mode |= O_BINARY;
		} else {
		    open_mode |= O_TEXT;
		}
	    #endif
	    #ifndef __NETWARE__
		fp->_handle = __F_NAME(sopen,_wsopen)( name, open_mode,
						      shflag, PMODE );
	    #else
		open_mode |= O_BINARY;
		open_mode &= ~O_TEXT;
		__NewFileHandle( &( fp->_handle ) );
		if( fp->_handle == -1 ) {
		    __freefp( fp );
		    return( NULL );
		}
		__FileHandles[ fp->_handle ] = __F_NAME(sopen,_wsopen)(
					    name, open_mode, shflag, PMODE );
	    #endif
	}
	if( fileno( fp ) == -1 ) {
	    #ifdef __NETWARE__
		__FileHandles[ fp->_handle ] = NULLHANDLE;
	    #endif
	    __freefp( fp );
	    return( NULL );
	}
	fp->_cnt = 0;
	fp->_bufsize = 0;			/* was BUFSIZ JBS 31-may-91 */
	#ifndef __NETWARE__
	    _FP_ORIENTATION(fp) = _NOT_ORIENTED; /* initial orientation */
	    _FP_EXTFLAGS(fp) = extflags;
	#endif
	#if defined(__NT__) || defined(__OS2__)
	    _FP_PIPEDATA(fp).isPipe = 0;	/* not a pipe */
	#endif
	_FP_BASE(fp) = NULL;
	_InitAccess( fp );
	if( file_flags & _APPEND ) {
	    fseek( fp, 0L, SEEK_END );
	}
	__chktty( fp );				/* JBS 28-aug-90 */
	return( fp );
    }


_WCRTLINK FILE *__F_NAME(_fsopen,_wfsopen)( const CHAR_TYPE *name,
				const CHAR_TYPE *access_mode, int shflag )
    {
	FILE *		fp;
	int		file_flags;
#ifndef __NETWARE__
	int		extflags;
#endif

	/* validate access_mode */
	#ifdef __NETWARE__
	    file_flags = __F_NAME(__open_flags,__wopen_flags)( access_mode );
	#else
	    file_flags = __F_NAME(__open_flags,__wopen_flags)( access_mode,
							       &extflags );
	#endif
	if( file_flags == 0 ) {
	    return( NULL );
	}

	/* specify dummy handle 0 */
	fp = __allocfp( 0 );			/* JBS 30-aug-91 */
	if( fp != NULL ) {
	    #ifdef __NETWARE__
		fp = __F_NAME(__doopen,__wdoopen)( name, *access_mode,
						   file_flags, shflag, fp );
	    #else
		fp = __F_NAME(__doopen,__wdoopen)( name, *access_mode,
						   file_flags, extflags,
						   shflag, fp );
	    #endif
	}
	return( fp );
    }


_WCRTLINK FILE *__F_NAME(fopen,_wfopen)( const CHAR_TYPE *name, const CHAR_TYPE *access_mode )
    {
	return( __F_NAME(_fsopen,_wfsopen)( name, access_mode,
					    OPENMODE_DENY_COMPAT ) );
    }

static FILE *close_file( FILE *fp )
    {
	__stream_link * link;
	__stream_link **owner;

	_AccessIOB();
 	/*
	    See if the file pointer is a currently open file.
	*/
	link = _RWD_ostream;
	for( ;; ) {
	    if( link == NULL ) break;
	    if( link->stream == fp ) {
		if( fp->_flag & (_READ|_WRITE) ) {
		    __doclose( fp, 1 );
		}
		_ReleaseIOB();
		return( fp );
	    }
	    link = link->next;
	}
	/*
	   It's not on the list of open files, so check the list of
	   recently closed ones.
	*/
	owner = &_RWD_cstream;
	for( ;; ) {
	    link = *owner;
	    if( link == NULL ) break;
	    if( link->stream == fp ) {
		/* remove from closed list and put on open */
		*owner = link->next;
		link->next = _RWD_ostream;
		_RWD_ostream = link;
		_ReleaseIOB();
		return( fp );
	    }
	    owner = &link->next;
	}
	/*
	    We ain't seen that file pointer ever. Leave things be.
	*/
	__set_errno( EBADF );
	_ReleaseIOB();
	return( NULL );
    }


_WCRTLINK FILE *__F_NAME(freopen,_wfreopen)( const CHAR_TYPE *name,
				const CHAR_TYPE *access_mode, FILE *fp )
    {
	int		hdl;
	int		file_flags;
#ifndef __NETWARE__
	int		extflags;
#endif

	_ValidFile( fp, 0 );

	/* validate access_mode */
	#ifdef __NETWARE__
	    file_flags = __F_NAME(__open_flags,__wopen_flags)( access_mode );
	#else
	    file_flags = __F_NAME(__open_flags,__wopen_flags)( access_mode,
							       &extflags );
	#endif
	if( file_flags == 0 ) {
	    return( NULL );
	}

	hdl = fileno( fp );
	_AccessFileH( hdl );

	#ifdef DEFAULT_WINDOWING
	    if( _WindowsRemoveWindowedHandle != 0 ) {
		_WindowsRemoveWindowedHandle( hdl );
	    }
	#endif
        fp = close_file( fp );
	if( fp != NULL ) {
	    fp->_flag &= _DYNAMIC;			/* 24-jul-92 */
	    #ifdef __NETWARE__
		fp = __F_NAME(__doopen,__wdoopen)( name, *access_mode,
						   file_flags, 0, fp );
	    #else
		fp = __F_NAME(__doopen,__wdoopen)( name, *access_mode,
						   file_flags, extflags,
						   0, fp );
	    #endif
	}
	_ReleaseFileH( hdl );
	return( fp );
    }
