/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985,1986 by WATCOM Systems Inc. All rights %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  20-May-86	A.C.C.Wai	Converted from /370 Waterloo C printf,
				and restructured to allow for wprintw()
  02-Jun-86	A.C.C.Wai	disabled floating point formating for PC
				(code far too big)
  04-Jun-86	A.C.C.Wai	separated __prtf() from this file
  18-dec-86	F.W.Crigger	Converted to ANSI standard
  23-feb-87	F.W.Crigger	Added vprintf and vfprintf
-------------------------------------------------------Version 6.0---------
  28-apr-88	F.W.Crigger	switch to temporary buffering if _IONBF set
-------------------------------------------------------Version 6.5---------
  12-jul-88	F.W.Crigger	put in prototype for _fprtf
-------------------------------------------------------Version 7.0---------
  25-may-89	John Dahms	split out printf fprintf vprintf vfprintf
  30-may-89	John Dahms	Added multi-thread support
  26-jul-89	John Dahms	Added FILE pointer validation
  12-sep-89	F.W.Crigger	return -1 if an error has occurred
  26-oct-90	B.J.Stecher	Added shared library support
  06-sep-91	J.B.Schueler	Proper ANSI handling of error flags
  13-feb-92	B.J. Stecher	changed _ERR to _SFERR (IBM used it too)
  12-Jun-92	D. Wellstood	Unicode support
  28-dec-92	B.J. Stecher	call __flush for better modularity
--------------------------------------------------------Version-11.0---------
  23-feb-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  09-mar-95	M. Hildebrand	Implemented Normative Addendum:  Changed to
  				reflect stream orientation (wide/byte/none).
  29-may-95	Greg Bentz	variety.h and dll linkage
  03-aug-95	T. Schiller	Updated SPECS reference to use new struct.
  01-nov-95     Greg Bentz	QNX -3s register convention
  05-jun-96	M. Hildebrand	Fixed multithreading issues.
*/

/*
 * printf -- CLIB formatted output
 */

#include "variety.h"
#include "widechar.h"
#include <stdio.h>
#include <stdarg.h>
#include "rtdata.h"
#include "fileacc.h"
#include "printf.h"

extern	void	__ioalloc( FILE * );
extern  int	__flush( FILE * );


/*
 * file_putc -- write a character to a file
 */
static slib_callback_t file_putc; // setup calling convention
static void __SLIB_CALLBACK file_putc( SPECS __SLIB *specs, int op_char )
{
    __F_NAME(fputc,fputwc)( op_char, (FILE *) specs->_o._dest );
    specs->_o._output_count++;
}


_WCRTLINK int __F_NAME(__fprtf,__fwprtf)( FILE *fp, const CHAR_TYPE *format, va_list arg )
{
    int 	not_buffered;
    int 	amount_written;
    unsigned	oflag;
    slib_callback_t *tmp;

    _ValidFile( fp, 0 );
    _AccessFile( fp );

    /*** Deal with stream orientation ***/
    #ifndef __NETWARE__
	#ifdef __WIDECHAR__
	    if( _FP_ORIENTATION(fp) != _WIDE_ORIENTED ) {
		if( _FP_ORIENTATION(fp) == _NOT_ORIENTED ) {
		    _FP_ORIENTATION(fp) = _WIDE_ORIENTED;
		} else {
		    _ReleaseFile( fp );
		    return( 0 );		/* error return */
		}
	    }
	#else
	    if( _FP_ORIENTATION(fp) != _BYTE_ORIENTED ) {
		if( _FP_ORIENTATION(fp) == _NOT_ORIENTED ) {
		    _FP_ORIENTATION(fp) = _BYTE_ORIENTED;
		} else {
		    _ReleaseFile( fp );
		    return( 0 );		/* error return */
		}
	    }
	#endif
    #endif

    oflag = fp->_flag & (_SFERR|_EOF);			/* 06-sep-91 */
    fp->_flag &= ~(_SFERR|_EOF);

    if( _FP_BASE(fp) == NULL ) {
	__ioalloc( fp );	/* allocate buffer */
    }
    not_buffered = 0;
    if( fp->_flag & _IONBF ) {
	not_buffered = 1;
	fp->_flag &= ~_IONBF;
	fp->_flag |= _IOFBF;
    }
#if defined(__386__) && defined(__QNX__)
	/* avoid some segment relocations for 32-bit QNX */
	tmp = (void (*)())file_putc;
#else
	tmp = file_putc;
#endif
    amount_written = __F_NAME(__prtf,__wprtf)( fp, format, arg, tmp );
    if( not_buffered ) {
	fp->_flag &= ~_IOFBF;
	fp->_flag |= _IONBF;
	__flush( fp );
    }
    if( ferror( fp ) ) amount_written = -1;		/* 06-sep-91 */
    fp->_flag |= oflag;

    _ReleaseFile( fp );
    return( amount_written );
}
