/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  27-dec-85	F.W.Crigger	Initial implementation
  29-May-86	A.C.C.Wai	added big buffer support
  27-may-87	J.Trabandt	changed use of "_cnt" field to allow proper
				file positioning
				initialize fp->_ptr after ioalloc call
  31-may-87	J.Trabandt	added check in fputc for the very first put
  26-jul-87	J.Trabandt	made 31-may-87 fix work for '\n' on binary file
  12-jan-88	F.W.Crigger	check for _WRITE permission at top of fputc
  04-feb-88	F.W.Crigger	remove logic concerning '\n' being
				translated into \r\n since write now does it
-----------------------------------------------------------Version 6.0-------
  26-apr-88	F.W.Crigger	added "fp->_cnt++" before fflush in _flushbuf
  28-apr-88	F.W.Crigger	reorganized the code
-----------------------------------------------------------Version 6.5-------
  17-aug-88	A.F.Scian	new buffered I/O method used
  23-sep-88	F.W.Crigger	must set _DIRTY flag after fflushing '\r'
-----------------------------------------------------------Version 7.0-------
  30-may-89	John Dahms	Added multi-thread support
  26-jul-89	John Dahms	Added FILE pointer validation
  09-mar-90	J.B.Schueler	Added QNX support, no TEXT/BINARY for QNX
  15-aug-91	J.B.Schueler	Added PENPOINT support, no TEXT/BINARY
  30-aug-91	J.B.Schueler	Revamped fputc, line buffer on both TEXT and
				BINARY files now supported
  26-jan-92	B.J. Stecher	set errno if file not open for write
  13-feb-92	B.J. Stecher	changed _ERR to _FSERR (IBM used it too)
  11-Jun-92	D. Wellstood	Unicode support
  13-dec-92	B.J. Stecher	POSIX conformance
  28-dec-92	B.J. Stecher	call __flush for better modularity
--------------------------------------------------------Version-11.0---------
  23-feb-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  09-mar-95	M. Hildebrand	Implemented Normative Addendum:  Changed to
  				reflect stream orientation (wide/byte/none).
  29-may-95	Greg Bentz	variety.h and dll linkage
  19-sep-95	M. Hildebrand	fputwc writes wide chars in binary mode and
				MBCS in text mode.
  05-jun-96	M. Hildebrand	Fixed multithreading issues.
*/
#include "variety.h"
#include "widechar.h"
#include <stdio.h>
#include <errno.h>
#include "fileacc.h"
#include "rtdata.h"
#include "seterrno.h"
#ifdef __WIDECHAR__
    #include <mbstring.h>
    #include <wchar.h>
#endif


extern void __ioalloc( FILE * );
extern int  __flush( FILE * );


#ifndef __WIDECHAR__

_WCRTLINK int fputc( int c, FILE *fp )
{
    int flags;

    _ValidFile( fp, EOF );
    _AccessFile( fp );

    /*** Deal with stream orientation ***/
    #ifndef __NETWARE__
	if( _FP_ORIENTATION(fp) != _BYTE_ORIENTED ) {
	    if( _FP_ORIENTATION(fp) == _NOT_ORIENTED ) {
		_FP_ORIENTATION(fp) = _BYTE_ORIENTED;
	    } else {
		_ReleaseFile( fp );
		return( EOF );		    /* error return */
	    }
	}
    #endif

    if( !(fp->_flag & _WRITE) ) {
	__set_errno( EBADF );
	fp->_flag |= _SFERR;
	_ReleaseFile( fp );
	return( EOF );
    }
    if( _FP_BASE(fp) == NULL ) {
	__ioalloc( fp );
    }
    flags = _IONBF;
    if( c == '\n' ) {
	flags = _IONBF | _IOLBF;
	#if !defined(__PENPOINT__)  &&	!defined(__QNX__)
	    if( !(fp->_flag & _BINARY) ) {
		fp->_flag |= _DIRTY;
		*(char*)fp->_ptr = '\r';   /* '\n' -> '\r''\n' */
		fp->_ptr++;
		fp->_cnt++;
		if( fp->_cnt == fp->_bufsize ) {
		    if( __flush( fp ) ) {
			_ReleaseFile( fp );
			return( EOF );
		    }
		}
	    }
	#endif
    }
    fp->_flag |= _DIRTY;
    *(char*)fp->_ptr = c;
    fp->_ptr++;
    fp->_cnt++;
    if( (fp->_flag & flags)  ||  (fp->_cnt == fp->_bufsize) ) {
	if( __flush( fp ) ) {
	    _ReleaseFile( fp );
	    return( EOF );
	}
    }
    _ReleaseFile( fp );
    return( (UCHAR_TYPE) c );
}


#else


static int __write_wide_char( FILE *fp, wchar_t wc )
/**************************************************/
{
    if( fp->_flag & _BINARY ) {
	/*** Dump the wide character ***/
	return( fwrite( &wc, sizeof(wchar_t), 1, fp ) );
    } else {
	char		mbc[MB_CUR_MAX];
	int		rc;

	/*** Convert the wide character to multibyte form and write it ***/
	rc = wctomb( mbc, wc );
	if( rc > 0 ) {
	    return( fwrite( mbc, rc, 1, fp ) );
	} else {
	    __set_errno( EILSEQ );
	    return( 0 );
	}
    }
}


_WCRTLINK wint_t fputwc( wint_t c, FILE *fp )
{
    _ValidFile( fp, WEOF );
    _AccessFile( fp );

    /*** Deal with stream orientation ***/
    #ifndef __NETWARE__
	if( _FP_ORIENTATION(fp) != _WIDE_ORIENTED ) {
	    if( _FP_ORIENTATION(fp) == _NOT_ORIENTED ) {
		_FP_ORIENTATION(fp) = _WIDE_ORIENTED;
	    } else {
		_ReleaseFile( fp );
		return( WEOF );		    /* error return */
	    }
	}
    #endif

    /*** Write the character ***/
    if( !__write_wide_char( fp, c ) ) {
	_ReleaseFile( fp );
	return( WEOF );
    } else {
	_ReleaseFile( fp );
	return( c );
    }
}

#endif
