/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  27-dec-85	F.W.Crigger	Initial implementation from K. & R.
--------------------------------------------------------Version 6.0------
  26-apr-88	F.W.Crigger	switch to _IOLBF temporarily if _IONBF
--------------------------------------------------------Version 7.0------
  30-may-89	John Dahms	Added multi-thread support
  26-jul-89	John Dahms	Added FILE pointer validation
  06-sep-91	J.B.Schueler	Proper ANSI handling of error flags
  23-oct-91	J.B.Schueler	Redid error handling from fputc call
  11-Jun-92	D. Wellstood	Unicode support
  28-dec-92	B.J. Stecher	call __flush for better modularity
  29-Apr-94	Greg Bentz	convention says to return # chars written,
  				this is consistent with ANSI
--------------------------------------------------------Version-11.0---------
  23-feb-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  14-mar-95	M. Hildebrand	Modified to reflect changes in __iobuf.
  20-mar-95	M. Hildebrand	fputws returns -1 on error, not WEOF.
  29-may-95	Greg Bentz	variety.h and dll linkage
  21-sep-95	M. Hildebrand	Added _u version, which calls wide version.
  03-jan-96	Greg Bentz	remove stubs that are now aliases
*/
#include "variety.h"
#include "widechar.h"
#include <stdio.h>
#include "fileacc.h"
#include "rtdata.h"


extern void __ioalloc( FILE * );
extern int  __flush( FILE * );


_WCRTLINK int __F_NAME(fputs,fputws)( const CHAR_TYPE *s, FILE *fp )	/* put string s to file fp */
{
    const CHAR_TYPE	*start;
    int 	c;
    int 	not_buffered;
    int 	rc;

    #ifndef __WIDECHAR__
	_ValidFile( fp, EOF );
    #else
	_ValidFile( fp, WEOF );
    #endif
    _AccessFile( fp );

    if( _FP_BASE(fp) == NULL ) {
	__ioalloc( fp );		/* allocate buffer */
    }
    not_buffered = 0;
    if( fp->_flag & _IONBF ) {
	not_buffered = 1;
	fp->_flag &= ~_IONBF;
	fp->_flag |= _IOLBF;
    }
    rc = 0;
    start = s;
    while( c = *s ) {
	s++;
	#ifndef __WIDECHAR__
	    if( (fputc)( c, fp ) == EOF ) {	    /* 23-oct-91 */
		rc = EOF;
		break;
	    }
	#else
	    if( (fputwc)( c, fp ) == WEOF ) {	    /* 23-oct-91 */
		rc = -1;
		break;
	    }
	#endif
    }
    if( not_buffered ) {
	fp->_flag &= ~_IOLBF;
	fp->_flag |= _IONBF;
	if( rc == 0 ) rc = __flush( fp );		/* 23-oct-91 */
    }
    if( rc == 0 ) {
	/* return the number of items written */
	/* this is ok by ANSI which says that success is */
	/* indicated by a non-negative return value */
	rc = s - start;
    }
    _ReleaseFile( fp );
    return( rc );
}
