/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  20-mar-87     F.W.Crigger     Initial implementation
  09-jun-87     F.W.Crigger     Check for EOF condition
  09-oct-87     F.W.Crigger     Added call to fflush.
  23-nov-87     F.W.Crigger     Check for n * size == 0.
  25-feb-88     F.W.Crigger     change to use fgetc (which has buffering)
--------------------------------------------------------------Version 7.0-----
  07-may-89     F.W.Crigger     Do buffered read for large read of binary files
  25-may-89     John Dahms      change _qread() to read() for portability
  30-may-89     John Dahms      Added multi-thread support
  26-jul-89     John Dahms      Added FILE pointer validation
  04-sep-89     F.W.Crigger     change read() back to __qread() for TEXT I/O
  07-oct-89     F.W.Crigger     added improvements suggested by Harvey Broffman
                                Also duplicated code from fgetc to speed up
                                text mode I/O.
  09-mar-90     J.B.Schueler    Added QNX support, no TEXT/BINARY or EOF_CHAR
                                for QNX
  10-mar-90     F.W.Crigger     Only read exact # of bytes when mode is _IONBF
  28-apr-90     F.W.Crigger     Read directly into user's buffer if it is larger
  06-feb-91     A.F.Scian       checked for _READ in _flag field
  15-aug-91     J.B.Schueler    Added PENPOINT support, no TEXT/BINARY
--------------------------------------------------------------Version 8.5-----
  29-aug-91     J.B.Schueler    Revamped __fgetc so it looks like fgetc
                                Changed __fill_buffer reference to _filbuf
  11-sep-91     J.B.Schueler    Pre rather than post decrement cnt
  19-sep-91     F.W.Crigger     if buffer not allocated call __ioalloc
  30-oct-91     F.W.Crigger     A file open for "rb" and with _IONBF will
                                miss reading any prebuffered data
  28-jan-92     J.B.Schueler    Renamed _filbuf to __filbuf
  13-feb-92	B.J. Stecher	changed _ERR to _SFERR (IBM used it too)
  05-apr-92	B.J. Stecher	QNX really has to call __qread (32K limit)
  13-dec-92	B.J. Stecher	POSIX conformance
  10-feb-93	A. Gupta	defined __qread as read for Netware
  16-aug-93	B.J.Stecher	work better with interactive devices
  14-oct-93	Greg Bentz	fixup for TEXT interactive devices to handle
  				temporal nature of comparison 
  15-nov-93	Greg Bentz	remove changes for interactive devices
--------------------------------------------------------Version-11.0---------
  14-mar-95	M. Hildebrand	Modified to reflect changes in __iobuf.
  11-may-95	T. Schiller     Fixed bug which left the FILE struct in an
				incorrect state after doing a large
				(>_bufsize) read.
  29-may-95	Greg Bentz	variety.h and dll linkage
  22-dec-95	M. Hildebrand	Automatically resync if the buffer is _DIRTY.
*/

#include "variety.h"
#include <stdio.h>
#if defined(__PENPOINT__)
#include <unistd.h>
#elif !defined(__QNX__)
#include <io.h>
#endif
#include "fileacc.h"
#include <string.h>
#include <errno.h>
#include "rtdata.h"
#include "seterrno.h"


#if defined(__PENPOINT__)
 #define __qread( h, b, l ) read( h, b, l )
#elif defined(__QNX__)
 extern int  __qread( int handle, char *buffer, unsigned len );
#elif defined(__NETWARE__)
 #define __qread( h, b, l ) read( h, b, l )             
 #define DOS_EOF_CHAR   0x1a
#else
 #define DOS_EOF_CHAR   0x1a
 extern int  __qread( int handle, char *buffer, unsigned len );
#endif

extern int  __fill_buffer( FILE * );    /* located in fgetc */
extern void __ioalloc( FILE *fp );


_WCRTLINK size_t fread( char *buf, size_t size, size_t n, FILE *fp )
{
    size_t len_read;

    _ValidFile( fp, 0 );
    _AccessFile( fp );
    if(( fp->_flag & _READ ) == 0 ) {
	__set_errno( EBADF );
        fp->_flag |= _SFERR;
        _ReleaseFile( fp );
        return( 0 );
    }

#if 0
    /*** If the buffer is _DIRTY, resync it before reading ***/
    if( fp->_flag & (_WRITE|_UNGET) ) {
	if( fp->_flag & _DIRTY ) {
	    fseek( fp, 0L, SEEK_CUR );
	}
    }
#endif

    n *= size;
    if( n == 0 ) {
        _ReleaseFile( fp );
        return( n );
    }
    if( _FP_BASE(fp) == NULL ) {
        __ioalloc( fp );                        /* allocate buffer */
    }
    len_read = 0;
#if !defined(__PENPOINT__)  &&  !defined(__QNX__)
    if( fp->_flag & _BINARY )
#endif
    {
        size_t bytes_left = n, bytes;
        for(;;) {
            if( fp->_cnt != 0 ) {
                bytes = fp->_cnt;
                if( bytes > bytes_left )  bytes = bytes_left;
                memcpy( buf, fp->_ptr, bytes );
                fp->_ptr += bytes;
                buf += bytes;
                fp->_cnt -= bytes;
                bytes_left -= bytes;
                len_read += bytes;
            }
            if( bytes_left == 0 ) break;

            /* if user's buffer is larger than our buffer, OR
               _IONBF is set, then read directly into user's buffer. */

            if ((bytes_left >= fp->_bufsize)  /* 28-apr-90 */
						|| (fp->_flag & _IONBF))         /* 30-oct-91 */
			{
                bytes = bytes_left;
				fp->_ptr = _FP_BASE(fp);
				fp->_cnt = 0;
                if( ! (fp->_flag & _IONBF) )
				{
                    /* if more than a sector, set to multiple of sector size*/
                    if( bytes > 512 )  bytes &= -512;
                }
                n = __qread( fileno(fp), buf, bytes );
                if( n == -1 ) {
                    fp->_flag |= _SFERR;
                    break;
                } else if( n == 0 ) {
                    fp->_flag |= _EOF;
                    break;
                }
                buf += n;
                bytes_left -= n;
                len_read += n;
            } else {
                if( __fill_buffer( fp ) == 0 )  break;
            }
        } /* end for */
#if !defined(__PENPOINT__)  &&  !defined(__QNX__)
    } else {
	for(;;) {
	    int c;

	    // ensure non-empty buffer
	    if( fp->_cnt == 0 ) {
		if( __fill_buffer( fp ) == 0 ) break;
	    }
	    // get character
	    --fp->_cnt;
	    c = *fp->_ptr++ & 0xff;
	    // perform new-line translation
	    if( c == '\r' ) {
		// ensure non-empty buffer
		if( fp->_cnt == 0 ) {
		    if( __fill_buffer( fp ) == 0 ) break;
		}
		// get character
		--fp->_cnt;
		c = *fp->_ptr++ & 0xff;
	    }
	    // check for DOS end of file marker
	    if( c == DOS_EOF_CHAR ) {
		fp->_flag |= _EOF;
		break;
	    }
	    // store chracter
	    buf[len_read] = (char)c;
	    ++len_read;
	    if( len_read == n ) break;
	}
#endif
    }
    _ReleaseFile( fp );
    return( len_read / size );
}
