/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  27-dec-85	F.W.Crigger	Initial implementation
  07-apr-87	M.A.Glaum	fseek returns zero if carry not set( no error )
  27-may-87	J.Trabandt	flush or reset buffer
  29-may-87	J.Trabandt	clear end-of-file flag
  08-oct-87	F.W.Crigger	adjust offset if origin == SEEK_CUR
  25-nov-87	F.W.Crigger	turn off the _UNGET flag
  13-jan-88	F.W.Crigger	check for -ve offset when origin == SEEK_SET
--------------------------------------------------Version 6.0----------
  11-apr-88	F.W.Crigger	use tinyio
--------------------------------------------------Version 6.5----------
  17-aug-88	A.F.Scian	cleaned up & added comments
--------------------------------------------------Version 7.0----------
  26-may-89	John Dahms	call lseek instead of TinySeek for portability
  30-may-89	John Dahms	Added multi-thread support
  26-jul-89	John Dahms	Added FILE pointer validation
  09-dec-89	F.W.Crigger	Optimize seeks to spot within current buffer
  10-mar-90	F.W.Crigger	Can't optimize seek if _UNGET flag is on
  28-apr-90	F.W.Crigger	flush dirty buffer in error case
--------------------------------------------------Version 8.0----------
  05-apr-91	F.W.Crigger	Don't fill buffer for TTY devices
				Make sure _UNGET flag is cleared
  01-jan-92	B.J. Stecher	report error if ftell fails
--------------------------------------------------Version 9.0----------
  03-jun-92	F.W.Crigger	for SEEK_END seek to end of current buffer
  18-aug-92	A.F.Scian	- optimized SEEK_CUR with +ve offset
  				- fixed a ReleaseFile omission
  28-dec-92	B.J. Stecher	call __flush for better modularity
  19-jan-93	A. Gupta	Modified for Netware
  25-oct-93	Greg Bentz	optimize for speed, taking into account need
  				for read/write switching and taking advantage
				of read-only file buffering
--------------------------------------------------------Version-11.0---------
  14-mar-95	M. Hildebrand	Modified to reflect changes in __iobuf.
  29-may-95	Greg Bentz	variety.h and dll linkage
  16-oct-96	T. Schiller	Wait to reset FP until after the seek has
				succeeded.
*/
#include "variety.h"
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include "fileacc.h"
#include "rtdata.h"
#include "seterrno.h"
extern  int	__flush( FILE * );

static int __update_buffer( long diff, FILE *fp )
{
	/*
	  diff is relative to fp->_ptr
	  if diff is within the buffer update the pointers and return 0
	  otherwise update the pointers and return 1
	*/
	if( diff <= fp->_cnt  &&  diff >= (_FP_BASE(fp) - fp->_ptr) ) {
	    fp->_flag &= ~(_EOF);
	    fp->_ptr += diff;
	    fp->_cnt -= diff;
	    return( 0 );
	}
	return( 1 );
}


/*
 * This used to be in __update_buffer(), but we don't want to do this until
 * AFTER we've made certain that lseek() will be a successful one.
 */
static void __reset_buffer(FILE *fp)
{
	fp->_flag &= ~(_EOF);
	fp->_ptr = _FP_BASE(fp);
	fp->_cnt = 0;
}


_WCRTLINK int fseek( FILE *fp, long offset, int origin )
    {
	_ValidFile( fp, -1 );
	_AccessFile( fp );
	/*
	  if the file is open for any sort of writing we must ensure that
	  the buffer is flushed when dirty so that the integrity of the
	  data is preserved.
	  if there is an ungotten character in the buffer then the data must
	  be discarded to ensure the integrity of the data
	*/
	if( fp->_flag & (_WRITE|_UNGET) ) {
	    if( fp->_flag & _DIRTY ) {
		/*
		  the __flush routine aligns the DOS file pointer with the
		  start of the resulting cleared buffer, as such, the SEEK_CUR
		  code used for the non-_DIRTY buffer case is not required
		*/
		if( __flush( fp ) ) {
		    // assume __flush set the errno value
		    // if erroneous input, override errno value
		    if( origin == SEEK_SET && offset < 0 ) {
			__set_errno( EINVAL );
		    }
		    _ReleaseFile( fp );
		    return( -1 );
		}
	    } else {
		if( origin == SEEK_CUR ) {
		    offset -= fp->_cnt;
		}
		fp->_ptr = _FP_BASE(fp);
		fp->_cnt = 0;
	    }
	    fp->_flag &= ~(_EOF|_UNGET);
	    if( lseek( fileno( fp ), offset, origin ) == -1 ) {
		_ReleaseFile( fp );
		return( -1 );
	    }
	} else {
	    // file is open for read only,
	    // no characters have been ungotten
	    // the OS file pointer is at fp->_ptr + fp->_cnt relative to the
	    // FILE* buffer
	    switch( origin ) {
	    case SEEK_CUR:
	    {   long ptr_delta = fp->_cnt;
		if( __update_buffer( offset, fp ) ) {
		    offset -= ptr_delta;
		    if( lseek( fileno( fp ), offset, origin ) == -1 ) {
			_ReleaseFile( fp );
			return( -1 );
		    }
		    __reset_buffer(fp);
		}
	    }   break;
	    case SEEK_SET:
	    {   long file_ptr = tell( fileno( fp ) );
		file_ptr -= fp->_cnt;
		if( __update_buffer( offset - file_ptr, fp ) ) {
		    if( lseek( fileno( fp ), offset, origin ) == -1 ) {
			_ReleaseFile( fp );
			return( -1 );
		    }
		    __reset_buffer(fp);
		}
	    }   break;
	    case SEEK_END:
		fp->_flag &= ~(_EOF);
		fp->_ptr = _FP_BASE(fp);
		fp->_cnt = 0;
		if( lseek( fileno( fp ), offset, origin ) == -1 ) {
		    _ReleaseFile( fp );
		    return( -1 );
		}
		break;
	    default:
		__set_errno( EINVAL );
		_ReleaseFile( fp );
		return( -1 );
	    }
	}
	_ReleaseFile( fp );
	return( 0 );		/* indicate success */
    }
