/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  20-mar-87	F.W.Crigger	Initial implementation
  09-oct-87	F.W.Crigger	Added call to fflush.
  23-nov-87	F.W.Crigger	Check for n * size == 0.
  25-feb-88	F.W.Crigger	change to use fputc (which has buffering)
---------------------------------------------------------Version 6.0--------
  27-apr-88	F.W.Crigger	do temporary buffering if _IONBF in effect
--------------------------------------------------------------Version 7.0-----
  07-may-89	F.W.Crigger	Do buffered write for large write of binary data
  25-may-89	John Dahms	change _qread() to read() for portability
  30-may-89	John Dahms	Added multi-thread support
  26-jul-89	John Dahms	Added FILE pointer validation
  04-sep-89	F.W.Crigger	change write() back to __qwrite() for TEXT I/O
  07-oct-89	F.W.Crigger	Added improvements suggested by Harvey Broffman
  09-mar-90	J.B.Schueler	Added QNX support, no TEXT/BINARY for QNX
  28-apr-90	F.W.Crigger	Write directly from user's buffer if it's larger
  06-feb-91	A.F.Scian	checked for _WRITE in _flag field
  15-aug-91	J.B.Schueler	Added PENPOINT support, no TEXT/BINARY
  10-sep-91	J.B.Schueler	Use ferror macro to test _ERR flag
  27-jan-92	J.B.Schueler	Attempt output even though error may be set
  13-feb-92	B.J. Stecher	renamed _ERR to _SFERR (IBM used it as well)
  05-apr-92	B.J. Stecher	QNX has to call __qwrite (32K limit)
  04-nov-92	B.J. Stecher	POSIX wants to return 0 on a write error
  13-dec-92	B.J. Stecher	POSIX conformance
  28-dec-92	B.J. Stecher	call __flush for better modularity
  10-feb-93	A. Gupta	defined __qwrite as write for Netware
  16-jun-93	B.J. Stecher	Error check after __qwrite call was messed
  				up... Have to check for -1 *first*.
  07-oct-94	B.J. Stecher	handle partial writes from __qwrite (POSIX)
--------------------------------------------------------Version-11.0---------
  14-mar-95	M. Hildebrand	Modified to reflect changes in __iobuf.
  29-may-95	Greg Bentz	variety.h and dll linkage
  13-mar-96	Greg Bentz	tweak typing inconsistencies
  05-jun-96	M. Hildebrand	Return 0, not -1, on _ValidFile failure.
*/

#include "variety.h"
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "fileacc.h"
#include "rtdata.h"
#include "seterrno.h"
#include "qwrite.h"

extern void __ioalloc( FILE *fp );
extern int __flush( FILE * );


_WCRTLINK size_t fwrite( const void *buf, size_t size, size_t n, FILE *fp )
{
    size_t count;
    unsigned	oflag;

    _ValidFile( fp, 0 );
    _AccessFile( fp );
    if(( fp->_flag & _WRITE) == 0 ) {
	__set_errno( EBADF );
	fp->_flag |= _SFERR;
	_ReleaseFile( fp );
	return((size_t) 0 ); /* POSIX says return 0 */
    }
    n *= size;
    if( n == 0 ) {
	_ReleaseFile( fp );
	return( n );
    }
    if( _FP_BASE(fp) == NULL ) {
	__ioalloc( fp );		    /* allocate buffer */
    }
    oflag = fp->_flag & (_SFERR|_EOF);	    /* JBS 27-jan-92 */
    fp->_flag &= ~(_SFERR|_EOF);		    /* JBS 27-jan-92 */
    count = 0;
#if !defined(__PENPOINT__)  &&	!defined(__QNX__)
    if( fp->_flag & _BINARY ) { 	    /* binary I/O */
#else
    {
#endif
	size_t bytes_left = n, bytes;

	do
	{
	    /* if our buffer is empty, and user's buffer is larger,
	       then write directly from user's buffer.  28-apr-90 */

	    if( fp->_cnt == 0  &&  bytes_left >= fp->_bufsize ) {
		bytes = bytes_left & -512;	    /* multiple of 512 */
		if( bytes == 0 ) bytes = bytes_left;/* bufsize < 512   */
		n = __qwrite( fileno(fp), buf, bytes );
		if( n == -1 ) {
		    fp->_flag |= _SFERR;
		}
		#if !defined(__QNX__)
		    else if( n == 0 ) {
			_RWD_errno = ENOSPC;
			fp->_flag |= _SFERR;
		    }
		#endif
		bytes = n;
	    } else {
		bytes = fp->_bufsize - fp->_cnt;
		if( bytes > bytes_left )  bytes = bytes_left;
		memcpy( fp->_ptr, buf, bytes );
		fp->_ptr += bytes;
		fp->_cnt += bytes;
		fp->_flag |= _DIRTY;
		if(( fp->_cnt == fp->_bufsize ) || ( fp->_flag & _IONBF )) {
		    __flush(fp);
		}
	    }
	    buf = ((const char *)buf) + bytes;
	    count += bytes;
	    bytes_left -= bytes;
	} while( bytes_left && !ferror( fp ) );
#if !defined(__PENPOINT__)  &&	!defined(__QNX__)
    } else {	    /* text I/O */
	const char *	bufptr;
	int		not_buffered;
	#ifndef __NETWARE__
	    int		old_orientation;
	#endif
	/* temporarily enable buffering saving the previous setting */
	not_buffered = 0;
	if( fp->_flag & _IONBF ) {
	    not_buffered = 1;
	    fp->_flag &= ~_IONBF;
	    fp->_flag |= _IOFBF;
	}

	/*** Use fputc, and make it think the stream is byte-oriented ***/
	#ifndef __NETWARE__
	    old_orientation = _FP_ORIENTATION(fp);
	    _FP_ORIENTATION(fp) = _BYTE_ORIENTED;
	#endif
	bufptr = (const char *)buf;
	do {
	    fputc( *(bufptr++), fp );
	    if( fp->_flag & (_EOF|_SFERR) ) break;
	    ++count;
	} while( count != n );
	#ifndef __NETWARE__
	    _FP_ORIENTATION(fp) = old_orientation;
	#endif
	
	if( not_buffered ) {	    /* if wasn't buffered, then reset */
	    fp->_flag &= ~_IOFBF;
	    fp->_flag |= _IONBF;
	    __flush( fp );
	}
#endif
    }
    if( fp->_flag & _SFERR ) {
	/*
	 * Quantum 11-17-92 Temporary buffering confuses the return
	 *                  value if the call is interrupted.
	 *                  kludge: return 0 on error
	 */
	count = 0;
    }
    fp->_flag |= oflag; 		    /* JBS 27-jan-92 */
    _ReleaseFile( fp );
    return( count / size );
}
