/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  18-mar-87	F.W.Crigger	Initial implementation
-------------------------------------------------------------Version 7.0-----
  27-may-89	John Dahms	got rid of _tmpfname variable
  02-aug-90	F.W.Crigger	Moved __remove_tmpfile logic back to fclose
  21-feb-91	A.F.Scian	fixed problem when variable character in temp
				file name becomes an invalid file name char
  12-nov-92	F.W.Crigger	Rewritten to work better in multi-tasking OS
  05-jan-93	F.W.Crigger	Check for EMFILE - no more file handles
--------------------------------------------------------Version-11.0---------
  14-mar-95	M. Hildebrand	Modified to reflect changes in __iobuf.
  12-may-95	T. Schiller	Modified tmpfile() to include the path to
				the tmpfile.
  29-may-95	Greg Bentz	variety.h and dll linkage
  27-jul-95	T. Schiller	Fixed infinite loop bug if we try to create
  				a tmpfile on a read-only device.
*/

#include "variety.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include "rtdata.h"
#include "tmpfname.h"
#include "seterrno.h"

#ifdef __NETWARE__
#define W_OK ACCESS_WR
#endif

extern	void	__MkTmpFile( char *, int);
char *__GetTmpPath(void);

_WCRTLINK FILE *tmpfile(void)		/* create a temporary file */
{
    FILE	*fp;
    int		old_errno;
    int		suffix1;
    int		suffix2;
    char	name1[PATH_MAX + _TMPFNAME_LENGTH + 1];
    char	name2[PATH_MAX + _TMPFNAME_LENGTH + 1];

    old_errno = _RWD_errno;
    suffix1 = 0;
    for(;;) {
	for(;;) {
	    __MkTmpFile( name1, suffix1 );
	    suffix1++;
	    if( access( name1, W_OK ) != 0 ) {	/* if file does not exist */
		fp = fopen( name1, "wb+" );	/* - try to create it     */
		if( fp != NULL ) break;		/* - quit if successful   */
#if !defined(__PENPOINT__)
		if ((errno == EMFILE) || (errno == EACCES))
			return NULL;
#endif
	    }					/* endif */
	}
	fclose( fp );
	/* we now have a empty file. Let's try to rename it
	   rename should be an indivisable operation in the operating system
	   so if it succeeds we can be sure no one else has this file.
	   Consider the following sequence:

	   task1: access x.y => file does not exist
	   task2: access x.y => file does not exist
	   task1: fopen  x.y => succeeds
	   task2: fopen  x.y => succeeds (now have both tasks with x.y open)
	   task1: rename x.y to y.y => succeeds, can use this file
	   task2: rename x.y to y.y => fails (because x.y no longer exists)
	   task2: start over again to get a new file name
	   task2: succeeds second time around since no more race condition
	   	  with task1.
	 */
	suffix2 = _RWD_tmpfnext;
	for(;;) {
	    __MkTmpFile( name2, suffix2 );
	    if( rename( name1, name2 ) == 0 ) {	/* if rename worked */
		fp = fopen( name2, "wb+" );	/* - create file for update */
		if( fp != NULL ) {
		    fp->_flag |= _TMPFIL;
		    _FP_TMPFCHAR(fp) = suffix2;
		    _RWD_tmpfnext = suffix2;
		    __set_errno( old_errno );
		    return( fp );
		}
#if !defined(__PENPOINT__)
		if( errno == EMFILE ) return( NULL ); /* - no file handles*/
#endif
	    }
	    ++suffix2;
	    if( access( name1, W_OK ) != 0 ) break;/* name1 does not exist */
	}
    }
}


#include <direct.h>
#include <string.h>

char *__GetTmpPath(void)
{
	#ifndef __NETWARE__
	static char *evars[] = { "TMP", "TEMP", "TMPDIR", "TEMPDIR", "" };
	char        **evar;
	#endif
	static char  tmppath[PATH_MAX + 1] = "";
	char        *tmp;

	/*
	 * If the `tmppath' var is empty then lets find it.
	 */
	if (!*tmppath)
	{
		#ifndef __NETWARE__
		    for (evar = evars; **evar; ++evar)
		    {
			    tmp = getenv(*evar);
			    if (tmp && (strlen(tmp) <= PATH_MAX))
			    {
				    _fullpath(tmppath, tmp, PATH_MAX);
				    break;
			    }
		    } /* for */
		#endif

		/*
		 * If `tmppath' is still empty then we didn't match on any
		 * environment vars.
		 */
		if (!*tmppath)
			strcpy(tmppath, getcwd(NULL, 0));

		/*
		 * Make sure the path ends in '\\'.
		 */
		tmp = &tmppath[strlen(tmppath)-1];
		if ((*tmp != '\\') && (*tmp != '/'))
		{
			*++tmp = '\\';
			*++tmp = 0x00;
		}
	} /* if */

	return tmppath;
} /* __GetTmpPath() */
