/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1985, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  27-dec-85     F.W.Crigger     Initial implementation
  06-feb-88     F.W.Crigger     Don't set EOF flag for EOF character
---------------------------------------------------------------Version 7.0----
  30-may-89     John Dahms      Added multi-thread support
  26-jul-89     John Dahms      Added FILE pointer validation
  10-mar-90     F.W.Crigger     Turn on _UNGET flag if char not same from buf
  12-Jun-92	D. Wellstood	Unicode support
  27-Aug-92	D. Wellstood	fixed bug where ungetc ing a character into
  				an empty buffer caused a page fault
--------------------------------------------------------Version-11.0---------
  23-feb-95	M. Hildebrand	Changed from UNICODE.H to WIDECHAR.H.
  09-mar-95	M. Hildebrand	Implemented Normative Addendum:  Changed to
  				reflect stream orientation (wide/byte/none),
				and presence of both _ungotten and _wungotten.
  14-mar-95	M. Hildebrand	Fixed ungetwc syntax.
  29-may-95	Greg Bentz	variety.h and dll linkage
  19-sep-95	M. Hildebrand	ungetwc will store either a wide char or a
				DBCS char, based on mode (text or binary).
  05-jun-96	M. Hildebrand	Fixed multithreading issues.
*/
#include "variety.h"
#include "widechar.h"
#include <errno.h>
#include <stdio.h>
#include "rtdata.h"
#include "fileacc.h"
#include "exitwmsg.h"
#include "liballoc.h"
#include "seterrno.h"
#ifdef __WIDECHAR__
    #include <mbstring.h>
    #include <string.h>
    #include <wchar.h>
#endif

extern void __ioalloc( FILE * );


#ifndef __WIDECHAR__
    _WCRTLINK int    (ungetc)( int c, FILE *fp ) /* push one character back onto file fp */
#else
    _WCRTLINK wint_t (ungetwc)( wint_t c, FILE *fp )
#endif
    {
	if( c == __F_NAME(EOF,WEOF) ) {   /* cannot push EOF */
            return( c );
        }
        _ValidFile( fp, __F_NAME(EOF,WEOF) );
        _AccessFile( fp );

	/*** Deal with stream orientation ***/
	#ifndef __NETWARE__
	    #ifdef __WIDECHAR__
		if( _FP_ORIENTATION(fp) != _WIDE_ORIENTED ) {
		    if( _FP_ORIENTATION(fp) == _NOT_ORIENTED ) {
			_FP_ORIENTATION(fp) = _WIDE_ORIENTED;
		    } else {
			_ReleaseFile( fp );
			return( WEOF );
		    }
		}
	    #else
		if( _FP_ORIENTATION(fp) != _BYTE_ORIENTED ) {
		    if( _FP_ORIENTATION(fp) == _NOT_ORIENTED ) {
			_FP_ORIENTATION(fp) = _BYTE_ORIENTED;
		    } else {
			_ReleaseFile( fp );
			return( EOF );
		    }
		}
	    #endif
	#endif

        if( fp->_flag & _DIRTY ) {        /* cannot unget after a put */
            _ReleaseFile( fp );
            return( __F_NAME(EOF,WEOF) );
        }
        if(( fp->_flag & _READ ) == 0 ) { /* not open for input */
            _ReleaseFile( fp );
            return( __F_NAME(EOF,WEOF) );
        }
	if( _FP_BASE(fp) == NULL ) {	  /* no buffer allocated */
            __ioalloc( fp );
        }
	#ifdef __WIDECHAR__
	    if( fp->_flag & _BINARY ) {
		/*** Leave the character in wide form ***/
		if( fp->_cnt == 0 ) {		/* read buffer is empty */
		    fp->_cnt = sizeof(wchar_t);
		    fp->_ptr = _FP_BASE(fp) + fp->_bufsize - sizeof(wchar_t);
		    fp->_flag |= _UNGET;		    /* 10-mar-90 */
		    memcpy( fp->_ptr, &c, sizeof(wchar_t) );
		} else if( fp->_ptr != _FP_BASE(fp) ) {
		    fp->_cnt += sizeof(wchar_t);
		    fp->_ptr -= sizeof(wchar_t);
		    fp->_flag |= _UNGET;
		    memcpy( fp->_ptr, &c, sizeof(wchar_t) );
		} else {			/* read buffer is full */
		    _ReleaseFile( fp );
		    return( WEOF );
		}
	    } else {
		char	mbc[MB_CUR_MAX];
		int	mbcLen;

		/*** Convert the character to multibyte form ***/
		if( wctomb( mbc, c ) == -1 ) {
		    __set_errno( EILSEQ );
		    return( WEOF );
		}
		mbcLen = _mbclen( mbc );
    
		/*** Store the converted character ***/
		if( fp->_cnt == 0 ) {		/* read buffer is empty */
		    fp->_cnt = mbcLen;
		    fp->_ptr = _FP_BASE(fp) + fp->_bufsize - mbcLen;
		    fp->_flag |= _UNGET;		    /* 10-mar-90 */
		    _mbccpy( fp->_ptr, mbc );
		} else if( fp->_ptr != _FP_BASE(fp) ) {
		    fp->_cnt += mbcLen;
		    fp->_ptr -= mbcLen;
		    fp->_flag |= _UNGET;
		    _mbccpy( fp->_ptr, mbc );
		} else {			/* read buffer is full */
		    _ReleaseFile( fp );
		    return( WEOF );
		}
	    }
	#else
	    if( fp->_cnt == 0 ) {		/* read buffer is empty */
		fp->_cnt = CHARSIZE;
		fp->_ptr = _FP_BASE(fp) + fp->_bufsize - CHARSIZE;
		fp->_flag |= _UNGET;                                 /* 10-mar-90 */
		*(CHAR_TYPE*)(fp->_ptr) = c;
	    } else if( fp->_ptr != _FP_BASE(fp) ) {
		fp->_cnt += CHARSIZE;
		fp->_ptr -= CHARSIZE;
		if( *(CHAR_TYPE*)(fp->_ptr) != c )  fp->_flag |= _UNGET; /* 10-mar-90 */
		*(CHAR_TYPE*)(fp->_ptr) = c;
	    } else {				/* read buffer is full */
		_ReleaseFile( fp );
		return( EOF );
	    }
	#endif
        fp->_flag &= ~ _EOF;

        _ReleaseFile( fp );
        return( (UCHAR_TYPE) c );
    }
