/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %     Copyright (C) 1996, by Watcom International Inc.  All rights    %
   %     reserved.  No part of this software may be reproduced or        %
   %     used in any form or by any means - graphic, electronic or       %
   %     mechanical, including photocopying, recording, taping or        %
   %     information storage and retrieval systems - except with the     %
   %     written permission of Watcom International Inc.                 %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
   Date         By              Reason
   ----         --              ------
--------------------------------------------------------Version-11.0---------
   23-apr-96    A. Kucharczyk   Initial implementation.
   25-apr-96    A. Kucharczyk   Introduced macros.
   24-may-96	M. Hildebrand	Added wide character support.
*/

#include "widechar.h"
#include <string.h>
#include "riscstr.h"


int __F_NAME(strncmp,wcsncmp)( const CHAR_TYPE *s1, const CHAR_TYPE *s2,
			       size_t n )
/**********************************************************************/
{
    RISC_DATA_LOCALREF;
    UINT *           dw1 = ROUND(s1); // round down to dword
    UINT *           dw2 = ROUND(s2);
    UINT             dword1, dword2;
    INT              shr1, shr2, shl1, shl2;
    INT              tmpchar, tmpdword1, tmpdword2;

    #ifdef __WIDECHAR__
	if( OFFSET(s1) % 2  ||  OFFSET(s2) % 2 ) {
	    return( __simple_wcsncmp( s1, s2, n ) );
	}
    #endif

    /*** Initialize locals ***/
    shr1 = OFFSET(s1) << 3; // sift right  = offset * 8
    shr2 = OFFSET(s2) << 3;
    shl1 = INT_SIZE - shr1; // shift left = 32 - shift right
    shl2 = INT_SIZE - shr2;
     
    if (shr1 != 0) {
        dword1 = *dw1++;
    }
    if (shr2 != 0) {
        dword2 = *dw2++;
    }

    /*** Scan in aligned 4-byte groups ***/
    for( ;; ) {
        if( n == 0 )  return( 0 ); // nothing left to compare
            
        if( shr1 == 0 ) {
            tmpdword1 = *dw1++;
        } else {
            tmpdword1 = dword1 >> shr1;
            dword1 = *dw1++;
            tmpdword1 |= (dword1 << shl1);
        }
        
        if( shr2 == 0 ) {
            tmpdword2 = *dw2++;
        } else {
            tmpdword2 = dword2 >> shr2;
            dword2 = *dw2++;
            tmpdword2 |= (dword2 << shl2);
        }

	if( n < CHARS_PER_WORD ) {	// some bytes can be set to 00
            tmpdword1 = FRONT_CHRS(tmpdword1,n);
            tmpdword2 = FRONT_CHRS(tmpdword2,n);
            break;
        }

        /*** Did s1 end already? ***/
        if( GOT_NIL(tmpdword1) )  break;
        /*** Are s1 and s2 still the same? ***/
	#ifdef __WIDECHAR__
	    if( tmpdword1 != tmpdword2 ) {
		tmpchar = CHR1(tmpdword1) - CHR1(tmpdword2);
		if( tmpchar )  return( tmpchar );

		return( CHR2(tmpdword1) - CHR2(tmpdword2) );
	    }
	#else
	    if( tmpdword1 != tmpdword2 ) {
		tmpchar = CHR1(tmpdword1) - CHR1(tmpdword2);
		if( tmpchar )  return( tmpchar );
    
		tmpchar = CHR2(tmpdword1) - CHR2(tmpdword2);
		if( tmpchar )  return( tmpchar );
    
		tmpchar = CHR3(tmpdword1) - CHR3(tmpdword2);
		if( tmpchar )  return( tmpchar );
    
		return( CHR4(tmpdword1) - CHR4(tmpdword2) );
	    }
	#endif

	n -= CHARS_PER_WORD;
    }


    /*** Scan the last byte(s) in the string ***/
    if( tmpdword1 == tmpdword2 )  return( 0 );

    /* we have a null char somewhere in the last dword */
    #ifdef __WIDECHAR__
	tmpchar = CHR1(tmpdword1);
	if( tmpchar == 0 ) {				/* 1st char is null */
	    return( tmpchar - CHR1(tmpdword2) );
	}

	tmpchar = CHR1(tmpdword1) - CHR1(tmpdword2);	/* 2nd char is null */
	if( tmpchar )  return( tmpchar );
	return( CHR2(tmpdword1) - CHR2(tmpdword2) );
    #else
	tmpchar = CHR1(tmpdword1);
	if( tmpchar == 0 ) {            // first char in the dword is null
	    return( tmpchar - CHR1(tmpdword2));
	}
    
	if( CHR2(tmpdword1) == 0 ) {     // second char in the dword is null
	    tmpchar = tmpchar - CHR1(tmpdword2);
	    if (tmpchar) return ( tmpchar) ;
	    
	    return ( CHR2(tmpdword1) - CHR2(tmpdword2) ) ;
	}
    
	if( CHR3(tmpdword1) == 0 ) {     // third char in the dword is null
	    tmpchar = tmpchar - CHR1(tmpdword2);
	    if (tmpchar) return ( tmpchar) ;
	    
	    tmpchar = CHR2(tmpdword1) - CHR2(tmpdword2);
	    if (tmpchar) return ( tmpchar) ;
	    
	    return ( CHR3(tmpdword1) - CHR3(tmpdword2) ) ;
	}
	
	// 4th char in the dword is null
	tmpchar = tmpchar - CHR1(tmpdword2);
	if (tmpchar) return ( tmpchar );
    
	tmpchar = CHR2(tmpdword1) - CHR2(tmpdword2);
	if (tmpchar) return ( tmpchar );
    
	tmpchar = CHR3(tmpdword1) - CHR3(tmpdword2);
	if (tmpchar) return ( tmpchar );
    
	return ( CHR4(tmpdword1) - CHR4(tmpdword2) );
    #endif
}
