/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %     Copyright (C) 1996, by Watcom International Inc.  All rights    %
   %     reserved.  No part of this software may be reproduced or        %
   %     used in any form or by any means - graphic, electronic or       %
   %     mechanical, including photocopying, recording, taping or        %
   %     information storage and retrieval systems - except with the     %
   %     written permission of Watcom International Inc.                 %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
   Date         By              Reason
   ----         --              ------
--------------------------------------------------------Version-11.0---------
   06-may-96    M. Hildebrand   Initial implementation.
   27-may-96	M. Hildebrand	Added wide character support.
*/

#include "widechar.h"
#include <string.h>
#include "riscstr.h"


CHAR_TYPE *__F_NAME(strncpy,wcsncpy)( CHAR_TYPE *dest, const CHAR_TYPE *src,
				      size_t n )
/**************************************************************************/
{
    CHAR_TYPE *		destStart = dest;
    UINT *		dwSrc;
    UINT *		dwDest;
    UINT		dword, tmpdword;
    int			shl, shr;
    int			offset = OFFSET(dest);

    #ifdef __WIDECHAR__
	if( offset % 2  ||  OFFSET(src) % 2 ) {
	    return( __simple_wcsncpy( dest, src, n ) );
	}
    #endif

    if( n <= 0 )  return( destStart );

    /*** Copy any unaligned bytes at the start (align dest pointer) ***/
    while( offset != 0  &&  n >= 1 ) {
	*dest++ = *src++;
	offset = MOD_BYTES_PER_WORD(offset+CHARSIZE); /* find new offset */
	n--;
    }

    /*** Initialize locals ***/
    dwSrc = ROUND(src);			/* round down to dword */
    dwDest = (UINT*)dest;		/* already rounded */
    shr = OFFSET(src) * 8;
    shl = INT_SIZE - shr;
    if( shr != 0 ) {
	dword = *dwSrc++;
    }

    /*** Copy in aligned 4-byte groups ***/
    for( ; n>=CHARS_PER_WORD; n-=CHARS_PER_WORD ) {
	/*** Form the dword containing the next four characters ***/
	if( shr == 0 ) {
	    tmpdword = *dwSrc++;
	} else {
	    tmpdword = dword >> shr;
	    dword = *dwSrc++;
	    tmpdword |= (dword << shl);
	}

        /*** Copy the characters unless the string has ended ***/
        if( GOT_NIL(tmpdword) ) {
	    break;
	} else {
	    *dwDest++ = tmpdword;
	}
    }

    /*** Either end of string reached or n is too small ***/
    dest = (CHAR_TYPE*)dwDest;
    if( shr == 0 ) {
	src = (CHAR_TYPE*)dwSrc;
    } else {
	src = (CHAR_TYPE*)dwSrc - (shl/8);
    }
    while( n >= 1  &&  *src != '\0' ) {
	*dest++ = *src++;
	n--;
    }

    /*** If necessary, pad with more null bytes ***/
    if( n > 0 ) {
	memset( dest, '\0', n*CHARSIZE );
    }

    return( destStart );
}
