/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1992, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 * This version of the clock routine uses DOS calls instead of BIOS
 * calls to make it compatable with Japanese machines

    Modified:	By:		Reason:
    ---------	---		-------

    92/05/13	D. Wellstood	Initial implementation based on assembly
				code from clock.asm
    92/07/31	J.B.Schueler	Fixed for midnight rollover
--------------------------------------------------------Version-11.0---------
    95/05/08    T. Schiller     Collapsed existing clock functions into
				one generic function using __getctime().
    95/05/18	Greg Bentz	use variety.h and _WCRTLINK
    95/09/22	T. Schiller	Changed precision from hundredths to
				to milliseconds.
 */

#include "variety.h"
#include <time.h>
#ifdef __QNX__
#include <sys/types.h>
#include <sys/timers.h>
#endif
#include <rtinit.h>
#include "timedata.h"

#define MAX_CLOCK_T   ~((clock_t)0)
#define MAX_SECONDS   ((time_t)(MAX_CLOCK_T / CLOCKS_PER_SEC) - 1)

static clock_t init_milliseconds;
static time_t  init_seconds;

static void get_clock_time(time_t *secs, clock_t *milliseconds)
{
#ifdef __QNX__
    struct timespec timer;

    getclock(TIMEOFDAY, &timer);
    *secs       = (time_t)timer.tv_sec;
    *milliseconds = (clock_t)(timer.tv_nsec / (1000000000 / CLOCKS_PER_SEC));
#else
    struct tm t;

    *milliseconds = (clock_t)__getctime(&t);
    *secs       = mktime(&t);
#endif
} /* get_clock_time() */


_WCRTLINK clock_t clock(void)
{
    time_t  new_seconds;
    clock_t ticks;

    /*
     * Get the change in seconds and milliseconds of seconds since startup.
     */
    get_clock_time(&new_seconds, &ticks);
    ticks       -= init_milliseconds;
    new_seconds -= init_seconds;

    /*
     * Make sure we won't overflow.
     */
    if (new_seconds > MAX_SECONDS)
	    return -1;

    /*
     * `ticks' right now contains the number of milliseconds of seconds since
     * startup.  We still need to account for the number of full seconds that
     * may have passed.
     */
    ticks += (clock_t)(new_seconds * CLOCKS_PER_SEC);

    return ticks;
} /* clock() */


static void __clock_init(void)
{
    get_clock_time(&init_seconds, &init_milliseconds);
} /* __clock_init() */

AXI( __clock_init, INIT_PRIORITY_LIBRARY )
