/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  18-jan-87	F.W.Crigger	initial implementation
  04-feb-88	F.W.Crigger	change time to base 1970.
----------------------------------------------------------Version 6.0------
  13-apr-88	F.W.Crigger	_cnvtime renamed to __brktime
				__isindst made global
----------------------------------------------------------Version 7.0------
  25-may-89	John Dahms	split out localtime
  27-mar-90	F.W.Crigger	add special check for year == 0
----------------------------------------------------------Version 8.0------
  31-oct-90	J.B.Schueler	__isindst moved to timeutil.c since not needed
				in gmtime.c
  12-jul-91	J.B.Schueler	macrotized R/W variable names
----------------------------------------------------------Version 8.5------
  18-dec-91	F.W.Crigger	Added _gmtime
----------------------------------------------------------Version 9.0------
  26-apr-92	B.J. Stecher	handle time_t being a signed type (QNX)
  24-sep-92	F.W.Crigger	Use much simpler method for computing tm_wday
  23-dec-92	F.W.Crigger	fix problem with 1904
   4-jan-93	B.J. Stecher	overflow problems from mktime
----------------------------------------------------------Version 9.5------
  17-jun-93	J.B.Schueler	corrected for fringe case (the 12 hours after
				Thu Jan  1 00:00:00 1970) for time zones west
				of UTC like EST, PST.
--------------------------------------------------------Version-11.0---------
  18-may-95	Greg Bentz	use variety.h and _WCRTLINK

*/

/*
 *	Convert time_t into struct tm
 */

#include "variety.h"
#include <time.h>
#include "rtdata.h"
#include "thetime.h"
#include "timedata.h"

/*  __brktime breaks down a calendar time (clock) into a struct tm t */

struct tm *__brktime( unsigned long days,
		      time_t clock,
		      long timezone,
		      struct tm *t )
{
    unsigned long	secs;
    unsigned		year;
    unsigned		day_of_year;
    int 		month;
    int 		days_in_year;
    short const 	*month_start;

    /*
	If date is Jan 1, 1970 0:00 to 12:00 UTC and we are west of UTC
	then add a day to clock, subtract the timezone value, and
	decrement the calculated days. This prevents local times
	such as "Wed Dec 31 19:00:00 1969 (EST)" from being
	erroneously reported as "Sun Feb 6 01:28:16 2106 (EST)"
	since (clock - timezone) wraps (i.e., clock < timezone).
    */
    if( (clock < 12*60*60UL) && (timezone > 0 ) ) {
	clock += SECONDS_PER_DAY;
	clock -= timezone;
	days += clock / SECONDS_PER_DAY;
	days--;
	/*
	    clock is now ahead one day but this doesn't
	    affect the time-of-day calculation
	*/
    } else {
	clock -= timezone;
	days += clock / SECONDS_PER_DAY;
    }
    secs = clock % SECONDS_PER_DAY;
    t->tm_hour = secs / 3600;
    secs = secs % 3600;
    t->tm_min = secs / 60;
    t->tm_sec = secs % 60;
    year = days / 366;	    /* approximate # of years */
    day_of_year = days - year * 365;
    if( year > 0 ) day_of_year -= (year - 1) / 4;	    /* 27-mar-90 */
    days_in_year = __leapyear( year + 1900 ) + 365;
    while( day_of_year >= days_in_year ) {
	day_of_year -= days_in_year;
	++year;
    }
    t->tm_yday = day_of_year;
    t->tm_year = year;
    month_start = __diyr;
    if( __leapyear( year + 1900 ) )  month_start = __dilyr;
    month = day_of_year / 31;		    /* approximate month */
    if( day_of_year >= month_start[ month + 1 ] ) ++month;
    t->tm_mon  = month;
    t->tm_mday = day_of_year - month_start[ month ] + 1;

/*  Calculate the day of the week */
/*   Jan 1,1900 is a Monday */

    t->tm_wday = (days + 1) % 7;		/* 24-sep-92 */
    return( t );
}

_WCRTLINK struct tm *_gmtime( const time_t *timer, struct tm *tm )
{
    tm->tm_isdst = 0;	       /* assume not */
    return __brktime( DAYS_FROM_1900_TO_1970, *timer, 0L, tm );
}

_WCRTLINK struct tm *gmtime( const time_t *timer )
{
    _INITTHETIME;
    return( _gmtime( timer, &_THE_TIME ) );
}
