/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  20-mar-87     F.W.Crigger     Initial implementation
  04-feb-88     F.W.Crigger     Make time_t relative to 1970
----------------------------------------------------------Version 6.0------
  13-apr-88     F.W.Crigger     mktime fixed to account for timezone & d.s.t.
  25-apr-88     F.W.Crigger     check daylight variable before __isindst()
  17-aug-90     B.J. Stecher    added __dst_adjust variable
  01-nov-90     J.B.Schueler    placed daylight test check in __isindst()
  12-jul-91     J.B.Schueler    Macrotized references to R/W data
----------------------------------------------------------Version 9.5------
   4-jan-93     B.J. Stecher    overflow problems when year > 2036
  17-jun-93     J.B.Schueler    added new parm to __brktime (timezone adjust)
  18-jun-93     J.B.Schueler    corrected for fringe case (the 12 hours after
                                Thu Jan  1 00:00:00 1970) for time zones west
                                of UTC like EST, PST.
  24-mar-94	Greg Bentz	__brktime() requires the seconds to be > 0
--------------------------------------------------------Version-11.0---------
  18-may-95	Greg Bentz	use variety.h and _WCRTLINK
  19-jun-96	Greg Bentz	adjust days in years for 2000 and 2100
*/

#include "variety.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <limits.h>
#include "rtdata.h"
#include "timedata.h"

#define MONTH_YR	(12)
#define DAY_YR		(365)
#define HOUR_YR		(DAY_YR*24)
#define MINUTE_YR	(HOUR_YR*60)
#define SECOND_YR	(MINUTE_YR*60)
#define __MONTHS	(INT_MIN/MONTH_YR)
#define __DAYS		(INT_MIN/DAY_YR)

// these ones can underflow in 16bit environments,
// so check the relative values first
#if (HOUR_YR) < (INT_MAX/60)
 #define __MINUTES	(INT_MIN/MINUTE_YR)
 #if (MINUTE_YR) < (INT_MAX/60)
  #define __SECONDS	(INT_MIN/SECOND_YR)
 #else
  #define __SECONDS	(0)
 #endif
#else
 #define __MINUTES	(0)
 #define __SECONDS	(0)
#endif

#define SMALLEST_YEAR_VALUE (__MONTHS+__DAYS+__MINUTES+__SECONDS)

_WCRTLINK time_t mktime( struct tm *t )
{
    int         month;
    int         year;
    long        days;
    long        seconds;
    short const *month_start;

    month_start = __diyr;
    month = t->tm_mon % 12;		/* put tm_mon into range */
    year = t->tm_year;
    if( year < SMALLEST_YEAR_VALUE ) {
	return( (time_t)-1 );
    }
    year += t->tm_mon / 12;
    while( month < 0 ) {
        --year;
        month += 12;
    }
    if( year < 0 ) return( (time_t)-1 );
    if( __leapyear( year + 1900 ) ) month_start = __dilyr;
    days = (unsigned)year * 365         /* # of days in the years */
	 + ((year + 3) / 4)		/* add # of leap years before year */
	 - ((year + 99) / 100)		/* sub # of leap centuries */
	 + ((year + 399 - 100) / 400)	/* add # of leap 4 centuries */
					/* adjust for 1900 offset */
					/* note: -100 == 300 (mod 400) */
         + month_start[ month ]         /* # of days to 1st of month*/
         + t->tm_mday - 1;              /* day of the month */
    seconds = (((long)(t->tm_hour))*60L + (long)(t->tm_min))*60L + t->tm_sec;
    /* seconds needs to be positive for __brktime */
    while( seconds < 0 ) {
        days -= 1;
	seconds += SECONDS_PER_DAY;
    }
    __brktime( days, seconds, 0L, t );
    tzset();
    seconds += _RWD_timezone;       /* add in seconds from GMT */
    /* if we are in d.s.t. then subtract __dst_adjust from seconds */
    if( t->tm_isdst < 0 ) {         /* if we are not sure */
        __isindst( t );             /* - determine if we are in d.s.t. */
    }
    if( t->tm_isdst > 0 ) {
        seconds -= _RWD_dst_adjust;
    }
    while( seconds < 0 ) {
        days -= 1;
        seconds += SECONDS_PER_DAY;
    }

    if( days < (DAYS_FROM_1900_TO_1970 - 1) ) {
        return( (time_t)-1 );
    }
    if( days == (DAYS_FROM_1900_TO_1970 - 1) ) {
        seconds -= SECONDS_PER_DAY;
        if( (_RWD_timezone <= 0) || (seconds < 0) ) {
            return( (time_t)-1 );
        }
    } else {
        seconds += (days - DAYS_FROM_1900_TO_1970) * SECONDS_PER_DAY;
    }
    return( seconds );
}

#if defined(__PENPOINT__)
_WCRTLINK time_t _gmmktime( struct tm *t )
{
    int         month;
    int         year;
    long        days;
    long        seconds;
    short const *month_start;

    month_start = __diyr;
    month = t->tm_mon % 12;                 /* put tm_mon into range */
    year = t->tm_year + t->tm_mon / 12;
    while( month < 0 ) {
        --year;
        month += 12;
    }
    if( year < 0 ) return( (time_t)-1 );
    if( __leapyear( year + 1900 ) ) month_start = __dilyr;
    days = (unsigned)year * 365         /* # of days in the years */
         + ((year + 3 ) >> 2)           /* # of leap years before year*/
         + month_start[ month ]         /* # of days to 1st of month*/
         + t->tm_mday - 1;              /* day of the month */
    if( year !=0 ) --days;              /* 1900 is not a leap year */
    seconds = (((long)(t->tm_hour))*60L + (long)(t->tm_min))*60L + t->tm_sec;
    /* seconds needs to be positive for __brktime */
    while( seconds < 0 ) {
	days -= 1;
	seconds += SECONDS_PER_DAY;
    }
    __brktime( days, seconds, 0L, t );
    while( seconds < 0 ) {
        days -= 1;
        seconds += SECONDS_PER_DAY;
    }
    if( days < (DAYS_FROM_1900_TO_1970 - 1) ) {
        return( (time_t)-1 );
    }
    if( days == (DAYS_FROM_1900_TO_1970 - 1) ) {
        seconds -= SECONDS_PER_DAY;
        if( seconds < 0 ) {
            return( (time_t)-1 );
        }
    } else {
        seconds += (days - DAYS_FROM_1900_TO_1970) * SECONDS_PER_DAY;
    }
    return( seconds );
}
#endif
