/*
 ********************************************************************
 *	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     *
 *	  reserved. No part of this software may be reproduced	    *
 *	  in any form or by any means - graphic, electronic or	    *
 *	  mechanical, including photocopying, recording, taping     *
 *	  or information storage and retrieval systems - except     *
 *	  with the written permission of WATCOM Systems Inc.	    *
 ********************************************************************

  Modified:	By:		Reason:
  ---------	---		-------
  18-dec-87	H.W.Hillis	Initial implementation
  19-feb-88	B.J.Stecher	set daylight to 0 (matches documentation now)
  25-feb-88	F.W.Crigger	daylight set back to 1
------------------------------------------------------------Version 6.5-----
  08-dec-88	F.W.Crigger	Removed calls to atol and isdigit
  26-oct-90	B.J.Stecher	Added support for POSIX 1003.1 TZ format
  30-oct-90	J.B.Schueler	Daylight savings (daylight) cannot be set
				simply because TZ specifies the D.S.T. name.
  30-oct-90	J.B.Schueler	Added support for daylight savings time rule
  02-nov-90	J.B.Schueler	Forgot to skip over "," in dst rule strings
  02-nov-90	J.B.Schueler	dst_adjust = timezone - dayzone (e.g.,EST5DST4)
				(i.e., by default, we add 1hr to EST for DST)
  02-nov-90	J.B.Schueler	dayzone was clobbered if = 0, changed logic
  12-jul-91	J.B.Schueler	modified to support PenPoint R/W variables
  16-aug-91	F.W.Crigger	calc. of __dst_adjust was backwards
  04-apr-92	B.J. Stecher	changed dtzone from DST to proper name of EDT
  18-jun-93	John Dahms  	all library allocations use lib_xxx routines
  07-oct-93	A.F.Scian	added beginnings of setting timzezone info
  				from system for Windows/NT
  11-nov-93	Greg Bentz	rationalize exit_with_msg
  28-jan-94	Greg Bentz	change exit_with_msg to __fatal_runtime_error
  17-oct-94	Greg Bentz	deal with boundary condition at change
  				from EDT to EST, end of EDT is now in terms
				of EST
  05-Dec-94	A.F.Scian	setup mechanism for caching OS TZ info
--------------------------------------------------------Version-11.0---------
  18-may-95	Greg Bentz	use variety.h and _WCRTLINK
  28-jun-96	M. Hildebrand	Fixed time zone name handling.
  12-Sep-96	A.F.Scian	by default cache OS TZ info
*/
#include "variety.h"
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <limits.h>
#ifdef __NT__
 #include <windows.h>
#endif
#include "rtdata.h"
#include "timedata.h"
#include "exitwmsg.h"

/*
 * TZSET - sets the values of the variables 'timezone', 'daylight'
 *	   and 'tzname' according to the setting of the environment
 *	   variable "TZ". The "TZ" variable has the format
 *	    :characters
 *		or
 *	    <std><offset>[<dst>[<offset>]][,<start>[/<time>],<end>[/<time>]]]
 */

//#define TZNAME_MAX	128	/* defined in <limits.h> */

/* PenPoint FXX: Move these to protected storage */

struct tm __start_dst = 			/* start of daylight savings */
	{ 0, 0, 2,				/* M4.1.0/02:00:00 default */
	  1, 3, 0,				/* i.e., 1st Sunday of Apr */
	  0, 0, 0
	};

struct tm __end_dst =				/* end of daylight savings */
	{ 0, 0, 1,				/* M10.5.0/02:00:00 default */
	  5, 9, 0,				/* i.e., last Sunday of Oct */
	  0, 0, 0				/* note that this is specified*/
	};					/* in terms of EST */
						/* i.e. 02:00 EDT == 01:00 EST*/

static char stzone[TZNAME_MAX+1] = "EST";	/* place to store names */
static char dtzone[TZNAME_MAX+1] = "EDT";	/* place to store names */

_WCRTLINK char	*tzname[2]	= { stzone, dtzone };

_WCRTLINK long	timezone     = 5L * 60L * 60L;	/* seconds from GMT */
_WCRTLINK int	daylight     = 1;		/* d.s.t. indicator */
	 int	__dst_adjust = 60L * 60L;	/* daylight adjustment */

static struct {
    unsigned	cache_OS_TZ : 1;
    unsigned	have_OS_TZ : 1;
} tzFlag = { 1, 0 };

#if defined(__PENPOINT__)

#undef tzset
#include "liballoc.h"

void __alloctime()
{
    if( _RWD_tzinfo != NULL ) return;

    _RWD_tzinfo = lib_malloc( sizeof( struct _tzdata ) );
    if( _RWD_tzinfo == NULL ) {
	__fatal_runtime_error(
	    "Not enough memory to allocate time structures\r\n", 1 );
    }
}

#endif

int __DontCacheOSTZ( void )
{
    int old_flag;

    old_flag = tzFlag.cache_OS_TZ;
    tzFlag.cache_OS_TZ = 0;
    tzFlag.have_OS_TZ = 0;
    return( old_flag );
}

int __CacheOSTZ( void )
{
    int old_flag;

    old_flag = tzFlag.cache_OS_TZ;
    tzFlag.cache_OS_TZ = 1;
    tzFlag.have_OS_TZ = 0;
    return( old_flag );
}

static void tryOSTimeZone( void )
{
    if( tzFlag.cache_OS_TZ && tzFlag.have_OS_TZ ) {
	/* calling the OS can be expensive and many programs don't care */
	return;
    }
    tzFlag.have_OS_TZ = 1;
#ifdef __NT__
    {
	auto TIME_ZONE_INFORMATION  tz_info;
	size_t			    rc;

	switch( GetTimeZoneInformation( &tz_info ) ) {
	case TIME_ZONE_ID_DAYLIGHT:
	    _RWD_daylight = 1;
	    _RWD_dst_adjust = tz_info.DaylightBias * -60L;
	    break;
	case TIME_ZONE_ID_STANDARD:
	    _RWD_daylight = 0;
	    break;
	case TIME_ZONE_ID_UNKNOWN:
	    return;
	default:
	    return;
	}
	_RWD_timezone = ( tz_info.StandardBias + tz_info.Bias ) * 60L;

	rc = wcstombs( stzone, tz_info.StandardName, TZNAME_MAX );
	if( rc == (size_t)-1 ) {	/* cannot convert string */
	    stzone[0] = '\0';
	} else {			/* ensure null-terminated */
	    stzone[TZNAME_MAX] = '\0';
	}

	rc = wcstombs( dtzone, tz_info.DaylightName, TZNAME_MAX );
	if( rc == (size_t)-1 ) {	/* cannot convert string */
	    dtzone[0] = '\0';
	} else {			/* ensure null-terminated */
	    dtzone[TZNAME_MAX] = '\0';
	}

	return;
    }
#endif
}

_WCRTLINK void tzset( void )
/*************************/
{
    #ifndef __NETWARE__
	char    *tz;

	tz = getenv( "TZ" );
	if( tz == NULL ) {
 	    tryOSTimeZone();
	} else {
	    __parse_tz( tz );
	}
    #endif
}

static char *parse_time( char *tz, int *val )
{
    int value;

    value = 0;
    while( *tz >= '0'  &&  *tz <= '9' ) {
	value = value * 10 + *tz - '0';
	++tz;
    }
    *val = value;
    return( tz );
}

static char *parse_offset( char *tz, char *name, long *offset )
{
	int	hours;
	int	minutes;
	int	seconds;
	int	neg;
	int	len;
	char	ch;
	char	*tzstart;

	if( *tz == ':' ) tz++;
	/* remember where time zone name string begins */
	tzstart = tz;
	/* parse time zone name (should be 3 or more characters) */
	/* examples:	PST8, EDT+6, Central Standard Time+7:00:00 */
	for( ;; ) {
	    ch = *tz;
	    if( ch == '\0' ) break;
	    if( ch == ',' ) break;
	    if( ch == '-' ) break;
	    if( ch == '+' ) break;
	    if( ch >= '0' && ch <= '9' ) break;
	    ++tz;
	}
	len = tz - tzstart;
	if( len > TZNAME_MAX ) len = TZNAME_MAX;
	memcpy( name, tzstart, len );
	name[ len ] = '\0';

	neg = 0;
	if( ch == '-' ) {
	    neg = 1;
	    ++tz;
	} else if( ch == '+' ) {
	    ++tz;
	}
	ch = *tz;
	if( ch >= '0' && ch <= '9' ) {
	    hours = minutes = seconds = 0;
	    tz = parse_time( tz, &hours );
	    if( *tz == ':' ) {
		tz = parse_time( tz + 1, &minutes );
		if( *tz == ':' ) {
		    tz = parse_time( tz + 1, &seconds );
		}
	    }
	    *offset = seconds + ((minutes + (hours * 60)) * 60L);
	    if( neg ) *offset = -*offset;
	}
	return( tz );
}

static char *parse_rule( char *tz, struct tm *timeptr )
{
	int	date_form;
	int	days;
	int	hours;
	int	minutes;
	int	seconds;

	date_form = -1; 			/* n  0-365 */
	if( *tz == 'J' ) {			/* Jn 1-365 (no leap days) */
	    date_form = 1;
	    tz++;
	}
	if( *tz == 'M' ) {			/* Mm.n.d n'th day of month */
	    date_form = 0;
	    tz++;
	}
	timeptr->tm_isdst = date_form;
	tz = parse_time( tz, &days );
	if( date_form != 0 ) {
	    timeptr->tm_yday = days;
	} else {
	    timeptr->tm_mon = days - 1; 	    /* 1-12 for M form */
	    if( *tz == '.' ) {
		tz = parse_time( tz + 1, &days );   /* 1-5 for M form */
		timeptr->tm_mday = days;
		if( *tz == '.' ) {
		    tz = parse_time( tz + 1, &days );/* 0-6 for M form */
		    timeptr->tm_wday = days;
		}
	    }
	    timeptr->tm_yday = 0;
	}

	hours = 2;
	minutes = seconds = 0;
	if( *tz == '/' ) {
	    tz = parse_time( tz + 1, &hours );
	    if( *tz == ':' ) {
		tz = parse_time( tz + 1, &minutes );
		if( *tz == ':' ) {
		    tz = parse_time( tz + 1, &seconds );
		}
	    }
	}
	timeptr->tm_sec = seconds;
	timeptr->tm_min = minutes;
	timeptr->tm_hour = hours;
	return( tz );
}

void __parse_tz( char * tz )
/**************************/
{
    long	    dayzone;

    _RWD_daylight = 0;
    tz = parse_offset( tz, stzone, &_RWD_timezone );
    if( *tz == '\0' ) {
	dtzone[0] = '\0';
	return;
    }
    _RWD_daylight = 1;
    dayzone = _RWD_timezone - (60*60);		    /* 16-aug-91 */
    tz = parse_offset( tz, dtzone, &dayzone );
    _RWD_dst_adjust = _RWD_timezone - dayzone;

    /* parse daylight changing rules */
    if( *tz == ',' ) tz = parse_rule( tz + 1, &__start_dst );
    if( *tz == ',' ) {
	tz = parse_rule( tz + 1, &__end_dst );
	/* convert rule to be in terms of Standard Time */
	/* rather than Daylight Time */
	__end_dst.tm_hour -= _RWD_dst_adjust/3600;
	__end_dst.tm_min -= (_RWD_dst_adjust/60)%60;
	__end_dst.tm_sec -= _RWD_dst_adjust%60;
    }
}
