/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %       Copyright (C) 1991, by WATCOM Systems Inc. All rights     %
  %       reserved. No part of this software may be reproduced      %
  %       in any form or by any means - graphic, electronic or      %
  %       mechanical, including photocopying, recording, taping     %
  %       or information storage and retrieval systems - except     %
  %       with the written permission of WATCOM Systems Inc.        %
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  Date		By		Reason
  ====		==		======
  19-apr-92	Craig Eisler	initial implementation
  16-jul-92	Craig Eisler	32-bit DLL's had pointer trunctations
  20-apr-93	Craig Eisler	mmioOpen has to alias its data buffer;
  				mmioClose has to free the alias.
  06-may-93	F.W.Crigger	removed index parm from BackPatch routine
  02-jun-93	F.W.Crigger	modified to run in 32-bit mode
  22-apr-96	Greg Bentz	tweak typing inconsistencies
*/
#include "alloc.h"
#include <stddef.h>
#include <malloc.h>


extern HMMIO PASCAL __mmioOpen(LPSTR szFileName, LPMMIOINFO lpmmioinfo, DWORD dwOpenFlags);
extern UINT PASCAL __mmioClose(HMMIO hmmio, UINT flags);
extern LONG PASCAL __mmioRead(HMMIO hmmio, const char *pch, LONG cch);
extern LONG PASCAL __mmioWrite(HMMIO hmmio, HPSTR pch, LONG cch);


typedef struct mminfo_list {
    struct mminfo_list		*next;
    HMMIO			handle;
    unsigned short		filler;
    DWORD			size;
    DWORD			alias;
} mminfo_list;
typedef mminfo_list *LPMMINFOLIST;

static LPMMINFOLIST	mminfoListHead;

/*
 * mmioOpen - cover for mmsystem function mmioOpen
 */
HMMIO PASCAL mmioOpen( LPSTR szFileName, LPMMIOINFO lpmmioinfo,
			DWORD dwOpenFlags)
{
    HMMIO		rc;
    DWORD		alias;
    LPMMINFOLIST	curr;

    alias = NULL;
    if( !(dwOpenFlags & (MMIO_ALLOCBUF | MMIO_DELETE | MMIO_PARSE | MMIO_EXIST
    			 	| MMIO_GETTEMP) ) ) {
	if( lpmmioinfo->cchBuffer != 0 && lpmmioinfo->pchBuffer != NULL ) {
	    DPMIGetHugeAlias( (DWORD) lpmmioinfo->pchBuffer, &alias,
	    			lpmmioinfo->cchBuffer );
	    lpmmioinfo->pchBuffer = (LPVOID) alias;
	}
    }

    rc = __mmioOpen( szFileName, lpmmioinfo, dwOpenFlags );

    if( rc != NULL && alias != NULL ) {
	curr = malloc( sizeof( mminfo_list ) );
	if( curr != NULL ) {
	    curr->next = mminfoListHead;
	    curr->handle = rc;
	    curr->alias = alias;
	    curr->size = lpmmioinfo->cchBuffer;
	    mminfoListHead = curr;
	}
    }
    return( rc );

} /* mmioOpen */

/*
 * mmioClose - cover for mmsystem function mmioClose
 */
UINT PASCAL mmioClose( HMMIO hmmio, UINT flags )
{
    UINT		rc;
    LPMMINFOLIST	curr,prev;

    rc = __mmioClose( hmmio, flags );

    curr = mminfoListHead;
    prev = NULL;
    while( curr != NULL ) {
	if( curr->handle == hmmio ) {
	    DPMIFreeHugeAlias( curr->alias, curr->size );
	    if( prev == NULL ) {
		mminfoListHead = curr->next;
	    } else {
		prev->next = curr->next;
	    }
	    free( curr );
	    break;
	}
	prev = curr;
	curr = curr->next;
    }
    return( rc );

} /* mmioOpen */

/*
 * mmioRead - cover for mmsystem function mmioRead
 */
LONG PASCAL mmioRead( HMMIO hmmio, HPSTR pch, LONG cch )
{
    LONG	rc;
    DWORD	alias;

    DPMIGetHugeAlias( (DWORD) pch, &alias, cch );
    rc = __mmioRead( hmmio, (HPSTR) alias, cch );
    DPMIFreeHugeAlias( alias, cch );
    return( rc );

} /* mmioRead */

/*
 * mmioWrite - cover for mmsystem function mmioWrite
 */
LONG PASCAL mmioWrite( HMMIO hmmio, const char *pch, LONG cch )
{
    LONG	rc;
    DWORD	alias;

    DPMIGetHugeAlias( (DWORD) pch, &alias, cch );
    rc = __mmioWrite( hmmio, (HPSTR) alias, cch );
    DPMIFreeHugeAlias( alias, cch );
    return( rc );

} /* mmioWrite */
