/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1987, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  31-oct-87	F.W.Crigger	Initial implementation
--------------------------------------------------------------------------
  27-oct-89	F.W.Crigger	if error code >= 0x0100 errno value in high byte
  28-dec-89	F.W.Crigger	Added __EINVAL()
--------------------------------------------------Version 8.5----------
  20-nov-91	G. Coschi	support OS/2 2.0
--------------------------------------------------Version 9.x----------
  05-mar-92	Craig Eisler	support NT
  27-may-92	J.B.Schueler	added support for "File already exists" error
  03-may-94	Greg Bentz	support EEXIST under NT
  				this is a stop-gap measure, the real solution
				is to set up a table to map the most common
				NT error codes to errno values
  06-apr-95	Greg Bentz	support EEXIST under OS/2 1.x
--------------------------------------------------------Version-11.0---------
  19-may-95	Greg Bentz	dll linkage
  19-jun-95	T. Schiller	Changed returncode from EXDEV to EINVAL when
                                there is an invalid drive in the search path.
*/
#include "variety.h"
#if   defined(__NT__)
#include <windows.h>
#elif defined(__OS2__)
#define INCL_DOSERRORS
#include <wos2.h>
#endif

#include <errno.h>
#include <stdlib.h>
#include "seterrno.h"
#include "rtdata.h"

#define E_ACCESS	5	/* access denied */
#define E_EXIST 	14	/* file already exists, DOS 3.0 */
#define E_MAXERR2	19	/* unknown error, DOS 2.0 */

#define DOS_SHARERR	0x20	/* DOS 3.0, sharing violation */
#define DOS_MAXERR	0x22	/* DOS 3.0, invalid disk change */
#define DOS_EXIST	0x50	/* DOS 3.0, file already exists */


/* translation table of DOS error numbers to errno codes */

static signed char xlat[] = {
/*	errno values	   DOS error values */
	0,		/* 00 - no error */
	EINVAL, 	/* 01 - invalid function number */
	ENOENT, 	/* 02 - file not found */
	ENOENT, 	/* 03 - path not found */
	EMFILE, 	/* 04 - too many open files (no handles left) */
	EACCES, 	/* 05 - access denied */
	EBADF,		/* 06 - invalid handle */
	ENOMEM, 	/* 07 - memory control blocks destroyed */
	ENOMEM, 	/* 08 - insufficient memory */
	ENOMEM, 	/* 09 - invalid memory block address */
	E2BIG,		/* 10 - invalid environment */
	ENOEXEC,	/* 11 - invalid format */
	-1,		/* 12 - invalid access code */
	-1,		/* 13 - invalid data */
	EEXIST, 	/* 14 - (reserved) */
	EINVAL,		/* 15 - invalid drive was specified */
	EACCES, 	/* 16 - attempt to remove current directory */
	EXDEV,		/* 17 - not same device */
	ENOENT, 	/* 18 - no more files */
	-1		/*    - unknown error */
};


int _dosret0( int ax, int carry )
    {
	if( carry == 0 ) return( 0 );
	return( _dosretax( ax, carry ) );
    }


int _dosretax( int ax, int carry )
    {
	if( carry != 0 ) {
	    return( __set_errno_dos( ax ) );
	}
	return( ax );
    }



_WCRTLINK int __set_errno_dos( unsigned int err )
    {

#if defined __NT__
	__set_doserrno( err );
	if( err == ERROR_INVALID_NAME ) {
	    __set_errno( ENOENT );
	} else if( err == ERROR_FILENAME_EXCED_RANGE ) {
	    __set_errno( EINVAL );
	} else if( err == ERROR_ALREADY_EXISTS ) {
	    __set_errno( EEXIST );
	} else {
	    if( err > E_MAXERR2 ) err = E_MAXERR2;
	    __set_errno( xlat[ err ] );
	}
#elif defined( __OS2__ )
	__set_doserrno( err );
	// map error codes so that we get a meaningful "errno" value
	if( ( err == ERROR_OPEN_FAILED ) || ( err == ERROR_INVALID_NAME ) ) {
	    __set_errno( ENOENT );
	} else if( err == ERROR_FILENAME_EXCED_RANGE ) {
	    __set_errno( EINVAL );
	} else if( err == ERROR_ALREADY_EXISTS ) {
	    __set_errno( EEXIST );
	} else if( (err == ERROR_SHARING_VIOLATION) ||
		   (err == ERROR_LOCK_VIOLATION) ) {
	    __set_errno( EACCES );
	} else  {
	    if( err > E_MAXERR2 ) err = E_MAXERR2;
	    __set_errno( xlat[ err ] );
	}
#else
	register unsigned char index;

	index = err & 0xff;
	__set_doserrno( index );
	if( err < 0x100 ) {
	    if( _RWD_osmajor >= 3 ) {
		if( index == DOS_EXIST ) {
		    index = E_EXIST;
		} else if( index >= DOS_MAXERR ) {
		    index = E_MAXERR2;
		} else if( index >= DOS_SHARERR ) {
		    index = E_ACCESS;
		}
	    }
	    if( index > E_MAXERR2 ) index = E_MAXERR2;
	    __set_errno( xlat[ index ] );
	} else {
	    __set_errno( (err >> 8) & 0xff );
	}
#endif
	return( -1 );
    }


#ifdef __NT__
int __set_errno_nt( void )
{
    DWORD	err;

    err = GetLastError();
    return( __set_errno_dos( err ) );

}
#endif
