/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1991, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  29-apr-91	F.W.Crigger	Initial implementation for OS/2
--------------------------------------------------Version 8.5----------
  11-nov-91     G. Coschi	support OS/2 2.0
--------------------------------------------------Version 9.0----------
  03-jul-92	D.J.Gaudet	HPFS filename support
--------------------------------------------------Version 9.5----------
  15-feb-94     G. Coschi	don't call DosFindClose() if DosFindFirst()
  				fails
  07-mar-94	Kevin Hui	FF_BUFFER should map to FILEFINDBUF for
  				16-bit.
--------------------------------------------------------Version-11.0---------
  19-may-95	Greg Bentz	dll linkage
*/

#include "variety.h"
#include <errno.h>
#include <stddef.h>
#include <stdlib.h>
#include <direct.h>
#include <dos.h>
#define INCL_ERRORS
#include <wos2.h>
#include "tinyio.h"
#include "rtdata.h"
#include "seterrno.h"

#if defined(__WARP__)
  #define FF_LEVEL	1
  #define FF_BUFFER	FILEFINDBUF3
#else
  #define FF_LEVEL	0
  #define FF_BUFFER	FILEFINDBUF
#endif


/* we'll use this to copy from a FILEFINDBUF to a DIR in copydir() */
struct name {
    char buf[ NAME_MAX + 1 ];
};

#define HANDLE_OF( __find )	( *( HDIR * )( &(__find)->reserved[0] ) )
#define BAD_HANDLE		((HDIR)(~0))


static	void	copydir( struct find_t *buf, FF_BUFFER *dir_buff ) {
/******************************************************************/

    buf->attrib  = dir_buff->attrFile;
    buf->wr_time = *(unsigned short *)&dir_buff->ftimeLastWrite;
    buf->wr_date = *(unsigned short *)&dir_buff->fdateLastWrite;
    buf->size	 = dir_buff->cbFile;

    *(struct name *)buf->name = *(struct name *)dir_buff->achName;
}


_WCRTLINK unsigned _dos_findfirst( const char *path, unsigned attr,
                                  struct find_t *buf ) {
/******************************************************/

    APIRET  rc;

#if defined(__OS2_286__)
    if( _RWD_osmode == OS2_MODE ) {
#endif
	FF_BUFFER	dir_buff;
	HDIR	    handle = BAD_HANDLE;
	OS_UINT	    searchcount;

	searchcount = 1;	/* only one at a time */
	rc = DosFindFirst( (PSZ)path, (PHFILE)&handle, attr,
		(PVOID)&dir_buff, sizeof( dir_buff ), &searchcount, FF_LEVEL );

	if( rc != 0 && rc != ERROR_EAS_DIDNT_FIT ) {
	    HANDLE_OF( buf ) = BAD_HANDLE;
	    __set_errno_dos( rc );
	    return( rc );
	}
	HANDLE_OF( buf ) = handle;
	copydir( buf, &dir_buff );	/* copy in other fields */

#if defined(__OS2_286__)
    } else {			/* real mode */
	TinySetDTA( buf );	/* set our DTA */

	rc = TinyFindFirst( path, attr );
	if( rc > 0 ) {
	    __set_errno_dos( rc );
	    return( rc );
	}
    }
#endif
    return( 0 );
}


_WCRTLINK unsigned _dos_findnext( struct find_t *buf ) {
/*****************************************************/

    APIRET  rc;

#if defined(__OS2_286__)
    if( _RWD_osmode == OS2_MODE ) {	   /* protected mode */
#endif
	FF_BUFFER	dir_buff;
	OS_UINT		searchcount = 1;

	rc = DosFindNext( HANDLE_OF( buf ), (PVOID)&dir_buff,
		    sizeof( dir_buff ), &searchcount );
	if( rc != 0 ) {
	    __set_errno_dos( rc );
	    return( rc );
	}

	copydir( buf, &dir_buff );

#if defined(__OS2_286__)
    } else {		/* real mode */
	TinySetDTA( buf );
	rc = TinyFindNext();
	if( rc > 0 ) {
	    __set_errno_dos( rc );
	    return( rc );
	}
    }
#endif
    return( 0 );
}


_WCRTLINK unsigned _dos_findclose( struct find_t *buf ) {
/******************************************************/

    APIRET	rc;
    
#if defined(__OS2_286__)
    if( _RWD_osmode == OS2_MODE ) {	   /* protected mode */
#endif
	if( HANDLE_OF( buf ) != BAD_HANDLE ) {
	    rc = DosFindClose( HANDLE_OF( buf ) );
	    if( rc != 0 ) {
		__set_errno_dos( rc );
		return( rc );
	    }
	}
#if defined(__OS2_286__)
    } else {
	buf = buf;
    }
#endif
    return( 0 );
}
