#ifndef __LANGENV_H__
#define __LANGENV_H__

// LangEnv.h -- establish common stuff for a target environment for
//              language-related processors
//
// This file uses __TGT_SYS to figure out the required target.
//
//      LangEnvD.h contains the possible values.
//
//      __TGT_SYS_X86 is the default
//
// 95/02/23 -- J.W.Welch        -- defined
// 95/03/07 -- Greg Bentz	-- modify xi,yi for AXP to ensure ordering

#include "LangEnvD.h"

#ifndef __TGT_SYS
    #define __TGT_SYS __TGT_SYS_X86
#endif

#if __TGT_SYS == __TGT_SYS_X86

    #define TS_SEG_CODE     "_TEXT"
    #define TS_SEG_CONST    "CONST"
    #define TS_SEG_CONST2   "CONST2"
    #define TS_SEG_DATA     "_DATA"
    #define TS_SEG_TIB      "TIB"
    #define TS_SEG_TI       "TI"
    #define TS_SEG_TIE      "TIE"
    #define TS_SEG_XIB      "XIB"
    #define TS_SEG_XI       "XI"
    #define TS_SEG_XIE      "XIE"
    #define TS_SEG_YIB      "YIB"
    #define TS_SEG_YI       "YI"
    #define TS_SEG_YIE      "YIE"
    #define TS_SEG_YC	    "YC"
    #define TS_SEG_BSS      "_BSS"
    #define TS_SEG_STACK    "STACK"
    #define TS_SEG_DEPENDS  "not used"
    #define TS_SEG_TLSB	    ".tls"
    #define TS_SEG_TLS	    ".tls$"
    #define TS_SEG_TLSE	    ".tls$ZZZ"
    #define TS_SEG_TLS_CLASS	    "TLS"

    #define TS_MAX_OBJNAME	256
    #define TS_DATA_MANGLE	"_*"
    #define TS_CODE_MANGLE	"*_"

#elif __TGT_SYS == __TGT_SYS_AXP_NT || __TGT_SYS == __TGT_SYS_PPC_NT

    #define TS_SEG_CODE     ".text"
    #define TS_SEG_CONST    ".const"
    #define TS_SEG_CONST2   ".const2"
    #define TS_SEG_DATA     ".data"
    #define TS_SEG_TIB      ".rtl$tib"
    #define TS_SEG_TI       ".rtl$tid"
    #define TS_SEG_TIE      ".rtl$tie"
    #define TS_SEG_XIB      ".rtl$xib"
    #define TS_SEG_XI       ".rtl$xid"
    #define TS_SEG_XIE      ".rtl$xie"
    #define TS_SEG_YIB      ".rtl$yib"
    #define TS_SEG_YI       ".rtl$yid"
    #define TS_SEG_YIE      ".rtl$yie"
    #define TS_SEG_YC	    ".rtl$yc"
    #define TS_SEG_BSS      ".bss"
    #define TS_SEG_STACK    ".stack"
    #define TS_SEG_DEPENDS  ".depend"
    #define TS_SEG_TLSB	    ".tls"
    #define TS_SEG_TLS	    ".tls$"
    #define TS_SEG_TLSE	    ".tls$ZZZ"
    #define TS_SEG_TLS_CLASS	    NULL

    #define TS_MAX_OBJNAME	1024
    #define TS_DATA_MANGLE	"*"
    #define TS_CODE_MANGLE	"*"

#else

    #error Invalid target system

#endif

#undef __TGT_SYS

#endif
