/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1994.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

DBGIO -- debug i/o (for use from debugger)

      -- redirects uses of stdout to a file

      -- between calls to DbgRedirectBeg, DbgRedirectEnd

      -- when logging not active, file contains stuff each time

      -- logging causes the output to be cumulative

      -- logging activated by DbgLogBeg or DbgLogFile, deactivated by
         DbgLogEnd

Modified        By              Reason
--------        --              ------
92/10/02        Greg Bentz      Initial version, shipped as "alpha"
                Jim Randall
                Anthony Scian
                Jim Welch
93/05/01        Greg Bentz      Version 9.5 shipped
                Anthony Scian
                Jim Welch
93/08/31        Greg Bentz      Version 9.5 patch level A
                Anthony Scian
                Jim Welch
94/02/04        Greg Bentz      Version 9.5 patch level B
                Anthony Scian
                Jim Welch
94/03/21        Greg Bentz      Version 10.0 LA Version
                Anthony Scian
                Jim Welch
94/05/31        Greg Bentz      Version 10.0 GA Version
                Anthony Scian
                Jim Welch
94/08/31        Greg Bentz      Version 10.0 patch level A
                Anthony Scian
                Jim Welch
*/
#include "stdio.h"
#include "stdlib.h"
#include "string.h"

#define MX_FSTK         10
#define default_file    "_CPPDBG_."

static FILE* fstk[ MX_FSTK ];   // suspended files
static unsigned index;          // top of files stack
static int logging;             // true ==> logging at level 0


static void reDirSwitch         // SWITCH TWO FILE AREAS
    ( void )
{
    FILE temp;                  // - temporary area
    FILE* fp;                   // - file
    
    fflush( stdout );
    temp = *stdout;
    fp = fstk[ index ];
    *stdout = *fp;
    *fp = temp;
}


static void reDirBeg            // START REDIRECTION FOR A FILE
    ( void )
{
    if( index >= MX_FSTK ) {
        puts( "DBGIO -- too many log files active" );
        fflush( stdout );
    } else {
        char fname[32];
        FILE* fp;
        strcpy( fname, default_file );
        itoa( index, &fname[ sizeof( default_file ) - 1 ], 10 );
        fp =  fopen( fname, "wt" );
        if( NULL == fp ) {
            puts( "DBGIO -- failure to open file" );
            puts( fname );
            fstk[ index ] = 0;
        } else {
            fstk[ index ] = fp;
            reDirSwitch();
        }
    }
    ++ index;
}


static void reDirEnd            // COMPLETE REDIRECTION FOR A FILE
    ( void )
{
    if( index == 0 ) {
        puts( "DBGIO -- too many files closed" );
    } else {
        -- index;
        if( index < MX_FSTK ) {
            FILE* fp = fstk[ index ];
            if( fp != 0 ) {
                reDirSwitch();
                fclose( fstk[ index ] );
            }
        }
    }
}


void DbgRedirectBeg             // START REDIRECTION
    ( void )
{
    if( index == 0 ) {
        reDirBeg();
        logging = 0;
    }
}


int DbgRedirectEnd              // COMPLETE REDIRECTION
    ( void )
{
    int retn;                   // - # of file to view
 
    retn = index - 1;
    if( index > 1 || logging ) {
        fflush( stdout );
    } else {
        reDirEnd();
    }
    return retn;
}


void DbgLogBeg                  // START LOGGING
    ( void )
{
    if( index == 0 ) {
        logging = 1;
    }
    reDirBeg();
}


int DbgLogEnd                   // END LOGGING
    ( void )
{
    if( index > 0 ) {
        reDirEnd();
    }
    return index;
}
