/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

    Complex class functions and operators

   The formula for some of these algorithms was found in:
     "Complex Variables and Applications, 2nd Edition",
     Ruel V. Churchill,
     McGraw-Hill, 1960
   (Look in Jim Welch's office)
   These formulae include:
     acos, asin, atan,
     cosh, sinh, tanh

    Modified    By              Reason
    ========    ==              ======
    92/01/09    Steve McDowell  Initial implementation.
    92/10/09	Greg Bentz	Cleanup.
    93/02/16	Greg Bentz	Removed some numerical instability.
    93/10/19    Raymond Tang    Split into separate files.
    95/05/19	Greg Bentz	dll linkage
*/


#include "variety.h"
#include <complex>
#include "cplx.h"

_WPRTLINK Complex acos( const Complex &z ) {
/*****************************************/
// Arccos of a complex number.
// From "Complex Variables and Applications" (pg. 62).
//    acos( z ) = -i log( z + sqrt( z*z - 1 ) )
//              = -i log( z + sqrt( (z + 1) * (z - 1) ) )
//  I'm not sure why, but both Borland C++ and Maple seem to give
//  answers only in the 1st and 4th quadrants, so any answer in the
//  2nd or 3rd quadrant is changed to the 4th or 1st quadrant by
//  negating both parts of the complex number, since
//    cos x == cos -x
    dcomplex value;

    value = _IF_C16Mul( z.real() + 1.0, z.imag(), z.real() - 1.0, z.imag() );
    value = _IF_CDSQRT( value.realpart, value.imagpart );
    value = _IF_CDLOG( value.realpart + z.real(), value.imagpart + z.imag() );
    Complex c_result( value.imagpart, -value.realpart );

// Any negative real part means 2nd or 3rd quadrant.
// If the real part is zero and the imaginary part is negative, the
// vector is between the 3rd and 4th quadrants (angle -pi/2), so convert
// to the 1st quadrant (angle pi/2)

    if( c_result.real() < 0.0
            ||  (c_result.real() == 0.0  &&  c_result.imag() < 0.0) ) {
	c_result = -c_result;
    }
    return( c_result );
}
