/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

    Complex class functions and operators

   The formula for some of these algorithms was found in:
     "Complex Variables and Applications, 2nd Edition",
     Ruel V. Churchill,
     McGraw-Hill, 1960
   (Look in Jim Welch's office)
   These formulae include:
     acos, asin, atan,
     cosh, sinh, tanh

    Modified    By              Reason
    ========    ==              ======
    92/01/09    Steve McDowell  Initial implementation.
    92/10/09	Greg Bentz	Cleanup.
    93/02/16	Greg Bentz	Removed some numerical instability.
    93/10/19    Raymond Tang    Split into separate files.
    95/05/19	Greg Bentz	dll linkage
*/

#include "variety.h"
#include <complex>
#include <math.h>

#define PI     3.141592653589793

_WPRTLINK double arg( const Complex &z ) {
/***************************************/
// The angle of the complex vector z.
// In the 1st and 4th quadrants:
//      angle( z ) = asin( z.i/sqrt(z.r*z.r + z.i*z.i) )
// In the 2nd quadrant:
//      angle( z ) = pi - (above (positive value))
// In the 3rd quadrant:
//      angle( z ) = -pi - (above (negative value))
// (Derived from basic principles.)
// The angle of the (0, 0) is 0.
    double mag_squared;
    double angle;

    mag_squared = z.real()*z.real() + z.imag()*z.imag();
    if( mag_squared == 0.0 ) {
        return( 0.0 );
    }
    angle = asin( z.imag() / sqrt( mag_squared ) );
    if( z.real() < 0.0 ) {                    // 2nd or 3rd quadrant?
        angle = -angle;
        if( z.imag() >= 0 ) {                   //   2nd!
            angle += PI;
        } else {                           //   3rd!
            angle -= PI;
        }
    }
    return( angle );
}
