/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

    Complex class functions and operators

   The formula for some of these algorithms was found in:
     "Complex Variables and Applications, 2nd Edition",
     Ruel V. Churchill,
     McGraw-Hill, 1960
   (Look in Jim Welch's office)
   These formulae include:
     acos, asin, atan,
     cosh, sinh, tanh

    Modified    By              Reason
    ========    ==              ======
    92/01/09    Steve McDowell  Initial implementation.
    92/10/09	Greg Bentz	Cleanup.
    93/02/16	Greg Bentz	Removed some numerical instability.
    93/10/19    Raymond Tang    Split into separate files.
    95/05/19	Greg Bentz	dll linkage
*/

#include "variety.h"
#include <complex>
#include <math.h>

_WPRTLINK Complex cosh( const Complex &z ) {
/*****************************************/
// Hyperbolic cosine of a complex number.
// From "Complex Variables and Applications" (pg. 53 and 54).
//    cosh( z ) = (exp( z ) + exp( -z ))/2
//    realpart  = cosh( z.r ) * cos( z.i )
//    imagpart  = sinh( z.r ) * sin( z.i )
    return Complex( cosh( z.real() ) * cos( z.imag() ),
                    sinh( z.real() ) * sin( z.imag() ) );
}
