/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

    Complex class functions and operators

   The formula for some of these algorithms was found in:
     "Complex Variables and Applications, 2nd Edition",
     Ruel V. Churchill,
     McGraw-Hill, 1960
   (Look in Jim Welch's office)
   These formulae include:
     acos, asin, atan,
     cosh, sinh, tanh

    Modified    By              Reason
    ========    ==              ======
    92/01/09    Steve McDowell  Initial implementation.
    92/10/09	Greg Bentz	Cleanup.
    93/02/16	Greg Bentz	Removed some numerical instability.
    93/10/19    Raymond Tang    Split into separate files.
    95/05/19	Greg Bentz	dll linkage
    95/08/10	T. Schiller	Changed to use Geno's C16PowI() mathlib
    				support.
*/

#include "variety.h"
#include <complex>
#include "cplx.h"

_WPRTLINK Complex pow( const Complex &base, int power ) {
/******************************************************/
    dcomplex result;
    result = _IF_C16PowI( base.real(), base.imag(), power );
    return Complex( result.realpart, result.imagpart );
}

#ifdef blah  // the way it was ...
    Complex     cresult;
    int         abs_power;
    Complex     base_2nd;       // "base" squared
    Complex     base_4th;       // "base" to the 4th power

    abs_power = abs( power );
    if( abs_power >= 2 ) {
        base_2nd = base * base;
        if( abs_power >= 4 ) {
            base_4th = base_2nd * base_2nd;
        }
    }
    switch( abs_power ) {
      case 0:
        return Complex( 1, 0 );
      case 1:
        cresult = base;
        break;
      case 2:
        cresult = base_2nd;
        break;
      case 3:
        cresult = base_2nd * base;
        break;
      case 4:
        cresult = base_4th;
        break;
      case 5:
        cresult = base_4th * base;
        break;
      case 6:
        cresult = base_4th * base_2nd;
        break;
      case 7:
        cresult = base_4th * base_2nd * base;
        break;
      case 8:
        cresult = base_4th * base_4th;
        break;
      default:
        return( pow( base, (double) power ) );
    }
    if( power < 0 ) {
        cresult = 1.0 / cresult;
    }
    return( cresult );
#endif
