/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

    Complex class functions and operators

   The formula for some of these algorithms was found in:
     "Complex Variables and Applications, 2nd Edition",
     Ruel V. Churchill,
     McGraw-Hill, 1960
   (Look in Jim Welch's office)
   These formulae include:
     acos, asin, atan,
     cosh, sinh, tanh

    Modified    By              Reason
    ========    ==              ======
    92/01/09    Steve McDowell  Initial implementation.
    92/10/09	Greg Bentz	Cleanup.
    93/02/16	Greg Bentz	Removed some numerical instability.
    93/10/19    Raymond Tang    Split into separate files.
    95/05/19	Greg Bentz	dll linkage
*/

#include "variety.h"
#include <complex>
#include "cplx.h"

_WPRTLINK Complex tan( const Complex &z ) {
/****************************************/
// Tangent of a complex number.
// From "Complex Variables and Applications" (pg. 51).
//    tan( z ) = sin( z ) / cos( z )
//             = ((exp(iz) - exp(-iz))/2i)/((exp(iz) + exp(-iz))/2)
//             = (exp(iz) - exp(-iz))/(i*(exp(iz) + exp(-iz)))
#if 0
    dcomplex result;
    dcomplex exp_iz;
    dcomplex exp_neg_iz;

    exp_iz     = CDEXP( z.imag(), -z.real() );
    exp_neg_iz = C16Div( 1.0, 0.0, exp_iz.realpart, exp_iz.imagpart );
    result     = C16Div( exp_iz.realpart - exp_neg_iz.realpart,
                         exp_iz.imagpart - exp_neg_iz.imagpart,
                         exp_iz.imagpart + exp_neg_iz.imagpart,
                       -(exp_iz.realpart + exp_neg_iz.realpart) );
    Complex c_result( result.realpart, result.imagpart );
#elif 0
    double   sin_r;
    double   cos_r;
    double   exp_i;
    double   sin_r_by_exp_i;
    double   cos_r_by_exp_i;
    double   sin_r_times_exp_i;
    double   cos_r_times_exp_i;
    dcomplex result;

    sin_r = sin( z.real() );
    cos_r = cos( z.real() );
    exp_i = exp( z.imag() );
    sin_r_by_exp_i    = sin_r/exp_i;
    cos_r_by_exp_i    = cos_r/exp_i;
    sin_r_times_exp_i = sin_r*exp_i;
    cos_r_times_exp_i = cos_r*exp_i;
    result = C16Div( cos_r_by_exp_i - cos_r_times_exp_i,
                     sin_r_by_exp_i + sin_r_times_exp_i,
                    -sin_r_by_exp_i + sin_r_times_exp_i,
                     cos_r_by_exp_i + cos_r_times_exp_i );
    Complex c_result( result.realpart, result.imagpart );
#endif

// From Maple:
//    tan( a+ib ):
//        real part = sin(2a)  / (cos(2a) + cosh(2b))
//        imag part = sinh(2a) / (cos(2a) + cosh(2b))

    double  two_r;
    double  two_i;
    double  denom;


    two_r = z.real() * 2.0;
    two_i = z.imag() * 2.0;
    denom = cos( two_r ) + cosh( two_i );
    return Complex( sin( two_r ) / denom, sinh( two_i ) / denom );
}
