// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %     Copyright (C) 1992, by WATCOM International Inc.  All rights    %
// %     reserved.  No part of this software may be reproduced or        %
// %     used in any form or by any means - graphic, electronic or       %
// %     mechanical, including photocopying, recording, taping or        %
// %     information storage and retrieval systems - except with the     %
// %     written permission of WATCOM International Inc.                 %
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//
//  Modified    By              Reason
//  ========    ==              ======
//  92/02/19    Steve McDowell  Initial implementation.
//  92/02/28    ...             Modified to delay allocation of buffers
//                              until overflow/underflow called.
//  92/09/08	Greg Bentz	Cleanup.
//  93/03/22	Greg Bentz	modify filebuf::open() and filebuf::attach()
//				to assume ios::binary unless ios::text is
//				found in the fstat() so that QNX is supported.
//  93/07/22	Greg Bentz	Make sure overflow() function sets up the
//				put area
//  93/09/13	Greg Bentz	filebuf::~filebuf() must close if !__attached
//  93/10/15	Greg Bentz	let __plusplus_open() determine if default
//				file mode is TEXT or BINARY
//  93/10/15	Raymond Tang	Modify filebuf::open() to return NULL if both
//				ios::noreplace and ios::nocreate are specified
//  93/10/22	Raymond Tang	Split into separate files.
//  94/04/06	Greg Bentz	combine header files

#ifdef __SW_FH
#include "iost.h"
#else
#include "variety.h"
#include <fstream.h>
#endif
#include "ioutil.h"


filebuf *filebuf::attach( filedesc fd ) {
/***************************************/
// Attach an open file (described by fd) to this filebuf.
// If the filebuf is already attached, return NULL, otherwise return
// a pointer to the filebuf.
// Note that we must extract the "mode" information via the file's handle.

    __lock_it( __b_lock );
    if( __file_handle != EOF ) {
        return( NULL );
    }

    // Derive the value of "mode" from the file handle:
    if( ::__plusplus_fstat( fd, &__file_mode ) != 0 ) {
        return( NULL );
    }
    __file_handle = fd;
    __attached    = 1;
    return( this );
}
