// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %     Copyright (C) 1992, by WATCOM International Inc.  All rights    %
// %     reserved.  No part of this software may be reproduced or        %
// %     used in any form or by any means - graphic, electronic or       %
// %     mechanical, including photocopying, recording, taping or        %
// %     information storage and retrieval systems - except with the     %
// %     written permission of WATCOM International Inc.                 %
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//
//  Modified    By              Reason
//  ========    ==              ======
//  92/02/19    Steve McDowell  Initial implementation.
//  92/02/28    ...             Modified to delay allocation of buffers
//                              until overflow/underflow called.
//  92/09/08	Greg Bentz	Cleanup.
//  93/03/22	Greg Bentz	modify filebuf::open() and filebuf::attach()
//				to assume ios::binary unless ios::text is
//				found in the fstat() so that QNX is supported.
//  93/07/22	Greg Bentz	Make sure overflow() function sets up the
//				put area
//  93/09/13	Greg Bentz	filebuf::~filebuf() must close if !__attached
//  93/10/15	Greg Bentz	let __plusplus_open() determine if default
//				file mode is TEXT or BINARY
//  93/10/15	Raymond Tang	Modify filebuf::open() to return NULL if both
//				ios::noreplace and ios::nocreate are specified
//  93/10/22	Raymond Tang	Split into separate files.
//  94/04/06	Greg Bentz	combine header files

#ifdef __SW_FH
#include "iost.h"
#else
#include "variety.h"
#include <fstream.h>
#endif
#include "ioutil.h"

streambuf *filebuf::setbuf( char *buf, int len ) {
/************************************************/
// Set up the filebuf using the specified buffer.
// The buffer can be used if:
//   (1) the filebuf is not attached to a file (may or may not have a buffer
//       already),
//   (2) the file is attached to a file and does not have a buffer.
// If the buffer is too small (<= DEFAULT_PUTBACK_SIZE), then it cannot be
// used.

    __lock_it( __b_lock );
    if( (fd() != EOF) && (base() != NULL) ) {
        return( NULL );
    }
    if( (buf == NULL) || (len <= 0) ) {
        setb( NULL, NULL, FALSE );
    } else {
        if( len <= DEFAULT_PUTBACK_SIZE ) {
            setb( NULL, NULL, FALSE );
            return( NULL );
        }
        setb( buf, buf + len, FALSE );
    }
    return( this );
}
