// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %     Copyright (C) 1992, by WATCOM International Inc.  All rights    %
// %     reserved.  No part of this software may be reproduced or        %
// %     used in any form or by any means - graphic, electronic or       %
// %     mechanical, including photocopying, recording, taping or        %
// %     information storage and retrieval systems - except with the     %
// %     written permission of WATCOM International Inc.                 %
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//
//  Modified    By              Reason
//  ========    ==              ======
//  92/02/04    Steve McDowell  Initial implementation.
//  92/09/08	Greg Bentz	Cleanup.
//  93/03/15	Greg Bentz	fix uninitialized state variables.
//  93/07/29	Greg Bentz	- change istream::op>>(streambuf &) to
//				  istream::op>>( streambuf * )
//				- fix istream::getline() to not set failbit
//				  if no input stored in user buffer
//  93/09/15	Greg Bentz	change getline() back to set ios::failbit
//				if not input stored in user buffer
//  93/10/21	Greg Bentz	change get() and getline() to not set failbit
//				if the delim character has been seen
//  93/10/28	Raymond Tang	Split into separate files.
//  93/11/08	Raymond Tang    Add the conditions to check the validity of
//				a number.
//  93/11/16	Raymond Tang    Check if a hexadecimal is right after a sign.
//  94/04/06	Greg Bentz	combine header files

#ifdef __SW_FH
#include "iost.h"
#else
#include "variety.h"
#include <ctype.h>
#include <iostream.h>
#include <streambu.h>
#endif
#include "ioutil.h"
#include "isthdr.h"

ios::iostate __getunsignedlong( streambuf *sb, unsigned long &value, 
/******************************************************************/
    unsigned long maxval, signed long minval, ios::fmtflags format )
{

    unsigned long  number;
    ios::iostate   state;
    char           sign;
    int		   base;
    int		   offset;
    int		   ch;

    state = ios::goodbit;
    offset = 0;
    format &= ios::basefield;
    if( format ) {
	base = __FlagsToBase( format );
    } else {
	base = 0;
    }
    state = __getsign( sb, sign );
    if( !state && sign ) {
	ch = sb->sbumpc();
	int low_char = sb->speekc();
	if( (base == 10 || base == 0) && isdigit( low_char ) ) {
	    offset++;
	} else if( base == 16 && isxdigit( low_char ) ) {
	    offset++;
	} else if( base == 8 && isdigit( low_char ) && low_char < '8' ) {
	    offset++;
	} else {
	    // this will catch (low_char == EOF) too
	    sb->sputbackc( (char)ch );
	    state |= ios::failbit;
	}
    }
    if( !state && !base ) {
	state = __getbase( sb, base, offset );
    }
    if( !state ) {
	state = __getnumber( sb, number, base, offset );
    }
    if( !state && offset > 0 ) {
	if( number <= maxval ) {
	    if( number == 0 || sign != '-' ) {
		value = number;
	    } else {
		if( minval == 0 || -number >= minval ) {
		    value = -number;
	        } else {
		    state |= ios::failbit;
		}
	    }
 	} else {
	    state |= ios::failbit;
	}
    }
    if( offset == 0 ) {
	state |= ios::failbit;
    }
    return( state );
}
