// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %     Copyright (C) 1992, by WATCOM International Inc.  All rights    %
// %     reserved.  No part of this software may be reproduced or        %
// %     used in any form or by any means - graphic, electronic or       %
// %     mechanical, including photocopying, recording, taping or        %
// %     information storage and retrieval systems - except with the     %
// %     written permission of WATCOM International Inc.                 %
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//
//  Modified    By              Reason
//  ========    ==              ======
//  92/02/04    Steve McDowell  Initial implementation.
//  92/09/08	Greg Bentz	Cleanup.
//  93/03/15	Greg Bentz	fix uninitialized state variables.
//  93/07/29	Greg Bentz	- change istream::op>>(streambuf &) to
//				  istream::op>>( streambuf * )
//				- fix istream::getline() to not set failbit
//				  if no input stored in user buffer
//  93/09/15	Greg Bentz	change getline() back to set ios::failbit
//				if not input stored in user buffer
//  93/10/21	Greg Bentz	change get() and getline() to not set failbit
//				if the delim character has been seen
//  93/10/28	Raymond Tang	Split into separate files.
//  94/04/06	Greg Bentz	combine header files

#ifdef __SW_FH
#include "iost.h"
#else
#include "variety.h"
#include <ctype.h>
#include <limits.h>
#include <iostream.h>
#include <streambu.h>
#endif
#include "ioutil.h"
#include "lock.h"
#include "isthdr.h"

// Used by getnumber.
// Multiplication by  8 is done using a left-shift of three bits.
// Multiplication by 16 is done using a left-shift of four bits.
// Multiplication by 10 is done using a left-shift of three bits plus
// a left-shift of one bit.
// This table is used to determine if a shift will overflow. The number of
// bits to shift is used to index into the table. The table entry is anded
// with the unsigned long number and if any bits are on, then the shift
// will overflow.
static unsigned long const overFlowMasks[] = {
    0x00000000,
    0x80000000,
    0xC0000000,
    0xE0000000,
    0xF0000000
};

ios::iostate __getnumber( streambuf *sb, unsigned long &number,
/******************************************************************/
    int base, int &offset ) {
// Extract digits from the stream.
// Stop when a non-digit is found, leaving the non-digit in the stream.
// As digits are read, convert to an "unsigned long".

    unsigned long  result;
    unsigned long  overflow;
    ios::iostate   state;
    int            ch;
    int            is_digit;
    int            digit_value;
    int            shift1;
    int            shift2;

    if( base == 8 ) {
	shift1 = 3;	// *8
	shift2 = 0;
    } else if( base == 16 ) {
	shift1 = 4;	// *16
	shift2 = 0;
    } else {
	shift1 = 3;	// *8
	shift2 = 1;	// *2
    }
    state    = ios::goodbit;
    result   = 0;
    overflow = 0;
    is_digit = TRUE;
    __lock_it( sb->__b_lock );
    while( is_digit ) {
	ch = sb->speekc();
        if( ch == EOF ) {
	    if( offset == 0 ) {
		state |= ios::eofbit;
	    }
	}
        digit_value = ch - '0';
        if( base == 8 ) {
            is_digit = (ch >= '0'  &&  ch <= '7');
        } else {
            is_digit = isdigit( ch );
            if( base == 16 && !is_digit ) {
		char low_char;
		low_char = (char)tolower( ch );
		if( low_char >= 'a'  &&  low_char <= 'f' ) {
		    digit_value = low_char - 'a' + 10;
		    is_digit    = TRUE;
		}
            }
        }
        if( is_digit ) {
	    unsigned long tmp_result = result;
            overflow  |= (result & overFlowMasks[shift1]);
            result   <<= shift1;
            if( shift2 > 0 ) {
                tmp_result <<= shift2;
                if( result > (ULONG_MAX - tmp_result) ) overflow = 1;
                result += tmp_result;
            }
            if( result > ULONG_MAX - digit_value ) overflow = 1;
            result += digit_value;
	    sb->sbumpc();
	    offset++;
        }
    }
    if( overflow != 0 ) {
	state |= ios::failbit;
    }
    number = result;
    return( state );
}
