/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1994.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

PDCTXAXP -- context support for DEC ALPHA AXP

Modified        By              Reason
--------        --              ------
92/10/02        Greg Bentz      Initial version, shipped as "alpha"
                Jim Randall
                Anthony Scian
                Jim Welch
93/05/01        Greg Bentz      Version 9.5 shipped
                Anthony Scian
                Jim Welch
93/08/31        Greg Bentz      Version 9.5 patch level A
                Anthony Scian
                Jim Welch
94/02/04        Greg Bentz      Version 9.5 patch level B
                Anthony Scian
                Jim Welch
94/03/21        Greg Bentz      Version 10.0 LA Version
                Anthony Scian
                Jim Welch
94/05/31        Greg Bentz      Version 10.0 GA Version
                Anthony Scian
                Jim Welch
94/08/31        Greg Bentz      Version 10.0 patch level A
                Anthony Scian
                Jim Welch

95/08/24        J.W.Welch       Defined
*/


//--------------------------------------------------------------------------
// The following routine receives control at the end of either an
// RtlUnwind or RtlUnwindRfp invocation. Its purpose is to restore
// the SP, FP registers and to return to the same location as did
// the corresponding call to CPPLIB( PdCtx ).
//
// 0 is returned so that the caller can distinguish between an initial
// invocation and a return from unwinding.
//
// Upon entry, the SP and FP registers correspond to those when
// the exception was raised, presumeably up the stack.  Consequently,
// there is an assumption that that intervening stack contents would
// not be modified while the routine is active.
//
// Input:
//
//      v0 - address of save area
//      sp - for code
//      fp - for code
//      ra - an old value
//
// Output:
//
//      v0 - zero (return value)
//      sp - same as for corresponding call to CPPLIB( PdCtx )
//      fp - same as for corresponding call to CPPLIB( PdCtx )
//      ra - same as for corresponding call to CPPLIB( PdCtx )
//
//--------------------------------------------------------------------------
            
_internal_restore:
    ldq     $ra,0x10($v0)       // restore RA
    ldq     $sp,0x00($v0)       // restore SP
    ldq     $fp,0x08($v0)       // restore FP
    clr     $v0                 // return 0
    ret                         // return

//--------------------------------------------------------------------------
//  CPPLIB( PdCtx ) is called to save the context to be restored
//  when RtlUnwind or RtlUnwindRfp are called.  It returns the address
//  of the restoration routine, to be used as the continuation address
//  on completion of these functions.
//        
//  Input:
//
//      a0 - address of _CONTEXT record
//      a1 - address of 3 quads to be used as a save area
//
//  Output:
//
//      v0 - address of _internal_restore
//
//--------------------------------------------------------------------------

    .globl  __wcpp_4_PdCtx_

__wcpp_4_PdCtx_:
    stq     $ra,0x10($a1)       // save RA address
    stq     $sp,0x00($a1)       // save SP address
    stq     $fp,0x08($a1)       // save FP address
    lda     $v0,_internal_restore // return value
    ret                         // return


 
//--------------------------------------------------------------------------
// Thunk for exception handler
//
// This thunk is used so we can identify our handlers from other ones.  We
// can't just use an address comparison, since our handler might be
// duplicated in a DLL.
//
// This is accomplished by placing "WAT++COM" in two words behind a
// branch to the handler code.
//--------------------------------------------------------------------------
    
    .globl  __wcpp_4_pd_handler_
    .extern __wcpp_4_pd_handler_rtn_


__wcpp_4_pd_handler_:
    br      __wcpp_4_pd_handler_rtn_
    .long   0x2B544157  // "WAT+"
    .long   0x4D4F432B  // "+COM"
