/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

CTORARST.CPP -- CTOR ARRAY_STORAGE

             -- set up function registration so that exception
                processing will delete the array

             -- the array is a new[]'ed array which is found in ARRAY_STORAGE

             -- the address of the array is RETURNED

Modified        By              Reason
--------        --              ------
93/12/21        J.W.Welch       Defined

95/03/21        J.W.Welch       Only have array storage when operator delete
                                    has size arg. and a destructor is involved

95/04/18        J.W.Welch       Misnamed ctor_array_storage_2s

95/05/24        J.W.Welch       Convert to real C++

95/05/26        J.W.Welch       Optimized
*/

#include "cpplib.h"
#include "rtexcept.h"
#include "rtdtor.h"

#ifdef RT_EXC_ENABLED

struct _CTAS;
typedef void (*pFunDealloc)( _CTAS* );

struct _CTAS                    // CTAS -- controls CTOR or ARRAY_STORAGE
    : public _RTDTOR
{
    pFunDelete op_del;          // - operator delete[] to be used
    void* memory;               // - addr[ memory to delete ]
    RT_TYPE_SIG sig;            // - type signature
    pFunDealloc dealloc;        // - deallocation function


    ~_CTAS();                   // - destructor
};


_CTAS::~_CTAS()                 // DESTRUCTOR
{
    if( alive() ) {
        (*dealloc)( this );
    }
}


// The following are fake DTORs which are called only during exception
// handling.
//
// Their purpose is to delete the memory acquired by new[] when the CTOR'ing
// of the memory causes an exception to be thrown which is caught past
// the new[] point.
//
// When CTOR'ing succeeds, the run-time support pops the registration
// without destructing.
//
static void ctas_delete_1(          // DELETION FUNCTION - 1 ARG.
    _CTAS* ctas )                   // - control data
{
    (*ctas->op_del.delete_1)( ctas->memory );
}


static void ctas_delete_2(          // DELETION FUNCTION - 2 ARGS.
    _CTAS* ctas )                   // - control data
{
    ARRAY_STORAGE* ar               // - array storage for array
        = (ARRAY_STORAGE*)ctas->memory;
    
    (*ctas->op_del.delete_2)( ar
                            , ( CPPLIB( ts_size )( ctas->sig ) *
                                ar->element_count ) +
                                sizeof( ar->element_count )
                            );
}


#endif


static void* ctor_array             // CTOR THE ARRAY
    ( void *array                    // - array memory
    , unsigned count                // - number of elements
    , RT_TYPE_SIG sig               // - type signature for array type
#ifdef RT_EXC_ENABLED
    , void (*op_del)()              // - operator delete[] to be used
    , pFunDealloc dealloc           // - deallocation function
    , void *mem                     // - allocated memory
#endif
    )
{
    void* (*rtn)( void*, unsigned, RT_TYPE_SIG ); // - ctor R/T routine
#ifdef RT_EXC_ENABLED
    _CTAS ctas;                     // - controls CTOR or ARRAY_STORAGE

    ctas.memory = mem;
    ctas.op_del.delete_1 = (pFunDelete1)op_del;
    ctas.dealloc = dealloc;
    ctas.sig = sig;
#else
    if( 0 != array ) {
#endif
        if( sig->hdr.type == THROBJ_CLASS_VIRT ) {
            rtn = &CPPLIB( ctor_varray );
        } else {
            rtn = &CPPLIB( ctor_array );
        }
        array = (*rtn)( array, count, sig );
#ifdef RT_EXC_ENABLED
        ctas.complete();
#else
    }
#endif
    return array;
}


#ifdef RT_EXC_ENABLED
static void* ctor_array_memory(     // CTOR ARRAY MEMORY
    void* array,                    // - array memory
    unsigned count,                 // - number of elements
    RT_TYPE_SIG sig,                // - type signature for array type
    void (*op_del)(),               // - operator delete[] to be used
    pFunDealloc dealloc )           // - deallocation function
{
    if( array != 0 ) {
        array = ctor_array( array
                          , count
                          , sig
                          , op_del
                          , dealloc
                          , array
                          );
    }
    return array;
}
#endif


static void* ctor_array_storage     // CTOR ARRAY_STORAGE
    ( ARRAY_STORAGE *base           // - array storage
    , unsigned count                // - number of elements
    , RT_TYPE_SIG sig               // - type signature for array type
#ifdef RT_EXC_ENABLED
    , void (*op_del)()              // - operator delete[] to be used
    , pFunDealloc dealloc           // - deallocation function
#endif
    )
{
    void* array;                    // - array

    if( base == 0 ) {
        array = 0;
    } else {
        base->element_count = count;
        array = ctor_array( base->apparent_address
                          , count
                          , sig
#ifdef RT_EXC_ENABLED
                          , op_del
                          , dealloc
                          , base
#endif
                          );
    }
    return array;
}


extern "C"
_WPRTLINK
void* CPPLIB( ctor_array_storage_1m )( // CTOR ARRAY MEMORY, SUPPLIED DELETE[]
    void* array,                    // - array memory
    unsigned count,                 // - number of elements
    RT_TYPE_SIG sig,                // - type signature for array type
    void (*op_del)()                // - operator delete[] to be used
    )
{
#ifdef RT_EXC_ENABLED
    return ctor_array_memory( array
                            , count
                            , sig
                            , op_del
                            , &ctas_delete_1
                            );
#else
    op_del = op_del;
    return ctor_array( array, count, sig );
#endif
}


extern "C"
_WPRTLINK
void* CPPLIB( ctor_array_storage_1s )( // CTOR ARRAY_STORAGE, SUPPLIED DELETE[]
    ARRAY_STORAGE *base,            // - array storage
    unsigned count,                 // - number of elements
    RT_TYPE_SIG sig,                // - type signature for array type
    void (*op_del)() )              // - operator delete[] to be used
{
#ifndef RT_EXC_ENABLED
    op_del = op_del;
#endif
    return ctor_array_storage( base
                             , count
                             , sig
#ifdef RT_EXC_ENABLED
                             , op_del
                             , &ctas_delete_1
#endif
                             );
}


extern "C"
_WPRTLINK
void* CPPLIB( ctor_array_storage_2s )( // CTOR ARRAY_STORAGE, SUPPLIED DELETE[]
    ARRAY_STORAGE *base,            // - array storage
    unsigned count,                 // - number of elements
    RT_TYPE_SIG sig,                // - type signature for array type
    void (*op_del)() )              // - operator delete[] to be used
{
#ifndef RT_EXC_ENABLED
    op_del = op_del;
#endif
    return ctor_array_storage( base
                             , count
                             , sig
#ifdef RT_EXC_ENABLED
                             , op_del
                             , &ctas_delete_2
#endif
                             );
}


extern "C"
_WPRTLINK
void* CPPLIB( ctor_array_storage_gm )( // CTOR ARRAY MEMORY, default
    void* array,                    // - array memory
    unsigned count,                 // - number of elements
    RT_TYPE_SIG sig )               // - type signature for array type
{
#ifdef RT_EXC_ENABLED
    auto void (*op_del)();          // - operator delete[] to be used

    op_del = ( __typeof(op_del) )&operator delete[];
    return ctor_array_memory( array
                            , count
                            , sig
                            , op_del
                            , &ctas_delete_1
                            );
#else
    return ctor_array( array, count, sig );
#endif
}


extern "C"
_WPRTLINK
void* CPPLIB( ctor_array_storage_gs )( // CTOR ARRAY_STORAGE, default
    ARRAY_STORAGE *base,            // - array storage
    unsigned count,                 // - number of elements
    RT_TYPE_SIG sig )               // - type signature for array type
{
#ifdef RT_EXC_ENABLED
    auto void (*op_del)();          // - operator delete[] to be used

    op_del = ( __typeof(op_del) )&operator delete[];
#endif
    return ctor_array_storage( base
                             , count
                             , sig
#ifdef RT_EXC_ENABLED
                             , op_del
                             , &ctas_delete_1
#endif
                             );
}
