/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1994.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

LJMPDTOR -- support for destruction on longjmp's

Modified        By              Reason
--------        --              ------
92/10/02        Greg Bentz      Initial version, shipped as "alpha"
                Jim Randall
                Anthony Scian
                Jim Welch
93/05/01        Greg Bentz      Version 9.5 shipped
                Anthony Scian
                Jim Welch
93/08/31        Greg Bentz      Version 9.5 patch level A
                Anthony Scian
                Jim Welch
94/02/04        Greg Bentz      Version 9.5 patch level B
                Anthony Scian
                Jim Welch
94/03/21        Greg Bentz      Version 10.0 LA Version
                Anthony Scian
                Jim Welch
94/05/31        Greg Bentz      Version 10.0 GA Version
                Anthony Scian
                Jim Welch
94/08/31        Greg Bentz      Version 10.0 patch level A
                Anthony Scian
                Jim Welch
*/

#include "cpplib.h"
#include "rtexcept.h"
extern "C" {
    #include "ljmphdl.h"
};


#ifndef SYSIND_REGISTRATION
    #error not system-independent registration
#endif


// The following provides protection against longjmp's.  The routine is
// called at the start of a longjmp to fix up the run-time data structure
// in case there are stacked exceptions or function state tables.
//
static void
#ifdef M_PC_INTEL
__pragma( "__arg_convention" )
#endif
longjmpDtoring                  // longjmp INTERFACE
    ( void __far *stack_bound ) // - bound for stack
{
    _RTCTL rt_ctl;              // - R/T control
    DISPATCH_EXC dispatch;      // - dispatch control
    RW_DTREG* rw;               // - R/W block: search block
    RW_DTREG* last;             // - R/W block: last block
    auto FsExcRec excrec;       // - system exception record

    // locate the R/W block for the corresponding setjmp
    //  last == NULL: all must be popped (setjmp is before everything)
    //  last != NULL: pop all blocks before "last"
    //      - setjmp is located:
    //          (a) in routine without registration called by last rtn
    //          (b) in last C++ rtn (we compiled destruction code
    //              after the setjmp to ensure the state is always the
    //              state before the setjmp)
    //
    CPPLIB( exc_setup )( &dispatch, 0, FALSE, &rt_ctl, 0, &excrec );
    rw = rt_ctl.thr->registered;
    last = rw;
    for( ; ; rw = rw->base.prev ) {
        if( rw == NULL ) {
            if( NULL == last ) {
                // nothing registered
                return;
            } else {
                // everything must be popped
                last = rw;
                dispatch.state_var = 0;
                break;
            }
        }
        last = rw;
        if( rw >= stack_bound ) {
            // unwindStack will do nothing for last
            dispatch.state_var = last->base.state_var;
            break;
        }
    }
    dispatch.rw = last;
    FS_UNWIND_GLOBAL( dispatch.rw, NULL, &excrec );
}


// The handler can be in a DLL, so we need to extract its address via
// a function call.
//
extern "C"
_WPRTLINK
pFUNVOIDVOID CPPLIB( lj_handler ) // GET ADDRESS OF longjmp HANDLER
    ( void )
{
    return (pFUNVOIDVOID)&longjmpDtoring;
}
