/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

RTTISUPP -- RTTI support routines

Modified        By              Reason
--------        --              ------
95/12/14	A.F.Scian	defined
95/12/21	A.F.Scian	compressed the size of the RTTI structures
96/07/05	A.F.Scian	upgraded to use type_info instead of typesig
*/
#include <stddef.h>
#include <stdio.h>
#include <string.h>

#include "cpplib.h"
#include "rt_rtti.h"

extern "C"
void *__CalcMostDerived( rtti_exec *data ) {
    char *pc;
    void *vftable;
    unsigned from_here;
    rtti_adjust *adjust;

    vftable = *((void **) data->vfptr );
    adjust = ((rtti_adjust **) vftable )[-1];
    if( adjust == NULL ) {
	/* vftable has no RTTI descriptor */
	return( NULL );
    }
    pc = (char *) data->vfptr;
    pc -= adjust->delta;
    from_here = adjust->from_here;
    if( from_here & RA_CDISP ) {
	pc = pc - ((unsigned*)pc)[-1];
    }
    from_here &= ~RA_RUNTIME;
    data->md_class = (rtti_class *) ( from_here + (char*) adjust );
    data->md_addr = pc;
    return( pc );
}

extern "C"
rtti_leap const *__MakeSureTidIsPresent( rtti_exec *data, type_info const *find, rtti_leap *pleap ) {
    rtti_class const *rt_class;
    rtti_leap const *leap;

    rt_class = data->md_class;
    if( *find == *(rt_class->id) ) {
	pleap->vb_index = 0;
	pleap->offset = 0;
	pleap->control = RL_NULL;
	pleap->id = find;
	return( pleap );
    }
    if( rt_class->vbptr != -1 ) {
	leap = rt_class->leaps;
	for(;;) {
	    if( *find == *(leap->id) ) {
		return( leap );
	    }
	    if( leap->control & RL_LAST ) break;
	    ++leap;
	}
    }
    return( NULL );
}
