/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1993.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

STABACPV.C -- support to call copy constructor for each element in a newly
              allocated array
           -- elements in array have a virtual base(s)

Modified        By              Reason
--------        --              ------
92/10/02        Greg Bentz      Initial version, shipped as "alpha"
                Jim Randall
                Anthony Scian
                Jim Welch

92/12/02        J.W.Welch       Added exceptions

93/05/01        Greg Bentz      Version 9.5 shipped
                Anthony Scian
                Jim Welch

94/10/18        J.W.Welch       Adapt for fs registration

95/05/25        J.W.Welch       Use RTARCTOR
*/

#include "cpplib.h"
#include "rtexcept.h"
#include "rtarctor.h"


extern "C"
_WPRTLINK
void * CPPLIB(copy_varray)(     // CALL COPY CONSTRUCTORS FOR ARRAY
    void *tgt_array,            // - target array
    void *src_array,            // - source array
    unsigned count,             // - number of elements
    RT_TYPE_SIG sig )           // - signature of array
{
    pFUNcopyV ctor;             // - constructor for an element
    size_t size;                // - size of an array element
    _RTARCTOR ar_ctor           // - array-ctor control
        ( 0, sig, tgt_array );

    ctor = sig->clss_v.copyctor;
    size = sig->clss_v.size;
    while( count > 0 ) {
        (*ctor)( tgt_array, CTOR_NULL, src_array );
        ++ ar_ctor;
        tgt_array = (void *)( ((char*)tgt_array) + size );
        src_array = (void *)( ((char*)src_array) + size );
        -- count;
    }
    ar_ctor.complete();
    return ar_ctor._array;
}
