/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

STABADT -- destruct array

Modified        By              Reason
--------        --              ------
93/09/01        J.W.Welch       Defined

94/10/18        J.W.Welch       Adapt for fs registration

95/05/25        J.W.Welch       Use RTARCTOR

*/

#include "cpplib.h"
#include "rtexcept.h"
#include "rtarctor.h"


void * dtor_the_array(          // CALL DESTRUCTORS FOR ARRAY ELEMENTS
    _RTARCTOR* cta )            // - array-control
{
    void *array;                // - array
    RT_TYPE_SIG sig;            // - signature of array
    pFUNdtor dtor;              // - destructor for an element
    size_t size;                // - size of an array element

    sig = cta->_sig;
    size = sig->clss.size;
    array = ( (char*)cta->_array + size * cta->_count );
    dtor = sig->clss.dtor;
    while( cta->_count > 0 ) {
        -- *cta;
        array = ( (char*)array - size );
        (*dtor)( array, 0 );
    }
    cta->complete();
    return cta->_array;
}


#ifdef RT_EXC_ENABLED
_RTARCTOR::~_RTARCTOR           // DESTRUCTOR FOR R/T AR-CTOR
    ( void )
{
    if( alive() ) {
        dtor_the_array( this );
    }
}
#endif

#ifndef NDEBUG
_RTARCTOR::_RTARCTOR            // COPY CTOR FOR R/T AR-CTOR
    ( _RTARCTOR const & )
    : _count( 0 )
    , _sig( 0 )
    , _array( 0 )
{
    GOOF_EXC( "Attempt to copy RTARCTOR" );
}
#endif


extern "C"
_WPRTLINK
void * CPPLIB(dtor_array)(      // CALL DESTRUCTORS FOR ARRAY ELEMENTS
    void *array,                // - array
    unsigned count,             // - number of elements
    RT_TYPE_SIG sig )           // - signature of array
{
    _RTARCTOR ar_ctor           // - array-ctor control
        ( count, sig, array );
        
    dtor_the_array( &ar_ctor );
    return array;
}
