/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1994.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

STABINIT -- initialization of dispatch and exception records

Modified        By              Reason
--------        --              ------
92/10/02        Greg Bentz      Initial version, shipped as "alpha"
                Jim Randall
                Anthony Scian
                Jim Welch
93/05/01        Greg Bentz      Version 9.5 shipped
                Anthony Scian
                Jim Welch
93/08/31        Greg Bentz      Version 9.5 patch level A
                Anthony Scian
                Jim Welch
94/02/04        Greg Bentz      Version 9.5 patch level B
                Anthony Scian
                Jim Welch
94/03/21        Greg Bentz      Version 10.0 LA Version
                Anthony Scian
                Jim Welch
94/05/31        Greg Bentz      Version 10.0 GA Version
                Anthony Scian
                Jim Welch
94/08/31        Greg Bentz      Version 10.0 patch level A
                Anthony Scian
                Jim Welch

95/10/03        J.W.Welch       Defined

*/

#if defined( __USE_FS ) || defined( __USE_PD )
#define __NEED_SYSTEM_HEADER
#endif

#include "cpplib.h"
#include "rtexcept.h"


inline
static void initFsExcRec(       // INITIALIZE FsExcRec
    void *object,               // - address of object
    FsExcRec* rec,              // - the record
    DISPATCH_EXC* dispatch )    // - dispatch info
{
    rec->code = EXCREC_CODE;
    rec->flags = EXCREC_FLAGS;
    rec->rec = 0;
    rec->addr = 0;
    rec->parm_count = EXCREC_PARM_COUNT;
    rec->object = object;
    rec->dispatch = dispatch;
}


inline
static void dispatch_init(      // INITIALIZE DISPATCH BLOCK
    DISPATCH_EXC *dispatch,     // - dispatch control
    THROW_RO *throw_ro,         // - thrown R/O block
    rboolean is_zero,           // - TRUE ==> thrown object is zero constant
    _RTCTL* rtc )               // - R/T control
{
    dispatch->ro = throw_ro;
    dispatch->rtc = rtc;
    dispatch->zero = is_zero;
    ThreadLookup( &dispatch->fs_last );
    dispatch->try_cmd = NULL;
    dispatch->exc = rtc->thr->excepts;
    if( 0 == throw_ro ) {
        dispatch->rethrow = TRUE;
    } else {
        dispatch->rethrow = FALSE;
    }
    dispatch->popped = FALSE;
    dispatch->non_watcom = FALSE;
    dispatch->fnexc_skip = NULL;
    dispatch->srch_ctl = rtc->thr->exc_pr;
}


void CPPLIB( exc_setup )        // SETUP DISPATCH, EXCEPTION RECORDS
    ( DISPATCH_EXC* disp        // - dispatch record
    , THROW_RO* thr_ro          // - throw r/o block
    , rboolean is_zero          // - TRUE ==> thrown object is zero constant
    , _RTCTL* rt_ctl            // - R/T control
    , void* object              // - thrown object
    , FsExcRec* rec )           // - exception record
{
    dispatch_init( disp, thr_ro, is_zero, rt_ctl );
    initFsExcRec( object, rec, disp );
}
