/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

STABMOD -- DTOR program statics (state-variable implementation)

Modified        By              Reason
--------        --              ------
93/09/01        J.W.Welch       Defined 

94/10/20        J.W.Welch       Add fs dtoring
95/05/17        J.W.Welch       System-independent re-implementation
95/09/22        J.W.Welch       Re-do method of destruction

00000 - increment to force compile
*/
#include "cpplib.h"
#include "rtexcept.h"
#include "rtinit.h"


extern "C"
void CPPLIB( module_dtor )      // DTOR STATICS FOR PROGRAM
    ( void )
{
    THREAD_CTL *thr;            // - thread control ptr.
    RW_DTREG* rw;               // - read/write entry
    union {
        RW_DTREG inlined;       // - inlined copy for destruction
        RW_DTREG_INITLS ls;     // - inlined copy for destruction
    };
        

#ifdef RW_REGISTRATION
    inlined.base.handler = & CPPLIB( fs_handler_rtn );
    thr = 0;
#else
    thr = &_RWD_ThreadData;
#endif
    for( ; ; ) {
#ifdef __MT__
        _RWD_StaticInitSema.p();
#endif
        rw = _RWD_ModuleInit;
        if( rw != 0 ) {
            _RWD_ModuleInit = rw->base_st.prev;
        }
#ifdef __MT__
        _RWD_StaticInitSema.v();
#endif
        if( rw == 0) break;
        RO_STATE* state = &rw->base_st.ro->init_ls.state_table[0];
        (*state->dtor)( state->u.data_addr, DTOR_NULL );
    }
}


#if 0


// *************************************************************************
// Support for state table and destruction of file-scope objects
// *************************************************************************


static RO_STATE* fs_state_table( // POINT AT STATE TABLE
    RO_DTREG* ro )              // - R/O entry
{
    return ro->init_fs.state_table;
}


static void* fs_point_object(   // POINT AT OBJECT FOR ENTRY
    RW_DTREG* rw,               // - R/W entry
    RT_STATE_VAR state_var )    // - state variable
{
    return (void*)( rw->base.ro
                      ->init_fs.state_table[ state_var - 1 ].u.data_addr );
}


static void fs_init_traverse(   // INITIALIZE TRAVERSAL DATA STRUCTURE
    STAB_TRAVERSE* traverse,    // - traversal information
    RW_DTREG* )                 // - read/write data
{
    traverse->obj_type = OBT_OBJECT;
}


static DTREG_VFTBL fs_vftable =// virtual functions for DTRG_FUN
{   &fs_state_table
,   &fs_point_object
,   &fs_init_traverse
};



// *************************************************************************
// Registration and Initialization
// *************************************************************************

static void init(               // PROGRAM INITIALIZATION
    void )
{
    CPPDATA( dtreg_vfptrs )[ DTRG_STATIC_INITFS ] = &fs_vftable;
}

XI( CPPDATA( stab_init_mod ), init, INIT_PRIORITY_RUNTIME )

#endif

extern "C"
void CPPLIB( mod_register )(    // REGISTRATION FOR MODULE INITIALIZED OBJECTS
    RW_DTREG* rw )              // - R/W block
{
#ifdef __MT__
    _RWD_StaticInitSema.p();
#endif
    rw->base_st.prev = _RWD_ModuleInit;
    _RWD_ModuleInit = rw;
#ifdef __MT__
    _RWD_StaticInitSema.v();
#endif
}
