/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

STABTHRW -- support for throw

Modified        By              Reason
--------        --              ------
92/11/10        J.W.Welch       Defined
93/01/29        J.W.Welch       Set bound in dispatch block before unwinding
                                    stack
93/01/29        J.W.Welch       Adjust for unwind_stack virtual call returning
                                    object addr. to indicate DTOR required
93/02/04        J.W.Welch       Add CPPLIB( corrupted_stk ).
93/02/10        J.W.Welch       Put support routines in their own files.
93/07/29        A.F.Scian       added aborts pragma using __pragma to remove
                                kludgey previous method that broke when prefix
                                changed
93/09/01        J.W.Welch       Adapt for 10.0
93/12/14        J.W.Welch       Add longjmp support
94/01/28        J.W.Welch       Remove aborts from run-time
94/06/09        J.W.Welch       Use "terminated" flag in THREAD_CTL
94/08/16        J.W.Welch       Support CTOR-TEST
94/10/11        J.W.Welch       Support for throwing function ptr.
94/10/14        J.W.Welch       use PgmThread, implement fs registration
94/10/31        J.W.Welch       Moved __compiled_under_... to CPPDATA.CPP
                                ACTIVE_EXC changed
94/11/22        J.W.Welch       Use throw-conversion signature of thrower on
                                    a catch (the catcher signature is not
                                    always filled in when DLLs are in play)
94/12/06        J.W.Welch       Got rid of ctored-obj flags
95/02/08        J.W.Welch       Was not handling longjmp when everthing was
                                    to be popped
95/02/20        A.F.Scian       fixed longjmpDtoring so that it pops the RW block
                                just under a C++ function that does a setjmp
95/02/22        A.F.Scian       fixed unwindStack so that it saves the state
                                of the current exception while it unwinds the
                                stack in case there is nothing to unwind and
                                the caller was just making sure
95/03/20        J.W.Welch       Pop exception at catch end only when it is
                                    owned by the try block in question
95/03/21        J.W.Welch       Support new type signatures
95/03/22        J.W.Welch       Save cmd when FNEXC dispatchable
95/03/23        J.W.Welch       Support new command definitions
95/03/25        J.W.Welch       Change naming convention
95/06/14        J.W.Welch       Always do lookup for fs_last in dispatch_init
95/06/14        Greg Bentz      dll linkage
95/07/18        J.W.Welch       Allow longjmp dtoring on alpha
95/09/04        J.W.Welch       Eliminated call_unexpected
95/09/13        J.W.Welch       Use find_active, dtor_free_exc
95/09/14        J.W.Welch       Use EXC_PR
95/09/26        J.W.Welch       catch_done was not resetting state
95/09/27        J.W.Welch       longjmp-handler initialization moved to
                                    LJMPINIT.CPP
95/09/29        J.W.Welch       Fix obsolete code with longjmp's
95/10/03        J.W.Welch       Removed longjmp support to LJMPDTOR
                                Removed initialization to STABINIT
95/10/10        J.W.Welch       Removed CPPLIB( dispatchable ) to STABDISP
95/12/05        J.W.Welch       Use SETJMPEX ON AXP

00001 - increment to force compile
*/

#if defined( __USE_FS ) || defined( __USE_PD )
#define __NEED_SYSTEM_HEADER
#endif

#include "cpplib.h"

#include <exception.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#ifdef __USE_PD
#   include <setjmpex.h>
#   pragma extref _setjmpex;
#else
#   include <setjmp.h>
#endif

#include "rtexcept.h"
#include "exc_pr.h"
#include "rtmsgs.h"


static void fneDispatch(        // DISPATCH "unexpected"
    DISPATCH_EXC *dispatch )    // - dispatch control
{
    THREAD_CTL *ctl;            // - thread-specific data
    ACTIVE_EXC *exc;            // - active exception
    ACTIVE_EXC *srch;           // - previous exception for this FNEXC

    ctl = dispatch->rtc->thr;
    ctl->abort_msg = RTMSG_FNEXC;
    exc = ctl->excepts;
    exc->dispatch = dispatch;
    for( srch = exc; ; ) {
        if( srch == NULL ) {
            // first time through fn_exc
            exc->dispatch = dispatch;
            exc->fnexc_state = exc->state;
            exc->state = EXCSTATE_UNEXPECTED;
            _EXC_PR_FNEXC marker( dispatch->rtc
                                , 0
                                , dispatch->rw
                                , dispatch->rethrow ? 0 : exc );
            unexpected();
            marker._state = EXCSTATE_TERMINATE;
            CPPLIB( call_terminate )( RTMSG_RET_UNEXPECT, ctl );
        }
        DISPATCH_EXC *cand = srch->dispatch;
        if( NULL != cand
         && dispatch->rw        == cand->rw
         && dispatch->state_var == cand->state_var ) {
            if( srch->state == EXCSTATE_UNEXPECTED ) {
                // throw/rethrow did not get through fn-exc
                exc->state = EXCSTATE_BAD_EXC;
                _EXC_PR_FNEXC marker( dispatch->rtc
                                    , 0
                                    , dispatch->rw
                                    , dispatch->rethrow ? 0 : exc );
                throw bad_exception();
            }
            if( srch->state == EXCSTATE_BAD_EXC ) {
                // throw of bad_exception did not get through fn-exc
                _EXC_PR_DTOR marker( dispatch->rtc
                                   , 0
                                   , EXCSTATE_TERMINATE
                                   , dispatch->rethrow ? 0 :exc );
                CPPLIB( call_terminate )( RTMSG_FNEXC, ctl );
            }
            if( srch != exc ) {
                CPPLIB( corrupted_stack )();
            }
        }
        srch = srch->prev;
        if( exc == srch ) {
            CPPLIB( corrupted_stack )();
        }
    }
}


#define throwCnvSig( d ) CPPLIB( ts_refed )( (d)->cnv_try->signature )


static void catchDispatch(      // DISPATCH A CATCH BLOCK
    DISPATCH_EXC *dispatch )    // - dispatch control
{
    RW_DTREG* blk;              // - function block for dispatch
    DTOR_CMD* cmd;              // - try command
    jmp_buf *env;               // - addr[ jmp_buf ]
    ACTIVE_EXC *active;         // - active exception
    void *tgt;                  // - target try variable
    void *src;                  // - source address
    RT_TYPE_SIG sig;            // - signature for catch
    unsigned cnv_offset;        // - conversion offset

    blk = dispatch->rw;
    cmd = dispatch->try_cmd;
    active = dispatch->exc;
    active->state = EXCSTATE_DISPATCH;
    env = (jmp_buf*)( (char*)blk + cmd->try_cmd.jmp_buf );
    sig = cmd->try_cmd.sigs[ dispatch->catch_no ];
    if( sig != NULL ) {
        cnv_offset = dispatch->cnv_try->offset;
        src = cnv_offset + (char*)active->data;
        tgt = (char*)blk + cmd->try_cmd.offset;
        switch( sig->hdr.type ) {
          case THROBJ_PTR_CLASS :
            if( dispatch->zero ) {
                *(void**)tgt = NULL;
            } else {
                *(void**)tgt = *(char**)active->data + cnv_offset;
            }
            break;
          case THROBJ_VOID_STAR :
          case THROBJ_PTR_SCALAR :
          case THROBJ_PTR_FUN :
            if( dispatch->zero ) {
                memset( tgt, 0, sig->scalar.size );
                break;
            }
          case THROBJ_SCALAR :
            memcpy( tgt, src, sig->scalar.size );
            break;
          case THROBJ_CLASS :
            sig = throwCnvSig( dispatch );
            (*sig->clss.copyctor)( tgt, src );
            break;
          case THROBJ_CLASS_VIRT :
            sig = throwCnvSig( dispatch );
            (*sig->clss_v.copyctor)( tgt, CTOR_NULL, src );
            break;
          case THROBJ_REFERENCE :
            sig = CPPLIB( ts_refed )( sig );
            switch( sig->hdr.type ) {
              case THROBJ_PTR_CLASS :
                if( dispatch->zero ) {
                    active->extra_object = NULL;
                } else {
                    active->extra_object = *(char**)active->data + cnv_offset;
                }
                *(void**)tgt = &active->extra_object;
                break;
              case THROBJ_PTR_SCALAR :
              case THROBJ_PTR_FUN :
              case THROBJ_VOID_STAR :
                if( dispatch->zero ) {
                    active->extra_object = NULL;
                    src = &active->extra_object;
                }
              case THROBJ_SCALAR :
              case THROBJ_CLASS :
              case THROBJ_CLASS_VIRT :
                *(void**)tgt = src;
                break;
            }
            break;
          default :
            GOOF_EXC( "unexpected exception type" );
        }
    }
    longjmp( *env, dispatch->catch_no + 1 );
}


static void processThrow(       // PROCESS A THROW
    void *object,               // - address of object
    THROW_RO *throw_ro,         // - thrown R/O block
    rboolean is_zero )          // - TRUE ==> thrown object is zero constant
{
    _RTCTL rt_ctl;              // - R/T control
    DISPATCH_EXC dispatch;      // - dispatch control
    auto FsExcRec excrec;       // - system exception record
    volatile rboolean unwound;
    void *force_this_routine_to_have_an_EBP_frame;

//  CPPLIB( DbgRtDumpAutoDtor )();

#if 0
//
// Check for throws under bad circumstances
//
    ACTIVE_EXC *exc = rt_ctl.thr->excepts;
    if( exc == NULL ) {
        if( rt_ctl.thr->flags.terminated ) {
            CPPLIB( fatal_runtime_error )( RTMSG_THROW_TERMIN, 1 );
        }
    } else {
        switch( exc->state ) {
          case EXCSTATE_DISPATCH :
          case EXCSTATE_UNEXPECTED :
          case EXCSTATE_BAD_EXC :
            break;
          case EXCSTATE_UNWIND :
          { CPPLIB( free_exc )( &rt_ctl, exc );
            _EXC_PR marker( &rt_ctl, 0, EXCSTATE_TERMINATE );
            CPPLIB( call_terminate )( RTMSG_THROW_DTOR, rt_ctl.thr );
          }
          case EXCSTATE_TERMINATE :
            CPPLIB( free_exc )( &rt_ctl, exc );
            CPPLIB( fatal_runtime_error )( RTMSG_THROW_TERMIN, 1 );
          case EXCSTATE_CTOR :
            CPPLIB( free_exc )( &rt_ctl, exc );
            CPPLIB( fatal_runtime_error )( RTMSG_THROW_CTOR, 1 );
          case EXCSTATE_DTOR :
            CPPLIB( free_exc )( &rt_ctl, exc );
            CPPLIB( fatal_runtime_error )( RTMSG_EXC_DTOR, 1 );
          default:
            GOOF_EXC( "getActiveExc: unexpected exception state" );
        }
    }
#endif
//
// Setup for dispatch
//
    rt_ctl.thr->abort_msg = NULL;
    CPPLIB( exc_setup )( &dispatch
                       , throw_ro
                       , is_zero
                       , &rt_ctl
                       , object
                       , &excrec );
    force_this_routine_to_have_an_EBP_frame = alloca(16);
    unwound = FALSE;
//
// raising exception locates the catch/fn-exception to be dispatched
// also fills in excrec.addr
//
    if( 0 == dispatch.fs_last ) {
        // no search when no R/W blocks
        dispatch.type = DISPATCHABLE_NO_CATCH;
    } else {
        FS_RAISE_EXCEPTION( &excrec );
    }
//
// comes here twice, unless an error situation:
//  unwound == 0 : after catch located, before unwinding
//  unwound == 1 : after unwinding
//
    switch( dispatch.type ) {
      case DISPATCHABLE_STOP :
        if( NULL == dispatch.srch_ctl ) {
            GOOF_EXC( "DISPATCHABLE_STOP: no srch_ctl" );
        }
        switch( dispatch.srch_ctl->_state ) {
          case EXCSTATE_UNWIND :
          { _EXC_PR_FREE marker( &rt_ctl
                               , 0
                               , EXCSTATE_TERMINATE
                               , dispatch.rethrow ? 0 : rt_ctl.thr->excepts );
            CPPLIB( call_terminate )( RTMSG_THROW_DTOR, rt_ctl.thr );
          }
          case EXCSTATE_TERMINATE :
            CPPLIB( fatal_runtime_error )( RTMSG_THROW_TERMIN, 1 );
          case EXCSTATE_CTOR :
            CPPLIB( fatal_runtime_error )( RTMSG_THROW_CTOR, 1 );
          case EXCSTATE_DTOR :
            CPPLIB( fatal_runtime_error )( RTMSG_EXC_DTOR, 1 );
          default:
            GOOF_EXC( "DISPATCHABLE_STOP: unexpected exception state" );
        }
      case DISPATCHABLE_FNEXC :
      case DISPATCHABLE_CATCH :
        if( ! unwound ) {
            ACTIVE_EXC *active; // - saved exception
            unwound = TRUE;
            if( dispatch.rethrow ) {
// do we still need fnexc_state ?
                dispatch.exc->fnexc_state = dispatch.exc->state;
                if( dispatch.popped ) {
                    active = dispatch.exc;
                    active->cat_try = dispatch.try_cmd;
                    active->rw = dispatch.rw;
                }
            } else {
                active = CPPLIB( alloc_exc )( object
                                            , dispatch.ro
                                            , &rt_ctl );
                dispatch.exc = active;
                active->cat_try = dispatch.try_cmd;
                active->rw = dispatch.rw;
            }
            if( dispatch.type == DISPATCHABLE_FNEXC ) {
                if( dispatch.rethrow ) {
                    dispatch.exc->state = dispatch.exc->fnexc_state;
                }
                break;
            }
            dispatch.exc->state = EXCSTATE_UNWIND;
#ifdef RW_REGISTRATION
            FS_UNWIND_GLOBAL( dispatch.rw, excrec.addr, &excrec );
#endif
#ifdef PD_REGISTRATION
            CPPLIB( PdUnwind )( &excrec );
#endif
        }
        // NT returns here instead of after RAISE_EXCEPTION if
        // there are no blocks to pop
        // nb. need to execute same code as after RAISE_EXCEPTION
        CPPLIB( destruct_internal )( dispatch.state_var, dispatch.rw );
        break;
    }
//
// dispatch the exception
//
    switch( dispatch.type ) {
      case DISPATCHABLE_FNEXC :
        fneDispatch( &dispatch );
      case DISPATCHABLE_CATCH :
        catchDispatch( &dispatch );
      case DISPATCHABLE_NO_CATCH :
        if( dispatch.rethrow ) {
            CPPLIB( fatal_runtime_error )( RTMSG_RETHROW, 1 );
        } else {
          { _EXC_PR marker( &rt_ctl, 0, EXCSTATE_TERMINATE );
            CPPLIB( call_terminate )( RTMSG_NO_HANDLER, rt_ctl.thr );
          }
        }
#if 0 // not now
      case DISPATCHABLE_SYS_EXC :
      {
        char buffer[ sizeof( RTMSG_SYS_EXC ) ];
        ::memcpy( buffer, RTMSG_SYS_EXC, sizeof( buffer ) );
        ltoa( dispatch.system_exc, buffer + sizeof( buffer) - 9, 16 );
        CPPLIB( fatal_runtime_error )( buffer, 1 );
      }
#endif
      default :
        GOOF_EXC( "throw: invalid DISPATCHABLE" );
    }
}


extern "C"
_WPRTLINK
void CPPLIB( rethrow )(        // RE-THROW AN EXCEPTION
    void )
{
    processThrow( NULL, NULL, FALSE );
}


extern "C"
_WPRTLINK
void CPPLIB( catch_done )(      // COMPLETION OF CATCH
#ifdef RW_REGISTRATION
    void
#else
    RW_DTREG *rw                // - current R/W block
#endif
    )
{
    _RTCTL rt_ctl;              // - R/T control
#ifdef RW_REGISTRATION
    RW_DTREG* rw = RwTop( rt_ctl.thr );
#endif
    rt_ctl.setRwRo( rw );
    RO_STATE* state = CPPLIB( stab_entry )( rw->base.ro
                                          , rw->fun.base.state_var );
    DTOR_CMD* cmd = TryFromCatch( state->u.cmd_addr );
    ACTIVE_EXC* exc = CPPLIB( find_active )( &rt_ctl, rw, cmd );
    CPPLIB( dtor_free_exc )( exc, &rt_ctl );
    STAB_TRAVERSE traverse;
    CPPLIB( stab_trav_init )( &traverse, &rt_ctl );
    CPPLIB( stab_trav_next )( &traverse );
    rw->fun.base.state_var = traverse.state_var;
}


extern "C"
_WPRTLINK
void CPPLIB( throw )(           // THROW AN EXCEPTION OBJECT (NOT CONST ZERO)
    void *object,               // - address of object
    THROW_RO *throw_ro )        // - throw R/O block
{
    processThrow( object, throw_ro, FALSE );
}


extern "C"
_WPRTLINK
void CPPLIB( throw_zero )(      // THROW AN EXCEPTION OBJECT (CONST ZERO)
    void *object,               // - address of object
    THROW_RO *throw_ro )        // - throw R/O block
{
    processThrow( object, throw_ro, TRUE );
}
