/*
**********************************************************************
*       Copyright by WATCOM International Corporation, 1987, 1992.   *
*       All rights reserved. No part of this software may be         *
*       reproduced in any form or by any means - graphic, electronic,*
*       mechanical or otherwise, including, without limitation,      *
*       photocopying, recording, taping or information storage and   *
*       retrieval systems - except with the written permission of    *
*       WATCOM International Corporation.                            *
**********************************************************************

TSEQUIV -- test if equivalent type signatures

Modified        By              Reason
--------        --              ------
92/11/13        J.W.Welch       Defined
94/11/21        J.W.Welch       Upgrade for mangled name in type sig.s
95/03/21        J.W.Welch       Support new type signatures
95/05/17        J.W.Welch       System-independent re-implementation
95/12/13        J.W.Welch       When 0 thrown, was not checking scalars
*/


#include "cpplib.h"
#include <string.h>
#include "rtexcept.h"


RT_TYPE_SIG CPPLIB( ts_refed )( // POINT PAST REFERENCE TYPE-SIG, IF REQ'D
    RT_TYPE_SIG sig )           // - the signature
{
    if( sig->hdr.type == THROBJ_REFERENCE ) {
        sig = (RT_TYPE_SIG)( (char*)sig - offsetof( TS_HDR, hdr_ref ) );
        if( sig->base.indirect ) {
            sig = sig->indirected.sig;
        }
    }
    return sig;
}


RT_TYPE_SIG CPPLIB( ts_pnted )( // POINT PAST POINTER TYPE-SIG
    RT_TYPE_SIG sig )           // - the signature
{
    sig = (RT_TYPE_SIG)( (char*)sig - offsetof( TS_HDR, hdr_ptr ) );
    if( sig->base.indirect ) {
        sig = sig->indirected.sig;
    }
    return sig;
}


rboolean CPPLIB( ts_equiv )(    // TEST IF TYPE SIG.S ARE EQUIVALENT
    RT_TYPE_SIG tgt,            // - target type signature
    RT_TYPE_SIG src,            // - source type signature
    rboolean zero_thrown )      // - TRUE ==> zero was thrown
{
    rboolean retn;              // - TRUE ==> conversion possible

    if( tgt == NULL ) {
        retn = TRUE;
    } else {
        tgt = CPPLIB( ts_refed )( tgt );
        retn = FALSE;
        if( zero_thrown ) {
            switch( tgt->hdr.type ) {
              case THROBJ_PTR_SCALAR :
              case THROBJ_VOID_STAR :
              case THROBJ_PTR_CLASS :
                retn = TRUE;
                break;
            }
        }
        if( ! retn ) {
            src = CPPLIB( ts_refed )( src );
            if( src == tgt ) {
                retn = TRUE;
            } else if( tgt->hdr.type != src->hdr.type ) {
                retn = FALSE;
            } else switch( tgt->hdr.type ) {
              case THROBJ_VOID_STAR :
              case THROBJ_ANYTHING :
                retn = TRUE;
                break;
              case THROBJ_PTR_SCALAR :
                if( tgt->base.indirect != src->base.indirect ) {
                    retn = FALSE;
                    break;
                }
                // drops thru
              case THROBJ_PTR_CLASS :
                retn = CPPLIB( ts_equiv )( CPPLIB( ts_pnted )( tgt )
                                         , CPPLIB( ts_pnted )( src )
                                         , zero_thrown );
                break;
              case THROBJ_CLASS :
              case THROBJ_CLASS_VIRT :
                if( tgt->clss.size == src->clss.size ) {
                    if( 0 == strcmp( tgt->clss.name, src->clss.name ) ) {
                        retn = TRUE;
                    } else {
                        retn = FALSE;
                    }
                } else {
                    retn = FALSE;
                }
                break;
              case THROBJ_PTR_FUN :
              case THROBJ_SCALAR :
                if( tgt->scalar.size == src->scalar.size ) {
                    if( 0 == strcmp( tgt->scalar.name, src->scalar.name ) ) {
                        retn = TRUE;
                    } else {
                        retn = FALSE;
                    }
                } else {
                    retn = FALSE;
                }
                break;
              default :
                GOOF( "TsEquiv -- bad THROBJ_..." );
            }
        }
    }
    return retn;
}
