// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %     Copyright (C) 1992, by WATCOM International Inc.  All rights    %
// %     reserved.  No part of this software may be reproduced or        %
// %     used in any form or by any means - graphic, electronic or       %
// %     mechanical, including photocopying, recording, taping or        %
// %     information storage and retrieval systems - except with the     %
// %     written permission of WATCOM International Inc.                 %
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//
//  Modified    By              Reason
//  ========    ==              ======
//  92/01/09    Steve McDowell  Initial simple implementation using C library
//                              string functions wherever possible.
//  92/01/20    ...             Added checking for invalid Strings, including
//                              adding the ! operator, the == and !=
//                              operators with integer arguments, and the
//                              "valid" member and friend functions.
//  92/03/11    ...             Re-implemented using StringReps (reference
//                              counts, offset and length).
//  92/10/08	Greg Bentz	Cleanup.
//  93/08/31	Greg Bentz	- make extractor skip whitespace if
//				  ios::skipws is set
//				- correct offset computation in substring
//  93/10/07	Greg Bentz	- alloc_mult_size() not setting new value right
//  93/10/19	Raymond Tang	split into separate files
//  94/04/06	Greg Bentz	combine header files

#include "strng.h"

char String::__ref_char         = '\0';

char &String::operator () ( size_t pos ) {
/****************************************/
// Reference-to-character operator: return a reference to the character
// at position "pos".
// If the String is not well-formed, or if pos is not valid,
// return a reference to String::__ref_char:

    if( __srep == NULL  ||  pos >= __slength ) {
        return( __ref_char );
    }

    // If there are no other Strings referring to this StringRep, then
    // return a reference to the appropriate character:
    if( __srep->__refcount == 1 ) {
        return( __srep->__value[__offset + pos] );
    }

    // Make a copy of the StringRep so that a modification of the
    // returned character won't affect other Strings:
    String::StringRep *new_srep = __AllocStringRep( NULL, __slength );
    if( new_srep == NULL ) {
        return( __ref_char );
    }
    if( __slength > 0 ) {
        ::memcpy( new_srep->__value, __srep->__value + __offset, __slength );
    }
    __srep->__refcount--;
    __srep   = new_srep;
    __offset = 0;
    return( __srep->__value[pos] );
}
