;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;%       Copyright (C) 1985,1989 by WATCOM Systems Inc. All rights %
;%       reserved. No part of this software may be reproduced      %
;%       in any form or by any means - graphic, electronic or      %
;%       mechanical, including photocopying, recording, taping     %
;%       or information storage and retrieval systems - except     %
;%       with the written permission of WATCOM Systems Inc.        %
;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;
;<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
;<>
;<> __FLDC - long double comparison
;<>     input:  AX - pointer to operand 1
;<>             DX - pointer to operand 2
;<>       if op1 > op2,  1 is returned in AX
;<>       if op1 < op2, -1 is returned in AX
;<>       if op1 = op2,  0 is returned in AX
;<>	  if either opnd is NaN, then 2 is returned in AX
;<>
;<>  modified:    by:           reason:
;<>  ---------    ---           -------
;<>  29-sep-86  F.W.Crigger     F8cmp adapted to new code generator
;<>  26-oct-87  John Dahms      Added __EDC
;<>--------------------------------------------------------Version 6.0-----
;<>  15-apr-88  F.W.Crigger     __EDC didn't set si before calling __FDC
;<>  22-may-88  F.W.Crigger     Save some memory references
;<>  10-mar-89  F.W.Crigger     Adapted for 386
;<>  29-may-89  F.W.Crigger     Adapted for long doubles
;<>  08-jul-89  F.W.Crigger     Added chkfor0 for comparing +0 with -0
;<>  16-nov-89  F.W.Crigger     Re-adapted for 8086
;<>--------------------------------------------------------Version 9.0-----
;<>  07-jun-92	F.W.Crigger	Support added for NaNs
;<>  11-jul-92	F.W.Crigger	Check for Infinity
;<>--------------------------------------------------------Version 11.0----
;<>  24-nov-95  Greg Bentz      add support for 16bit fpc mathlib
;<>
;<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

        xdefp   __FLDC

        defp    __FLDC
ifdef _BUILDING_MATHLIB
	push	DS		; save DS
	push	SS		; fpc code assumes parms are relative to SS
	pop	DS		; ...
endif
        push    DI              ; save DI
        push    SI              ; save SI
        mov     SI,AX           ; point to op1
        mov     DI,DX           ; point to op2
	_guess			; guess: op1 is not a NaN
	  mov	AX,8[SI]	; - get sign and exponent of op1
	  or	AH,80H		; - turn on sign bit (all 1's if a NaN)
	  inc	AX		; - will be 0 if a NaN
	  _quif	ne		; - quit if not a NaN
	  mov	AX,4[SI]	; - or in rest of fraction
	  or	AX,2[SI]	; - ...	
	  or	AX,[SI]		; - ...	
	  jne	cmpNaN		; - non-zero fraction means its a NaN
	  mov	AX,6[SI]	; - get high order word of fraction
	  cmp	AX,8000h	; - check for infinity
	  jne	cmpNaN		; - 8000h => Inf, otherwise NaN
	_endguess		; endguess
	_guess			; guess: op2 is not a NaN
	  mov	AX,8[DI]	; - get sign and exponent of op2
	  or	AH,80H		; - turn on sign bit (all 1's if a NaN)
	  inc	AX		; - will be 0 if a NaN
	  _quif	ne		; - quit if not a NaN
	  mov	AX,4[DI]	; - or in rest of fraction
	  or	AX,2[DI]	; - ...	
	  or	AX,[DI]		; - ...	
	  jne	cmpNaN		; - non-zero fraction means its a NaN
	  mov	AX,6[DI]	; - get high order word of fraction
	  cmp	AX,8000h	; - check for infinity
	  jne	cmpNaN		; - 8000h => Inf, otherwise NaN
	_endguess		; endguess
        mov     DX,8[SI]        ; get sign and exponent of op1
        mov     AX,8[DI]        ; get sign and exponent of op2
        xor     AX,DX           ; see about signs of the operands
        mov     AX,0            ; clear result
        js      short chkfor0   ; quif arg1 & arg2 have diff signs
        _guess                  ; guess
          cmp   DX,8[DI]        ; - compare exponents
          _quif ne              ; - quif not equal
          mov   AX,6[SI]        ; - get high part of fraction
          cmp   AX,6[DI]        ; - compare them
          _quif ne              ; - quif not equal
          mov   AX,4[SI]        ; - get next part of fraction
          cmp   AX,4[DI]        ; - compare them
          _quif ne              ; - quif not equal
          mov   AX,2[SI]        ; - get next part of fraction
          cmp   AX,2[DI]        ; - compare them
          _quif ne              ; - quif not equal
          mov   AX,[SI]         ; - get low part of fraction
          cmp   AX,[DI]         ; - compare them
        _endguess               ; endguess
        mov     AX,0            ; clear result
        _if     ne              ; if arg1 <> arg2
          rcr   AX,1            ; - save carry in AX
          xor   DX,AX           ; - sign of DX is sign of result

cmpdone:  _shl  DX,1            ; - get sign of result into carry
          sbb   AX,0            ; - AX gets sign of result
          _shl  AX,1            ; - double AX
          inc   AX              ; - make AX -1 or 1
        _endif                  ; endif
        pop     SI              ; restore SI
        pop     DI              ; restore DI
ifdef _BUILDING_MATHLIB
	pop	DS		; restore DS
endif
        ret                     ; return to caller

cmpNaN:	mov	AX,2		; indicate NaN
        pop     SI              ; restore SI
        pop     DI              ; restore DI
ifdef _BUILDING_MATHLIB
	pop	DS		; restore DS
endif
        ret                     ; return to caller

chkfor0:or      AX,[SI]         ; see if both operands are 0
        or      AX,[DI]         ; ...
        or      AX,2[SI]        ; ...
        or      AX,2[DI]        ; ...
        or      AX,4[SI]        ; ...
        or      AX,4[DI]        ; ...
        or      AX,6[SI]        ; ...
        or      AX,6[DI]        ; ...
        and     DX,7FFFh        ; get rid of sign from op1
        or      AX,DX           ; check for zero
        mov     DX,8[DI]        ; get exponent of op2
        and     DX,7FFFh        ; get rid of sign from op2
        or      AX,DX           ; check for zero
        mov     DX,8[SI]        ; get sign of op1
        mov     AX,0            ;
        jne     cmpdone         ; if not zero
        pop     SI              ; restore SI
        pop     DI              ; restore DI
ifdef _BUILDING_MATHLIB
	pop	DS		; restore DS
endif
        ret                     ; return to caller

        endproc __FLDC

