/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1986, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  01-oct-86	F.W.Crigger	translated to WSL
  16-jan-87	A.J.Stein	Converted from WSL to C
  22-jan-88	F.W.Crigger	accuracy of root3 extended
-----------------------------------------------------------Version 6.5------
  06-dec-88	F.W.Crigger	special case 1.0 for exact answer
--------------------------------------------------------------Version 8.0-----
  20-dec-90	F.W.Crigger	IF@DATAN entry added
  91-jan-04	John Dahms	Made IF@ routines conform to new code generator
  03-sep-91	J.B.Schueler	macrotized read/write data reference
  09-sep-91	J.B.Schueler	changed _8087 to _real87
  03-sep-92	J.B.Schueler	Added "const" attribute to AtanPoly
--------------------------------------------------------Version-11.0---------
  06-jun-95	Greg Bentz	variety.h and dll linkage
  05-jul-95	Greg Bentz	better input checking for max values
  19-sep-95	Greg Bentz	modify name of IF@ functions for alpha
  28-may-96	Greg Bentz	fixup _IF_ function modifiers
*/

/*
	atan - compute arctangent

	Uses the following identities
	*****************************

	arctan(x) = - arctan( - x )

	arctan(x) = pi/2 - arctan( 1/x )

	arctan(x) = arctan( (x*sqrt(3)-1) / (x+sqrt(3)) ) + pi/6
*/

#include "variety.h"
#include <math.h>
#include <float.h>
#include "pi.h"
#include <ifprag.h>
#include "rtdata.h"
#include "pdiv.h"

#define FALSE		0
#define TRUE		1

#define tan15		0.26794919243112270647
#define sqrt3		1.73205080756887729353
#define sqrt3m1 	0.73205080756887729353


#if defined(_M_IX86)
  extern	double	_atan87(double);
  #if defined(__386__)
    #pragma aux	_atan87  "_*" parm [edx eax] value [edx eax];
  #else
    #pragma aux	_atan87  "_*" parm [ax bx cx dx] value [ax bx cx dx];
  #endif
#endif

extern	int		__sgn(double);	 /* get sign of x */
extern	double		_OddPoly(double, const double *, int);


#define MIN_POLY_CONST 0.04
static	const double	AtanPoly[] = {
	 0.0443895157187,
	-0.06483193510303,
	 0.0767936869066,
	-0.0909037114191074,  //	-0.09090371144275426,
	 0.11111097898051048,
	-0.14285714102825545,
	 0.1999999999872945,
	-0.3333333333332993,
	 1.0
};

_WMRTLINK extern double _IF_datan( double );
#if defined(_M_IX86)
  #pragma aux (if_rtn) _IF_atan "IF@ATAN";
  #pragma aux (if_rtn) _IF_datan "IF@DATAN";
#endif

_WMRTLINK float _IF_atan( float x )
/*********************/
    {
	return( _IF_datan( x ) );
    }

_WMRTLINK double (atan)( double x )
/***********************/
    {
	return( _IF_datan( x ) );
    }

_WMRTLINK double _IF_datan( double x )
/*************************/
    {
	register char	add_piby2;
	register char	add_piby6;
	register int	sgnx;
	double	tmp;

	#if defined(_M_IX86)
	    if( _RWD_real87 ) return( _atan87(x) );
	#endif
	add_piby2 = FALSE;
	add_piby6 = FALSE;
	sgnx = __sgn( x );
	x = fabs( x );
	if( x == 1.0 ) {	/* 06-dec-88 */
	    x = Pi / 4;
	} else if( x > (1.0 / DBL_MIN) ) {
	    x = Pi / 2;
	} else if( x < (DBL_MIN/MIN_POLY_CONST) ) {
	    x = DBL_MIN;
	} else {
	    if( x > 1.0 ) {
		x = PDIV( 1.0 , x );
		add_piby2 = TRUE;
	    }
	    if( x > tan15 ) {
		tmp = sqrt3m1 * x - 0.5;
		tmp -= 0.5;
		x = PDIV( (tmp + x) , (x + sqrt3) );
//		  x = (x * sqrt3 - 1.0) / (x + sqrt3);
		add_piby6 = TRUE;
	    }
	    x = _OddPoly( x, AtanPoly, 8 );
	    if( add_piby6 ) {
		x += Piby6;
	    }
	    if( add_piby2 ) {
		x = Piby2 - x;
	    }
	}
	if( sgnx < 0 ) {
	    x = -x;
	}
	return( x );
    }
