/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1986, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  01-oct-86     F.W.Crigger     Translated to WSL
  29-oct-86     B.J. Stecher    bug fixes
  16-jan-87     A.J.Stein       Converted from WSL to C
  24-aug-87     F.W.Crigger     Exception handling added
  11-dec-87     F.W.Crigger     Fixed
------------------------------------------------------Version 6.0---------
  21-mar-88     F.W.Crigger     special cases 1 & 2 were backwards.
  04-jan-91     John Dahms      Made IF@ routines conform to new code generator
--------------------------------------------------------Version-11.0---------
  06-jun-95	Greg Bentz	variety.h and dll linkage
  05-jul-95	Greg Bentz	better checking on input values
  19-sep-95	Greg Bentz	modify name of IF@ functions for alpha
  28-may-96	Greg Bentz	fixup _IF_ function modifiers
*/

/*

atan2( y : double, x : double ) : double - compute atan(y/x)

NOTE: x,y cannot both be 0

case 1: y = 0    if x < 0 return pi, otherwise return 0

case 2: x = 0    if y < 0 return -pi/2, otherwise return pi/2

otherwise: compute z = atan( y/x )
               if y >= 0 and z < 0 return z + pi
               if y <  0 and z > 0 return z - pi
               else return z

*/
#include "variety.h"
#include <stddef.h>
#include <math.h>
#include <float.h>
#include "pi.h"
#include "ifprag.h"
#include "mathcode.h"
#include "pdiv.h"

extern  int             __sgn(double);   /* determine sgn(x) */

_WMRTLINK extern double _IF_datan2( double, double );

#if defined(_M_IX86)
  #pragma aux (if_rtn) _IF_atan2 "IF@ATAN2";
  #pragma aux (if_rtn) _IF_datan2 "IF@DATAN2";
#endif

_WMRTLINK float _IF_atan2( float y, float x )
/********************************/
    {
        return( _IF_datan2( y, x ) );
    }

_WMRTLINK double (atan2)( double y, double x )
/**********************************/
    {
        return( _IF_datan2( y, x ) );
    }

_WMRTLINK double _IF_datan2( double y, double x )
/************************************/
    {
        register int    sgnx;   /* sgn(x) */
        register int    sgny;   /* sgn(y) */
        register int    sgnz;   /* sgn(z) */

        sgny = __sgn( y );
        sgnx = __sgn( x );
        if( sgny == 0 ) {               /* case 1 */
            if( sgnx == 0 ) {
//                x = _matherr( DOMAIN, "atan2", &y, &x, 0.0 );
                x = __math2err( FUNC_ATAN2 | M_DOMAIN | V_ZERO, &y, &x );
            } else if( sgnx < 0 ) {
                x = Pi;
            } else {
                x = 0.0;
            }
        } else if( sgnx == 0 ) {        /* case 2 */
            if( sgny < 0 ) {
                x = -Piby2;
            } else {
                x = Piby2;
            }
	} else {
	    int exp_x;
	    int exp_y;
	    int diff_exp_y_x;

	    frexp( y, &exp_y );
	    frexp( x, &exp_x );

	    diff_exp_y_x = exp_y - exp_x;

	    if( diff_exp_y_x > DBL_MAX_EXP ) {
		if( sgnx == sgny ) {
		    x = +DBL_MAX;
		} else {
		    x = -DBL_MAX;
		}
	    } else if( diff_exp_y_x < DBL_MIN_EXP ) {
		if( sgnx == sgny ) {
		    x = +DBL_MIN;
		} else {
		    x = -DBL_MIN;
		}
	    } else {
		x = PDIV( y, x );
	    }
	    x = atan( x );
            sgnz = __sgn( x );
            if( sgny >= 0 ) {
                if( sgnz < 0 ) {
                    x += Pi;
                }
            } else {
                if( sgnz > 0 ) {
                    x -= Pi;
                }
            }
        }
        return( x );
    }
