/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1986, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  01-oct-86	F.W.Crigger	Translated to WSL
  16-jan-87	A.J.Stein	Converted from WSL to C
------------------------------------------------------------Version 7.0-----
  07-jun-89	F.W.Crigger	extended precision of polynomial constants
  04-jan-91	John Dahms	Made IF@ routines conform to new code generator
  03-sep-91	J.B.Schueler	macrotized read/write data reference
  09-sep-91	J.B.Schueler	changed _8087 to _real87
  03-sep-92	J.B.Schueler	Added "const" attribute to ExpConsts
--------------------------------------------------------Version-11.0---------
  06-jun-95	Greg Bentz	variety.h and dll linkage
  19-sep-95	Greg Bentz	modify name of IF@ functions for alpha
  28-may-96	Greg Bentz	fixup _IF_ function modifiers
*/
#include "variety.h"
#include <stddef.h>
#include <math.h>
#include "ifprag.h"
#include "mathcode.h"
#include "rtdata.h"

#if defined(_M_IX86)
  extern	double	_exp87(double);
  #if defined(__386__)
    #pragma aux	_exp87	"_*" parm [edx eax] value [edx eax];
  #else
    #pragma aux	_exp87	"_*" parm [ax bx cx dx] value [ax bx cx dx];
  #endif
#endif

extern	int	__sgn(double);

#define log2e		1.4426950408889633
#define const0		20.813771196523035
#define const1		0.057761135831801924
#define const2		7.213503410844819


static	const double	ExpConsts[] = {
	1.0442737824274138403,	/*   sqrt(sqrt(sqrt(sqrt(2))))	*/
	1.0905077326652576592,	/*   sqrt(sqrt(sqrt(2)))	*/
	1.1892071150027210667,	/*   sqrt(sqrt(2))		*/
	1.4142135623730950488	/*   sqrt(2)			*/
};




_WMRTLINK extern double _IF_dexp( double );

#if defined(_M_IX86)
  #pragma aux (if_rtn) _IF_exp "IF@EXP";
  #pragma aux (if_rtn) _IF_dexp "IF@DEXP";
#endif

_WMRTLINK float _IF_exp( float x )
/*********************/
{
    return( _IF_dexp( x ) );
}

_WMRTLINK double (exp)( double x )
/**********************/
{
    return( _IF_dexp( x ) );
}

_WMRTLINK double _IF_dexp( double x )
/************************/
{
    register int	sgnx;
    register int	exp;
    register int	exp2;
    register const double *poly;
    double		ipart;
    double		a;
    double		b;
    double		ee;

    if( fabs( x ) < 4.445e-16 ) {	/* if argument is too small */
	x = 1.0;
    } else if( fabs( x ) > 709.782712893384 ) {/* if argument is too large */
	if( x < 0.0 ) { 		/* - if argument < 0 */
	    /* FPStatus = FPS_UNDERFLOW;   - - underflow */
	    x = 0.0;			/* - - set result to 0 */
	} else {			/* - else */
//	      x = _matherr( OVERFLOW, "exp", &x, &x, HUGE_VAL );
	    x = __math1err( FUNC_EXP | M_OVERFLOW | V_HUGEVAL, &x );
	}				/* - endif */
#if defined(_M_IX86)
    } else if( _RWD_real87 ) {
	x = _exp87( x );
#endif
    } else {
	x *= log2e;
	sgnx = __sgn( x );
	x = modf( fabs( x ), &ipart );
	exp = ipart;
	if( sgnx < 0 ) {
	    exp = -exp;
	    if( x != 0.0 ) {
		--exp;
		x = 1.0 - x;
	    }
	}
	exp2 = 0;
	if( x != 0.0 ) {
	    x = modf( ldexp( x, 4 ), &ipart );
	    if( x != 0.0 ) {
		x = ldexp( x, -4 );
	    }
	    exp2 = ipart;
	}
	ee = x * x;
	a = ee + const0;
	b = (ee * const1 + const2) * x;
	x = (a + b) / (a - b);
	poly = ExpConsts;
	while( exp2 != 0 ) {
	    if( exp2 & 1 ) {
		x *= *poly;
	    }
	    ++poly;
	    exp2 = exp2 >> 1;
	}
	x = ldexp( x, exp );
    }
    return( x );
}
