/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %     Copyright (C) 1995, by WATCOM International Inc.  All rights    %
  %     reserved.  No part of this software may be reproduced or        %
  %     used in any form or by any means - graphic, electronic or       %
  %     mechanical, including photocopying, recording, taping or        %
  %     information storage and retrieval systems - except with the     %
  %     written permission of WATCOM International Inc.                 %
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  Date		By		Reason
  ----		--		------
--------------------------------------------------------Version-11.0---------
  01-jun-95	Greg Bentz	variety.h and dll linkage
  23-jun-95	Greg Bentz	alpha support
  18-Oct-95	A.F.Scian	added support so more extreme values don't
  				cause FP over/underflow errors
*/
#include <math.h>
#include <float.h>
#include <limits.h>

#include "variety.h"
#include "xfloat.h"

void __fprem( double x, double y, int *quot, double *rem ) {
    int sign_x;
    int sign_y;
    int q;

    if( y == 0.0 ) {				    // if modulus is zero
	*quot = 0;				    // set quot to zero
	*rem = 0.0;				    // set rem to zero
	return;
    }

    sign_x = 1;
    if( x < 0.0 ) {
	sign_x = -1;
	x = -x;
    }
    sign_y = 1;
    if( y < 0.0 ) {
	sign_y = -1;
	y = -y;
    }
    q = 0;
    while( x >= y ) {
	int ex;
	double mx = frexp( x, &ex );
	int ey;
	double my = frexp( y, &ey );
	double ny;

	#if 0	// use this code when modf doesn't call __fprem
	if( ( ex - ey ) <= DBL_MAX_EXP ) {
	    double d;
	    double ipart;

	    d = x / y;
	    modf( d, &ipart );
	    q = ipart;
	    x -= ipart * y;
	    break;
	}
	#endif
	if( mx >= my ) {
	    ny = ldexp( my, ex );
	} else {
	    --ex;
	    ny = ldexp( my, ex );
	}
	ex -= ey;
	if( ex <= ( sizeof( *quot ) * CHAR_BIT ) ) {
	    q += 1 << ex;
	}
	x -= ny;
    }
    if( sign_x == -1 ) {
	x = -x;
    }
    *rem = x;
    if( sign_x != sign_y ) {
	q = -q;
    }
    *quot = q;
}
