/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %     Copyright (C) 1995, by WATCOM International Inc.  All rights    %
  %     reserved.  No part of this software may be reproduced or        %
  %     used in any form or by any means - graphic, electronic or       %
  %     mechanical, including photocopying, recording, taping or        %
  %     information storage and retrieval systems - except with the     %
  %     written permission of WATCOM International Inc.                 %
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  

 IFCDSQRT     : square root function for COMPLEX*16 argument

  Date		By		Reason
  ----		--		------
  88/01/14	G. Coschi	initial implementation
  94/04/08	Kevin Hui	merge C++ and Fortran Complex support
--------------------------------------------------------Version-11.0---------
  06-jun-95	Greg Bentz	variety.h and dll linkage
  26-sep-95	Greg Bentz	intrinsic support for alpha
*/
#include "variety.h"
#include "cplx.h"
#include "pdiv.h"

_WMRTLINK dcomplex _IF_CDSQRT( double rp, double ip ) {
//====================================================

// Return the complex square root of "arg".

    double      r;
    double      t;
    double      x;
    double      y;
    dcomplex    result;

    // treat the imaginary number as a vector [magnitude m,angle 2*theta]
    // square root is equal to [magnitude sqrt(m),angle theta]

    if( ip == 0.0 ) {
        if( rp < 0.0 ) {
            // arg is (a,0) where a is negative
            result.realpart = 0.0;
            result.imagpart = sqrt( -rp );
        } else {
            // arg is (a,0) where a is positive
            result.realpart = sqrt( rp );
            result.imagpart = 0.0;
        }
    } else if( rp == 0.0 ) {
        if( ip < 0.0 ) {
            // arg is (0,a) where a is negative
            t = sqrt( (-ip) / 2 );
            result.imagpart = -t;
        } else {
            // arg is (0,a) where a is positive
            t = sqrt( ip / 2 );
            result.imagpart = t;
        }
        result.realpart = t;
    } else {
        x = fabs( rp );
        y = fabs( ip );
        if( x >= y ) {
            r = PDIV( y , x );
            t = sqrt( x ) * sqrt( ( 1 + sqrt( 1 + r * r ) ) / 2 );
        } else {
            r = PDIV( x , y );
            t = sqrt( y ) * sqrt( ( r + sqrt( 1 + r * r ) ) / 2 );
        }
        if( rp >= 0.0 ) {
            result.realpart = t;
            result.imagpart = PDIV( ip , ( 2 * t ) );
        } else {
            if( ip >= 0 ) {
                result.imagpart = t;
            } else {
                result.imagpart = -t;
            }
            result.realpart = PDIV( ip , ( 2 * result.imagpart ) );
        }
    }
    return( result );
}
