/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1986, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  01-oct-86	F.W.Crigger	Translated to WSL
  16-jan-87	A.J.Stein	Converted from WSL to C
  24-aug-87	F.W.Crigger	Exception handling added
--------------------------------------------------------------Version 7.0-----
  31-mar-89	F.W.Crigger	adapted for 386
  22-jun-89	F.W.Crigger	New implementation based on Cody & Waite
--------------------------------------------------------------Version 8.0-----
  20-dec-90	F.W.Crigger	IF@DLOG entry added
  04-jan-91	John Dahms	Made IF@ routines conform to new code generator
  30-aug-91	J.B.Schueler	Added "const" attribute to tables
  03-sep-91	J.B.Schueler	macrotized read/write data reference
  09-sep-91	J.B.Schueler	changed _8087 to _real87
--------------------------------------------------------Version-11.0---------
  06-jun-95	Greg Bentz	variety.h and dll linkage
  19-sep-95	Greg Bentz	modify name of IF@ functions for alpha
  28-may-96	Greg Bentz	fixup _IF_ function modifiers
*/
#include "variety.h"
#include <stddef.h>
#include <math.h>
#include "ifprag.h"
#include "mathcode.h"
#include "rtdata.h"

_WMRTLINK extern double _IF_dlog( double );
#if defined(_M_IX86)
  #pragma aux (if_rtn) _IF_log "IF@LOG";
  #pragma aux (if_rtn) _IF_dlog "IF@DLOG";
#endif

_WMRTLINK float _IF_log( float x )
/*********************/
{
    return( _IF_dlog( x ) );
}

_WMRTLINK double (log)( double x )
/**********************/
{
    return( _IF_dlog( x ) );
}

#if defined(_M_IX86)
  extern	double	_log87(double);
  #if defined(__386__)
    #pragma aux	_log87	"_*" parm [edx eax] value [edx eax];
  #else
    #pragma aux	_log87	"_*" parm [ax bx cx dx] value [ax bx cx dx];
  #endif
#endif

extern	double	_EvalPoly( double, const double *, int );

#define sqrt_of_half	0.7071067811865475244
#define C1		(355./512.)
#define C2		(-2.121944400546905827679e-4)


static const double		APoly[] = {
	-0.78956112887491257267,
	 0.16383943563021534222e+2,
	-0.64124943423745581147e+2
};
static const double		BPoly[] = {
	 1.0,
	-0.35667977739034646171e+2,
	 0.31203222091924532844e+3,
	-0.76949932108494879777e+3
};

_WMRTLINK double _IF_dlog( double x )
/********************/
{
    int     exp;
    double  z;

    if( x <= 0.0 ) {
//	x = _matherr( x == 0.0 ? SING : DOMAIN, "log", &x, &x, -HUGE_VAL );
	x = __log87_err( x, FUNC_LOG );
#if defined(_M_IX86)
    } else if( _RWD_real87 ) {
	x = _log87( x );
#endif
    } else {
	x = frexp( x, &exp );
	z = x - .5;
	if( x > sqrt_of_half ) {
	    z -= .5;
	} else {
	    x = z;
	    --exp;
	}
	z = z / (x * .5 + .5);
	x = z * z;
	x = z + z * x * _EvalPoly( x, APoly, 2 ) / _EvalPoly( x, BPoly, 3 );
	if( exp != 0 ) {
	    z = exp;
	    x = (z * C2 + x) + z * C1;
	}
    }
    return( x );
}
