/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1986, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  02-oct-86     F.W.Crigger     Initial implementation in WSL
  16-jan-87     A.J.Stein       Converted from WSL to C
  04-feb-88     F.W.Crigger     check for overflow conditions
-------------------------------------------------------------Version 7.0------
  01-mar-89     F.W.Crigger     x ** y where x and y < 0, must check to see
                                if y is odd or even to get sign of result.
  27-mar-89     F.W.Crigger     Use shift and multiply if exponent is integer
  18-aug-89     John Dahms      include "newerrno.h" so errno is ok in DOS
                                as well as multi-threaded OS/2 environment
  15-feb-90     F.W.Crigger     check exponent for LONG_MIN
  28-feb-90     F.W.Crigger     call __set_ERANGE() to set errno
  04-jan-91     John Dahms      Made IF@ routines conform to new code generator
  15-feb-91     G. Coschi       arguments to if_dpow were backwards
  24-mar-91     F.W.Crigger     Use new __math2err instead of _matherr
--------------------------------------------------------Version-11.0---------
  06-jun-95	Greg Bentz	variety.h and dll linkage
  19-sep-95	Greg Bentz	modify name of IF@ functions for alpha
  28-may-96	Greg Bentz	fixup function prototypes and add libsupp.h
  28-may-96	Greg Bentz	fixup _IF_ function modifiers
*/
#include "variety.h"
#include "libsupp.h"
#include <stddef.h>
#include <math.h>
#include <limits.h>
#include "ifprag.h"
#include "mathcode.h"

extern  int     __sgn(double);


_WMRTLINK extern double _IF_dpow( double, double );
#if defined(_M_IX86)
  #pragma aux (if_rtn) _IF_pow "IF@POW";
  #pragma aux (if_rtn) _IF_dpow "IF@DPOW";
#endif

_WMRTLINK float _IF_pow( float x, float y )
/********************************/
    {
        return( _IF_dpow( x, y ) );
    }

_WMRTLINK double (pow)( double x, double y )
/********************************/
    {
        return( _IF_dpow( x, y ) );
    }

_WMRTLINK double _IF_dpow( double x, double y )
/**********************************/
{
        unsigned int    err_code;
        register int    sign;
        double          z;
        double          fraction;
        long            exponent;

    sign = __sgn( x );
    if( sign == 0 ) {
        if( y < 0.0 ) {
//            x = _matherr( DOMAIN, "pow", &x, &y, HUGE_VAL );
            x = __math2err( FUNC_POW | M_DOMAIN | V_HUGEVAL, &x, &y );
        } else if( y == 0.0 ) {
//            x = _matherr( DOMAIN, "pow", &x, &y, 1.0 );
            x = __math2err( FUNC_POW | M_DOMAIN | V_ONE, &x, &y );
        }
    } else {
        fraction = modf( y, &z );
        exponent = z;
        if( sign < 0 ) {
            if( fraction != 0.0 ) {     /* negative ** fraction */
//                return( _matherr( DOMAIN, "pow", &x, &y, 0.0 ) );
                return( __math2err( FUNC_POW | M_DOMAIN | V_ZERO, &x, &y ) );
            }
            x = -x;
            if( (exponent & 1) == 0 ) {         /* 01-mar-89 */
                sign = 0;                       /* if # is even, change sign */
            }
        }
        z = log( x ) * y;
        if( fabs( z ) > 709.782712893384 ) {/* if argument is too large */
            if( z < 0.0 ) {                 /* - if argument < 0 */
                /* FPStatus = FPS_UNDERFLOW;   - - underflow */
                __set_ERANGE();
                x = 0.0;                    /* - - set result to 0 */
            } else {                        /* - else */
//                z = (sign > 0) ? HUGE_VAL : - HUGE_VAL;
//                x = _matherr( OVERFLOW, "pow", &x, &y, z );
                if( sign > 0 ) {
                    err_code = FUNC_POW | M_OVERFLOW | V_HUGEVAL;
                } else {
                    err_code = FUNC_POW | M_OVERFLOW | V_NEG_HUGEVAL;
                }
                x = __math2err( err_code, &x, &y );
            }
        } else {
            if( fraction == 0.0 && exponent != LONG_MIN ) {  /* x ** integer */
                if( exponent < 0 ) {
                    exponent = - exponent;
                    x = 1 / x;
                }
                if( exponent != 1 ) {
                    z = 1.0;
                    while( exponent != 0 ) {
                        if( exponent & 1 )  z *= x;
                        x *= x;
                        exponent >>= 1;
                    }
                    x = z;
                }
            } else {
                x = exp( z );
            }
            if( sign < 0 ) {
                x = -x;
            }
        }
    }
    return( x );
}
