/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %     Copyright (C) 1995, by WATCOM International Inc.  All rights    %
  %     reserved.  No part of this software may be reproduced or        %
  %     used in any form or by any means - graphic, electronic or       %
  %     mechanical, including photocopying, recording, taping or        %
  %     information storage and retrieval systems - except with the     %
  %     written permission of WATCOM International Inc.                 %
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  

 POWC16          : complex power run-time routine

  Date		By		Reason
  ----		--		------
  88/01/14	G. Coschi	Initial implementation.
  94/04/08	Kevin Hui	merge C++ and Fortran Complex support
  94/04/12	Kevin Hui	fixup error-handling
--------------------------------------------------------Version-11.0---------
  06-jun-95	Greg Bentz	variety.h and dll linkage
  26-sep-95	Greg Bentz	intrinsic support for alpha
*/
#include "variety.h"
#include "cplx.h"
#include "mathcode.h"


_WMRTLINK dcomplex _IF_C16Pow( double a, double b, double c, double d ) {
//======================================================================

// Return the real part of the result of taking one double precision
// complex number to the power of another.
// ( a, b ) ** ( c, d ) = exp( ( c,d ) * log( a,b ) )

    double      u;
    double      v;
    dcomplex    arg1;
    dcomplex    arg2;
    dcomplex    res;

    if( ( a == 0 ) && ( b == 0 ) ) {
        if( d == 0 ) {
            if( c > 0 ) {
                res.realpart = 0;
                res.imagpart = 0;
		return( res );
	    }
	}
	arg1.realpart = a;
	arg1.imagpart = b;
	arg2.realpart = c;
	arg2.imagpart = d;
	return( __qmath2err( FUNC_POW | M_DOMAIN | V_ZERO, &arg1, &arg2 ) );
    } else {
        u = DLOG( a*a + b*b ) / 2;
        v = DATAN2( b, a );
        res.realpart = DEXP( u*c - v*d );
        res.imagpart = res.realpart;
        u = u*d + v*c;
        res.realpart *= DCOS( u );
        res.imagpart *= DSIN( u );
	return( res );
    }
}
