/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1986, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:	By:		Reason:
  ---------	---		-------
  23-jan-85	F.W.Crigger	initial implementation
  24-nov-88	F.W.Crigger	split off from ftos.c
  31-jan-91	F.W.Crigger	check for +inf and -inf
  30-aug-91	J.B.Schueler	Added "const" attribute to table
  29-aug-92	B.J. Stecher	got rid of some pointer trunc warnings
  03-sep-92	J.B.Schueler	Added "const" attribute to PlusInf, etc.
--------------------------------------------------------Version-11.0---------
  28-apr-95	Greg Bentz	variety.h for near/far and composite macros
  06-jun-95	Greg Bentz	dll linkage
*/

#include "variety.h"

typedef struct pow10table {
	short	exponent;
	double	powerof10;
} pow10table;


extern	const pow10table    _BigPow10Table[8];
extern	void		    _Rnd2Int(double _WCNEAR *,int _WCNEAR *);
extern	int		    _CmpBigInt(int, int _WCNEAR *);
extern	double		    _Scale10V( double, int );

#if defined(__386__)
#pragma aux	_Rnd2Int	"_*" parm routine [eax] [edx];
#pragma aux	_CmpBigInt	"_*" parm caller [eax] [edx];
#elif defined(M_I86)
#pragma aux	_Rnd2Int	"_*" parm routine [ax] [dx];
#pragma aux	_CmpBigInt	"_*" parm caller [ax] [dx];
#endif


/**************************************************************************/
/***									***/
/*** Scale - scale a number up or down so that it lies between 10**k	***/
/***	and 10**(k-1).							***/
/***									***/
/**************************************************************************/

int _Scale( exponent, realnum, sigdigits, bigint )
/************************************************/
	register int		exponent;
		 double 	realnum;
	register int		sigdigits;
	register int	_WCNEAR	*bigint;
    {
    if( exponent != sigdigits ) {
	realnum = _Scale10V( realnum, sigdigits - exponent );
    }
    _Rnd2Int( (double _WCNEAR *)&realnum, bigint );
    return( _CmpBigInt( sigdigits, bigint ) );
}


double _Scale10V( x, scale )
/**************************/
	double	x;
	register int	scale;
    {
	register const pow10table   *table;
	register int		    n;
	auto	 double 	    z;
	union u {
	    unsigned short int ui[4];
	    double   x;
	};
	static const union u PlusInf  = { 0x0000, 0x0000, 0x0000, 0x7ff0 };
	static const union u NegInf   = { 0x0000, 0x0000, 0x0000, 0xfff0 };
	static const union u PlusInf1 = { 0xffff, 0xffff, 0xffff, 0x7fef };
	static const union u NegInf1  = { 0xffff, 0xffff, 0xffff, 0xffef };

	if( x == PlusInf.x ) {				/* 31-jan-91 */
	    x = PlusInf1.x;
	} else if( x == NegInf.x ) {
	    x = NegInf1.x;
	}
	table = _BigPow10Table;
	if( scale < 0 ) {
	    n = -scale;
	    if( n > 308 ) {
		x /= table->powerof10;
		n -= 216;
	    }
	} else {
	    n = scale;
	    if( n > 308 ) {
		x *= table->powerof10;
		n -= 216;
	    }
	}
	z = 1.0;
	for( ;; ) {
	    if( n >= table->exponent ) {
		n -= table->exponent;
		z *= table->powerof10;
	    }
	    if( n == 0 ) break;
	    if( table->exponent != 1 ) {
		++table;
	    }
	}
	if( scale < 0 ) {
	    x /= z;
	} else {
	    x *= z;
	}
	return( x );
    }
