/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%       Copyright (C) 1986, by WATCOM Systems Inc. All rights     %
 *%       reserved. No part of this software may be reproduced      %
 *%       in any form or by any means - graphic, electronic or      %
 *%       mechanical, including photocopying, recording, taping     %
 *%       or information storage and retrieval systems - except     %
 *%       with the written permission of WATCOM Systems Inc.        %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  Modified:     By:             Reason:
  ---------     ---             -------
  06-oct-86     F.W.Crigger     Initial implementation in WSL
  12-dec-86     F.W.Crigger     only place 16 characters into buffer
  16-jan-87     A.J.Stein       Converted from WSL to C
  06-jan-88     F.W.Crigger     interface to _ZBuf2F changed
  22-jan-88     F.W.Crigger     # of significant digits bumped by 1 to 17
------------------------------------------------------------Version 6.5-----
  08-dec-88     F.W.Crigger     isspace() replaced with inline check
------------------------------------------------------------Version 7.0-----
  18-aug-89     John Dahms      include "newerrno.h" so errno is ok in DOS
                                as well as multi-threaded OS/2 environment
  07-sep-89     F.W.Crigger     strip trailing '0's from buffer so that
                                numbers like 100e+300 convert to the same
                                value as 1e+302.
  27-nov-89     F.W.Crigger     bufptr one too many past digits of exponent
  28-feb-90     F.W.Crigger     call __set_ERANGE() to set errno
  07-mar-91     F.W.Crigger     0.0e99999 should return 0 (ANSI ruling)
  29-jun-92	D. Wellstood	Unicode support
  29-aug-92	B.J. Stecher	got rid of some pointer trunc warnings
  08-feb-93	F.W.Crigger	Allow exponents with 'd' and 'D'
  23-jun-94	F.W.Crigger	Added support for long doubles
  19-sep-94	Greg Bentz	improve range checking on exponent for errno
  21-dec-94	B.J. Stecher	undid 8-feb-93 change. not needed, and incorrect for C
  09-mar-95	F.W.Crigger	support both doubles and long doubles
  20-apr-95	F.W.Crigger	set MAXDIGITS to 19 for _LONG_DOUBLE_
--------------------------------------------------------Version-11.0---------
  28-apr-95	Greg Bentz	variety.h for near/far and composite macros
  06-jun-95	Greg Bentz	dll linkage
  29-sep-95	M. Hildebrand	Added _u version, which calls wide version.
  03-oct-95	M. Hildebrand	Fixed bug in original wcstod code.
  28-may-96	Greg Bentz	fixup function prototypes and add libsupp.h
  09-may-97	Greg Bentz	improve range checking for JCK and to
  				accurately reflect that operation is done
				on a long double, but returned in a double
  02-jun-97	Greg Bentz	tune above improvement for FORTRAN
  08-oct-97	Greg Bentz	make sure wcstombs() call includes nullchar
*/
#include "variety.h"
#include "widechar.h"
#include "libsupp.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <wchar.h>
#include "xfloat.h"

#define NEGATIVE_NUMBER         0x01
#define NEGATIVE_EXPONENT       0x02
#define DIGITS_PRESENT          0x04
#define DOT_FOUND               0x08

enum {
    _ZERO	= 0,
    _NONZERO,
    _UNDERFLOW,
    _OVERFLOW
};

#ifdef _LONG_DOUBLE_
 #define MAX_DIGITS	19
#else
 #define MAX_DIGITS	18
#endif

#if !(defined(_LONG_DOUBLE_) || defined(__WIDECHAR__))
void __ZBuf2LD( char _WCNEAR *buf, long_double _WCNEAR *ld )
{
    int		i;
    int		n;
    long	high;
    long	low;
    CHAR_TYPE	*ptr;

    ptr = (CHAR_TYPE *)buf;
    n = __F_NAME(strlen,wcslen)( ptr );
    high = 0;
    for( i = 0; i <= 8; i++ ) {		// collect high 9 significant digits
	if( n <= 9 ) break;
	--n;
	high = high * 10 + (*ptr++ - '0');
    }
    low = 0;
    for( i = 0; i <= 8; i++ ) {		// collect low 9 significant digits
	if( n == 0 ) break;
	--n;
	low = low * 10 + (*ptr++ - '0');
    }
    if( high == 0 && low == 0 ) {
	ld->value = 0.0;
    } else {
	ld->value = (double)high * 1e9 + (double)low;
    }
}
#endif

_WMRTLINK int __F_NAME(__Strtold,__wStrtold)( const CHAR_TYPE *bufptr,
				    long_double *pld,
				    CHAR_TYPE **endptr )
/*******************************************************/
{
    CHAR_TYPE	chr;
    int		sigdigits;
    int 	fdigits;
    int 	exponent;
    const CHAR_TYPE *oldptr;
    long_double	ld;
    CHAR_TYPE	digitflag;
    CHAR_TYPE	flags;
    CHAR_TYPE	buffer[20];

    oldptr = bufptr;
    for(;;) {
	chr = *bufptr;
	if( chr != ' ' && ( chr < '\t' || chr > '\r' ) ) break;
	++bufptr;
    }
    flags = 0;
    if( chr == '+' ) {
	++bufptr;
    } else if( chr == '-' ) {
	flags |= NEGATIVE_NUMBER;
	++bufptr;
    }
    sigdigits = 0;
    fdigits = 0;
    digitflag = '0';
    for(;;) {
	chr = *bufptr++;
	if( chr == '.' ) {
	    if( flags & DOT_FOUND ) break;
	    flags |= DOT_FOUND;
	} else {
	    if( chr < '0' || chr > '9' ) break;
	    if( flags & DOT_FOUND ) ++fdigits;
	    digitflag |= chr;
	    if( digitflag != '0' ) {        /* if a significant digit */
		if( sigdigits < MAX_DIGITS ) {
		    buffer[ sigdigits ] = (CHAR_TYPE)chr;
		}
		++sigdigits;
	    }
	    flags |= DIGITS_PRESENT;
	}
    }
    exponent = 0;
    if( flags & DIGITS_PRESENT ) {
	if( chr == 'e' || chr == 'E' ) {
	    oldptr = bufptr - 1;
	    chr = *bufptr;
	    if( chr == '+' ) {
		++bufptr;
	    } else if( chr == '-' ) {
		flags |= NEGATIVE_EXPONENT;
		++bufptr;
	    }
	    flags &= ~ DIGITS_PRESENT;
	    for( ;; ) {
		chr = *bufptr;              /* don't increment bufptr yet */
		if( chr < '0' || chr > '9' )  break;
		if( exponent < 1000 ) {
		    exponent = exponent * 10 + chr - '0';
		}
		flags |= DIGITS_PRESENT;
		++bufptr;                           /* 27-nov-89 */
	    }
	    if( flags & NEGATIVE_EXPONENT ) {
		exponent = -exponent;
	    }
	    if( (flags & DIGITS_PRESENT) == 0 ) {
		bufptr = oldptr;
	    }
	} else {    /* digits found, but no e or E */
	    --bufptr;
	}
	oldptr = bufptr;
    }
    if( endptr != NULL ) *endptr = (CHAR_TYPE *)oldptr;
    exponent -= fdigits;
    if( sigdigits > MAX_DIGITS ) {
	exponent += sigdigits - MAX_DIGITS;
	sigdigits = MAX_DIGITS;
    }
    while( sigdigits > 0 ) {
	if( buffer[ sigdigits - 1 ] != '0' )  break;
	++exponent;
	--sigdigits;
    }
    if( sigdigits == 0 ) {          /* number is 0.0   07-mar-91 */
#ifdef _LONG_DOUBLE_
	pld->exponent  = 0;
	pld->high_word = 0;
	pld->low_word  = 0;
#else
	pld->word[0] = 0;
	pld->word[1] = 0;
#endif
	return( _ZERO );		// indicate number is 0
    } else {
	buffer[ sigdigits ] = '\0';
	#ifdef __WIDECHAR__
	{
	    // convert wide string of digits to skinny string of digits
	    char tmp[20];
	    wcstombs( tmp, buffer, sigdigits+1 );
	    __ZBuf2LD( (char _WCNEAR *)tmp, (long_double _WCNEAR *)&ld );
	}
	#else
	    __ZBuf2LD( (char _WCNEAR *)&buffer[0], (long_double _WCNEAR *)&ld );
	#endif
	if( exponent != 0 ) {
	    _LDScale10x( (long_double _WCNEAR *)&ld, exponent );
	}
	if( flags & NEGATIVE_NUMBER ) {
#ifdef _LONG_DOUBLE_
	    ld.exponent |= 0x8000;
#else
	    ld.word[1] |= 0x80000000;
#endif
	}
    }
#ifdef _LONG_DOUBLE_
    pld->exponent  = ld.exponent;
    pld->high_word = ld.high_word;
    pld->low_word  = ld.low_word;
#else
    pld->word[0] = ld.word[0];
    pld->word[1] = ld.word[1];
#endif
    if( (exponent+sigdigits-1) > 308 ) {          /* overflow */
	return( _OVERFLOW );
    } else if( (exponent+sigdigits-1) < -308 ) {  /* underflow */
	return( _UNDERFLOW );
    }
    return( _NONZERO );			// indicate number is non-zero
}


_WMRTLINK double __F_NAME(strtod,wcstod)( const CHAR_TYPE *bufptr, CHAR_TYPE **endptr )
/*************************************************************************************/
{
    int		rc;
    double	x;
    long_double	ld;

    rc = __F_NAME(__Strtold,__wStrtold)( bufptr, &ld, endptr );
    if( rc == _ZERO ) {
	x = 0.0;
    } else {
#ifdef _LONG_DOUBLE_
	int exponent;
	exponent = ld.exponent & 0x7FFF;
	if( exponent >= 0x3FFF + 0x400 ) {	// if exponent too large
	    __set_ERANGE();			// - overflow
	    if( ld.exponent & 0x8000 ) {
		x = - HUGE_VAL;
	    } else {
		x = HUGE_VAL;
	    }
	} else if( exponent < 0x3FFF - 0x3FF - 51 ) {  /* underflow */
	    if( exponent < 0x3FFF - 0x3FFF - 52 ) {
		__set_ERANGE();	// we're sure it's zero
	        x = 0.0;
	    } else {
		union {
		    double	    value;	// - double value
		    unsigned long   word[2];	// - so we can access bits
		} d;
		// not sure if it's really going to give back zero
	        __iLDFD( (long_double _WCNEAR *)&ld, (double _WCNEAR *)&x );
		// retest for underflow given that we are really
		// returning a double
		d.value = x;
		if( x == 0.0 || (d.word[1] & 0x7FF00000) == 0 ) {
		    __set_ERANGE();
		}
	    }
	} else {
	    __iLDFD( (long_double _WCNEAR *)&ld, (double _WCNEAR *)&x );
	    // retest for overflow or underflow given that we are really
	    // returning a double
	    if( rc == _OVERFLOW ) {
		__set_ERANGE();
	    } else {
		union {
		    double	    value;	// - double value
		    unsigned long   word[2];	// - so we can access bits
		} d;
		d.value = x;
		if( (d.word[1] & 0x7FF00000) == 0 ) {
		    __set_ERANGE();
		}
	    }
	}
#else
	if( rc == _OVERFLOW ) {			// if exponent too large
	    __set_ERANGE();			// - overflow
	    if( ld.word[1] & 0x80000000 ) {
		x = - HUGE_VAL;
	    } else {
		x = HUGE_VAL;
	    }
	} else if( (ld.word[1] & 0x7FFFFFFF) == 0 && ld.word[0] == 0 ) {
	    /* underflow */
	    __set_ERANGE();
	    x = 0.0;
	} else {
	    x = ld.value;
	}
#endif
    }
    return( x );
}

