/*
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *%	  Copyright (C) 1990, by WATCOM Systems Inc. All rights     %
 *%	  reserved. No part of this software may be reproduced	    %
 *%	  in any form or by any means - graphic, electronic or	    %
 *%	  mechanical, including photocopying, recording, taping     %
 *%	  or information storage and retrieval systems - except     %
 *%	  with the written permission of WATCOM Systems Inc.	    %
 *%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     Algorithm taken from "Numerical Recipes in C" page 188.
     Polynomial constants are from "Computer Approximations" by
     John F. Hart, et al. 1968, page 141.

  Modified:	By:		Reason:
  ---------	---		-------
  28-jan-90	F.W.Crigger	Initial implementation
  30-aug-91	J.B.Schueler	Added "const" attribute to tables
--------------------------------------------------------Version-11.0---------
  06-jun-95	Greg Bentz	variety.h and dll linkage
*/
#include "variety.h"
#include <math.h>
#include "mathcode.h"
#include "pdiv.h"

extern	double _EvalPoly( double, const double *, int );

#define Two_over_pi  0.63661977236758134307
#define PI	     3.14159265358979323846
#define PIby4	     (PI/4.)


static const double _Y0P[] = {
		 0.27295402301235549348e+7,
		-0.27725001364628093891e+10,
		 0.10869647137874907869e+13,
		-0.20430408868635630787e+15,
		 0.18702004643462447944e+17,
		-0.75173702582406051323e+18,
		 0.96736537146262430222e+19,
		-0.40421653126104925544e+19
	     };

static const double _Y0Q[] = {
		1.0,
		0.86456632023518903220e+3,
		0.50069757439413061879e+6,
		0.22400416108570432901e+9,
		0.80363630456265291118e+11,
		0.22900727639481849036e+14,
		0.49441312329373299270e+16,
		0.72620404690767951505e+18,
		0.54768700204477752393e+20
	     };

extern const double _P0P[5];
extern const double _P0Q[6];
extern const double _Q0P[5];
extern const double _Q0Q[6];


_WMRTLINK double y0( double x )		/* Bessel function y0(x) */
/*******************/
    {
	double	xx, y, z;

	if( x < 0.0 ) {
//	      z = _matherr( DOMAIN, "y0", &x, &x, - HUGE_VAL );
	    z = __math1err( FUNC_Y0 | M_DOMAIN | V_NEG_HUGEVAL, &x );
	} else if( x < 8.0 ) {
	    y = x * x;
	    z = PDIV( _EvalPoly( y, _Y0P, 7 )  ,  _EvalPoly( y, _Y0Q, 8 ) )
		  + Two_over_pi * j0(x) * log(x);
	} else {
	    z = PDIV( 8.0 , x );
	    y = z * z;
	    xx = x - PIby4;
	    z = sqrt( PDIV( Two_over_pi , x ) ) *
		( PDIV( sin(xx) * _EvalPoly( y, _P0P, 4 ) , _EvalPoly( y, _P0Q, 5 ) )
	    + z * cos(xx) *(PDIV(_EvalPoly( y, _Q0P, 4 ) , _EvalPoly( y, _Q0Q, 5 ))));
	}
	return( z );
    }
